/*
 * Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <dlfcn.h>
#include <stdarg.h>
#include <unordered_set>

#include <infiniband/ibdiag/ibdiag.h>
#include <infiniband/ibutils/release_containers.h>
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h"
#include "infiniband/ibdm/cable/CombinedCable.h"


/****************************************************/
#include <infiniband/ibdiag/ibdiag_version.h>

string get_ibdiag_version()
{
    string version = "\"";
    #ifdef IBDIAG_VERSION
        version += IBDIAG_VERSION;
    #else
        version += "unknown";
    #endif

    //add dependency libraries
    version += "\",\"";
    version += get_ibdm_version();
    version += "\",\"";
    version += get_ibis_version();
    version += "\"";
    return version;
}

u_int8_t get_operational_vl_num(u_int8_t opvl_code)
{
    if ((opvl_code < 1) | (opvl_code > 5))
        return 0;
    if (opvl_code == 5)
        return 15;
    u_int8_t operational_vl = 1;
    opvl_code--;
    operational_vl = (u_int8_t)(operational_vl << opvl_code);
    return operational_vl;
}

vec_p_fabric_err& IBDiag::GetErrorsByAPort(APort* p_aport)
{
    if (errors_by_aport.find(p_aport) == errors_by_aport.end()) {
        // construct empty vector
        errors_by_aport[p_aport];
    }

    return errors_by_aport[p_aport];
}

/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::SetLastError(const char *fmt, ...)
{
    IBDIAG_ENTER;
    char buffer[IBDIAG_ERR_SIZE] = {0};
    va_list args;

    va_start(args, fmt);
    vsnprintf(buffer, sizeof(buffer), fmt, args);
    va_end(args);
    this->last_error.assign(buffer);
    IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "-E- %s\n", this->last_error.c_str());
    IBDIAG_RETURN_VOID;
}


void IBDiag::CleanUpInternalDB()
{
    IBDIAG_ENTER;

    std::unordered_set<direct_route_t*> DRs;

    //clean bad direct routes
    for (auto x : this->bad_direct_routes) {
        DRs.insert(x->direct_route);
        delete x;
    }

    //clean loop direct routes
    for (auto x : this->loop_direct_routes)
        DRs.insert(x);

    //clean bfs list
    for (auto x : this->bfs_list)
        DRs.insert(x);

    //clean known node direct routes, because good nodes was contain only parts of routes other can be in bad path
    for (auto & l : this->bfs_known_node_guids)
        for (auto x : l.second)
            DRs.insert(x);

    //release dr objects
    for (auto x : DRs)
        delete x;

    this->ibdiag_discovery_status = DISCOVERY_NOT_DONE;
    this->root_node = NULL;
    this->root_port_num = 0;
    this->fabric_extended_info.CleanUpInternalDB();

    //make sure all lists + maps are empty
    this->bfs_list.clear();
    this->good_direct_routes.clear();
    this->bad_direct_routes.clear();
    this->loop_direct_routes.clear();
    this->duplicated_guids_detection_errs.clear();
    this->bfs_known_node_guids.clear();
    this->bfs_known_port_guids.clear();
    this->port_routes_map.clear();
    this->errors.clear();

    IBDIAG_RETURN_VOID;
}

void IBDiag::PrintFileTimestamp(const string &file_path, const string &file_type)
{
    IBDIAG_ENTER;

    char timestamp_buf[100] = {0};
    if (!IBFabric::GetFileTimestamp(timestamp_buf, sizeof(timestamp_buf), file_path))
         INFO_PRINT("Load %s file: %s, Timestamp: %s\n", file_type.c_str(), file_path.c_str(),
                    timestamp_buf);
    else
         WARN_PRINT("Load %s file: %s - file is not accessible\n", file_type.c_str(),
                    file_path.c_str());

    IBDIAG_RETURN_VOID;
}

//Returns non zero on error
int IBDiag::OpenFile(const string& name, const OutputControl::Identity& identity, ofstream& sout, bool to_append, bool add_header)
{
    IBDIAG_ENTER;

    int                       rc = 0;
    string                    message;
    OutputControl::Properties properties(identity);
    bool                      isSkipped = false;


    if(properties.is_valid()) {
        if(properties.enabled())
            if(!OutputControl::CreateFolder(properties.path()))
            {
            	string err_message = "Cannot create directory for '" + properties.path() + "'.";
                rc = -1;
            }
            else
            {
			    rc = IBFabric::OpenFile(properties.path().c_str(), sout,
			    		                to_append, message, add_header);

			    if(rc)
				    if(message.empty())
				        string err_message = "Cannot open file '" + properties.path() + "'.";
            }
        else
            isSkipped = true;
	}
    else
    {
        string err_message = "Cannot retrieve output properties for '" + identity.text() + "'";
        rc = -1;
    }

    if (!isSkipped) {
        if (!message.empty())
            this->SetLastError(message.c_str());
        else
            message = properties.path();
    }

    if (properties.in_summary() && !isSkipped)
        AddGeneratedFile(name, message);

    IBDIAG_RETURN(rc);
}

void IBDiag::CloseFile(ofstream & sout) {
    if (sout.is_open())
        sout << endl << endl << "# File closed at : " << IBFabric::GetNowTimestamp() << endl;

    sout.close();
}

void IBDiag::AddGeneratedFile(const string &name, const string &file_name)
{
    if(!name.empty())
    {
        stringstream ss;
        ss << "-I- " << left << setw(35) << name << " : " << file_name << endl;
        this->generated_files_list += ss.str();
    }
}

// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::GetReverseDirectRoute(INOUT direct_route_t *p_reverse_direct_route,
        IN direct_route_t *p_direct_route, IN bool startWithZero)
{
    IBDIAG_ENTER;
    CLEAR_STRUCT(*p_reverse_direct_route);
    IBNode *p_curr_node = this->root_node;
    if (!p_curr_node) {         //don't have root node
        this->SetLastError("DB error - can't find reverse direct route for direct route=%s - null root node",
                        Ibis::ConvertDirPathToStr(p_direct_route).c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    if (p_direct_route->length <= 1) {
        *p_reverse_direct_route = *p_direct_route;
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
    }

    int i, j, loop_min;

    if (startWithZero) {
        p_reverse_direct_route->length = p_direct_route->length;
        // loop will exit before p_reverse_direct_route->path.BYTE[0] is changed
        loop_min = 1;
    } else {
        p_reverse_direct_route->length = (u_int8_t)(p_direct_route->length - 1);
        loop_min = 0;
    }

    for (i = 1, j = p_reverse_direct_route->length - 1; j >= loop_min; ++i, --j) {
        u_int8_t curr_port = p_direct_route->path.BYTE[i];
        if ((curr_port <= 0) || (curr_port > p_curr_node->numPorts)) {
            this->SetLastError("DB error - can't find reverse direct route for direct route=%s - original direct route with port out of range",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        if (!p_curr_node->getPort(curr_port) || !p_curr_node->getPort(curr_port)->p_remotePort) {
            this->SetLastError("DB error - can't find reverse direct route for direct route=%s - reached null port",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str());
            if (!p_curr_node->getPort(curr_port)) {
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Node %s port:%d is null\n",
                           p_curr_node->name.c_str(), curr_port);
            } else {
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Node %s the peer of port:%d is null\n",
                           p_curr_node->name.c_str(), curr_port);
            }
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        p_reverse_direct_route->path.BYTE[j] = p_curr_node->getPort(curr_port)->p_remotePort->num;
        p_curr_node = p_curr_node->getPort(curr_port)->p_remotePort->p_node;
        if (!p_curr_node) {
            this->SetLastError("DB error - can't find reverse direct route for direct route=%s - reached null node",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_EXCEEDS_MAX_HOPS
int IBDiag::ConcatDirectRoutes(IN direct_route_t *p_direct_route1,
        IN direct_route_t *p_direct_route2,
        OUT direct_route_t *p_direct_route_result)
{
    IBDIAG_ENTER;
    CLEAR_STRUCT(*p_direct_route_result);
    int i;
    if ((p_direct_route1->length + p_direct_route2->length) > IBDIAG_MAX_HOPS) {
        this->SetLastError("Concatenate of direct route1=%s with direct route2=%s exceeded max hops",
                Ibis::ConvertDirPathToStr(p_direct_route1).c_str(),
                Ibis::ConvertDirPathToStr(p_direct_route2).c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_EXCEEDS_MAX_HOPS);
    }
    //insert first direct route elements at the beginning
    for (i = 0; i < p_direct_route1->length; ++i)
        p_direct_route_result->path.BYTE[i] = p_direct_route1->path.BYTE[i];
    //insert second direct route elements at the end
    for (i = 0; i < p_direct_route2->length; ++i)
        p_direct_route_result->path.BYTE[i + p_direct_route1->length] = p_direct_route2->path.BYTE[i];
    //update length
    p_direct_route_result->length = (u_int8_t)(p_direct_route1->length + p_direct_route2->length);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::ResetAppData(bool force)
{
    static bool done = false;
    if (!force && done) {
        return;
    }
    //mark the node as not checked (xmit wait, port counters extended capability, ext speed counters)
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node)
            continue;
        p_node->appData1.val = 0;
    }
    done = true;
}

void IBDiag::DumpIbisStat(ofstream &sout)
{
    GetIbisPtr()->m_mads_stat.output_all_records(sout, false);
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
const char* IBDiag::GetLastError()
{
    IBDIAG_ENTER;
    if (this->last_error != "")
        IBDIAG_RETURN(this->last_error.c_str());
    IBDIAG_RETURN("Unknown");
}

IBDiag::IBDiag():
                curr_iteration(0), curr_control_session(NULL),
                retrieved_ext_node_info_ok(false), ext_node_info_check_results(0),
                ibdiag_status(NOT_INITILIAZED),
                ibdiag_discovery_status(DISCOVERY_NOT_DONE), check_duplicated_guids(true),
                check_switch_duplicated_guids(false), root_node(NULL), root_port_num(0),
                llr_active_cell_size(RETRANS_NO_RETRANS),
                ber_threshold(DEFAULT_BER_THRESHOLD),
                send_node_desc(false), send_port_info(false),
                max_node_info_mads_in_pack(DEAFULT_NODES_INFO_MADS_IN_PACK),
                rail_on(false), rail_data_collected(false), dump_simulator_cables(false),
                cable_exported(false),
                // Control API
                export_lib_handle(NULL),
                    pf_export_get_api_version(NULL),
                    pf_export_open_session(NULL),
                    pf_export_close_session(NULL),
                    pf_export_data_node(NULL),
                    pf_export_data_port(NULL),

                // Export API
                control_lib_handle(NULL),
                    pf_control_get_api_version(NULL),
                    pf_control_open_session(NULL),
                    pf_control_close_session(NULL),
                    pf_control_is_stage_enabled(NULL),
                    pf_control_get_stage_flags(NULL),
                    pf_control_get_scope(NULL)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}


IBDiag::NodeInfoSendData::NodeInfoSendData(list_route_and_node_info &in_list):
                        position_itr(in_list.begin()), end_itr(in_list.end())
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}

IBDiag::DirectRouteAndNodeInfo::DirectRouteAndNodeInfo(): p_direct_route(NULL),
                                                is_filled(false)
{
    IBDIAG_ENTER;
    CLEAR_STRUCT(node_info);
    IBDIAG_RETURN_VOID;
}

IBDiag::~IBDiag()
{
    IBDIAG_ENTER;
    //Make sure that there are no pending and on the wire MADs.
    this->ibis_obj.MadRecAll();

    this->CleanUpInternalDB();

    if (this->export_lib_handle)
        dlclose(this->export_lib_handle);

    if (this->control_lib_handle)
        dlclose(this->control_lib_handle);

    for (auto& aport_vec_pair : errors_by_aport) {
        // release the values (errors), don't release the keys (APorts)
        release_container_data(aport_vec_pair.second);
    }

    IBDIAG_RETURN_VOID;
}


// Returns: SUCCESS_CODE / ERR_CODE_INIT_FAILED
int IBDiag::Init()
{
    IBDIAG_ENTER;
    if (this->IsInit() == false) {
        IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Making IBDM using internal log\n");
        ibdmUseInternalLog();

        IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Initialize IBIS\n");
        if (this->ibis_obj.Init()) {
            this->SetLastError("Failed to init ibis object, err=%s", this->ibis_obj.GetLastError());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
        }

        if (this->capability_module.Init()) {
            this->SetLastError("Failed to init capability_module object");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
        }
        this->ibdiag_status = NOT_SET_PORT;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_INIT_FAILED
int IBDiag::SetPort(const char* device_name, phys_port_t port_num)
{
    IBDIAG_ENTER;
    if (this->IsInit() == false) {
        this->SetLastError("IBDiag initialize wasn't done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    if (this->IsReady() == true) {
        this->SetLastError("IBDiag set_port was already done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Setting IBDiag with (Device name=%s, Port number=%u)\n",
            device_name, port_num);
    if (this->ibis_obj.SetPort(device_name, port_num)) {
        this->SetLastError("Failed to set port of ibis object, err=%s", this->ibis_obj.GetLastError());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    this->ibdiag_status = READY;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_INIT_FAILED
int IBDiag::SetPort(u_int64_t port_guid)
{
    IBDIAG_ENTER;
    if (this->IsInit() == false) {
        this->SetLastError("IBDiag initialize wasn't done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    if (this->IsReady() == true) {
        this->SetLastError("IBDiag set_port was already done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Setting IBDiag with (GUID=" U64H_FMT ")\n",
            port_guid);
    if (this->ibis_obj.SetPort(ADB2C_CPU_TO_BE64(port_guid))) {
        this->SetLastError("Failed to set port of ibis object, err=%s", this->ibis_obj.GetLastError());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    this->ibdiag_status = READY;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


IBNode * IBDiag::GetNodeByDirectRoute(IN const direct_route_t *p_direct_route)
{
    IBDIAG_ENTER;
    //set root node
    IBNode *p_curr_node = this->root_node;
    if (!p_curr_node)
        IBDIAG_RETURN(NULL);

    for (int length = 1; length < p_direct_route->length; ++length) {
        u_int8_t curr_port = p_direct_route->path.BYTE[length];
        if ((curr_port <= 0) || (curr_port > p_curr_node->numPorts))
            IBDIAG_RETURN(NULL);
        if (!p_curr_node->getPort(curr_port) || !p_curr_node->getPort(curr_port)->p_remotePort)
            IBDIAG_RETURN(NULL);

        p_curr_node = p_curr_node->getPort(curr_port)->p_remotePort->p_node;
        if (!p_curr_node)
            IBDIAG_RETURN(NULL);
    }
    IBDIAG_RETURN(p_curr_node);
}

direct_route_t * IBDiag::GetDR(const IBNode *p_node)
{
   IBDIAG_ENTER;

    const list_p_direct_route &list_direct_routes =
                        this->GetDirectRoutesByNodeGuid(p_node->guid_get());
    if (list_direct_routes.empty())
        IBDIAG_RETURN(NULL);

    IBDIAG_RETURN(list_direct_routes.front());
}


/**
 * @brief return the last "hop" in the direct route - the port that connects to
 *        the destination node.
 */
IBPort* IBDiag::GetLastOutPortByDirectRoute(IN const direct_route_t *p_direct_route)
{
    IBDIAG_ENTER;
    if (!p_direct_route)
        IBDIAG_RETURN(nullptr);
    direct_route_t dr_copy = *p_direct_route;
    --dr_copy.length;
    IBNode* p_last_node = GetNodeByDirectRoute(&dr_copy);
    if (!p_last_node)
        IBDIAG_RETURN(nullptr);
    IBPort* last_port = p_last_node->getPort(dr_copy.path.BYTE[dr_copy.length]);

    IBDIAG_RETURN(last_port);
}

APort* IBDiag::GetLastOutAPortByDirectRoute(IN const direct_route_t *p_direct_route)
{
    IBDIAG_ENTER;
    IBPort* p_port = GetLastOutPortByDirectRoute(p_direct_route);
    if (!p_port)
        IBDIAG_RETURN(nullptr);

    IBDIAG_RETURN(p_port->p_aport);
}

IBPort* IBDiag::GetDestPortByDirectRoute(IN const direct_route_t *p_direct_route)
{
    IBDIAG_ENTER;
    if (!p_direct_route->length)
        IBDIAG_RETURN(nullptr);

    if (p_direct_route->length == 1)
        IBDIAG_RETURN(this->GetRootPort());

    IBPort* p_last_out_port = GetLastOutPortByDirectRoute(p_direct_route);
    if (!p_last_out_port)
        IBDIAG_RETURN(nullptr);

    IBDIAG_RETURN(p_last_out_port->p_remotePort);
}

APort* IBDiag::GetDestAPortByDirectRoute(IN const direct_route_t *p_direct_route)
{
    IBDIAG_ENTER;
    IBPort* p_port = GetDestPortByDirectRoute(p_direct_route);
    if (!p_port)
        IBDIAG_RETURN(nullptr);

    IBDIAG_RETURN(p_port->p_aport);
}

/**
 * @brief Checks if a list of direct routes all have the same APort destination
 *
 * @param p_routes - a list of direct routes
 * @return true if all the routes lead to the same APort, false otherwise.
 */
bool IBDiag::isRoutesToSameAPort(IN const list_p_direct_route& p_routes)
{
    IBDIAG_ENTER;
    // compare the destination APort of every direct route in the list to the
    // last APort of the first direct route (of length > 1) in the routes list.
    APort *p_first_dest_aport = nullptr;
    if (p_routes.size() < 2)
        IBDIAG_RETURN(true);

    for (const direct_route_t* p_route : p_routes) {
        // ignore a path if it's empty (end is the same as start)
        if (p_route->length == 1 && p_route->path.BYTE[0] == 0)
            continue;
        if (!p_first_dest_aport) {
            p_first_dest_aport = GetDestAPortByDirectRoute(p_route);
        }
        APort *p_curr_dest_aport = GetDestAPortByDirectRoute(p_route);
        if (!p_curr_dest_aport || p_curr_dest_aport != p_first_dest_aport)
            IBDIAG_RETURN(false);
    }

    IBDIAG_RETURN(true);
}

IBPort * IBDiag::GetRootPort()
{
    IBDIAG_ENTER;
    IBPort *p_port = NULL;

    //get root port
    if (!this->root_node) {
        this->SetLastError("DB error - root node is null");
        IBDIAG_RETURN(NULL);
    }
    p_port = this->root_node->getPort(this->root_port_num);
    if (!p_port) {
        this->SetLastError("DB error - root port is null");
        IBDIAG_RETURN(NULL);
    }
    IBDIAG_RETURN(p_port);
}


// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::GetLocalPortState(OUT u_int8_t& state)
{
    IBDIAG_ENTER;
    IBPort *p_root_port = this->GetRootPort();
    if (!p_root_port) {
        this->SetLastError("DB error - failed to found root port");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    //check if logical state in init
    struct SMP_PortInfo *p_root_port_info = this->fabric_extended_info.getSMPPortInfo(p_root_port->createIndex);
    if (!p_root_port_info) {
        this->SetLastError("DB error - failed to found port info for root port");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    state = p_root_port_info->PortState;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


int IBDiag::GetAllLocalPortGUIDs(OUT local_port_t local_ports_array[IBIS_MAX_LOCAL_PORTS],
            OUT u_int32_t *p_local_ports_num)
{
    IBDIAG_ENTER;
    if (this->ibis_obj.GetAllLocalPortGUIDs(local_ports_array, p_local_ports_num)) {
        this->SetLastError("Failed to get all local ports from ibis object, err=%s",
                this->ibis_obj.GetLastError());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::getLatestSupportedVersion(int page_number,
                                      unsigned int &latest_version)
{
    IBDIAG_ENTER;
    switch (page_number) {
        case VS_MLNX_CNTRS_PAGE0:
            latest_version = PAGE0_LATEST_VER;
            break;
        case VS_MLNX_CNTRS_PAGE1:
            latest_version = PAGE1_LATEST_VER;
            break;
        case VS_MLNX_CNTRS_PAGE255:
            latest_version = PAGE255_LATEST_VER;
            break;
        default:
            this->SetLastError("Failed to found latest version"  \
                               " for page %d",page_number);
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            break;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DiscoverFabricFromFile(IN const string &csv_file, IN bool build_direct_routes)
{
    int rc = IBDIAG_SUCCESS_CODE;

    IBDiagFabric diag_fabric(discovered_fabric, fabric_extended_info, capability_module);

    try {
        if((rc = diag_fabric.UpdateFabric(csv_file))) {
            this->SetLastError("Failed to parse CSV file");
            IBDIAG_RETURN(rc);
        }
    } catch (exception &e) {
        this->SetLastError("Failed to parse CSV file, error: %s", e.what());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    } catch (...) {
        this->SetLastError("Failed to parse CSV file");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    }

    // Collect APorts
    list_p_fabric_general_err aports_errors;
    INFO_PRINT("Build APorts Info\n");
    this->CollectAPortsData(aports_errors);

    if(build_direct_routes && (rc = BuildDirectRoutesDB()))
        IBDIAG_RETURN(rc);

    SCREEN_INFO_PRINT("Discovering ... " U32D_FMT " Nodes (" U32D_FMT " Switches & " U32D_FMT
                      " CAs) discovered.",
                      diag_fabric.getNodesFound(), diag_fabric.getSWFound(),
                      diag_fabric.getCAFound());

    IBDIAG_RETURN(rc);
}

IBSpecialPortType IBDiag::GetSpecialPortType(IBPort *p_port)
{
    if (!capability_module.IsSupportedSMPCapability(
            p_port->p_node, EnSMPCapIsSpecialPortsMarkingSupported))
        return IB_NOT_SPECIAL_PORT;

    struct SMP_MlnxExtPortInfo *p_mepi = fabric_extended_info.
            getSMPMlnxExtPortInfo(p_port->createIndex);

    if (p_mepi && p_mepi->IsSpecialPort)
       return (IBSpecialPortType)p_mepi->SpecialPortType;

    return IB_NOT_SPECIAL_PORT;
}

IBSpecialPortType IBDiag::GetSpecialCAPortType(IBNode *p_node)
{
    if (p_node->type == IB_SW_NODE)
        return IB_NOT_SPECIAL_PORT;

    // go over all ports of this node:
    for (u_int32_t i = 1; i <= p_node->numPorts; ++i) {
        IBPort *p_port = p_node->getPort((phys_port_t)i);
        if (!p_port || p_port->get_internal_state() !=  IB_PORT_STATE_ACTIVE)
            continue;

        if (!p_port->p_remotePort ||
            p_port->p_remotePort->get_internal_state() !=  IB_PORT_STATE_ACTIVE)
            continue;

        return GetSpecialPortType(p_port->p_remotePort);
    }

    return IB_NOT_SPECIAL_PORT;
}

int IBDiag::ParseSMDBFile()
{
    IBDIAG_ENTER;

    static int rc = -1;

    // This method should be called once at most.
    // In case of return code is equal to -1, it means that the function
    // wasn't called yet.
    // If we try to call this function more than once, the return code
    // will be the last return code (not -1) of this function.
    if (rc != -1)
        IBDIAG_RETURN(rc);

    rc = IBDIAG_SUCCESS_CODE;

    try {
        if((rc = this->ibdiag_smdb.ParseSMDB(this->smdb_path))) {
            this->SetLastError("Failed to parse SMDB file - %s", this->smdb_path.c_str());
            IBDIAG_RETURN(rc);
        }
    } catch (exception &e) {
        this->SetLastError("Failed to parse SMDB file - %s, error: %s", this->smdb_path.c_str(),
                           e.what());
        rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
        IBDIAG_RETURN(rc);
    } catch (...) {
        this->SetLastError("Failed to parse SMDB file - %s", this->smdb_path.c_str());
        rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
        IBDIAG_RETURN(rc);
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::WriteIbisStatFile(const std::string file_name)
{
    IBDIAG_ENTER;

    ofstream sout;
    int rc = this->OpenFile("Ibis Stat",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->DumpIbisStat(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::LoadSymbol(void *p_lib_handle, const char *name, void **symbol,
                       list_p_fabric_general_err &export_data_errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    char *err;

    // Since the value of the symbol could actually be NULL, the correct way to test dlsym() for an
    // error is to call dlerror() to clear any old error, then call dlsym(), and then call dlerror()
    // again.
    dlerror();
    *symbol = dlsym(p_lib_handle, name);
    if ((err = dlerror()) != NULL) {
        export_data_errors.push_back(new ExportDataErr(
            NULL, NULL, "Failed to load symbol %s: %s", name, err));
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::InitExportAPI(const string &export_lib_path)
{
    IBDIAG_ENTER;

    int rc = 0;
    list_p_fabric_general_err export_data_errors;

    if (this->export_lib_handle) {
        ERR_PRINT("Export library is already loaded\n");
        IBDIAG_RETURN(1);
    }

    this->export_lib_handle = dlopen(export_lib_path.c_str(), RTLD_LOCAL | RTLD_LAZY);
    if (!this->export_lib_handle) {
        char *err = dlerror();
        ERR_PRINT("Failed to load library - %s\n", err);
        IBDIAG_RETURN(1);
    }

    if ((rc = this->LoadSymbol(this->export_lib_handle, "export_get_api_version",
                              (void **)&this->pf_export_get_api_version, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->export_lib_handle, "export_open_session",
                              (void **)&this->pf_export_open_session, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->export_lib_handle, "export_close_session",
                              (void **)&this->pf_export_close_session, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->export_lib_handle, "export_data_node",
                              (void **)&this->pf_export_data_node, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->export_lib_handle, "export_data_port",
                              (void **)&this->pf_export_data_port, export_data_errors)))
        goto exit;

exit:
    if (rc) {
        for (list_p_fabric_general_err::iterator it = export_data_errors.begin();
                it != export_data_errors.end(); ++it) {
            ERR_PRINT("%s\n", (*it)->GetErrorLine().c_str());
            delete *it;
        }
        dlclose(this->export_lib_handle);
        this->export_lib_handle = NULL;
        this->pf_export_get_api_version = NULL;
        this->pf_export_open_session = NULL;
        this->pf_export_close_session = NULL;
        this->pf_export_data_node = NULL;
        this->pf_export_data_port = NULL;
    }

    IBDIAG_RETURN(rc);
}

export_session_handle_t IBDiag::ExportOpenSession(u_int64_t iteration)
{
    if (this->pf_export_open_session)
        return this->pf_export_open_session(iteration);

    return NULL;
}

void IBDiag::ExportCloseSession(export_session_handle_t session_handle, int rc)
{
    if (this->pf_export_close_session)
        this->pf_export_close_session(session_handle, rc);
}

int IBDiag::ExportData(export_session_handle_t session_handle,
                       list_p_fabric_general_err &export_data_errors)
{
    IBDIAG_ENTER;

    int rc2, rc = IBDIAG_SUCCESS_CODE;
    export_data_port_t export_data_port;
    export_data_node_t export_data_node;
    cable_record_data_t cable_record_data;

    for (map_str_pnode::iterator nI = this->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = nI->second;
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        CLEAR_STRUCT(export_data_node);

        export_data_node.node_guid = p_curr_node->guid_get();

        export_data_node.p_smp_node_info =
            this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        export_data_node.p_smp_switch_info =
            this->fabric_extended_info.getSMPSwitchInfo(p_curr_node->createIndex);
        export_data_node.p_vendor_spec_general_info =
            this->fabric_extended_info.getVSGeneralInfo(p_curr_node->createIndex);
        export_data_node.p_smp_temp_sensing =
            this->fabric_extended_info.getSMPTempSensing(p_curr_node->createIndex);
        export_data_node.p_switch_network_info =
             this->fabric_extended_info.getVSSwitchNetworkInfo(p_curr_node->createIndex);

        // Congestion Control
        export_data_node.cc_congestion_info =
                this->fabric_extended_info.getCCEnhancedCongestionInfo(p_curr_node->createIndex);
        export_data_node.curr_switch_general_settings =
                this->fabric_extended_info.getCCSwitchGeneralSettings(p_curr_node->createIndex);

        rc2 = this->pf_export_data_node(session_handle, &export_data_node);
        if (rc2) {
            ExportDataErr *p_curr_fabric_err
                = new ExportDataErr(p_curr_node, NULL, "Failed to export node data: %d", rc2);
            if (!p_curr_fabric_err) {
                this->SetLastError("Failed to allocate ExportDataErr");
                rc = IBDIAG_ERR_CODE_NO_MEM;
                goto exit;
            }
            export_data_errors.push_back(p_curr_fabric_err);
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
        }

        for (phys_port_t pI = 0; pI <= p_curr_node->numPorts; ++pI) {
            IBPort *p_curr_port = p_curr_node->getPort(pI);
            if (!p_curr_port || !p_curr_port->getInSubFabric())
                continue;

            CLEAR_STRUCT(export_data_port);

            export_data_port.node_guid = p_curr_port->p_node->guid_get();
            export_data_port.port_guid = p_curr_port->guid_get();
            export_data_port.port_num = p_curr_port->num;

            if (p_curr_port->p_remotePort) {
                export_data_port.remote_node_guid = p_curr_port->p_remotePort->p_node->guid_get();
                export_data_port.remote_port_guid = p_curr_port->p_remotePort->guid_get();
                export_data_port.remote_port_num = p_curr_port->p_remotePort->num;
            }

            export_data_port.p_smp_port_info =
                this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            export_data_port.p_smp_mlnx_ext_port_info =
                this->fabric_extended_info.getSMPMlnxExtPortInfo(p_curr_port->createIndex);

            export_data_port.p_pm_port_counters =
                this->fabric_extended_info.getPMPortCounters(p_curr_port->createIndex);
            export_data_port.p_pm_port_counters_extended =
                this->fabric_extended_info.getPMPortCountersExtended(p_curr_port->createIndex);
            export_data_port.p_pm_port_extended_speeds_counters =
                this->fabric_extended_info.getPMPortExtSpeedsCounters(p_curr_port->createIndex);
            export_data_port.p_pm_port_extended_speeds_rsfec_counters =
                this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(p_curr_port->createIndex);
            export_data_port.p_pm_port_calc_counters =
                this->fabric_extended_info.getPMPortCalcCounters(p_curr_port->createIndex);
            export_data_port.p_vendor_spec_port_llr_statistics =
                this->fabric_extended_info.getVSPortLLRStatistics(p_curr_port->createIndex);
            export_data_port.p_pm_port_rcv_error_details =
                this->fabric_extended_info.getPMPortRcvErrorDetails(p_curr_port->createIndex);
            export_data_port.p_pm_port_xmit_discard_details =
                this->fabric_extended_info.getPMPortXmitDiscardDetails(p_curr_port->createIndex);
            export_data_port.p_pm_port_samples_control =
                this->fabric_extended_info.getPMPortSamplesControl(p_curr_port->createIndex);

            // Congestion Control
            for (u_int8_t vl = 0; vl < IB_NUM_VL; ++vl)
                export_data_port.curr_port_profile_settings[vl] =
                    this->fabric_extended_info.getCCPortProfileSettings(p_curr_port->createIndex, vl);

            export_data_port.curr_sl_mapping_settings =
                    this->fabric_extended_info.getCCSLMappingSettings(p_curr_port->createIndex);
            export_data_port.curr_hca_general_settings =
                    this->fabric_extended_info.getCCHCAGeneralSettings(p_curr_port->createIndex);
            export_data_port.curr_hca_rp_parameters =
                    this->fabric_extended_info.getCCHCARPParameters(p_curr_port->createIndex);
            export_data_port.curr_hca_np_parameters =
                    this->fabric_extended_info.getCCHCANPParameters(p_curr_port->createIndex);
            export_data_port.curr_hca_statistics_query =
                    this->fabric_extended_info.getCCHCAStatisticsQuery(p_curr_port->createIndex);
            export_data_port.cc_hca_algo_config_sup =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigSup(p_curr_port->createIndex);

            for (int algo = 0; algo < MAX_CC_ALGO_SLOT; ++algo) {
                export_data_port.cc_hca_algo_config[algo] =
                    this->fabric_extended_info.getCC_HCA_AlgoConfig(p_curr_port->createIndex, algo);

                export_data_port.cc_hca_algo_config_params[algo] =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigParams(p_curr_port->createIndex, algo);

                export_data_port.cc_hca_algo_counters[algo] =
                    this->fabric_extended_info.getCC_HCA_AlgoCounters(p_curr_port->createIndex, algo);
            }

            if (p_curr_port->p_combined_cable) {
                if (p_curr_port->p_combined_cable->ExportData(&cable_record_data))
                    export_data_port.p_cable_record_data = &cable_record_data;
            }

            rc2 = this->pf_export_data_port(session_handle, &export_data_port);
            if (rc2) {
                export_data_errors.push_back(new ExportDataErr(
                    p_curr_node, p_curr_port, "Failed to export port data: %d", rc2));
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            }
        }
    }

exit:
    IBDIAG_RETURN(rc);
}

int IBDiag::InitControlAPI(const string &control_lib_path)
{
    IBDIAG_ENTER;

    int rc = 0;
    list_p_fabric_general_err export_data_errors;

    if (this->control_lib_handle) {
        ERR_PRINT("Control library is already loaded\n");
        IBDIAG_RETURN(1);
    }

    this->control_lib_handle = dlopen(control_lib_path.c_str(), RTLD_LOCAL | RTLD_LAZY);
    if (!this->control_lib_handle) {
        char *err = dlerror();
        ERR_PRINT("Failed to load library - %s\n", err);
        IBDIAG_RETURN(1);
    }

    if ((rc = this->LoadSymbol(this->control_lib_handle, "control_get_api_version",
                               (void **)&this->pf_control_get_api_version, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->control_lib_handle, "control_open_session",
                               (void **)&this->pf_control_open_session, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->control_lib_handle, "control_close_session",
                               (void **)&this->pf_control_close_session, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->control_lib_handle, "control_is_stage_enabled",
                               (void **)&this->pf_control_is_stage_enabled, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->control_lib_handle, "control_get_stage_flags",
                               (void **)&this->pf_control_get_stage_flags, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->control_lib_handle, "control_get_scope",
                               (void **)&this->pf_control_get_scope, export_data_errors)))
        goto exit;

exit:
    if (rc) {
        for (list_p_fabric_general_err::iterator it = export_data_errors.begin();
                it != export_data_errors.end(); ++it) {
            ERR_PRINT("%s\n", (*it)->GetErrorLine().c_str());
            delete *it;
        }
        dlclose(this->control_lib_handle);
        this->control_lib_handle = NULL;
        this->pf_control_get_api_version = NULL;
        this->pf_control_open_session = NULL;
        this->pf_control_close_session = NULL;
        this->pf_control_is_stage_enabled = NULL;
        this->pf_control_get_stage_flags = NULL;
        this->pf_control_get_scope = NULL;
    }

    IBDIAG_RETURN(rc);
}

control_session_handle_t IBDiag::ControlOpenSession(u_int64_t iteration)
{
    if (this->pf_control_open_session)
        return this->pf_control_open_session(iteration);

    return NULL;
}

int IBDiag::ControlCloseSession(control_session_handle_t session_handle, int rc)
{
    if (this->pf_control_close_session)
        return this->pf_control_close_session(session_handle, rc);

    return 1;
}

int IBDiag::ControlIsStageEnabled(control_session_handle_t session_handle, const char *stage_name)
{
    if (this->pf_control_is_stage_enabled)
        return this->pf_control_is_stage_enabled(session_handle, stage_name);

    return 1;
}

void IBDiag::ControlGetStageFlags(control_session_handle_t session_handle, const char *stage_name,
                                  control_stage_flags_t *stage_flags)
{
    memset(stage_flags, 0xff, sizeof(control_stage_flags_t));

    if (this->pf_control_get_stage_flags)
        this->pf_control_get_stage_flags(session_handle, stage_name, stage_flags);
}

bool IBDiag::ControlGetScope(control_session_handle_t session_handle, control_scope_t *scope)
{
    if (this->pf_control_get_scope) {
        this->pf_control_get_scope(session_handle, scope);
        return true;
    }

    return false;
}

int IBDiag::CollectAPortsData(list_p_fabric_general_err& errors_list)
{
    IBDIAGNET_ENTER;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        vec_p_aport &aports_vec = this->discovered_fabric.APortsBySysGuid[p_curr_node->system_guid_get()];

        for (phys_port_t port_num = 0; port_num <= p_curr_node->numPorts; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort(port_num);
            if (!p_curr_port || !p_curr_port->p_port_hierarchy_info)
                continue;

            const PortHierarchyInfo *p_hinfo = p_curr_port->p_port_hierarchy_info;
            if (p_hinfo->m_aport < 0 || p_hinfo->m_plane < 0 || p_hinfo->m_num_of_planes < 0)
                continue;

            size_t curr_size = aports_vec.size();
            size_t new_size = static_cast<size_t>(p_hinfo->m_aport + 1);
            if (curr_size < new_size) {
                aports_vec.resize(new_size, NULL);
            }

            //planes are numbered from 1 to ..., so we have nullptr at 0 index
            //and vector's size is actually (PortHierarchyInfo.m_num_of_planes)"number of planes" + 1
            if (!p_hinfo->m_plane) {
                errors_list.push_back(new APortInvalidPlaneNumError(p_curr_port));
                continue;
            }
            if (!aports_vec[p_hinfo->m_aport]) {
                // avoid creating APort if NumOfPlanes isn't consistent with the plane index
                if (p_hinfo->m_plane <= p_hinfo->m_num_of_planes) {
                    aports_vec[p_hinfo->m_aport] = new APort(p_hinfo->m_num_of_planes, p_hinfo->m_aport);
                } else {
                    errors_list.push_back(new APortInvalidPlaneNumError(p_curr_port,
                                                static_cast<size_t>(p_hinfo->m_num_of_planes)));
                    continue;
                }
            }

            APort *p_aport = aports_vec[p_hinfo->m_aport];
            size_t index = static_cast<size_t>(p_hinfo->m_plane);
            if (p_aport->ports.size() > index) {
                if (!p_aport->ports[index])
                    p_aport->ports[index] = p_curr_port;
                else
                    errors_list.push_back(new APortPlaneAlreadyInUseError(p_curr_port));
            } else {
                 errors_list.push_back(new APortInvalidPlaneNumError(p_curr_port,
                                            static_cast<size_t>(p_aport->ports.size() - 1)));
            }

            if (p_curr_port->p_aport)
                errors_list.push_back(new PlaneInMultipleAPorts(p_curr_port));
            else
                p_curr_port->p_aport = p_aport;
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

template<typename T>
void sm_valid_check_set(const set<T> & values_set, const string& name, list_p_fabric_general_err& errors)
{
    if (values_set.size() > 1) {
        stringstream ss;
        ss << "Field '" << name << "' has " << values_set.size()
           << " different values across the fabric [";

        size_t n = 5;

        for (auto val = values_set.begin(); val != values_set.end() && n; ++val, --n) {
            if (val != values_set.begin())
                ss << "; ";

            if (n == 1)
                ss << "...";
            else
                ss << +(*val);
        }

        ss << "]";

        errors.push_back(new SMConfigDiffValues(ss.str()));
    }
}

void IBDiag::PostReportsSMValidations_CC_algo(list_p_fabric_general_err& errors)
{
    IBDIAGNET_ENTER;

    vector <set <uint32_t>> algo_info_vec_set(MAX_CC_ALGO_SLOT);
    // CC Algo Config
    vector <set <uint8_t>>  algo_en_vec_set(MAX_CC_ALGO_SLOT);
    vector <set <uint8_t>>  algo_status_vec_set(MAX_CC_ALGO_SLOT);
    vector <set <uint8_t>>  algo_trace_en_vec_set(MAX_CC_ALGO_SLOT);
    vector <set <uint8_t>>  algo_counters_en_vec_set(MAX_CC_ALGO_SLOT);
    vector <set <uint16_t>> algo_sl_bitmask_vec_set(MAX_CC_ALGO_SLOT);
    vector <set <uint8_t>>  algo_encap_len_vec_set(MAX_CC_ALGO_SLOT);
    //CC Algo Config Params
    vector <set <uint8_t>>  algo_params_encap_len_vec_set(MAX_CC_ALGO_SLOT);
    vector <set <uint16_t>> algo_params_sl_bitmask_vec_set(MAX_CC_ALGO_SLOT);

    vector <vector <set <uint32_t>>> algo_params_vec_vec_set(MAX_CC_ALGO_SLOT);
    // init the params vec
    for (uint8_t algo = 0; algo < MAX_CC_ALGO_SLOT; ++algo) {
        algo_params_vec_vec_set[algo].resize(MAX_CC_ALGO_ARRAY_LAYOUT);
    }

    for (IBNode *p_node : this->discovered_fabric.HCAs) {

        if (!p_node || !p_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_node->numPorts + 1; ++pi) {

            IBPort *p_port = p_node->getPort(pi);
            if (!p_port || !p_port->is_data_worthy())
                continue;

            CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config_sup =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigSup(p_port->createIndex);
            if (!p_cc_hca_algo_config_sup)
                continue;

            const uint32_t* algo_info_vec = (uint32_t*)p_cc_hca_algo_config_sup->encapsulation;

            for (uint32_t algo = 0; algo < MAX_CC_ALGO_SLOT; ++algo) {

                algo_info_vec_set[algo].insert(algo_info_vec[algo]);

                CC_CongestionHCAAlgoConfig* p_cc_hca_algo_config = this->fabric_extended_info.
                    getCC_HCA_AlgoConfig(p_port->createIndex, algo);

                if (p_cc_hca_algo_config) {
                    algo_en_vec_set[algo].insert(p_cc_hca_algo_config->algo_en);
                    algo_status_vec_set[algo].insert(p_cc_hca_algo_config->algo_status);
                    algo_trace_en_vec_set[algo].insert(p_cc_hca_algo_config->trace_en);
                    algo_counters_en_vec_set[algo].insert(p_cc_hca_algo_config->counters_en);
                    algo_sl_bitmask_vec_set[algo].insert(p_cc_hca_algo_config->sl_bitmask);
                    algo_encap_len_vec_set[algo].insert(p_cc_hca_algo_config->encap_len);
                }

                CC_CongestionHCAAlgoConfigParams* p_cc_algo_config_params =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigParams(p_port->createIndex, algo);

                if (p_cc_algo_config_params) {
                    algo_params_encap_len_vec_set[algo].insert(p_cc_algo_config_params->encap_len);
                    algo_params_sl_bitmask_vec_set[algo].insert(p_cc_algo_config_params->sl_bitmask);

                    for (int p = 0; p < MAX_CC_ALGO_ARRAY_LAYOUT; ++p)
                        algo_params_vec_vec_set[algo][p].insert(p_cc_algo_config_params->encapsulation[p]);
                }

            } // algos
        } // ports
    } // nodes

    for (int i = 0; i < MAX_CC_ALGO_SLOT; i++) {

        stringstream s;
        s << "algo_info" << "_" << i;
        sm_valid_check_set(algo_info_vec_set[i], s.str(), errors);

        // CC Algo Config
        s.str("");
        s << "algo_en" << "_" << i;
        sm_valid_check_set(algo_en_vec_set[i], s.str(), errors);

        s.str("");
        s << "algo_status" << "_" << i;
        sm_valid_check_set(algo_status_vec_set[i], s.str(), errors);

        s.str("");
        s << "algo_trace_en" << "_" << i;
        sm_valid_check_set(algo_trace_en_vec_set[i], s.str(), errors);

        s.str("");
        s << "algo_counters_en" << "_" << i;
        sm_valid_check_set(algo_counters_en_vec_set[i], s.str(), errors);

        s.str("");
        s << "algo_sl_bitmask" << "_" << i;
        sm_valid_check_set(algo_sl_bitmask_vec_set[i], s.str(), errors);

        s.str("");
        s << "algo_encap_len" << "_" << i;
        sm_valid_check_set(algo_encap_len_vec_set[i], s.str(), errors);

        //CC Algo Config Params
        s.str("");
        s << "algo_params_encap_len" << "_" << i;
        sm_valid_check_set(algo_params_encap_len_vec_set[i], s.str(), errors);

        s.str("");
        s << "algo_params_sl_bitmask" << "_" << i;
        sm_valid_check_set(algo_params_sl_bitmask_vec_set[i], s.str(), errors);

        for (int p = 0; p < MAX_CC_ALGO_ARRAY_LAYOUT; ++p) {
            s.str("");
            s << "algo_params" << "_" << i << "_param" << p;
            sm_valid_check_set(algo_params_vec_vec_set[i][p], s.str(), errors);
        }
    }

    IBDIAGNET_RETURN_VOID;
}

void IBDiag::PostReportsSMValidations_CC_RP_NP_Params(list_p_fabric_general_err& errors)
{
    IBDIAGNET_ENTER;

    // CC
    set <uint8_t> cc_en_react_set;
    set <uint8_t> cc_en_notify_set;
    // RP
    set <uint8_t>  clamp_tgt_rate_after_time_inc_set;
    set <uint8_t>  clamp_tgt_rate_set;
    set <uint32_t> rpg_time_reset_set;
    set <uint32_t> rpg_byte_reset_set;
    set <uint8_t>  rpg_threshold_set;
    set <uint32_t> rpg_max_rate_set;
    set <uint16_t> rpg_ai_rate_set;
    set <uint16_t> rpg_hai_rate_set;
    set <uint8_t>  rpg_gd_set;
    set <uint8_t>  rpg_min_dec_fac_set;
    set <uint32_t> rpg_min_rate_set;
    set <uint32_t> rate_to_set_on_first_cnp_set;
    set <uint16_t> dce_tcp_g_set;
    set <uint32_t> dce_tcp_rtt_set;
    set <uint32_t> rate_reduce_monitor_period_set;
    set <uint16_t> initial_alpha_value_set;
    // NP
    set <uint16_t> min_time_between_cnps_set;
    set <uint8_t>  cnp_sl_set;
    set <uint8_t>  cnp_sl_mode_set;

    for (IBNode *p_node : this->discovered_fabric.HCAs) {

        if (!p_node || !p_node->getInSubFabric())
            continue;

        for (phys_port_t pi = 1; pi < p_node->numPorts + 1; ++pi) {

            IBPort *p_port = p_node->getPort(pi);
            if (!p_port || !p_port->is_data_worthy())
                continue;

            // CC
            struct CC_CongestionHCAGeneralSettings* p_curr_cc_hca_general_settings =
                    this->fabric_extended_info.getCCHCAGeneralSettings(p_port->createIndex);

            if (p_curr_cc_hca_general_settings) {
                cc_en_react_set.insert(p_curr_cc_hca_general_settings->en_react);
                cc_en_notify_set.insert(p_curr_cc_hca_general_settings->en_notify);
            }

            CC_CongestionHCARPParameters *p_cc_rp_parameters =
                this->fabric_extended_info.getCCHCARPParameters(p_port->createIndex);

            if (p_cc_rp_parameters) {
                clamp_tgt_rate_after_time_inc_set.insert(p_cc_rp_parameters->clamp_tgt_rate_after_time_inc);
                clamp_tgt_rate_set.insert(p_cc_rp_parameters->clamp_tgt_rate);
                rpg_time_reset_set.insert(p_cc_rp_parameters->rpg_time_reset);
                rpg_byte_reset_set.insert(p_cc_rp_parameters->rpg_byte_reset);
                rpg_threshold_set.insert(p_cc_rp_parameters->rpg_threshold);
                rpg_max_rate_set.insert(p_cc_rp_parameters->rpg_max_rate);
                rpg_ai_rate_set.insert(p_cc_rp_parameters->rpg_ai_rate);
                rpg_hai_rate_set.insert(p_cc_rp_parameters->rpg_hai_rate);
                rpg_gd_set.insert(p_cc_rp_parameters->rpg_gd);
                rpg_min_dec_fac_set.insert(p_cc_rp_parameters->rpg_min_dec_fac);
                rpg_min_rate_set.insert(p_cc_rp_parameters->rpg_min_rate);
                rate_to_set_on_first_cnp_set.insert(p_cc_rp_parameters->rate_to_set_on_first_cnp);
                dce_tcp_g_set.insert(p_cc_rp_parameters->dce_tcp_g);
                dce_tcp_rtt_set.insert(p_cc_rp_parameters->dce_tcp_rtt);
                rate_reduce_monitor_period_set.insert(p_cc_rp_parameters->rate_reduce_mionitor_period);
                initial_alpha_value_set.insert(p_cc_rp_parameters->initial_alpha_value);
            }

            CC_CongestionHCANPParameters *p_cc_np_parameters =
                    this->fabric_extended_info.getCCHCANPParameters(p_port->createIndex);

            if (p_cc_np_parameters) {
                min_time_between_cnps_set.insert(p_cc_np_parameters->min_time_between_cnps);
                cnp_sl_set.insert(p_cc_np_parameters->cnp_sl);
                cnp_sl_mode_set.insert(p_cc_np_parameters->cnp_sl_mode);
            }
        } // ports
    } // nodes

    sm_valid_check_set(cc_en_react_set, "CC en_react", errors);
    sm_valid_check_set(cc_en_notify_set, "CC en_notify", errors);
    // RP
    sm_valid_check_set(clamp_tgt_rate_after_time_inc_set, "CC_RP_clamp_tgt_rate_after_time_inc", errors);
    sm_valid_check_set(clamp_tgt_rate_set, "CC_RP_clamp_tgt_rate", errors);
    sm_valid_check_set(rpg_time_reset_set, "CC_RP_rpg_time_reset", errors);
    sm_valid_check_set(rpg_byte_reset_set, "CC_RP_rpg_byte_reset", errors);
    sm_valid_check_set(rpg_threshold_set, "CC_RP_rpg_threshold", errors);
    sm_valid_check_set(rpg_max_rate_set, "CC_RP_rpg_max_rate", errors);
    sm_valid_check_set(rpg_ai_rate_set, "CC_RP_rpg_ai_rate", errors);
    sm_valid_check_set(rpg_hai_rate_set, "CC_RP_rpg_hai_rate", errors);
    sm_valid_check_set(rpg_gd_set, "CC_RP_rpg_gd", errors);
    sm_valid_check_set(rpg_min_dec_fac_set, "CC_RP_rpg_min_dec_fac", errors);
    sm_valid_check_set(rpg_min_rate_set, "CC_RP_rpg_min_rate", errors);
    sm_valid_check_set(rate_to_set_on_first_cnp_set, "CC_RP_rate_to_set_on_first_cnp", errors);
    sm_valid_check_set(dce_tcp_g_set, "CC_RP_dce_tcp_g", errors);
    sm_valid_check_set(dce_tcp_rtt_set, "CC_RP_dce_tcp_rtt", errors);
    sm_valid_check_set(rate_reduce_monitor_period_set, "CC_RP_rate_reduce_monitor_period", errors);
    sm_valid_check_set(initial_alpha_value_set, "CC_RP_initial_alpha_value", errors);
    // NP
    sm_valid_check_set(min_time_between_cnps_set, "CC_NP_min_time_between_cnps", errors);
    sm_valid_check_set(cnp_sl_set, "CC_NP_cnp_sl", errors);
    sm_valid_check_set(cnp_sl_mode_set, "CC_NP_cnp_sl_mode", errors);

    IBDIAGNET_RETURN_VOID;
}

// Post Reports SM configuration Validations
int IBDiag::PostReportsSMValidations(list_p_fabric_general_err& errors, uint32_t& num_of_diff_val)
{
    IBDIAGNET_ENTER;

    set<uint16_t> ooo_sl_mask_set;
    set<uint16_t> adaptive_timeout_sl_mask_set;
    set<uint8_t> virt_enable_set;
    vector <set <uint8_t>> sl2vl_vec_set(IB_NUM_SL);
    vector <set <uint32_t>> vl_arbit_lo_vec_set(64);
    vector <set <uint32_t>> vl_arbit_hi_vec_set(64);

    // currently validation is only for HCAs
    for (IBNode *p_node : this->discovered_fabric.HCAs) {
        if (!p_node ||
                !p_node->getInSubFabric() ||
                p_node->isSpecialNode())
            continue;

        // skip UFM nodes by name
        if (p_node->name.find("UFM") != string::npos || p_node->name.find("ufm") != string::npos)
            continue;

        for (phys_port_t port_num = 1; port_num <= p_node->numPorts; ++port_num) {
            IBPort *p_port = p_node->getPort(port_num);

            if (!p_port || !p_port->is_data_worthy())
                continue;

            // Mlnx Ext Port Info
            struct SMP_MlnxExtPortInfo *p_mepi = fabric_extended_info.
                            getSMPMlnxExtPortInfo(p_port->createIndex);
            if (!p_mepi)
                continue;

            ooo_sl_mask_set.insert(p_mepi->OOOSLMask);
            adaptive_timeout_sl_mask_set.insert(p_mepi->AdaptiveTimeoutSLMask);

            // Virtualization Info
            struct SMP_VirtualizationInfo *p_virt_info = fabric_extended_info.
                            getSMPVirtualizationInfo(p_port->createIndex);

            if (p_virt_info)
                virt_enable_set.insert(p_virt_info->virtualization_enable);

            // sl2vl - SLVL[iport][oport][sl] = vl;
            // By ReadCASLVL func in-port is always 0
            if (!p_node->SLVL.empty())
                for (int i = 0; i < IB_NUM_SL; i++)
                    sl2vl_vec_set[i].insert(p_node->SLVL[0][p_port->num][i]);

            // VL Arbitration
            for (uint32_t block = LOWER_LOW_PRIORITY_VL_ARBITRATION_TABLE;
                    block <= UPPER_HIGH_PRIORITY_VL_ARBITRATION_TABLE; ++block) {

                SMP_VLArbitrationTable* p_curr_vl_arbitration_table =
                    this->fabric_extended_info.getSMPVLArbitrationTable(p_port->createIndex, block);

                if (p_curr_vl_arbitration_table) {
                    for (uint32_t entry = 0; entry < 32; ++entry) {

                        uint32_t vl_arb = p_curr_vl_arbitration_table->VLArb[entry].VL;
                        vl_arb <<= 8;
                        vl_arb |= p_curr_vl_arbitration_table->VLArb[entry].Weight;

                        switch (block) {
                            case LOWER_LOW_PRIORITY_VL_ARBITRATION_TABLE:
                                vl_arbit_lo_vec_set[entry].insert(vl_arb);
                                break;

                            case UPPER_LOW_PRIORITY_VL_ARBITRATION_TABLE:
                                vl_arbit_lo_vec_set[entry + 32].insert(vl_arb);
                                break;

                            case LOWER_HIGH_PRIORITY_VL_ARBITRATION_TABLE:
                                vl_arbit_hi_vec_set[entry].insert(vl_arb);
                                break;

                            case UPPER_HIGH_PRIORITY_VL_ARBITRATION_TABLE:
                                vl_arbit_hi_vec_set[entry + 32].insert(vl_arb);
                                break;
                        }
                    } // entries
                }
            } // VL Arbitration blocks

        } // ports
    } // nodes

    sm_valid_check_set(ooo_sl_mask_set, "OOOSLMask", errors);
    sm_valid_check_set(adaptive_timeout_sl_mask_set, "AdaptiveTimeoutSLMask", errors);
    sm_valid_check_set(virt_enable_set, "virtualization_enable", errors);

    for (int i = 0; i < IB_NUM_SL; i++) {
        stringstream ss;
        ss << "SL2VL_" << i;
        sm_valid_check_set(sl2vl_vec_set[i], ss.str(), errors);
    }

    for (int i = 0; i < 64; i++) {
        stringstream ss;
        ss << "VL Arbitration Low Priority " << i;
        sm_valid_check_set(vl_arbit_lo_vec_set[i], ss.str(), errors);
    }

    for (int i = 0; i < 64; i++) {
        stringstream ss;
        ss << "VL Arbitration High Priority " << i;
        sm_valid_check_set(vl_arbit_hi_vec_set[i], ss.str(), errors);
    }

    // CC funcs
    this->PostReportsSMValidations_CC_RP_NP_Params(errors);
    this->PostReportsSMValidations_CC_algo(errors);

    num_of_diff_val = (uint32_t)errors.size();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}
