/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <infiniband/misc/regexp/regexp.h>
#include "infiniband/ibdm/Fabric.h"
#include "infiniband/ibutils/output_stream_primitives.h"
#include "infiniband/ibutils/release_containers.h"
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h"
#include "infiniband/ibdiag/ibdiag_fat_tree.h"

#include <cstdlib>


using namespace std;

#define ERROR_PREFIX "-E- "
#define WARNING_PREFIX "-W- "
#define INFO_PREFIX "-I- "

#define MAX_2L_FT_DISTANCE 2
#define MAX_3L_FT_DISTANCE 4
#define MAX_4L_FT_DISTANCE 6

#define RANKS_2L_FT 2
#define RANKS_3L_FT 3
#define RANKS_4L_FT 4

#define SHOW_SYS_GUID  1
#define SHOW_NODE_GUID 2

//should be applied for switches only
static string GetSwitchASIC(const IBNode *p_switch)
{
    string asic = p_switch->getPrismaSwitchASIC();
    return (asic.empty () ? "" : "/" + asic);
}

int FTTopology::Show_GUID = 0;

static string GetNodeRecord(const IBNode *p_switch)
{
    IBDIAG_ENTER;

    stringstream out;

    if (FTTopology::Show_GUID == SHOW_SYS_GUID)
        out << PTR(p_switch->system_guid_get()) << GetSwitchASIC(p_switch)
            << " -- ";
    else if (FTTopology::Show_GUID == SHOW_NODE_GUID)
        out << PTR(p_switch->guid_get()) << " -- ";

    const string &name = p_switch->getAlternativeName();
    size_t found = name.find("/A");
    if (found == string::npos)
        found = name.find("/U");

    out << (found != string::npos ? name.substr(0, found) : name)
        << GetSwitchASIC(p_switch);

    IBDIAG_RETURN(out.str());
}

static const IBNode *GetRemoteSwitch(const IBPort *p_port)
{
    if (!p_port || p_port->isFNMPort() || !p_port->isSymmetricLink())
        IBDIAGNET_RETURN(NULL);

    const IBNode *p_remote = p_port->get_remote_node();
    if (!p_remote || p_remote->type != IB_SW_NODE)
        IBDIAGNET_RETURN(NULL);

    IBDIAGNET_RETURN(p_remote);
}

static double GetSpeed(const IBNode& node) {
    IBDIAGNET_ENTER;

    for (phys_port_t i = 1 ; i <= node.numPorts; i++) {
        const IBPort *p_port = node.getPort(i);

        if (p_port && p_port->p_remotePort)
            IBDIAGNET_RETURN(speed2double(p_port->get_internal_speed()));
    }

    IBDIAGNET_RETURN(0);
}

FTClassification::FTClassification(const FTTopology &inT):
                    maxDistance(0), maxThresholdDistance(0), topology(inT)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

const IBNode* FTClassification::GetLeafToClassify(const classifications_vec &classifications)
{
    IBDIAGNET_ENTER;

    const IBNode* p_ret = NULL;
    switch(this->maxThresholdDistance) {
        case MAX_2L_FT_DISTANCE:
        case MAX_3L_FT_DISTANCE:
        case MAX_4L_FT_DISTANCE:
        {
            distance_to_nodes_map::const_iterator itr =
                                    this->distanceToNodesMap.find(this->maxThresholdDistance);
            if (itr == this->distanceToNodesMap.end()) {
                this->lastError << "Wrong Classification. There are no nodes at the distance: "
                                    << maxThresholdDistance;
                IBDIAGNET_RETURN(NULL);
            }

            p_ret = this->GetLeafToClassify(classifications, itr->second);
            break;
        }
        default:
            this->lastError << "Wrong Classification. Unexpected maximal distance: "
                            << maxThresholdDistance;
            IBDIAGNET_RETURN(NULL);
    }

    IBDIAGNET_RETURN(p_ret);
}

//This function has to calculate how many switches with wrong distances we may have to proceed
//We try to find next possible leaf switch. If Max distance differs from possible
//values for (2, 3, 4) level Fat Tree it may be because of missing links and we should have
//relatively small number of switches with such distance comparing to other distances.
int FTClassification::CalculateThreshold() const
{
    IBDIAGNET_ENTER;

    size_t size = this->nodesByDistanceMap.size();
    if (size > 1000)
        IBDIAGNET_RETURN(static_cast<int>(this->nodesByDistanceMap.begin()->first->numPorts/2));

    if (size > 100)
        IBDIAGNET_RETURN(10);

    if (size > 50)
        IBDIAGNET_RETURN(4);

    if (size > 25)
        IBDIAGNET_RETURN(2);

    if (size > 5)//just for simple testing topologies
        IBDIAGNET_RETURN(1);

    IBDIAGNET_RETURN(0);
}

int FTClassification::GetMaxThresholdDistance() const
{
    IBDIAGNET_ENTER;

    int treshold = this->CalculateThreshold();
    int maxDistanceTreshold = 0;

    for (distance_to_nodes_map::const_iterator itr = this->distanceToNodesMap.begin();
            itr != this->distanceToNodesMap.end(); itr++) {
        //we move from smaller to larger distances
        //and we end with "max distance" which holds more nodes than treshold.

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "on distance: %d there are %d nodes\n",
                itr->first, static_cast<int>(itr->second.size()));

        if (static_cast<int>(itr->second.size()) > treshold)
            maxDistanceTreshold = itr->first;
    }

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "classification: %p maxDistanceTreshold: %d and treshold %d\n",
               this, maxDistanceTreshold, treshold);

    IBDIAGNET_RETURN(maxDistanceTreshold);
}

const IBNode* FTClassification::GetLeafToClassify(const classifications_vec &classifications,
                                                  const nodes_list &nodes)
{
    IBDIAGNET_ENTER;
 
    nodedata_to_nodes_map nodesToStart;
    for (nodes_list::const_iterator itr_nodes = nodes.begin(); itr_nodes != nodes.end();
            itr_nodes++) {
        const IBNode* p_node = *itr_nodes;
        if (!p_node) {
            this->lastError << "Cannot find a next leaf to classify. "
                            << "DB error - One of IB-Nodes is NULL";
            IBDIAGNET_RETURN(NULL);
        }

        const NodeData *p_data = this->topology.GetClassificationNodeData(p_node);
        if (!p_data) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                    "No classification NodeData was associated with the node GUID=" U64H_FMT "\n",
                    p_node->guid_get());
            continue;
        }

        nodesToStart[*p_data].push_back(p_node);
    }

    for (nodedata_to_nodes_map::const_reverse_iterator itr = nodesToStart.rbegin();
            itr !=  nodesToStart.rend(); itr++) {
        for (nodes_list::const_iterator itr_nodes = itr->second.begin();
                itr_nodes != itr->second.end(); itr_nodes++) {
            const IBNode* p_node = *itr_nodes;
            size_t inDistanceToAll = 0;
            for (classifications_vec::const_reverse_iterator itr_class = classifications.rbegin();
                    itr_class != classifications.rend(); itr_class++) {
                const FTClassification* p_cls = *itr_class;
                if (!p_cls) {
                    this->lastError << "Cannot find a next leaf to classify. "
                                    << "One of FT-Classifications is NULL";
                    IBDIAGNET_RETURN(NULL);
                }

                if (p_cls == this) {
                    inDistanceToAll++;
                    continue;
                }

                bool inDistance = false;
                if (p_cls->CheckDistanceTo(*p_node, inDistance)) {
                    this->lastError << "Cannot find a next leaf to classify. "
                                    << "Failed to find a distance to the Switch "
                                    << "(GUID: " << PTR(p_node->guid_get())
                                    << ')';
                    IBDIAGNET_RETURN(NULL);
                }

                if (inDistance)
                    inDistanceToAll++;
            }

            if (inDistanceToAll == classifications.size()) {
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                    "Next selected node GUID=" U64H_FMT " has ca: %u and links: %u\n",
                    p_node->guid_get(), itr->first.caNum, itr->first.linksNum);
                IBDIAGNET_RETURN(p_node);
            }
        }
    }

    this->lastError << "Failed to find next leaf to classify out of "
                    << nodes.size() << " possible nodes";

    IBDIAGNET_RETURN(NULL);
}

int FTClassification::CheckDistanceTo(const IBNode &node, bool &inDistance) const
{
    IBDIAGNET_ENTER;

    inDistance = false;

    nodes_by_distance_map::const_iterator itr = this->nodesByDistanceMap.find(&node);
    if (itr == this->nodesByDistanceMap.end())
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);

    if (itr->second >= this->maxThresholdDistance ||
            (this->maxThresholdDistance == MAX_4L_FT_DISTANCE && itr->second == 4))
        inDistance = true;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Checking distance. Classifications ptr1=%p, node's GUID: "
                            U64H_FMT " on distance: %d, maxTresholdDistance %d, inDistance: %s\n",
                            this, node.guid_get(), itr->second, this->maxThresholdDistance,
                            inDistance ? "True" : "False");

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTClassification::CountEquals(const classifications_vec &classifications) const
{
    IBDIAGNET_ENTER;

    int ret = 1;
    for (size_t i = 0; i < classifications.size(); i++) {
        if (this != classifications[i] && this->EqualsTo(*classifications[i]))
            ret++;
    }

    IBDIAGNET_RETURN(ret);
}

bool FTClassification::EqualsTo(const FTClassification &other) const
{
    IBDIAGNET_ENTER;

    if (this->nodesByRank.empty() ||
            other.nodesByRank.empty() ||
            this->nodesByRank.size() != other.nodesByRank.size()) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Classifications ptr1=%p and ptr2=%p are not equal\n",
                this, &other);
        IBDIAGNET_RETURN(false);
    }

    //we have ordered by pointer lists of nodes
    size_t leavesRank = this->nodesByRank.size() - 1;
    bool ret = (this->nodesByRank[0] == other.nodesByRank[0]) &&
               (this->nodesByRank[leavesRank] == other.nodesByRank[leavesRank]);

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Classifications ptr1=%p and ptr2=%p are %s\n",
                this, &other, ret ? "equal" : "not equal");

    IBDIAGNET_RETURN(ret);
}

int FTClassification::Classify(const IBNode &fromLeaf)
{
    IBDIAGNET_ENTER;

    this->ClassifyByDistance(fromLeaf);
    int rc = this->SetNodesRanks();

    //todo consider something better
    //IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "%s", this->ToString().c_str());
    IBDIAGNET_RETURN(rc);
}

//todo consider something better
string FTClassification::ToString () const
{
    IBDIAGNET_ENTER;

    stringstream strStream;

    strStream << "Classification (ptr) " << this << endl;
    for (distance_to_nodes_map::const_iterator itr_dist = this->distanceToNodesMap.begin();
            itr_dist != this->distanceToNodesMap.end(); itr_dist++) {
        strStream << "distance: " << itr_dist->first << " has " << itr_dist->second.size() << " [";

        for (nodes_list::const_iterator itr_nodes = itr_dist->second.begin();
                itr_nodes != itr_dist->second.end(); itr_nodes++) {
            const IBNode* p_node = *itr_nodes;
            if (!p_node) {
                strStream << " NULL_ptr";
                continue;
            }

            strStream << ' ' << PTR(p_node->guid_get());
        }

        strStream << " ]" << endl;
    }

    IBDIAGNET_RETURN(strStream.str());
}

int FTClassification::SetNodesRanks() {
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    switch (this->maxDistance) {
        case MAX_2L_FT_DISTANCE:
            rc = this->Set2L_FTRanks();
            break;
        case MAX_3L_FT_DISTANCE:
            rc = this->Set3L_FTRanks();
            break;
        case MAX_4L_FT_DISTANCE:
            rc = this->Set4L_FTRanks();
            break;
        default:
            nodesByRank.clear();
            //todo think of warning
    }

    for (size_t i = 0; i < this->nodesByRank.size(); i++)
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "on rank: %lu there are: %lu\n",
                i, this->nodesByRank[i].size());

    IBDIAGNET_RETURN(rc);
}

int FTClassification::Set4L_FTRanks()
{
    IBDIAGNET_ENTER;

    this->nodesByRank.resize(RANKS_4L_FT);

    this->SetRankToNodes(this->distanceToNodesMap[0], this->nodesByRank[3]);//leaves
    this->SetRankToNodes(this->distanceToNodesMap[6], this->nodesByRank[3]);//leaves

    this->SetRankToNodes(this->distanceToNodesMap[1], this->nodesByRank[2]);//#2 Level
    this->SetRankToNodes(this->distanceToNodesMap[5], this->nodesByRank[2]);//#2 Level

    int rc = this->Set4L_DistanceToRanks(2, 1);//#either leaves or #1 level
    if (rc)
       IBDIAGNET_RETURN(rc);

    rc = this->Set4L_DistanceToRanks(4, 3);//#either leaves or #1 level
    if (rc)
       IBDIAGNET_RETURN(rc);

    for (nodes_list::const_iterator itr = this->distanceToNodesMap[3].begin();
            itr != this->distanceToNodesMap[3].end(); itr++) {
        const IBNode* p_node = *itr;
        bool isRoot = true;
        for (phys_port_t i = 1 ; i <= p_node->numPorts; i++) {
            const IBNode *p_remote = GetRemoteSwitch(p_node->getPort(i));
            if (!p_remote)
                continue;

            nodes_set::const_iterator itr_remote = this->nodesByRank[1].find(p_remote);
            if (itr_remote == this->nodesByRank[1].end()) { //it is not in #1 level
                isRoot = false;
                this->nodesByRank[2].insert(p_node);//#2 Level
                break;
            }
        }

        if (isRoot)
            this->nodesByRank[0].insert(p_node);//roots
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTClassification::Set4L_DistanceToRanks(int distance, int compareDistance)
{
    IBDIAGNET_ENTER;

    for (nodes_list::const_iterator itr = this->distanceToNodesMap[distance].begin();
            itr != this->distanceToNodesMap[distance].end(); itr++) {
        const IBNode* p_node = *itr;
        bool isLeaf = true;
        for (phys_port_t i = 1 ; i <= p_node->numPorts ; i++) {
            const IBNode *p_remote = GetRemoteSwitch(p_node->getPort(i));
            if (!p_remote)
                continue;

            nodes_by_distance_map::const_iterator itr_remote = this->nodesByDistanceMap.find(p_remote);
            if (itr_remote == this->nodesByDistanceMap.end()) {
                //   this->lastError << "Failed to find distance for switch on 4L"
                //                   << "(GUID: " << PTR(p_remote->guid_get()) << ')';
                //   IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
                continue;
            }

            if (itr_remote->second != compareDistance) {
                isLeaf = false;
                this->nodesByRank[1].insert(p_node);//#1 Level
                break;
            }
        }

        if (isLeaf)
            this->nodesByRank[3].insert(p_node);//leaves
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTClassification::Set3L_FTRanks()
{
    IBDIAGNET_ENTER;

    this->nodesByRank.resize(RANKS_3L_FT);

    this->SetRankToNodes(this->distanceToNodesMap[0], this->nodesByRank[2]);//leaves
    this->SetRankToNodes(this->distanceToNodesMap[4], this->nodesByRank[2]);//leaves

    this->SetRankToNodes(this->distanceToNodesMap[1],  this->nodesByRank[1]);//#1 Level
    this->SetRankToNodes(this->distanceToNodesMap[3],  this->nodesByRank[1]);//#1 Level

    for (nodes_list::const_iterator itr = this->distanceToNodesMap[2].begin();
            itr != this->distanceToNodesMap[2].end(); itr++) {
        const IBNode* p_node = *itr;
        bool isRoot = false;
        for (phys_port_t i = 1 ; i <= p_node->numPorts ; i++) {
            const IBNode *p_remote = GetRemoteSwitch(p_node->getPort(i));
            if (!p_remote)
                continue;

            nodes_by_distance_map::const_iterator itr_remote = this->nodesByDistanceMap.find(p_remote);
            if (itr_remote == this->nodesByDistanceMap.end()) {
                   // this->lastError << "Failed to find distance for switch on 3L"
                   //                 << "(GUID: " << PTR(p_remote->guid_get()) << ')';
                   // IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
                continue;
            }

            if (itr_remote->second != 1) {
                isRoot = true;
                this->nodesByRank[0].insert(p_node);//roots
                break;
            }
        }

        if (!isRoot)
            this->nodesByRank[2].insert(p_node);//leaves
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void FTClassification::SetRankToNodes(const nodes_list &inList,  nodes_set &outSet) const
{
    IBDIAGNET_ENTER;

    for (nodes_list::const_iterator itr = inList.begin(); itr != inList.end(); itr++)
        outSet.insert(*itr);

    IBDIAGNET_RETURN_VOID;
}

int FTClassification::Set2L_FTRanks()
{
    IBDIAGNET_ENTER;

    this->nodesByRank.resize(RANKS_2L_FT);

    this->SetRankToNodes(this->distanceToNodesMap[0],   this->nodesByRank[1]);//leaves
    this->SetRankToNodes(this->distanceToNodesMap[2],   this->nodesByRank[1]);//leaves
    this->SetRankToNodes(this->distanceToNodesMap[1],   this->nodesByRank[0]);//Roots

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void FTClassification::ClassifyByDistance(const IBNode &fromLeaf)
{
    IBDIAGNET_ENTER;

    int startingPlane = fromLeaf.getSuitablePlane();
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Started from plane: %d; node: %s GUID:" U64H_FMT "\n",
                    startingPlane, fromLeaf.getName().c_str(), fromLeaf.guid_get());

    queue<SearchData> searchQueue;
    searchQueue.push(SearchData(&fromLeaf, 0));

    while (!searchQueue.empty()) {
        SearchData next = searchQueue.front();
        searchQueue.pop();

        nodes_by_distance_map::const_iterator itr_x =
                                    this->nodesByDistanceMap.find(next.p_node);
        if (itr_x != this->nodesByDistanceMap.end())
            continue;

        this->nodesByDistanceMap[next.p_node] = next.distance;
        this->distanceToNodesMap[next.distance].push_back(next.p_node);


        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Processing node: %s GUID:" U64H_FMT "\n",
                                next.p_node->getName().c_str(), next.p_node->guid_get());
        for (phys_port_t i = 1 ; i <= next.p_node->numPorts ; i++) {
            const IBNode *p_remote = GetRemoteSwitch(next.p_node->getPort(i));
            if (!p_remote)
                continue;

            if (next.p_node->isPrismaSwitch() && !p_remote->isOnSamePlane(startingPlane))
                continue;

            nodes_by_distance_map::const_iterator itr_y =
                                    this->nodesByDistanceMap.find(p_remote);
            if (itr_y != this->nodesByDistanceMap.end())
                continue;

            searchQueue.push(SearchData(p_remote, next.distance + 1));
        }
    }

    this->maxDistance = this->distanceToNodesMap.rbegin()->first;
    this->maxThresholdDistance = this->GetMaxThresholdDistance();

    IBDIAGNET_RETURN_VOID;
}

FTClassification* FTClassificationHandler::GetNewClassification(const FTTopology &topology)
{
   IBDIAGNET_ENTER;

   FTClassification* p_ret = new FTClassification(topology);
   this->classifications.push_back(p_ret);

   IBDIAGNET_RETURN(p_ret);
}

void FTClassification::SwapRanks(nodes_by_rank_vec &ranks)
{
    IBDIAGNET_ENTER;

    this->nodesByRank.swap(ranks);

    IBDIAGNET_RETURN_VOID;
}

FTClassificationHandler::~FTClassificationHandler()
{
    IBDIAGNET_ENTER;

    release_container_data(classifications);

    IBDIAGNET_RETURN_VOID;
}

const IBNode* FTTopology::GetFirstLeaf()
{
    IBDIAGNET_ENTER;

    map<FTClassification::NodeData, const IBNode*> nodesToStartMap;
    for (set_pnode::const_iterator it = this->fabric.Switches.begin();
            it != fabric.Switches.end(); it++) {
        const IBNode *p_switch = *it;
        if (!p_switch)
            IBDIAGNET_RETURN(NULL);

        FTClassification::NodeData nodeData;
        for (phys_port_t i = 1 ; i <= p_switch->numPorts ; i++) {
            const IBPort *p_port = p_switch->getPort(i);
            if (!p_port || p_port->isFNMPort() || !p_port->isSymmetricLink())
                continue;

            const IBNode *p_remote_node = p_port->get_remote_node();
            if (!p_remote_node)
                continue;

            if (p_remote_node->type == IB_CA_NODE && !p_remote_node->isSpecialNode())
                nodeData.caNum++;
            else if (p_remote_node->type == IB_SW_NODE)
                nodeData.linksNum++;
        }

        if (nodeData.caNum) {
            this->nodeDataMap[p_switch] = nodeData;
            nodesToStartMap[nodeData] = p_switch;
        }
    }

    if (!nodesToStartMap.empty()) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "first selected leaf for classification has ca: %u, links %u\n",
                nodesToStartMap.rbegin()->first.caNum, nodesToStartMap.rbegin()->first.linksNum);
        IBDIAGNET_RETURN(nodesToStartMap.rbegin()->second);
    }

    IBDIAGNET_RETURN(NULL);
}

FTTopology::~FTTopology()
{
    IBDIAGNET_ENTER;

    release_container_data(neighborhoodsByRank);

    IBDIAGNET_RETURN_VOID;
}

int FTTopology::Build(list_p_fabric_general_err &errors, string &lastError)
{
    IBDIAGNET_ENTER;

    const string error_prefix = "Cannot build Fat-Tree topology. ";

    this->outFile << INFO_PREFIX << "Taking roots from the SMDB file" << endl;

    FTClassification::nodes_set roots;
    int rc = this->GetRootsBySMDB(roots);
    if (rc) {
        lastError = error_prefix + "Failed to find roots by SMDB: " + this->lastErrorStream.str();
        IBDIAGNET_RETURN(rc);
    }

    rc = this->FillRanksFromRoots(roots);
    if (rc) {
        lastError = error_prefix + "Failed to fill tree's ranks by SMDB: "
                    + this->lastErrorStream.str();
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::GetRootsBySMDB(FTClassification::nodes_set &roots)
{
    IBDIAGNET_ENTER;

    int startingPlane = UNDEFINED_PLANE;

    for (set_pnode::const_iterator itr = this->fabric.Switches.begin();
            itr != fabric.Switches.end(); itr++) {
        const IBNode *p_switch = *itr;
        if (!p_switch) {
            this->lastErrorStream << "NULL pointer was found in fabric's switches";
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //roots have 0 rank
        if (p_switch->rank)
            continue;

        if (startingPlane == UNDEFINED_PLANE) {
            startingPlane = p_switch->getSuitablePlane();
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Started from plane: %d; node: %s GUID:" U64H_FMT "\n",
                    startingPlane, p_switch->getName().c_str(), p_switch->guid_get());

            roots.insert(p_switch);
        } else {
            if (p_switch->isOnSamePlane(startingPlane))
                roots.insert(p_switch);
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::Build(list_p_fabric_general_err &errors, string &lastError, regExp &rootsRegExp)
{
    IBDIAGNET_ENTER;

    const string error_prefix = "Cannot build Fat-Tree topology. ";

    this->outFile << INFO_PREFIX << "Detecting roots by the regular exp: "
                  << rootsRegExp.getExpr() << endl;

    FTClassification::nodes_set roots;
    int rc = this->GetNodes(roots, rootsRegExp);
    if (rc) {
        lastError = error_prefix + "Failed to find roots: " + this->lastErrorStream.str();
        IBDIAGNET_RETURN(rc);
    }

    rc = this->FillRanksFromRoots(roots);
    if (rc) {
        lastError = error_prefix + "Failed to fill tree's ranks by the regular expression: "
                    + this->lastErrorStream.str();
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::GetNodes(FTClassification::nodes_set &nodes, regExp &regExp)
{
    IBDIAGNET_ENTER;

    int startingPlane = UNDEFINED_PLANE;

    for (set_pnode::const_iterator itr = this->fabric.Switches.begin();
            itr != this->fabric.Switches.end(); itr++) {
        const IBNode *p_switch = *itr;
        if (!p_switch) {
            this->lastErrorStream << "NULL pointer was found in fabric's switches";
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        rexMatch *p_rexRes = regExp.apply(p_switch->description.c_str());
        if (p_rexRes) {
            delete p_rexRes;

            if (startingPlane == UNDEFINED_PLANE) {
                startingPlane = p_switch->getSuitablePlane();
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Started from plane: %d; node: %s GUID:" U64H_FMT "\n",
                    startingPlane, p_switch->getName().c_str(), p_switch->guid_get());

                nodes.insert(p_switch);
            } else {
                if (p_switch->isOnSamePlane(startingPlane))
                    nodes.insert(p_switch);
            }
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::FillRanksFromRoots(FTClassification::nodes_set &roots)
{
    IBDIAGNET_ENTER;

    if (roots.empty()) {
        this->lastErrorStream << "No root was provided for creating topology";
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    const IBNode *p_switch = *roots.begin();
    int startingPlane = p_switch->getSuitablePlane();
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Started from plane: %d; node: %s GUID:" U64H_FMT "\n",
                    startingPlane, p_switch->getName().c_str(), p_switch->guid_get());

    this->nodesByRank.clear();
    this->nodesByRank.push_back(FTClassification::nodes_set());
    this->nodesByRank[0].swap(roots);

    int rank = 0;
    bool keepDiscover = true;
    while (keepDiscover)  {
        const FTClassification::nodes_set &from = this->nodesByRank[rank];
        FTClassification::nodes_set nodes;
        for (FTClassification::nodes_set::const_iterator itr_from = from.begin();
                itr_from != from.end(); itr_from++) {
            const IBNode *p_switch = *itr_from;

            for (phys_port_t i = 1 ; i <= p_switch->numPorts ; i++) {
                const IBNode *p_remoteNode = GetRemoteSwitch(p_switch->getPort(i));
                if (!p_remoteNode)
                    continue;

                if (!p_remoteNode->isPrismaSwitch() && !p_remoteNode->isOnSamePlane(startingPlane))
                    continue;

                if (rank) {
                    const FTClassification::nodes_set &previous = this->nodesByRank[rank-1];
                    if (previous.find(p_remoteNode) != previous.end())
                        continue;
                }

                if (from.find(p_remoteNode) != from.end())
                    continue;

                nodes.insert(p_remoteNode);
            }
        }

        if (!nodes.empty()) {
            this->nodesByRank.push_back(FTClassification::nodes_set());
            this->nodesByRank[++rank].swap(nodes);
        }
        else
            keepDiscover = false;
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::Build(list_p_fabric_general_err &errors, string &lastError,
                      int retries, int equalResults)
{
    IBDIAGNET_ENTER;

    if (FTTopology::Show_GUID == SHOW_NODE_GUID || FTTopology::Show_GUID == SHOW_SYS_GUID)
        this->outFile << '#' << "Output format: "
                << (FTTopology::Show_GUID == SHOW_NODE_GUID ? "Node GUID" : "System GUID(/Switch ASIC)")
                << " -- Node Name"
                << endl << endl;

    const string error_prefix = "Cannot build Fat-Tree topology. ";

    this->outFile << INFO_PREFIX << "Detecting roots by distance classifications"  << endl;

    const IBNode* p_leaf = this->GetFirstLeaf();
    if (!p_leaf) {
        lastError = error_prefix + "Failed to find a leaf switch to classify";
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "first selected leaf switch with GUID: " U64H_FMT " and name: %s\n",
            p_leaf->guid_get(), p_leaf->name.c_str());

    FTClassificationHandler handler;
    FTClassification* p_cls = handler.GetNewClassification(*this);

    int rc = p_cls->Classify(*p_leaf);
    if (rc) {
       lastError = error_prefix + p_cls->GetLastError();
       IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    bool isOk = false;
    for (int i = 0; i < retries; i++) {
        p_leaf = p_cls->GetLeafToClassify(handler.GetClassifications());
        if (!p_leaf) {
            lastError = error_prefix + p_cls->GetLastError();
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
        }

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "i=%d;  next selected leaf switch with GUID: "
                U64H_FMT " and name: %s\n",
                i, p_leaf->guid_get(), p_leaf->name.c_str());

        p_cls = handler.GetNewClassification(*this);
        rc = p_cls->Classify(*p_leaf);
        if (rc) {
            lastError = error_prefix + p_cls->GetLastError();
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
        }

        if (p_cls->CountEquals(handler.GetClassifications()) == equalResults) {
            isOk = true;
            break;
        }
    }

    if (!isOk) {
        this->lastErrorStream <<  error_prefix << "Failed to find "
           << equalResults << " equal Classifications out of " << retries << " retries";
        lastError = this->lastErrorStream.str();
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    p_cls->SwapRanks(this->nodesByRank);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::Validate(list_p_fabric_general_err &errors, string &lastError)
{
    IBDIAGNET_ENTER;

    const string  error_prefix = "Cannot validate Fat-Tree topology. ";

    int rc = this->CreateNeighborhoods(errors);
    if (rc) {
       lastError = error_prefix + this->lastErrorStream.str();
       IBDIAGNET_RETURN(rc);
    }

    rc = this->CheckUpDownLinksAndAPorts(errors);
    if (rc) {
       lastError = error_prefix + this->lastErrorStream.str();
       IBDIAGNET_RETURN(rc);
    }

    rc = this->CheckFLIDs();
    if (rc) {
        lastError = "Cannot validate FLIDs distribution. " + this->lastErrorStream.str();
        IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::CreateNeighborhoods(list_p_fabric_general_err &errors)
{
    IBDIAGNET_ENTER;

    this->neighborhoodsByRank.resize(this->nodesByRank.size()-1);
    for (size_t rank = 0; rank < this->nodesByRank.size()-1; rank++) {
        int rc = this->CreateNeighborhoodsOnRank(errors, rank);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::CreateNeighborhoodsOnRank(list_p_fabric_general_err &errors, size_t rank)
{
    IBDIAGNET_ENTER;

    FTUpHopHistogram histogram(*this, rank);
    int rc = histogram.Init();
    if (rc) {
        this->lastErrorStream << histogram.GetLastError();
        IBDIAGNET_RETURN(rc);
    }

    rc = histogram.CreateNeighborhoods(errors);
    if (rc) {
        this->lastErrorStream << histogram.GetLastError();
        IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

//Looks like we do not need keeping connectivity map for topology.
//All errors are collected either in the following function or when neighborhoods are created.
int FTUpHopHistogram::Init()
{
    IBDIAGNET_ENTER;

    const FTClassification::nodes_set *p_upNodes = this->topology.GetNodesOnRank(this->rank);
    if (!p_upNodes)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);

    const FTClassification::nodes_set *p_downNodes = this->topology.GetNodesOnRank(this->rank + 1);
    if (!p_downNodes)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);

    this->bitSetMaxSize = p_upNodes->size();
    if (this->bitSetMaxSize > FT_UP_HOP_SET_SIZE) {
        this->lastErrorStream << "Bit-set is too short to story " << this->bitSetMaxSize
                              << "bits";
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    this->InitNodeToIndexConverters(*p_upNodes);

    for (FTClassification::nodes_set::const_iterator itr = p_downNodes->begin();
            itr != p_downNodes->end(); itr++) {
        const IBNode *p_downNode = *itr;
        if (!p_downNode) {
            this->lastErrorStream << "One of IBNodes is NULL."
                                   << "Cannot initialize the \"up-hop\" histogram";
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        FTUpHopSet::bit_set upNodesBitSet;
        upNodesBitSet.reset();
        for (phys_port_t i = 1 ; i <= p_downNode->numPorts ; i++) {
            const IBPort *p_port = p_downNode->getPort(i);
            const IBNode *p_remoteNode = GetRemoteSwitch(p_port);
            if (!p_remoteNode)
                continue;

            //two nodes on the same level are connected
            const size_t remoteRank = this->topology.GetNodeRank(p_remoteNode);
            if (remoteRank == this->rank + 1) {
                this->AddIllegalLinkIssue(FTLinkIssue(p_downNode, p_port->num,
                                                        p_remoteNode, p_port->p_remotePort->num,
                                                        remoteRank));
                continue;
            }

            if (remoteRank != this->rank)//not in UP nodes
                continue;

            if (!this->rank)//root's level has rank 0
                this->CheckRootSwitchConnections(*p_remoteNode);

            size_t index = 0;
            int rc = this->NodeToIndex(index, *p_remoteNode);
            if (rc)
                IBDIAGNET_RETURN(rc);

            upNodesBitSet[index] = true;
            this->connectedNodes.Add(p_remoteNode, p_downNode);
        }

        string hash = this->GetHashCode(upNodesBitSet);
        FTUpHopSet &hopSet = this->upHopSetsMap[hash];
        if (hopSet.downNodes.empty())//encountered for the first time
            hopSet.upNodesBitSet = upNodesBitSet;

        hopSet.encountered++;
        hopSet.downNodes.push_back(p_downNode);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void FTUpHopHistogram::CheckRootSwitchConnections(const IBNode &node)
{
    IBDIAGNET_ENTER;

    for (phys_port_t i = 1 ; i <= node.numPorts ; i++) {
        const IBPort *p_port = node.getPort(i);
        const IBNode *p_remoteNode = GetRemoteSwitch(p_port);
        if (!p_remoteNode)
            continue;

        //root's level has rank 0
        if (!this->topology.GetNodeRank(p_remoteNode))
            this->AddIllegalLinkIssue(FTLinkIssue(&node, p_port->num,
                                                p_remoteNode, p_port->p_remotePort->num, 0));
    }

    IBDIAGNET_RETURN_VOID;
}

void FTUpHopHistogram::InitNodeToIndexConverters(const FTClassification::nodes_set &nodes)
{
    IBDIAGNET_ENTER;

    size_t index = 0;
    for (FTClassification::nodes_set::const_iterator itr = nodes.begin();
            itr != nodes.end(); itr++) {

        nodeToIndexMap[*itr] = index;
        indexToNodeMap[index] = *itr;

        index++;
    }

    IBDIAGNET_RETURN_VOID;
}

std::string FTUpHopHistogram::GetHashCode(const FTUpHopSet::bit_set &bitSet) const
{
    IBDIAGNET_ENTER;

    string hashCode = bitSet.to_string();
    if (hashCode.size() > this->bitSetMaxSize)
        hashCode = hashCode.substr(hashCode.size() - this->bitSetMaxSize -1);

    IBDIAGNET_RETURN(hashCode);
}

int FTUpHopHistogram::NodeToIndex(size_t &index, const IBNode &node)
{
    IBDIAGNET_ENTER;

    node_to_index_map::const_iterator itr = this->nodeToIndexMap.find(&node);
    if (itr == this->nodeToIndexMap.end()) {
        this->lastErrorStream << "Failed to find index for the switch ( GUID: "
                              << PTR (node.guid_get()) << " )";
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    index = itr->second;

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

const IBNode *FTUpHopHistogram::IndexToNode(size_t index)
{
    IBDIAGNET_ENTER;

    index_to_node_map::const_iterator itr = this->indexToNodeMap.find(index);
    if (itr == this->indexToNodeMap.end()) {
        this->lastErrorStream << "Failed to find IBNode associated with the index: "
                              << index;
        IBDIAGNET_RETURN(NULL);
    }

    IBDIAGNET_RETURN(itr->second);
}

int FTUpHopHistogram::CreateNeighborhoods(list_p_fabric_general_err &errors)
{
    IBDIAGNET_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Creating neighborhoods on rank: %lu\n", this->rank);

    //init encountered map for all sets
    for (up_hop_sets_map::iterator itr_cur = this->upHopSetsMap.begin();
            itr_cur != this->upHopSetsMap.end(); itr_cur++) {
        itr_cur->second.InitEncounteredMap(this->bitSetMaxSize);

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "UpHopSet: %s \n", this->UpHopSetToString(itr_cur->second).c_str());
    }

    //merge sets
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Try merging sets on rank: %lu\n", this->rank);
    for (up_hop_sets_map::iterator itr_cur = this->upHopSetsMap.begin();
            itr_cur != this->upHopSetsMap.end();) {
        const FTUpHopSet &currentSet = itr_cur->second;

        bool isMerged = false;
        int rc = this->TryMergeSet_2(currentSet, isMerged);
        if (rc)
            IBDIAGNET_RETURN(rc);

        if (isMerged) {
            up_hop_sets_map::iterator itr_del = itr_cur++;
            this->upHopSetsMap.erase(itr_del);
            continue;
        }

        itr_cur++;
    }

    //check cross links
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Checking cross links on rank: %lu\n", this->rank);
    for (up_hop_sets_map::iterator itr_cur = this->upHopSetsMap.begin();
            itr_cur != this->upHopSetsMap.end(); itr_cur++) {
        int rc = this->CheckCrossLinks(itr_cur->second);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    int rc = this->SetsToNeighborhoods(errors);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTUpHopHistogram::CheckCrossLinks(FTUpHopSet &currentSet)
{
    IBDIAGNET_ENTER;

    for (up_hop_sets_map::const_iterator itr_other = this->upHopSetsMap.begin();
            itr_other != this->upHopSetsMap.end(); itr_other++) {
        const FTUpHopSet &other = itr_other->second;

        if (&currentSet == &other)
            continue;

        const FTUpHopSet::bit_set test = other.Intersect(currentSet);
        if (!test.count())
            continue;

        for (size_t i = 0; i < this->bitSetMaxSize; i++) {
            if (!test[i])
                continue;

            int currentEnc = 0;
            int otherEnc = 0;

            int rc = currentSet.GetNodeEncountered(i, currentEnc);
            if (rc) {
                this->lastErrorStream << "Failed to get how many times a node with "
                        << "index: " << i << " was encountered. Up-Hop set ptr: " << &currentEnc;
                IBDIAGNET_RETURN(rc);
            }

            rc = other.GetNodeEncountered(i, otherEnc);
            if (rc) {
                this->lastErrorStream << "Failed to get how many times a node with "
                        << "index: " << i << " was encountered. Up-Hop set ptr: " << &other;
                IBDIAGNET_RETURN(rc);
            }

            if (currentEnc*2 < otherEnc) {
                //hope we do not need to take out anything from down nodes
                rc = currentSet.TakeOutUpNode(i);
                if (rc) {
                    this->lastErrorStream << "Failed to take out a node with "
                                        << "index: " << i << ". Up-Hop set ptr: " << &currentSet;
                    IBDIAGNET_RETURN(rc);
                }

                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Cross link between Up-Hop sets found; "
                                                "ptr1=%p and ptr2=%p on rank: %lu; "
                                                "currentEnc: %d otherEnc %d; "
                                                "index: %lu is taken out\n",
                                                &currentSet, &other, this->rank + 1,
                                                currentEnc, otherEnc, i);

                rc = this->AddIllegalLinkIssues(i, currentSet.downNodes);
                if (rc)
                    IBDIAGNET_RETURN(rc);
            }
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTUpHopHistogram::TryMergeSet_2(const FTUpHopSet &currentSet, bool &isMerged)
{
    IBDIAGNET_ENTER;

    for (up_hop_sets_map::iterator itr_other = this->upHopSetsMap.begin();
            itr_other != this->upHopSetsMap.end(); itr_other++) {
        FTUpHopSet &other = itr_other->second;

        if (&currentSet == &other)
            continue;

        int rc = this->TryMergeTwoSets(currentSet, other, isMerged);
        if (rc)
            IBDIAGNET_RETURN(rc);

        if (isMerged) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Merge Up-Hop sets ptr1=%p and ptr2=%p\n",
                                    &currentSet, &other);
            break;
        }
    }    

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTUpHopHistogram::TryMergeTwoSets(const FTUpHopSet &currentSet, FTUpHopSet &other,
                                    bool &isMerged)
{
    IBDIAGNET_ENTER;

    //delta says: how many 1 "current set" has for not being a subset of "other set"
    FTUpHopSet::bit_set delta = other.Delta(currentSet, this->bitSetMaxSize);
    if (static_cast<int>(delta.count()*100/other.upNodesBitSet.count()) <
            (100 - this->topology.GetUpHopFitPercents())) {

        other.Merge(currentSet, this->bitSetMaxSize);
        //todo take care of Missing links

        isMerged = true;
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTUpHopHistogram::AddIllegalLinkIssues(size_t index, const FTClassification::nodes_list &downNodes) {

    IBDIAGNET_ENTER;

    const IBNode* p_upNode = this->IndexToNode(index);
    if (!p_upNode)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    bool found = false;
    for (FTClassification::nodes_list::const_iterator itr_down = downNodes.begin();
            itr_down != downNodes.end(); itr_down++) {
        const IBNode *p_downNode = *itr_down;

        //roll over all ports - in case there are few wrongly connected
        for (phys_port_t i = 1 ; i <= p_downNode->numPorts ; i++) {
            const IBPort *p_port = p_downNode->getPort(i);
            const IBNode *p_remoteNode = GetRemoteSwitch(p_port);
            if (!p_remoteNode || p_upNode != p_remoteNode)
                  continue;

            this->invalidLinkIssuesVec.push_back(
                FTLinkIssue(p_remoteNode, p_port->p_remotePort->num, this->rank,//up node
                            p_downNode, p_port->num, this->rank+1//down node
                            ));
            found = true;
        }
    }

    if (!found) {
        this->lastErrorStream << "The switch GUID: " << PTR(p_upNode->guid_get())
                            << "is not connected to any switch from the down nodes"
                            << " provided for Invalid Link Issue";
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void FTUpHopHistogram::AddIllegalLinkIssue(const FTLinkIssue &issue)
{
    IBDIAGNET_ENTER;

    //avoid reports on the same pairs but in different order
    if (this->topology.IsReportedLinkIssue(issue.p_node1, issue.p_node2))
        IBDIAGNET_RETURN_VOID;

    this->invalidLinkIssuesVec.push_back(issue);

    this->topology.AddNewLinkIssue(issue.p_node1, issue.p_node2);

    IBDIAGNET_RETURN_VOID;
}

int FTUpHopHistogram::SetsToNeighborhoods(list_p_fabric_general_err &errors)
{
    IBDIAGNET_ENTER;

    static size_t neighborhoodId = 0;
    static size_t groupId = 0;

    int rc = IBDIAG_SUCCESS_CODE;
    neighborhoods_vec neighborhoods;
    for (up_hop_sets_map::const_iterator itr_hist = this->upHopSetsMap.begin();
            itr_hist != this->upHopSetsMap.end(); itr_hist++) {
        const FTUpHopSet &hopSet = itr_hist->second;

        FTClassification::nodes_list upNodes;
        rc = BitSetToNodes(hopSet.upNodesBitSet, upNodes);
        if (rc)
            IBDIAGNET_RETURN(rc);

        FTNeighborhood *p_neighborhood = new FTNeighborhood(this->topology,
                (topology.IsLastRankNeighborhood(this->rank) ? neighborhoodId++ : groupId++),
                this->rank);
        neighborhoods.push_back(p_neighborhood);

        p_neighborhood->AddToUpNodes(upNodes);
        p_neighborhood->AddToDownNodes(hopSet.downNodes);
    }

    rc= this->InvalidLinksReport(errors, neighborhoods);
    if (rc)
        IBDIAGNET_RETURN(rc);

    for (neighborhoods_vec::const_iterator itr = neighborhoods.begin();
            itr != neighborhoods.end(); itr++) {
        rc = (*itr)->MissingLinksReport(errors, this->connectedNodes);
        if (rc) {
            this->lastErrorStream << "Failed to report on missing link for"
                        << (this->topology.IsLastRankNeighborhood(this->rank) ?
                            "neighborhood: " : "connectivity group: ")
                        << (*itr)->GetId();
            IBDIAGNET_RETURN(rc);
        }
    }

    rc = topology.SetNeighborhoodsOnRank(neighborhoods, this->rank);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTUpHopHistogram::InvalidLinksReport(list_p_fabric_general_err &errors,
                        const neighborhoods_vec &neighborhoods)
{
    IBDIAGNET_ENTER;

    for (size_t i = 0; i < this->invalidLinkIssuesVec.size(); i++) {
        const FTLinkIssue &issue = this->invalidLinkIssuesVec[i];

        const FTNeighborhood *p_n1 = this->FindNeighborhood(neighborhoods, issue.p_node1);
        if (!p_n1)
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

        const FTNeighborhood *p_n2 = this->FindNeighborhood(neighborhoods, issue.p_node2);
        if (!p_n2)
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

        //the errors created below will be processed when none of "neighborhoods" objects is alive.
        //never forward to "errors" any pointer on a "neighborhood" object.

        errors.push_back(new FTInvalidLinkError(p_n1->GetId(), p_n2->GetId(), issue,
                                    this->topology.IsLastRankNeighborhood(this->rank)));
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

const FTNeighborhood *FTUpHopHistogram::FindNeighborhood(const neighborhoods_vec &neighborhoods,
                                        const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    const FTNeighborhood *p_neighborhood = NULL;
    for (size_t i = 0; i < neighborhoods.size(); i++) {
        if (neighborhoods[i]->Contains(p_node)) {
            p_neighborhood = neighborhoods[i];
            break;
        }
    }

    if (!p_neighborhood)
        this->lastErrorStream << "The provided switch ( GUID: " << PTR(p_node->guid_get())
                              << " ) is not associated with any neighborhood on the rank: "
                              << this->rank;

    IBDIAGNET_RETURN(p_neighborhood);

}

int FTUpHopHistogram::BitSetToNodes(const FTUpHopSet::bit_set &bitSet, FTClassification::nodes_list &nodes) {

    IBDIAGNET_ENTER;

    for (size_t i = 0; i < this->bitSetMaxSize; i++) {
        if (!bitSet[i])
            continue;

        const IBNode* p_node = this->IndexToNode(i);
        if (!p_node)
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

        nodes.push_back(p_node);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void FTUpHopSet::AddDownNodes(const FTUpHopSet &other)
{
    IBDIAGNET_ENTER;

    for (FTClassification::nodes_list::const_iterator itr_down = other.downNodes.begin();
            itr_down != other.downNodes.end(); itr_down++)
        this->downNodes.push_back(*itr_down);

    IBDIAGNET_RETURN_VOID;
}

FTUpHopSet::bit_set FTUpHopSet::Delta(const FTUpHopSet &other, size_t maxSize) const
{
    IBDIAGNET_ENTER;

    bit_set delta;
    for (size_t i = 0; i < maxSize; i++) {
        if (other.upNodesBitSet[i] && !this->upNodesBitSet[i])
            delta[i] = 1;
    }

    IBDIAGNET_RETURN(delta);
}

void FTUpHopSet::Merge(const FTUpHopSet &other, size_t maxSize)
{
    IBDIAGNET_ENTER;

    for (size_t i = 0; i < maxSize; i++) {
        if (!other.upNodesBitSet[i])
            continue;

        index_to_encounter_map::const_iterator itr = other.encountered_map.find(i);
        ///todo take care about search in maps
        //todo take care of end
        if (this->upNodesBitSet[i])
            this->encountered_map[i] += itr->second;
        else
            this->encountered_map[i] = itr->second;
    }
 
    this->upNodesBitSet |= other.upNodesBitSet;

    //todo take care of missing links
    this->AddDownNodes(other);

    IBDIAGNET_RETURN_VOID;
}

int FTUpHopSet::TakeOutUpNode(size_t index)
{
    IBDIAGNET_ENTER;

    this->upNodesBitSet[index] = 0;

    FTUpHopSet::index_to_encounter_map::iterator itr = this->encountered_map.find(index);
    if (itr == this->encountered_map.end())
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    this->encountered_map.erase(itr);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTUpHopSet::GetNodeEncountered(size_t index, int &encountered) const {

    IBDIAGNET_ENTER;

    index_to_encounter_map::const_iterator itr = this->encountered_map.find(index);
    if (itr == this->encountered_map.end())
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    encountered = itr->second;

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

}

void FTUpHopSet::InitEncounteredMap(size_t maxSize)
{
    IBDIAGNET_ENTER;

    for (size_t i = 0; i < maxSize; i++) {
        if (!this->upNodesBitSet[i])
            continue;

        this->encountered_map[i] = this->encountered;
    }

    IBDIAGNET_RETURN_VOID;
}

string FTUpHopHistogram::UpHopSetToString(const FTUpHopSet &upHopSet)
{
    IBDIAGNET_ENTER;

    stringstream strStream;

    strStream << endl << "FTUpHopSet (ptr) " << &upHopSet << endl;
    strStream << "initiated from: " << PTR((*(upHopSet.downNodes.begin()))->guid_get()) << endl;
    strStream << "encountered: " << upHopSet.encountered << endl;
    //strStream << "upNodesBitSet: " << this->GetHashCode(upHopSet.upNodesBitSet) << endl;
    strStream << "up_set: " ;
    for (size_t index = 0; index < this->bitSetMaxSize; index++) {
        if (!upHopSet.upNodesBitSet[index])
            continue;

        const IBNode *p_node = this->IndexToNode(index);
        if (p_node)
            strStream << PTR(p_node->guid_get()) << " ";
    }

    #ifdef DOWN_SET
        strStream << endl << "down_set: " ;
        for (FTClassification::nodes_list::const_iterator itr = upHopSet.downNodes.begin();
                itr != upHopSet.downNodes.end(); itr++) {

            const IBNode *p_node = *itr;
            if (p_node)
                strStream << PTR(p_node->guid_get()) << " ";
        }
    #endif

    IBDIAGNET_RETURN(strStream.str());
}

const FTClassification::nodes_set *FTTopology::GetNodesOnRank(size_t rank)
{
    IBDIAGNET_ENTER;

    if (rank >= this->nodesByRank.size()) {
        this->lastErrorStream << "Cannot get nodes on rank: " << rank
                              << " Total ranks in the Fat-tree: "
                              << this->nodesByRank.size();
        IBDIAGNET_RETURN(NULL);
    }

    IBDIAGNET_RETURN(&this->nodesByRank[rank]);
}

int FTTopology::SetNeighborhoodsOnRank(neighborhoods_vec &neighborhoods, size_t rank)
{
    IBDIAGNET_ENTER;

    if (rank >= this->neighborhoodsByRank.size()) {
        this->lastErrorStream << "Cannot set neighborhoods/connectivity groups on rank: " << rank
                              << " container size: "
                              << this->neighborhoodsByRank.size();
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    this->neighborhoodsByRank[rank].swap(neighborhoods);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

size_t FTTopology::GetNodeRank(const IBNode* p_node) const
{
    IBDIAGNET_ENTER;

    size_t ret = -1;
    for (size_t rank = 0; rank < this->nodesByRank.size(); rank++) {
        if (this->nodesByRank[rank].find(p_node) != this->nodesByRank[rank].end()) {
            ret = rank;
            break;
        }
    }

    IBDIAGNET_RETURN(ret);
}

LinksData FTTopology::GetSwitchLinksData(size_t rank, const IBNode &node)
{
    IBDIAGNET_ENTER;

    node_to_links_map::const_iterator itr = this->nodeToLinksMap.find(&node);
    if (itr != this->nodeToLinksMap.end())
        IBDIAGNET_RETURN(itr->second);

    LinksData linksData = this->CalculateSwitchUpDownLinks(rank, node);
    nodeToLinksMap[&node] = linksData;

    IBDIAGNET_RETURN(linksData);
}

int FTTopology::CheckUpDownLinksAndAPorts(list_p_fabric_general_err &errors)
{
    IBDIAGNET_ENTER;

    int rc = CalculateUpDownLinksMinRatio();
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->CheckNeighborhoodsLinksAndAPorts(errors);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::CheckFLIDs()
{
    IBDIAGNET_ENTER;

    if (this->neighborhoodsByRank.empty()) {
        this->lastErrorStream << "Map is empty";
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    FTNeighborhood::lid_to_neighborhood_map flidsToNeighborhoods;

    //we are taking neighborhoods from the last level -> TOR switches
    for (FTNeighborhood *p_neighborhood : *(this->neighborhoodsByRank.rbegin())) {
        if (!p_neighborhood) {
            this->lastErrorStream << "NULL Neighborhood is found";
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        int rc = p_neighborhood->CollectFLIDs(flidsToNeighborhoods);
        if (rc)
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    this->outFile << endl << endl;
    bool found = false;
    for (const auto& entry : flidsToNeighborhoods){
        if (entry.second.empty() || entry.second.size() == 1)
            continue;

        found = true;
        this->IncrementInternalErrors();

        this->outFile << WARNING_PREFIX << "FLID=" << entry.first
                    << " appears in the following neighborhoods: ";
        bool firstTime = true;
        for (size_t id : entry.second) {
            if (firstTime) {
                this->outFile << id;
                firstTime = false;
            } else
                this->outFile << ", " << id;
        }
        this->outFile << endl;
    }

    if (found)
       this->outFile << INFO_PREFIX << "Additional information related to FLIDs "
                     << "could be found in the ibdiagnet2.flid file "
                     << "generated by the Router stage"
                     << endl;

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

//todo consider make it with neighborhoods
int FTTopology::CalculateUpDownLinksMinRatio()
{
    IBDIAGNET_ENTER;

    if (this->nodesByRank.size() < RANKS_2L_FT) {
        this->lastErrorStream << "Cannot calculate Up/Down links. It is not a Fat-Tree topology";
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    for (size_t rank  = 0; rank < this->nodesByRank.size(); rank++) {
        links_to_nodes_map linksToNodesMap;

        for (FTClassification::nodes_set::const_iterator itr =  this->nodesByRank[rank].begin();
                itr != this->nodesByRank[rank].end(); itr++) {
            const IBNode *p_node = *itr;
            if (!p_node) {
                this->lastErrorStream << "Cannot calculate Up/Down links. One of IB-Nodes is NULL";
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            LinksData linksData = this->GetSwitchLinksData(rank, *p_node);
            //calculate minimal ratio.
            if (linksData.first && linksData.second) {
                double currentRatio =
                                GetSpeed(*p_node)*(double)linksData.first/(double)linksData.second;

                if (!this->minimalRatio)
                    this->minimalRatio = currentRatio;
                else if (currentRatio < this->minimalRatio)
                    this->minimalRatio = currentRatio;
              }
         }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::CheckNeighborhoodsLinksAndAPorts(list_p_fabric_general_err &errors)
{
    IBDIAGNET_ENTER;

    if (this->neighborhoodsByRank.empty()) {
        this->lastErrorStream << "Cannot calculate Up/Down links. Map is empty";
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
    }

    for (size_t rank = 0; rank < this->neighborhoodsByRank.size(); rank++) {
        for (size_t i = 0; i < this->neighborhoodsByRank[rank].size(); i++) {
            FTNeighborhood  *p_neighborhood = this->neighborhoodsByRank[rank][i];
            if (!p_neighborhood) {
                this->lastErrorStream << "Cannot calculate Up/Down links. One of FTNeighborhoods is NULL";
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            int rc = p_neighborhood->CheckUpDownLinks(errors, this->outFile);
            if (rc) {
                this->lastErrorStream << p_neighborhood->GetLastError();
                IBDIAGNET_RETURN(rc);
            }

            rc = p_neighborhood->CheckAsymmetricAPorts(this->outFile);
            if (rc) {
                this->lastErrorStream << p_neighborhood->GetLastError();
                IBDIAGNET_RETURN(rc);
            }
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

LinksData FTTopology::CalculateSwitchUpDownLinks(size_t rank, const IBNode& node)
{
    IBDIAGNET_ENTER;

    LinksData linksData(0, 0);
    FTClassification::nodes_set seenCA;

    for (phys_port_t i = 1 ; i <= node.numPorts; i++) {
        const IBPort *p_port = node.getPort(i);
        if (!p_port || p_port->isFNMPort() || !p_port->isSymmetricLink())
            continue;

        const IBNode *p_remote = p_port->get_remote_node();
        if (!p_remote)
            continue;

        if (rank == 0) {//root
            if (p_remote->type != IB_SW_NODE)
                continue;

            FTClassification::nodes_set::const_iterator itr_down =
                                            this->nodesByRank[1].find(p_remote);
            if (itr_down != this->nodesByRank[1].end())
                linksData.second++;
        } else if (rank == this->nodesByRank.size() - 1) {//leaf
            if (p_remote->type == IB_CA_NODE && !p_remote->isSpecialNode()) {
                if (node.isPrismaSwitch()) {
                    const IBPort *p_port = node.getPort(i);
                    if (p_port && p_port->p_remotePort &&
                            p_port->p_remotePort->p_port_hierarchy_info &&
                            p_port->p_remotePort->p_port_hierarchy_info->m_num_of_planes > 0) {
                        std::pair<FTClassification::nodes_set::const_iterator, bool> itr_insert =
                              seenCA.insert(p_remote);
                        if (itr_insert.second)
                            linksData.second++;
                    }
                    else
                       linksData.second++;
                }
                else
                    linksData.second++;
            }
            else if (p_remote->type == IB_SW_NODE) {
                FTClassification::nodes_set::const_iterator itr_up =
                                                this->nodesByRank[rank - 1].find(p_remote);
                if (itr_up != this->nodesByRank[rank - 1].end())
                    linksData.first++;
            }
        } else {//spine
            if (p_remote->type != IB_SW_NODE)
                continue;

            FTClassification::nodes_set::const_iterator itr_up =
                                            this->nodesByRank[rank - 1].find(p_remote);
            if (itr_up != this->nodesByRank[rank - 1].end())
                linksData.first++;

            FTClassification::nodes_set::const_iterator itr_down =
                                            this->nodesByRank[rank + 1].find(p_remote);
            if (itr_down != this->nodesByRank[rank + 1].end())
                linksData.second++;
        }
    }

    IBDIAGNET_RETURN(linksData);
}

int FTTopology::Dump()
{
    IBDIAGNET_ENTER;

    this->outFile << endl << endl;

    int rc = this->DumpNodesToStream();
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->DumpNeighborhoodsToStream();
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::DumpNodesToStream()
{
    IBDIAGNET_ENTER;

    for (size_t i = 0; i < this->nodesByRank.size(); i++) {
        const char *p_name = " ";
        if (i == 0)
            p_name = " (Roots)";
        else if (i == this->nodesByRank.size() - 1)
            p_name = " (Leaves)";

        this->outFile << endl
               << "rank: " << i << p_name
               << "size: " << this->nodesByRank[i].size()
               << endl;

        for (FTClassification::nodes_set::const_iterator itr = this->nodesByRank[i].begin();
                itr != this->nodesByRank[i].end(); itr++) {
            const IBNode* p_node = *itr;
            if (!p_node) {
                ERR_PRINT("One of IBNodes is NULL. Cannot dump it\n");
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            this->outFile << '\t' << GetNodeRecord(p_node)
                          << endl;
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTTopology::DumpNeighborhoodsToStream()
{
    IBDIAGNET_ENTER;

    this->outFile << "\n\n\n";
    for (size_t rank = 0; rank < this->neighborhoodsByRank.size(); rank++) {
        this->outFile << "on ranks (" << rank << ", " << rank + 1
               <<  ") -- "
               << (this->IsLastRankNeighborhood(rank) ? "neighborhoods: " : "connectivity groups: ")
               << this->neighborhoodsByRank[rank].size()
               << endl;

        for (size_t i = 0; i < this->neighborhoodsByRank[rank].size(); i++) {
            const FTNeighborhood *p_neighborhood = this->neighborhoodsByRank[rank][i];
            if (!p_neighborhood) {
                ERR_PRINT("One of FTNeighborhoods is NULL. Cannot dump it\n");
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            int rc = p_neighborhood->DumpToStream(this->outFile);
            if (rc)
                IBDIAGNET_RETURN(rc);
        }

        this->outFile << endl;
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

bool FTTopology::IsReportedLinkIssue(const IBNode* p_node1, const IBNode* p_node2) const
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(this->reportedLinksIssues.Contains(p_node1, p_node2));
}

void FTTopology::AddNewLinkIssue(const IBNode* p_node1, const IBNode* p_node2)
{
    IBDIAGNET_ENTER;

    this->reportedLinksIssues.Add(p_node1, p_node2);

    IBDIAGNET_RETURN_VOID;
}

bool FTTopology::IsLastRankNeighborhood(size_t rank) const
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(rank + 1 == this->neighborhoodsByRank.size());
}

string FTTopology::LevelsReport() const
{
    IBDIAGNET_ENTER;

    stringstream strStream;
    strStream << this->nodesByRank.size() << " level Fat-Tree was discovered: " << endl;
    for (size_t i = 0; i < nodesByRank.size(); i++)
        strStream << "\t rank: " << i << (!i ? "(Roots)" : "")
                << " #switches: " << nodesByRank[i].size() << endl;

    IBDIAGNET_RETURN(strStream.str());
}

const FTClassification::NodeData *FTTopology::GetClassificationNodeData(const IBNode *p_node) const
{
    IBDIAGNET_ENTER;

    node_to_classification_data_map::const_iterator itr = this->nodeDataMap.find(p_node);
    if (itr == this->nodeDataMap.end())
        IBDIAGNET_RETURN(NULL);

    IBDIAGNET_RETURN(&itr->second);
}

int FTTopology::GetPlanesNumber() const
{
    IBDIAGNET_ENTER;

    const FTClassification::nodes_set &roots = this->nodesByRank[0];
    if (roots.empty())
        return 0;

    const IBNode *p_switch = *roots.begin();
    if (!p_switch)
        IBDIAGNET_RETURN(0);

    IBDIAGNET_RETURN(p_switch->getPlanesNumber());
}

int FTNeighborhood::CollectFLIDs(lid_to_neighborhood_map &flidsToNeighborhoods)
{
    IBDIAG_ENTER

    for (const IBNode *p_switch : this->down) {
        if (!p_switch) {
            this->lastErrorStream << "NULL IB Node is found in "
                                << "the neighborhood/connectivity group id="
                                << this->id;
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        for (lid_t flid : p_switch->flids)
            flidsToNeighborhoods[flid].insert(this->id);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTNeighborhood::MissingLinksReport(list_p_fabric_general_err &errors,
                                    const PairsContainer<const IBNode*> &connectedNodes)
{
    IBDIAGNET_ENTER;

    PairsContainer<const IBNode*> checkedPairs;

    for (FTClassification::nodes_set::const_iterator itr_up = this->up.begin();
            itr_up != this->up.end(); itr_up++) {
        const IBNode *p_node1 =  *itr_up;
        if (!p_node1)
            IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

        for (FTClassification::nodes_set::const_iterator itr_down = this->down.begin();
                itr_down != this->down.end(); itr_down++) {

            const IBNode *p_node2 =  *itr_down;
            if (!p_node2)
                IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

            if (p_node1 == p_node2 || checkedPairs.Contains(p_node1, p_node2))
                  continue;

            checkedPairs.Add(p_node1, p_node2);

            if (connectedNodes.Contains(p_node1, p_node2))
                continue;

            errors.push_back(new FTMissingLinkError(this->GetId(),
                                            FTLinkIssue(p_node1, p_node2),
                                            this->topology.IsLastRankNeighborhood(this->rank)));
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTNeighborhood::DumpToStream(ostream &stream) const
{
    IBDIAGNET_ENTER;

    stream << '\t'
           << (this->topology.IsLastRankNeighborhood(this->rank) ?
                    "neighborhood: " : "connectivity group: ")
           << this->id << endl;

    if (this->rank) {//is not on roots
        stream << "\t\t" << "total spine uplinks: " << this->totalUpLinks << endl
               << "\t\t" << "total internal links: " << this->totalInternalLinks << endl;
    }

    int rc = this->DumpNodesToStream(stream, this->up, "spines");
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->DumpNodesToStream(stream, this->down, "lines");
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void FTNeighborhood::AddNodes(const FTClassification::nodes_list &nodes, bool isUp)
{
    IBDIAGNET_ENTER;

    for (FTClassification::nodes_list::const_iterator itr = nodes.begin();
            itr != nodes.end(); itr++) {
        if (isUp)
            this->up.insert(*itr);
        else
            this->down.insert(*itr);
    }

    IBDIAGNET_RETURN_VOID;
}

int FTNeighborhood::DumpNodesToStream(ostream &stream,
                                const FTClassification::nodes_set &nodes, const char *p_name) const
{
    IBDIAGNET_ENTER

    stream << "\t\t"
           << p_name << ": " << nodes.size() << " switches" << endl;

    for (FTClassification::nodes_set::const_iterator itr = nodes.begin(); itr != nodes.end();
            itr++) {
        const IBNode *p_node = *itr;
        if (!p_node) {
            //todo think of error
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        stream << "\t\t\t" << GetNodeRecord(p_node) << endl;
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


int FTNeighborhood::CheckAsymmetricAPorts(ostream &outFile)
{
    IBDIAGNET_ENTER;

    //a neighborhood/(con. group) consists of UP and DOWN nodes sets.
    //for each neighborhood/(con group) we check APort related to a physical link
    //connecting two nodes from these sets

    int rc = this->CheckInternalAPorts(outFile);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (this->topology.IsLastRankNeighborhood(rank)) {
        rc = this->CheckExternalAPorts(outFile);
        if (rc)
           IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


int FTNeighborhood::CheckUpDownLinks(list_p_fabric_general_err &errors, ostream &outFile)
{
    IBDIAGNET_ENTER;

    outFile << endl;

    int rc = 0;
    if (this->rank) {//no reason to check roots uplinks
        rc = this->CheckBlockingConfiguration(errors, outFile);
        if (rc)
            IBDIAGNET_RETURN(rc);

        rc = this->CheckSetLinks(this->up, this->rank, true, errors, outFile);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    rc = this->CheckSetLinks(this->up, this->rank, false, errors, outFile);
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->CheckSetLinks(this->down, this->rank + 1, true, errors, outFile);
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->CheckSetLinks(this->down, this->rank + 1, false, errors, outFile);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTNeighborhood::CheckBlockingConfiguration(list_p_fabric_general_err &errors, ostream &outFile)
{
    IBDIAGNET_ENTER;

    for (FTClassification::nodes_set::const_iterator itr = this->up.begin();
            itr != this->up.end(); itr++) {

        const IBNode *p_node = *itr;
        if (!p_node) {
            this->lastErrorStream << "Cannot calculate Up/Down links for the "
                                << (this->topology.IsLastRankNeighborhood(this->rank) ?
                                    "neighborhood: " : "connectivity group: ")
                                << this->id
                                << ". One of its IB-Nodes is NULL";
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        LinksData currentData =
                    this->topology.GetSwitchLinksData(this->rank, *p_node);
        this->totalUpLinks += currentData.first;
        this->totalInternalLinks += currentData.second;
    }

    if (this->totalUpLinks >= this->totalInternalLinks)
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

    outFile << WARNING_PREFIX
            << (this->topology.IsLastRankNeighborhood(this->rank) ?
                                        "Neighborhood " : "Connectivity group ")
            << this->id
            << ": suspected blocking configuration "
            << " -- " << " total number of spine's uplinks " << this->totalUpLinks
            << " is less then total number of internal links " << this->totalInternalLinks
            << endl;

    this->topology.IncrementInternalErrors();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTNeighborhood::CheckInternalAPorts(ostream &outFile)
{
    IBDIAGNET_ENTER;

    int  warnings = 0;
    for (const IBNode *p_node : this->up) {
        if (!p_node) {
            this->lastErrorStream << "Cannot check APorts internal links symmetry for the "
                                << (this->topology.IsLastRankNeighborhood(this->rank) ?
                                    "neighborhood: " : "connectivity group: ")
                                << this->id
                                << ". One of its IB-Nodes is NULL";
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        for (phys_port_t i = 1 ; i <= p_node->numPorts ; i++) {
            const IBPort *p_port = p_node->getPort(i);
            if (!p_port || !p_port->is_data_worthy() || p_port->isFNMPort() ||
                    !p_port->p_remotePort || !p_port->p_remotePort->p_node)
                continue;

            if (this->down.find (p_port->p_remotePort->p_node) == this->down.end())
                continue;

            //p_port and p_port->p_remotePort do belong to the neighborhood/(con. group)
            if (p_port->p_aport && !p_port->p_aport->isSymmetric()) {
                warnings++;
                this->topology.IncrementInternalErrors();
            }

            if (p_port->p_remotePort->p_aport && !p_port->p_remotePort->p_aport->isSymmetric()) {
                warnings++;
                this->topology.IncrementInternalErrors();
            }
        }
    }

    const string name = (this->topology.IsLastRankNeighborhood(this->rank) ?
                    "Neighborhood " : "Connectivity group ");
    if (warnings)
        outFile << WARNING_PREFIX << (this->topology.IsLastRankNeighborhood(this->rank) ?
                    "Neighborhood " : "Connectivity group ")
                << this->id
                << ": found " << warnings
                << " asymmetric APort"  << (warnings == 1 ? "" : "s")
                << endl;


    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


int FTNeighborhood::CheckExternalAPorts(ostream & outFile)
{
    IBDIAGNET_ENTER;

    int  warnings = 0;
    for (const IBNode *p_node : this->down) {
        if (!p_node) {
            this->lastErrorStream << "Cannot check APorts external links symmetry for the "
                                  << "connectivity group: " << this->id
                                  << ". One of its IB-Nodes is NULL";
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        set<const APort*> aports;
        for (phys_port_t i = 1 ; i <= p_node->numPorts ; i++) {
            //check only ports that are connected to HCA
            const IBPort *p_port = p_node->getPort(i);
            if (!p_port || !p_port->is_data_worthy() || p_port->isFNMPort() ||
                    !p_port->p_remotePort || !p_port->p_remotePort->p_node ||
                    p_port->p_remotePort->p_node->type != IB_CA_NODE || !p_port->p_aport)
                continue;

            if (p_port->p_aport->isSymmetric() || aports.find(p_port->p_aport) != aports.end())
                continue;

            //XDR CA is connected to a Crocodile switch by all its ports.
            //We would like to avoid checking the same APort a few times.
            aports.insert(p_port->p_aport);

            warnings++;
            this->topology.IncrementInternalErrors();
        }
    }

    if (warnings)
        outFile << WARNING_PREFIX << "Neighborhood " << this->id
                << ": found " << warnings
                << " asymmetric APort"  << (warnings == 1 ? "" : "s")
                << " connected to CAs"
                << endl;

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int FTNeighborhood::CheckSetLinks(const FTClassification::nodes_set &nodes, size_t nodesRank, bool uplinks,
                                list_p_fabric_general_err &errors, ostream &outFile)
{
    IBDIAGNET_ENTER;

    link_to_nodes_map linkToNodesMap;
    for (FTClassification::nodes_set::const_iterator itr = nodes.begin();
            itr != nodes.end(); itr++) {
        const IBNode *p_node = *itr;
        if (!p_node) {
            this->lastErrorStream << "Cannot calculate Up/Down links for the "
                                << (this->topology.IsLastRankNeighborhood(this->rank) ?
                                    "neighborhood: " : "connectivity group: ")
                                << this->id
                                << ". One of its IB-Nodes is NULL";
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        LinksData currentData =
                    this->topology.GetSwitchLinksData(nodesRank, *p_node);
        if (uplinks)
            linkToNodesMap[currentData.first].push_back(p_node->guid_get());
        else
            linkToNodesMap[currentData.second].push_back(p_node->guid_get());
    }

    this->SetLinksReport(outFile, linkToNodesMap, nodesRank, uplinks);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void FTNeighborhood::SetLinksReport(ostream &outFile, const link_to_nodes_map &linksToNodesMap,
                                    size_t nodesRank, bool uplinks) const
{
    IBDIAGNET_ENTER;

    const string name = (this->topology.IsLastRankNeighborhood(this->rank) ?
                    "Neighborhood " : "Connectivity group ");
    const string switchType = (nodesRank == this->rank) ? "spines" : "lines";
    const string linkType =  (uplinks ? "uplinks" : "downlinks");

    if (linksToNodesMap.size() == 1) {
        outFile << INFO_PREFIX << name << this->id
                << ": all " << switchType
                << " have the same number of " << linkType
                << ": " << linksToNodesMap.begin()->first << endl;
        IBDIAGNET_RETURN_VOID;
    }

    if (linksToNodesMap.rbegin() == linksToNodesMap.rend())
        IBDIAGNET_RETURN_VOID;

    outFile << (this->IsWarning(nodesRank, uplinks) ? WARNING_PREFIX : ERROR_PREFIX)
            << name << this->id
            << ": " << switchType << " with different number of "<< linkType
            << " (expected  " << linksToNodesMap.rbegin()->first << ' ' << linkType << ')';

    this->ReportToStream(outFile, linksToNodesMap, 7, linkType);
    outFile << endl;

    this->topology.IncrementInternalErrors(linksToNodesMap.size());

    IBDIAGNET_RETURN_VOID;
}

void FTNeighborhood::ReportToStream(ostream &stream, const link_to_nodes_map &map,
                                    size_t maxInLine, const string &linkType) const
{
    IBDIAGNET_ENTER

    for (link_to_nodes_map::const_reverse_iterator itr_m = ++map.rbegin();
            itr_m != map.rend(); itr_m++) {
        stream << endl
               << "\t " << itr_m->first << ' ' << linkType << ": ";

        size_t inLineCount = 0;
        for (std::vector<uint64_t>::const_iterator itr_v = itr_m->second.begin();
                itr_v != itr_m->second.end(); itr_v++) {

            const IBNode *p_node = this->topology.fabric.getNodeByGuid(*itr_v);
            if (!p_node)
                continue;

            stream << GetNodeRecord(p_node) << ' ';

            if (inLineCount++ >= maxInLine) {
                stream << endl << "\t\t";
                inLineCount = 0;
            }
        }
    }

    IBDIAGNET_RETURN_VOID;
}

bool FTNeighborhood::IsWarning(size_t nodesRank, bool uplinks) const
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN((this->topology.IsLastRankNeighborhood(this->rank) &&
                      !uplinks                                          &&
                      nodesRank == this->rank + 1))
}

FTInvalidLinkError::FTInvalidLinkError(size_t id_1, size_t id_2,
                                    const FTLinkIssue &issue, bool isNeighborhood)
{
    IBDIAGNET_ENTER

    //if two root switches are connected - it is taken as a warning
    if (issue.rank1 == issue.rank2 && !issue.rank1)
        level = EN_FABRIC_ERR_WARNING;

    stringstream strStream;
    if (id_1 == id_2)
        strStream << (isNeighborhood ? "Neighborhood " : "Connectivity group ")
                  << id_1
                  << " : invalid link between switches ("
                  << GetNodeRecord(issue.p_node1)
                  << " port: " << +issue.port1 << ") and ("
                  << GetNodeRecord(issue.p_node2)
                  << " port: " << +issue.port2 << ')' ;
    else
        strStream << "Invalid link between "
                  << (isNeighborhood ? "neighborhood " : "connectivity group ")
                  << id_1
                  << " (" << GetNodeRecord(issue.p_node1)
                  << " port: " << +issue.port1 << ") and "
                  << (isNeighborhood ? "neighborhood " : "group ")
                  << id_2 << " (" << GetNodeRecord(issue.p_node2)
                  << " port: " << +issue.port2 << ')' ;
    this->description = strStream.str();

    IBDIAGNET_RETURN_VOID;
}

string FTInvalidLinkError::GetErrorLine()
{
    IBDIAGNET_ENTER
    IBDIAGNET_RETURN(this->description)
}

string FTInvalidLinkError::GetCSVErrorLine()
{
    IBDIAG_ENTER;

    stringstream ss;
    ss << SCOPE_CLUSTER       << ','
       << "0x0,0x0,0x0"       << ','
       << "INVALID_LINK"      << ','
       << '\"' << DescToCsvDesc(this->description) << '\"';

    IBDIAGNET_RETURN(ss.str());
}

FTMissingLinkError::FTMissingLinkError(size_t id, const FTLinkIssue &issue, bool isNeighborhood)
{
    IBDIAGNET_ENTER

    stringstream strStream;
    strStream << (isNeighborhood ? "Neighborhood " : "Connectivity group ")
              << id
              << ": missing link between switches ("
              << GetNodeRecord(issue.p_node1)
              << ") and ("
              << GetNodeRecord(issue.p_node2)
              << ')' ;

    this->description = strStream.str();

    IBDIAGNET_RETURN_VOID;
}

string FTMissingLinkError::GetErrorLine()
{
    IBDIAGNET_ENTER
    IBDIAGNET_RETURN(this->description);
}

string FTMissingLinkError::GetCSVErrorLine()
{
    IBDIAG_ENTER;

    stringstream ss;
    ss << SCOPE_CLUSTER       << ','
       << "0x0,0x0,0x0"       << ','
       << "MISSING_LINK"      << ','
       << '\"' << DescToCsvDesc(this->description) << '\"';

    IBDIAGNET_RETURN(ss.str());
}
