/*
 * Copyright (c) 2023-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


/*
 * All Functions in this file are for ibdiagnet2.iblinkinfo file only
 */

#include <infiniband/ibdiag/ibdiag.h>
#include "infiniband/ibutils/output_stream_primitives.h"

string nodeTypeToStr(IBNodeType type)
{
    switch (type) {
        case IB_SW_NODE:    return("Switch");
        case IB_CA_NODE:    return("CA");
        case IB_RTR_NODE:   return("Router");
        default:            return("??");
    }
}

string widthToStr(IBLinkWidth width)
{
    stringstream sstr;

    switch (width) {
        case IB_LINK_WIDTH_1X:
            sstr << "1X";
            break;
        case IB_LINK_WIDTH_4X:
            sstr << "4X";
            break;
        case IB_LINK_WIDTH_8X:
            sstr << "8X";
            break;
        case IB_LINK_WIDTH_12X:
            sstr << "12X";
            break;
        case IB_LINK_WIDTH_2X:
            sstr << "2X";
            break;
        default:
            sstr << "undefined (" << width << ")";
            break;
    }

    return sstr.str();
}

string speedToStr(IBLinkSpeed speed)
{
    stringstream sstr;

    switch (speed) {
        case 0:
            sstr << "Extended Speed";
            break;
        case IB_LINK_SPEED_2_5:
            sstr << "2.5 Gbps";
            break;
        case IB_LINK_SPEED_5:
            sstr << "5.0 Gbps";
            break;
        case IB_LINK_SPEED_10:
            sstr << "10.0 Gbps";
            break;
        case IB_LINK_SPEED_14:
            sstr << "14.0625 Gbps";
            break;
        case IB_LINK_SPEED_25:
            sstr << "25.78125 Gbps";
            break;
        case IB_LINK_SPEED_50:
            sstr << "53.125 Gbps";
            break;
        case IB_LINK_SPEED_100:
            sstr << "106.25 Gbps";
            break;
        case IB_LINK_SPEED_200:
            sstr << "212.5 Gbps";
            break;
        case IB_LINK_SPEED_FDR_10:
            sstr << "10.0 Gbps (FDR10)";
            break;
        default:
            sstr << "undefined (" << speed << ")";
            break;
    }

    return sstr.str();
}

string portStateToStr(IBPortState state)
{
    stringstream sstr;

    switch (state) {
        case IB_UNKNOWN_PORT_STATE:
            sstr << "NoChange";
            break;
        case IB_PORT_STATE_DOWN:
            sstr << "Down";
            break;
        case IB_PORT_STATE_INIT:
            sstr << "Initialize";
            break;
        case IB_PORT_STATE_ARM:
            sstr << "Armed";
            break;
        case IB_PORT_STATE_ACTIVE:
            sstr << "Active";
            break;
        default:
            sstr << "?(" << state << ")";
            break;
    }

    return sstr.str();
}

string physPortStateToStr(u_int8_t state)
{
    stringstream sstr;

    switch (state) {
        case 0:
            sstr << "NoChange";
            break;
        case 1:
            sstr << "Sleep";
            break;
        case 2:
            sstr << "Polling";
            break;
        case 3:
            sstr << "Disabled";
            break;
        case 4:
            sstr << "PortConfigurationTraining";
            break;
        case 5:
            sstr << "LinkUp";
            break;
        case 6:
            sstr << "LinkErrorRecovery";
            break;
        case 7:
            sstr << "PhyTest";
            break;
        default:
            sstr << "?(" << +state << ")";
            break;
    }

    return sstr.str();
}

void DumpDownPortIBLinkInfo(IBPort* p_port, u_int8_t PortPhyState, ofstream& sout)
{
    sout << setfill(' ')
         << setw(6)  << p_port->base_lid
         << setw(25) << p_port->numAsString()
         << "[  ] ==(             "
         << setw(7)  << portStateToStr(p_port->get_internal_state()) << "/"
         << setw(8)  << physPortStateToStr(PortPhyState)
         << ")==>             [  ] \"\" ( )";
}

void DumpPortIBLinkInfo(IBPort* p_port, u_int8_t PortPhyState, ofstream& sout)
{
    sout << setfill(' ')
         << setw(6)  << p_port->base_lid
         << setw(25) << p_port->numAsString()
         << "[  ] ==("
         << setw(3)  << widthToStr(p_port->get_internal_width())
         << setw(19) << speedToStr(p_port->get_internal_speed())
         << setw(7)  << portStateToStr(p_port->get_internal_state()) << "/"
         << setw(8)  << physPortStateToStr(PortPhyState)
         << ")";
}

void DumpRemotePortIBLinkInfo(IBPort* p_port, ofstream& sout)
{
    sout << setfill(' ')
         << setw(6)   << p_port->base_lid
         << setw(25)  << p_port->numAsString()
         << "[  ] \"" << p_port->p_node->description  << "\" (";

    if (p_port->expected_width != IB_UNKNOWN_LINK_WIDTH)
        sout << "Could be " << widthToStr(p_port->expected_width);

    sout << " ";

    if (p_port->expected_speed != IB_UNKNOWN_LINK_SPEED)
        sout << "Could be " << speedToStr(p_port->expected_speed);

    sout << ")";
}

int DumpIBLinkInfoNode(ofstream& sout, IBNode* p_curr_node, IBDMExtendedInfo& fabric_extended_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    sout << nodeTypeToStr(p_curr_node->type) << ": ";

    if (p_curr_node->type == IB_SW_NODE)
        sout << PTR(p_curr_node->guid_get()) << " ";

    sout << p_curr_node->description << ":" << endl;

    for (phys_port_t i = 1; i <= p_curr_node->numPorts; ++i) {

        IBPort* p_curr_port = p_curr_node->getPort(i);
        if (!p_curr_port || !p_curr_port->getInSubFabric())
            continue;

        // if not a switch - skip not active ports
        if (p_curr_node->type != IB_SW_NODE && !p_curr_port->is_data_worthy())
            continue;

        sout << "      ";

        if (p_curr_node->type != IB_SW_NODE)
            sout << PTR(p_curr_port->guid_get()) << " ";

        SMP_PortInfo *p_curr_port_info =
            fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);

        u_int8_t PortPhyState = p_curr_port_info? p_curr_port_info->PortPhyState : 0;

        if (p_curr_port->p_remotePort) {
            DumpPortIBLinkInfo(p_curr_port, PortPhyState, sout);
            sout << "==>  ";
            DumpRemotePortIBLinkInfo(p_curr_port->p_remotePort, sout);
        } else {
            DumpDownPortIBLinkInfo(p_curr_port, PortPhyState, sout);
        }

        sout << endl;
    } // ports

    IBDIAG_RETURN(rc);
}
