/*
 * Copyright (c) 2024-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#pragma once

#include <ostream>
#include <vector>
#include <set>
#include <map>

class IBNode;
class IBDiag;
struct SMP_PortInfo;
struct SMP_MlnxExtPortInfo;
struct SMP_PortInfoExtended;

class SimInfoData {
    public:
        static const std::vector<const char*>  SMPCapMaskBits;
        static const std::vector<const char*>  GMPCapMaskBits;
        static const std::vector<const char*>  PortInfoCapMaskBits;
        static const std::vector<const char*>  PortInfoCapMask2Bits;
};

class SimInfoDumpPy {
    public:
        IBDiag & m_ibdiag;

    public:
        SimInfoDumpPy(IBDiag & ibdiag)
            : m_ibdiag(ibdiag)
        {}

        int Dump(const std::string & filename);

    private:
        void DumpPortInfoAllCapMask(std::ostream & sout, IBNode & node);
        void DumpPortInfoCapMask(std::ostream & sout, phys_port_t index, SMP_PortInfo *p_info, bool is_cap_mask2);

    private:
        void DumpSMPCapMask(std::ostream & sout, IBNode & node);
        void DumpGMPCapMask(std::ostream & sout, IBNode & node);
};

class SimInfoDumpCPP {
public:
    SimInfoDumpCPP(IBDiag &ibdiag):
        m_ibdiag(ibdiag) {}

    int Generate(const std::string &class_name, const std::set< uint64_t > &simulator_guids);

private:
    typedef std::map < u_int8_t, const SMP_MlnxExtPortInfo* > ext_port_info_map;
    typedef std::map < u_int8_t, const SMP_PortInfo* > port_info_map;

    std::string GetDumpedNodeType(const IBNode& node) const;
    void PrintCapabilityMask(int offset, std::ostream &sout,
                             u_int32_t cap_mask, bool is_cap_mask2) const;
    void PrintCopyright(std::ostream &sout) const;
    void PrintIncludes(std::ostream &sout, const std::string &class_name) const;

    void GenerateNodeInfo(std::ostream &sout, const IBNode &node) const;
    void GeneratePortInfo(std::ostream &sout, const IBNode &node) const;
    void GeneratePortInfoCA(std::ostream &sout, const SMP_PortInfo *p_source) const;
    void GeneratePortInfoSW(std::ostream &sout, const SMP_PortInfo *p_zero,
                            const SMP_PortInfo *p_info,
                            const port_info_map &fnm_ports) const;
    void GeneratePortInfoSW(std::ostream &sout, const SMP_PortInfo *p_source,
                          int offset, bool print_cap = false) const;
    void GenerateSMPCapabilityMask(std::ostream &sout, const IBNode &node) const;
    void GenerateGMPCapabilityMask(std::ostream &sout, const IBNode &node) const;
    void GenerateHWInfo(std::ostream &sout, const IBNode &node) const;
    void GenerateFWInfo(std::ostream &sout, const IBNode &node) const;
    void GenerateSWInfo(std::ostream &sout, const IBNode &node) const;
    void GenerateExtendedNodeInfo(std::ostream &sout, const IBNode &node) const;
    void GenerateExtendedSwitchInfo(std::ostream &sout, const IBNode &node) const;
    void GenerateExtendedPortInfo(std::ostream &sout, const IBNode &node) const;
    void GenerateExtendedPortInfo(std::ostream &sout, const SMP_MlnxExtPortInfo *p_info,
                                  const ext_port_info_map &fnm_ports) const;
    void GenerateExtendedPortInfo(std::ostream &sout, int offset,
                                  const SMP_MlnxExtPortInfo *p_source) const;
    void GenerateARInfo(std::ostream &sout, const IBNode &node) const;
    void GeneratePortInfoExtended(std::ostream &sout, const IBNode &node) const;
    void GeneratePortInfoExtended(std::ostream &sout, const SMP_PortInfoExtended *p_source) const;
    void GenerateSwitchInfo(std::ostream &sout, const IBNode &node) const;

    IBDiag &m_ibdiag;
};
