 /*
 * Copyright (c) 2021-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <iomanip>
#include <string>

#include "uphy_json_loader.h"

namespace UPHY
{
    JsonLoader::JsonLoader(const std::string & filename)
        : m_filename(filename), m_fd(fopen(m_filename.c_str(), "r")), m_dataset(nullptr)
    {
        if( m_fd == nullptr )
        {
            m_errors << "UPHY JSON loader cannot open file '" << m_filename << "'" << UPHY_JSON_REASON_PREFIX << strerror(errno) << std::endl;
            return;
        }

        try
        {
            nlohmann::json json = nlohmann::json::parse(m_fd);
            read_json(json);
            DB::instance().add(m_dataset);
        }
        catch( std::exception & e )
        {
            m_errors << "UPHY JSON load '" << m_filename << "' failure\n\t--> " << e.what() << std::endl;
            delete m_dataset;
            m_dataset = nullptr;
        }
        catch( ... )
        {
            m_errors << "UPHY JSON load '" << m_filename << "' failure, unhandled exception." << std::endl;
            delete m_dataset;
            m_dataset = nullptr;
        }
    }

    JsonLoader::~JsonLoader()
    {
        if( m_fd != nullptr )
        {
            fclose( m_fd );
            m_fd = nullptr;
        }
    }

    void JsonLoader::read_json(const nlohmann::json & json)
    {
        read_header(json);
        read_enums(json);
        read_registers(json);
    }

    void JsonLoader::read_header(const nlohmann::json & json)
    {
        try {
            dataset_type_t          type = DataSetType_None;
            Version::ver_elem_t     major;
            Version::ver_elem_t     minor;
            std::string             name;
            std::string             release;

            read(json, "version_major", major);
            read(json, "version_minor", minor);
            read(json, "name", name);
            read(json, "release", release);

            if( name == "NV_UPHY_DLM" )
                type = DataSetType_DLN;
            else
                if( name == "NV_UPHY_CLM" )
                    type = DataSetType_CLN;
                else
                    throw std::out_of_range("Wrong dataset type='" + name + "'.");

            if( DB::instance().get(type, Version(major, minor)) )
                throw std::runtime_error( "DataSet '" + name + "' already exist, version='" + std::to_string(major) + "." + std::to_string(minor) + "'");

            m_dataset = new DataSet(m_filename, type, name, Version(major, minor), release);
        }
        catch ( std::exception & e ) {
            throw std::runtime_error( std::string() + "Failed to read JSON header" + UPHY_JSON_REASON_PREFIX + e.what() );
        }
    }

    void JsonLoader::read_enums(const nlohmann::json & json)
    {
        try
        {
            int                           index    = 0;
            DataSet::Enumerator *         enumerator  = nullptr;
            std::string                   name;
            DataSet::Enumerator::width_t  width;

            for( auto & x : json.at("enums").items() )
                try
                {
                    name.clear();
                    width = 0;

                    ++index;

                    read(x.value(), "name", name);
                    width = read_enum_width(x.value());

                    enumerator  = new DataSet::Enumerator(name, width);

                    read_enum_labels(enumerator, x.value());

                    if( m_dataset->add(enumerator) == nullptr )
                    {
                        m_errors << "Enum '" << name << "' skipped, because already exists in dataset enums '" << m_dataset->name() << "' filename '" << m_dataset->filename() << "'" << std::endl;
                        delete enumerator;
                    }

                    enumerator = nullptr;
                }
                catch( std::exception & e )
                {
                    m_errors << "Enum index " << index << " '" << name << "' skipped" << UPHY_JSON_REASON_PREFIX << e.what() << "." << std::endl;
                    delete enumerator;
                    enumerator = nullptr;
                }
                catch ( ... )
                {
                    m_errors << "Enum index " << index << " '" << name << "' skipped, unhandled exception" << "." << std::endl;
                    delete enumerator;
                    enumerator = nullptr;
                }
        }
        catch ( std::exception & e ) {
            throw std::runtime_error(std::string() + "Failed to read JSON enums section" + UPHY_JSON_REASON_PREFIX + e.what());
        }
        catch ( ... ) {
            throw std::runtime_error("Failed to read JSON enums section, unhandled exception");
        }
    }

    DataSet::Enumerator::width_t JsonLoader::read_enum_width(const nlohmann::json & json)
    {
        const nlohmann::json & j = json.at("width");

        return j.is_string() ?
                static_cast<DataSet::Enumerator::width_t>(std::stoi((std::string)j)) :
                static_cast<DataSet::Enumerator::width_t>(j);
    }

    void JsonLoader::read_enum_labels(DataSet::Enumerator * enumerator, const nlohmann::json & json)
    {
        int                                     index = 0;
        DataSet::Enumerator::Label *            label = nullptr;
        std::string                             name;
        DataSet::Enumerator::Label::value_t     value;

        for( auto & x : json.at("labels").items() )
            try
            {
                name.clear();
                value = 0;

                ++index;

                read(x.value(), "name", name);
                read(x.value(), "value", value);

                label = new DataSet::Enumerator::Label(name, value);

                if( enumerator->add(label) == nullptr)
                {
                    m_errors << "Label '" << name << "'(" << value << ") skipped, because already exists in enumerator '" << enumerator->name() << "'." << std::endl;
                    delete label;
                }
                label = nullptr;
            }
            catch( std::exception & e )
            {
                m_errors << "Label index " << index << " '" << name << "' skipped" << UPHY_JSON_REASON_PREFIX << e.what() << "." << std::endl;
                delete label;
                label = nullptr;
            }
            catch( ... )
            {
                m_errors << "Label index " << index << " '" << name << "' skipped, unhandled exception" << "." << std::endl;
                delete label;
                label = nullptr;
            }
    }

    register_access_flags_t JsonLoader::read_register_access(const nlohmann::json & json)
    {
        std::string access;
        read(json, "access", access);

        std::transform(access.begin(), access.end(), access.begin(), ::tolower);

        if( access == "ro" )
            return RegisterAccessFlags_ReadOnly;
        else
            if( access == "rw" )
                return RegisterAccessFlags_ReadWrite;
            else
                if( access == "proxy" )
                    return RegisterAccessFlags_Proxy;

        throw std::out_of_range("Wrong register access type='" + access + "'.");
    }

    register_secure_t JsonLoader::read_register_secure(const nlohmann::json & json)
    {
        const nlohmann::json & j = json.at("secure");

        return j.is_string() ?
                static_cast<register_secure_t>(std::stoi((std::string)j)) :
                static_cast<register_secure_t>(j);
    }

    register_unit_flags_t JsonLoader::read_register_units(const nlohmann::json & json)
    {
        register_unit_flags_t value = RegisterUnitFlags_None;

        for( auto x : json.at("units").items() )
        {
            std::string unit = x.value();
            std::transform(unit.begin(), unit.end(), unit.begin(), ::tolower);

            if( unit == "rx" )
                value = static_cast<register_unit_flags_t>( value | RegisterUnitFlags_RX );
            else
                if( unit == "tx" )
                    value = static_cast<register_unit_flags_t>( value | RegisterUnitFlags_TX );
                else
                    if( unit == "clm" )
                        value = static_cast<register_unit_flags_t>( value | RegisterUnitFlags_CLM );
                    else
                        throw std::out_of_range("Wrong register unit type='" + unit + "'.");
        }

        return value;
    }

    void JsonLoader::read_register_dumps(const nlohmann::json & json, DataSet::Register::dumps_collection_t & value)
    {
        for( auto x : json.at("dumps").items() )
        {
            std::string unit = x.value();

            if(!unit.empty())
            {
                std::transform(unit.begin(), unit.end(), unit.begin(), ::tolower);
                value.insert(unit);
            }
        }
    }

    void JsonLoader::read_registers(const nlohmann::json & json)
    {
        try
        {
            int                   index   = 0;
            DataSet::Register *   reg     = nullptr;
            std::string           name;

            for( auto & x : json.at("registers").items() )
                try
                {
                    name.clear();

                    ++index;

                    DataSet::Register::address_t      address;
                    register_access_flags_t           access;
                    register_unit_flags_t             units;
                    register_secure_t                 secure;

                    read(x.value(), "name", name);
                    read(x.value(), "addr", address);

                    access  = read_register_access(x.value());
                    units   = read_register_units(x.value());
                    secure  = read_register_secure(x.value());

                    DataSet::Register::dumps_collection_t dumps;
                    read_register_dumps(x.value(), dumps);

                    switch (m_dataset->type())
                    {
                        case DataSetType_DLN:
                            if( units != RegisterUnitFlags_RX && units != RegisterUnitFlags_TX && units != (RegisterUnitFlags_RX | RegisterUnitFlags_TX)  )
                                throw std::out_of_range(
                                        "Invalid unit type " + to_str(units) + " of register name='" + name +
                                        "' dataset type " + to_str(m_dataset->type()));
                            break;

                        case DataSetType_CLN:
                            if( units != RegisterUnitFlags_CLM )
                                throw std::out_of_range("Invalid unit type " + to_str(units) +
                                                        " of register name='" + name +
                                                        "' dataset type " + to_str(m_dataset->type()));

                            break;

                        default:
                            throw std::runtime_error("Invalid dataset type " + to_str(m_dataset->type()) +
                                                       ", dataset name='"+ m_dataset->name() +"'");
                            break;
                    }

                    // Validate dumps unit prefixes
                    for( auto & d : dumps )
                        if( d.rfind("rx_", 0) == 0 )
                        {
                            if( !(units & RegisterUnitFlags_RX) )
                                throw std::out_of_range("Invalid dump " + d + " of register name='" + name +
                                        "' doesn't contain RX unit, dataset type " + to_str(m_dataset->type()));

                        }
                        else
                            if( d.rfind("tx_", 0) == 0 )
                            {
                                if( !(units & RegisterUnitFlags_TX) )
                                    throw std::out_of_range("Invalid dump " + d + " of register name='" + name +
                                            "' doesn't contain TX unit, dataset type " + to_str(m_dataset->type()));

                            }
                            else
                                if( d.rfind("pll_", 0) == 0 )
                                {
                                    if( !(units & RegisterUnitFlags_CLM) )
                                        throw std::out_of_range("Invalid dump " + d + " of register name='" + name +
                                                "' doesn't contain CLM/PLL unit, dataset type " + to_str(m_dataset->type()));

                                }

                    reg = new DataSet::Register(name, address, access, units, secure, dumps);

                    read_register_fields(reg, x.value());

                    if( m_dataset->add(reg) == nullptr )
                    {
                        m_errors << "Register '" << name << "' skipped, because already exists in dataset '" << m_dataset->name()
                                << "' filename '" << m_dataset->filename() << "'" << std::endl;
                        delete reg;
                    }

                    reg = nullptr;
                }
                catch( std::exception & e )
                {
                    m_errors << "Register index " << index << " '" << name << "' skipped" << UPHY_JSON_REASON_PREFIX << e.what() << "." << std::endl;
                    delete reg;
                    reg = nullptr;
                }
                catch ( ... )
                {
                    m_errors << "Register index " << index << " '" << name << "' skipped, unhandled exception" << "." << std::endl;
                    delete reg;
                    reg = nullptr;
                }
        }
        catch ( std::exception & e ) {
            throw std::runtime_error(std::string() + "Failed to read JSON registers section" + UPHY_JSON_REASON_PREFIX + e.what());
        }
        catch ( ... ) {
            throw std::runtime_error("Failed to read JSON registers section, unhandled exception");
        }
    }

    void JsonLoader::read_register_fields(UPHY::DataSet::Register * reg, const nlohmann::json & json)
    {
        int                                     index = 0;
        DataSet::Register::Field *              field = nullptr;
        std::string                             name;

        for( auto & x : json.at("fields").items() )
            try
            {
                name.clear();

                ++index;

                const DataSet::Enumerator *             enumerator = nullptr;
                DataSet::Register::Field::width_t       width;
                DataSet::Register::Field::range_t       start;
                DataSet::Register::Field::range_t       end;
                std::string                             type;
                bool                                    is_signed = false;


                read(x.value(), "name", name);
                read(x.value(), "width", width);
                read(x.value(), "type", type);

                //TODO:: We invert start & end because json file have not regular logic for msb/lsb
                //read(x.value(), "range_start", start);
                //read(x.value(), "range_end", end);
                read(x.value(), "range_end", start);
                read(x.value(), "range_start", end);

                if( start > 15 )
                    throw std::out_of_range("Field range_start " + std::to_string(start) + "of field '" + name + "' out of range [0-15], register='" + reg->name());

                if( end > 15 )
                    throw std::out_of_range("Field range_end " + std::to_string(end) + " of field '" + name + "' out of range [0-15], register='" + reg->name());

                // TODO::IN jason file start > end, it's regular logic for msb/lsb, in previous step we invert, but message should be clear for json
                if( start > end )
                    throw std::out_of_range(""
                            "Invalid combination range_end " + std::to_string(start) + " bigger range_start " + std::to_string(end) +
                            " of field '" + name + "', register='" + reg->name());

                // TODO::As in previous step inverted start & end we print message inverted
                if( end - start + 1 != width )
                    throw std::out_of_range(
                            "Invalid combination range_start " + std::to_string(end) + ", range_end " + std::to_string(start) +
                            " and width " + std::to_string(width) + " of field '" + name + "', register='" + reg->name());

                std::transform(type.begin(), type.end(), type.begin(), ::tolower);

                if( type == "logic")
                    is_signed = false;
                else
                    if( type == "logic_2c")
                        is_signed = true;
                    else
                    {
                        auto z = m_dataset->enums().find( type );

                        if( z == m_dataset->enums().end() )
                            throw std::out_of_range("Enum '" + type + "' not found, requested by field '" + name + "' in register='" + reg->name() + "'.");

                        enumerator = z->second;
                    }

                field = enumerator != nullptr ?
                            new DataSet::Register::Field(name, width, start, end, enumerator) :
                            new DataSet::Register::Field(name, width, start, end, is_signed);

                if( reg->add(field) == nullptr)
                {
                    m_errors << "Field '" << name << " skipped, because already exists in register '" << reg->name() << "'." << std::endl;
                    delete field;
                }

                field = nullptr;
            }
            catch( std::exception & e )
            {
                m_errors << "Field index " << index << " '" << name << "' skipped" << UPHY_JSON_REASON_PREFIX << e.what() << "." << std::endl;
                delete field;
                field = nullptr;
            }
            catch( ... )
            {
                m_errors << "Field index " << index << " '" << name << "' skipped, unhandled exception" << "." << std::endl;
                delete field;
                field = nullptr;
            }
    }
}

