/*
 * Copyright (c) 2004-2013 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <string.h>

#include "ibis.h"
#include "packets/packets_layouts.h"


/****************************************************/
/******************private methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-15] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSMadGetSet(u_int16_t lid,
        u_int8_t method,
        u_int16_t attribute_id,
        u_int32_t attribute_modifier,
        const data_func_set_t & attribute_data,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    struct MAD_VendorSpec vs_mad;
    CLEAR_STRUCT(vs_mad);

    //build vs lid route header
    this->CommonMadHeaderBuild(&vs_mad.MAD_Header_Common,
            IBIS_IB_CLASS_VENDOR_MELLANOX,
            method,
            attribute_id,
            attribute_modifier);

    vs_mad.VS_Key = this->key_manager.GetKey(lid, IBIS_VS_KEY);

    //send the mad
    IBIS_RETURN(this->MadGetSet(lid, 1, 0, IBIS_IB_DEFAULT_QP1_QKEY,
            IBIS_IB_CLASS_VENDOR_MELLANOX,
            method, attribute_id, attribute_modifier,
            IBIS_IB_DATA_OFFSET_VENDOR_MELLANOX,
            data_func_set_t(&vs_mad, IBIS_FUNC_LST(MAD_VendorSpec)),
            attribute_data,
            p_clbck_data));
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSGeneralInfoGet(u_int16_t lid,
        struct VendorSpec_GeneralInfo *p_general_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_general_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending VSGeneralInfo Get MAD lid = %u\n", lid);
    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_GENERAL_INFO,
            0,
            data_func_set_t(p_general_info, IBIS_FUNC_LST(VendorSpec_GeneralInfo)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSSwitchNetworkInfoGet(u_int16_t lid,
        struct VS_SwitchNetworkInfo *p_switch_network_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_switch_network_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending VSSwitchNetworkInfo Get MAD lid = %u\n", lid);
    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_SWITCH_NETWORK_INFO,
            0,
            data_func_set_t(p_switch_network_info, IBIS_FUNC_LST(VS_SwitchNetworkInfo)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

#define LLR_SYMBOL_ERRORS_FLAG_OFFSET  31

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPortLLRStatisticsGet(u_int16_t lid,
        phys_port_t port_number,
        struct VendorSpec_PortLLRStatistics *p_port_llr_statistics,
        bool get_symbol_errors,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_llr_statistics);

    u_int32_t attribute_modifier = 0;

    p_port_llr_statistics->PortSelect = (u_int8_t)port_number;
    if (get_symbol_errors)
        attribute_modifier |= (1 << LLR_SYMBOL_ERRORS_FLAG_OFFSET);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortLLRStatistics Get MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_LLR_STATISTICS,
            attribute_modifier,
            data_func_set_t(p_port_llr_statistics, IBIS_FUNC_LST(VendorSpec_PortLLRStatistics)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPortLLRStatisticsClear(u_int16_t lid,
        phys_port_t port_number,
        bool clear_symbol_errors,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    u_int32_t attribute_modifier = 0;

    struct VendorSpec_PortLLRStatistics port_llr_statistics;
    CLEAR_STRUCT(port_llr_statistics);
    port_llr_statistics.PortSelect = port_number;
    if (clear_symbol_errors)
        attribute_modifier |= (1 << LLR_SYMBOL_ERRORS_FLAG_OFFSET);


    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortLLRStatistics Reset MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_LLR_STATISTICS,
            attribute_modifier,
            data_func_set_t(&port_llr_statistics, IBIS_FUNC_LST(VendorSpec_PortLLRStatistics)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPortLLRStatisticsSet(u_int16_t lid,
        phys_port_t port_number,
        struct VendorSpec_PortLLRStatistics *p_port_llr_statistics,
        bool set_symbol_errors,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_llr_statistics);

    u_int32_t attribute_modifier = 0;

    p_port_llr_statistics->PortSelect = (u_int8_t)port_number;
    if (set_symbol_errors)
        attribute_modifier |= (1 << LLR_SYMBOL_ERRORS_FLAG_OFFSET);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortLLRStatistics Set MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_LLR_STATISTICS,
            attribute_modifier,
            data_func_set_t(p_port_llr_statistics, IBIS_FUNC_LST(VendorSpec_PortLLRStatistics)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED /
 * IBIS_MAD_STATUS_RECV_FAILED / IBIS_MAD_STATUS_TIMEOUT /
 * IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSDiagnosticDataPageClear(u_int16_t lid,
        phys_port_t port_number,
        u_int8_t page_number,
        struct VS_DiagnosticData *p_dc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_dc);

    u_int32_t attribute_modifier = 0;

    attribute_modifier = port_number | (page_number << 16);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSDiagnosticCounters Set MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_DIAGNOSTIC_COUNTERS,
            attribute_modifier,
            data_func_set_t(p_dc, IBIS_FUNC_LST(VS_DiagnosticData)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED /
 * IBIS_MAD_STATUS_RECV_FAILED / IBIS_MAD_STATUS_TIMEOUT /
 * IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSDiagnosticDataPageClear_AM(u_int16_t lid,
        u_int32_t attribute_modifier,
        struct VS_DiagnosticData *p_dc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_dc);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSDiagnosticCounters Set MAD lid = %u\n",
             lid);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_DIAGNOSTIC_COUNTERS,
            attribute_modifier,
            data_func_set_t(p_dc, IBIS_FUNC_LST(VS_DiagnosticData)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSDiagnosticDataGet(u_int16_t lid,
        phys_port_t port_number,
        u_int8_t page_number,
        struct VS_DiagnosticData *p_dc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_dc);

    u_int32_t attribute_modifier = 0;

    attribute_modifier = (page_number << 16) | port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSDiagnosticCounters Get MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_DIAGNOSTIC_COUNTERS,
            attribute_modifier,
            data_func_set_t(p_dc, IBIS_FUNC_LST(VS_DiagnosticData)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSDiagnosticDataGet_AM(u_int16_t lid,
        u_int32_t attribute_modifier,
        struct VS_DiagnosticData *p_dc,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_dc);

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSDiagnosticCounters Get MAD lid = %u\n",
             lid);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_DIAGNOSTIC_COUNTERS,
            attribute_modifier,
            data_func_set_t(p_dc, IBIS_FUNC_LST(VS_DiagnosticData)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPortRNCountersClear(u_int16_t lid,
        phys_port_t port_number,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    struct port_rn_counters rn_counters;
    CLEAR_STRUCT(rn_counters);

    rn_counters.counter_select = 0xFFFF;
    rn_counters.port_select = port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortRNCounters Clear MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RN_COUNTERS,
            0,
            data_func_set_t(&rn_counters, IBIS_FUNC_LST(port_rn_counters)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPortRNCountersGet(u_int16_t lid,
        phys_port_t port_number,
        struct port_rn_counters *p_rn_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_rn_counters);

    p_rn_counters->port_select = port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortRNCounters Get MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_RN_COUNTERS,
            0,
            data_func_set_t(p_rn_counters, IBIS_FUNC_LST(port_rn_counters)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSPortRoutingDecisionCountersClear(u_int16_t lid,
            phys_port_t port_number,
            const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    struct port_routing_decision_counters routing_decision_counters;
    CLEAR_STRUCT(routing_decision_counters);

    routing_decision_counters.counter_select = 0xFFFF;
    routing_decision_counters.port_select = port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortRoutingDecisionCountersClear Clear MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_ROUTING_DECISION_COUNTERS,
            0,
            data_func_set_t(&routing_decision_counters, IBIS_FUNC_LST(port_routing_decision_counters)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSPortRoutingDecisionCountersGet(u_int16_t lid,
            phys_port_t port_number,
            struct port_routing_decision_counters *p_routing_decision_counters,
            const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_routing_decision_counters);
    p_routing_decision_counters->port_select = port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VSPortRoutingDecisionCountersGet MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_ROUTING_DECISION_COUNTERS,
            0,
            data_func_set_t(p_routing_decision_counters, IBIS_FUNC_LST(port_routing_decision_counters)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

/*
 * Returns: mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
 *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
 */
int Ibis::VSPerVLCounters(bool is_reset_cntr,
        u_int16_t lid,
        phys_port_t port_number,
        u_int32_t attr_id,
        struct PM_PortRcvXmitCntrsSlVl *p_pm_port_rcvxmit_data_slvl,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_pm_port_rcvxmit_data_slvl);

    if (is_reset_cntr)
        p_pm_port_rcvxmit_data_slvl->counter_select = 0xffff;
    p_pm_port_rcvxmit_data_slvl->port_select = port_number;
    IBIS_LOG(TT_LOG_LEVEL_MAD,
            "Sending %s attribute %u Get MAD lid = %u port = %u\n",
            (is_reset_cntr ? "Clear" : "Get"),
            attr_id, lid, port_number);

    int rc = this->VSMadGetSet(lid,
            is_reset_cntr ? IBIS_IB_MAD_METHOD_SET : IBIS_IB_MAD_METHOD_GET,
            attr_id,
            0,
            data_func_set_t(p_pm_port_rcvxmit_data_slvl, IBIS_FUNC_LST(PM_PortRcvXmitCntrsSlVl)),
            p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSMirroringInfoGet(u_int16_t lid,
        struct VS_MirroringInfo *p_mirroring_info,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_mirroring_info);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending VS_MirroringInfo Get MAD lid = %u\n", lid);
    int rc = this->VSMadGetSet(lid,
                IBIS_IB_MAD_METHOD_GET,
                IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_MIRRORING_INFO,
                0,
                data_func_set_t(p_mirroring_info, IBIS_FUNC_LST(VS_MirroringInfo)),
                p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);  // bits [8-15] are class specific - don't need them here
}

#define ENCAPSULATION_OFFSET 16 // See NVIDIA Mellanox Vendor Specific MAD Specification, 3.6.3 MirroringAgent Attribute

static void _VSMirroringAgentSet_pack(const struct VS_MirroringAgent *ptr_struct, u_int8_t* ptr_buff) {
    VS_MirroringAgent_pack(ptr_struct, ptr_buff);

    // We're overriding the encapsulation part
    switch (ptr_struct->span_type) {
    case 0x4: // LOCAL_IB
        {
            struct VS_Encapsulation_LocalSPAN local_span;
            memcpy(&local_span, &ptr_struct->encapsulation.local_span, sizeof(local_span));
            VS_Encapsulation_LocalSPAN_pack(&local_span, ptr_buff + ENCAPSULATION_OFFSET);
        }
        break;
    case 0xA: // REMOTE_IB_LOCAL_UD
        {
            struct VS_Encapsulation_RemoteSPANLocalUD remote_span_local_ud;
            memcpy(&remote_span_local_ud, &ptr_struct->encapsulation.remote_span_local_ud, sizeof(remote_span_local_ud));
            VS_Encapsulation_RemoteSPANLocalUD_pack(&remote_span_local_ud, ptr_buff + ENCAPSULATION_OFFSET);
        }
        break;
    case 0xB: // REMOTE_IB_GLOBAL_UD
        {
            struct VS_Encapsulation_RemoteSPANGlobalUD remote_span_global_ud;
            memcpy(&remote_span_global_ud, &ptr_struct->encapsulation.remote_span_global_ud, sizeof(remote_span_global_ud));
            VS_Encapsulation_RemoteSPANGlobalUD_pack(&remote_span_global_ud, ptr_buff + ENCAPSULATION_OFFSET);
        }
        break;
    default:
        IBIS_LOG(TT_LOG_LEVEL_ERROR,
            "Invalid span_type in VS_MirroringAgent (UNPACK) %x\n",
            ptr_struct->span_type);
        break;
    }
}

static void _VSMirroringAgentSet_unpack(struct VS_MirroringAgent *ptr_struct, const u_int8_t* ptr_buff) {
    VS_MirroringAgent_unpack(ptr_struct, ptr_buff);

    // We're overriding the encapsulation part
    switch (ptr_struct->span_type) {
    case 0x4: // LOCAL_IB
        VS_Encapsulation_LocalSPAN_unpack(&ptr_struct->encapsulation.local_span, ptr_buff + ENCAPSULATION_OFFSET);
        break;
    case 0xA: // REMOTE_IB_LOCAL_UD
        VS_Encapsulation_RemoteSPANLocalUD_unpack(&ptr_struct->encapsulation.remote_span_local_ud, ptr_buff + ENCAPSULATION_OFFSET);
        break;
    case 0xB: // REMOTE_IB_GLOBAL_UD
        VS_Encapsulation_RemoteSPANGlobalUD_unpack(&ptr_struct->encapsulation.remote_span_global_ud, ptr_buff + ENCAPSULATION_OFFSET);
        break;
    default:
        IBIS_LOG(TT_LOG_LEVEL_ERROR,
            "Invalid span_type in VS_MirroringAgent (SET) %x\n",
            ptr_struct->span_type);
        break;
    }
}

int Ibis::VSMirroringAgentGet(u_int16_t lid,
        u_int8_t mirror_agent_index,
        struct VS_MirroringAgent *p_mirroring_agent,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_mirroring_agent);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending VS_MirroringAgent Get MAD lid = %u\n", lid);
    int rc = this->VSMadGetSet(lid,
                IBIS_IB_MAD_METHOD_GET,
                IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_MIRRORING_AGENT,
                mirror_agent_index,
                data_func_set_t(p_mirroring_agent,
                                _VSMirroringAgentSet_pack,
                                _VSMirroringAgentSet_unpack,
                                VS_MirroringAgent_dump),
                p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSMirroringAgentSet(u_int16_t lid,
        u_int8_t mirror_agent_index,
        struct VS_MirroringAgent *p_mirroring_agent,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VS_MirroringAgent Set MAD lid = %u mirror_agent_index = %u\n",
             lid, mirror_agent_index);

    int rc = this->VSMadGetSet(lid,
                IBIS_IB_MAD_METHOD_SET,
                IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_MIRRORING_AGENT,
                mirror_agent_index,
                data_func_set_t(p_mirroring_agent,
                                _VSMirroringAgentSet_pack,
                                _VSMirroringAgentSet_unpack,
                                VS_MirroringAgent_dump),
                p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSMirroringGlobalTriggerGet(u_int16_t lid,
        u_int8_t trigger,
        struct VS_MirroringGlobalTrigger *p_mirroring_global_trigger,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_mirroring_global_trigger);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending VS_MirroringGlobalTrigger Get MAD lid = %u trigger=%u\n", lid, trigger);
    int rc = this->VSMadGetSet(lid,
                IBIS_IB_MAD_METHOD_GET,
                IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_MIRRORING_GLOBAL_TRIGGER,
                trigger,
                data_func_set_t(p_mirroring_global_trigger, IBIS_FUNC_LST(VS_MirroringGlobalTrigger)),
                p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);  // bits [8-15] are class specific - don't need them here
}

int Ibis::VSMirroringGlobalTriggerSet(u_int16_t lid,
        u_int8_t trigger,
        struct VS_MirroringGlobalTrigger *p_mirroring_global_trigger,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    u_int32_t attribute_modifier = 0;

    // See 3.6.4 MirroringGlobalTrigger Attribute, NVIDIA Mellanox Vendor Specific MAD Specification
    if (trigger > 0x0F) {
        IBIS_LOG(TT_LOG_LEVEL_ERROR, "Invalid trigger in VSMirroringGlobalTriggerSet 0x%x\n", trigger);
        return IBIS_MAD_STATUS_INVALID_FIELD;
    }

    // See Table 202 Attribute Modifier Fields, NVIDIA Mellanox Vendor Specific MAD Specification
    attribute_modifier = ((u_int32_t)trigger & 0x0F) << 15;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending VS_MirroringGlobalTrigger Set MAD lid = %u trigger=%u attribute_modifier=0x%08x\n", lid,
             trigger, attribute_modifier);
    int rc = this->VSMadGetSet(lid,
                IBIS_IB_MAD_METHOD_SET,
                IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_MIRRORING_GLOBAL_TRIGGER,
                attribute_modifier,
                data_func_set_t(p_mirroring_global_trigger, IBIS_FUNC_LST(VS_MirroringGlobalTrigger)),
                p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);  // bits [8-15] are class specific - don't need them here
}

int Ibis::VSCongestionMirroringGet(u_int16_t lid,
        bool global,
        u_int8_t port,
        struct VS_CongestionMirroring *p_congestion_mirroring,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    u_int32_t attribute_modifier = 0;

    CLEAR_STRUCT(*p_congestion_mirroring);
    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CongestionMirroring Get MAD lid = %u global = %d port = %u\n", lid, global, port);

    attribute_modifier = global ? 0x0100 : ((u_int32_t)port);

    int rc = this->VSMadGetSet(lid,
                IBIS_IB_MAD_METHOD_GET,
                IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_CONGESTION_MIRRORING,
                attribute_modifier,
                data_func_set_t(p_congestion_mirroring, IBIS_FUNC_LST(VS_CongestionMirroring)),
                p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);  // bits [8-15] are class specific - don't need them here
}

int Ibis::VSCongestionMirroringSet(u_int16_t lid,
        bool global,
        u_int8_t port,
        struct VS_CongestionMirroring *p_congestion_mirroring,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    u_int32_t attribute_modifier = 0;

    IBIS_LOG(TT_LOG_LEVEL_MAD, "Sending CongestionMirroring Set MAD lid = %u global = %d port = %u\n", lid, global, port);

    attribute_modifier = global ? 0x10000 : ((u_int32_t)port);

    int rc = this->VSMadGetSet(lid,
                IBIS_IB_MAD_METHOD_SET,
                IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_CONGESTION_MIRRORING,
                attribute_modifier,
                data_func_set_t(p_congestion_mirroring, IBIS_FUNC_LST(VS_CongestionMirroring)),
                p_clbck_data);
    IBIS_RETURN(rc & 0x00ff);  // bits [8-15] are class specific - don't need them here
}


/*** Fast Recovery Counters ***/

int Ibis::VSCreditWatchdogTimeoutCountersGet(u_int16_t lid,
        phys_port_t port_number,
        struct VS_CreditWatchdogTimeoutCounters *p_credit_wd_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_credit_wd_counters);
    u_int32_t attribute_modifier = 0;
    p_credit_wd_counters->PortSelect = (u_int8_t)port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VS_CreditWatchdogTimeoutCounters Get MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_CREDIT_WATCHDOG_TIMEOUT_COUNTERS,
            attribute_modifier,
            data_func_set_t(p_credit_wd_counters, IBIS_FUNC_LST(VS_CreditWatchdogTimeoutCounters)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSCreditWatchdogTimeoutCountersClear(u_int16_t lid,
        phys_port_t port_number,
        struct VS_CreditWatchdogTimeoutCounters *p_credit_wd_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_credit_wd_counters);
    u_int32_t attribute_modifier = 0;
    p_credit_wd_counters->PortSelect = (u_int8_t)port_number;
    memset(&p_credit_wd_counters->CounterSelect, 0xff, sizeof(p_credit_wd_counters->CounterSelect));

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VS_CreditWatchdogTimeoutCounters Reset MAD lid = %u port = %u\n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_CREDIT_WATCHDOG_TIMEOUT_COUNTERS,
            attribute_modifier,
            data_func_set_t(p_credit_wd_counters, IBIS_FUNC_LST(VS_CreditWatchdogTimeoutCounters)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSFastRecoveryCountersGet(u_int16_t lid,
        phys_port_t port_number,
        u_int8_t trigger,
        struct VS_FastRecoveryCounters *p_fast_recovery_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_fast_recovery_counters);
    u_int32_t attribute_modifier = 0;
    p_fast_recovery_counters->PortSelect = (u_int8_t)port_number;
    p_fast_recovery_counters->trigger = trigger;


    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VS_FastRecoveryCounters Get MAD lid = %u port = %u trigger = %u\n",
             lid, port_number, trigger);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_FAST_RECOVERY_COUNTERS,
            attribute_modifier,
            data_func_set_t(p_fast_recovery_counters, IBIS_FUNC_LST(VS_FastRecoveryCounters)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSFastRecoveryCountersClear(u_int16_t lid,
        phys_port_t port_number,
        u_int8_t trigger,
        struct VS_FastRecoveryCounters *p_fast_recovery_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_fast_recovery_counters);
    u_int32_t attribute_modifier = 0;
    p_fast_recovery_counters->PortSelect = (u_int8_t)port_number;
    p_fast_recovery_counters->trigger = trigger;

    memset(&p_fast_recovery_counters->FieldSelect, 0xff, sizeof(p_fast_recovery_counters->FieldSelect));

    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VS_FastRecoveryCounters Reset MAD lid = %u port = %u trigger = %u\n",
             lid, port_number, trigger);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_FAST_RECOVERY_COUNTERS,
            attribute_modifier,
            data_func_set_t(p_fast_recovery_counters, IBIS_FUNC_LST(VS_FastRecoveryCounters)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSPortGeneralCountersGet(u_int16_t lid, 
        phys_port_t port_number,
        struct VS_PortGeneralCounters* p_port_general_counters,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_general_counters);
    u_int32_t attribute_modifier = 0;
    p_port_general_counters->port_select = (u_int8_t)port_number;
    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VS_PortGeneralCounters Get MAD lid = %u port = %u \n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_GENERAL_COUNTERS,
            attribute_modifier,
            data_func_set_t(p_port_general_counters, IBIS_FUNC_LST(VS_PortGeneralCounters)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSPortGeneralCountersClear(u_int16_t lid, 
        phys_port_t port_number,
        struct VS_PortGeneralCounters* p_port_general_counters,
        const clbck_data_t *p_clbck_data )
{
    IBIS_ENTER;
    CLEAR_STRUCT(*p_port_general_counters);
    u_int32_t attribute_modifier = 0;
    p_port_general_counters->port_select = (u_int8_t)port_number;
    p_port_general_counters->counter_select = 0xffffffff;
    IBIS_LOG(TT_LOG_LEVEL_MAD,
             "Sending VS_PortGeneralCounters Clear MAD lid = %u port = %u \n",
             lid, port_number);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PORT_GENERAL_COUNTERS,
            attribute_modifier,
            data_func_set_t(p_port_general_counters, IBIS_FUNC_LST(VS_PortGeneralCounters)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSPerformanceHistogramInfoGet(u_int16_t lid,
        struct VS_PerformanceHistogramInfo *p_data,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_data);
    u_int32_t attribute_modifier = 0;


    IBIS_LOG(TT_LOG_LEVEL_MAD,"Sending VS_PerformanceHistogramInfo Get MAD lid = %u\n",lid);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PERFORMANCE_HISTOGRAM_INFO,
            attribute_modifier,
            data_func_set_t(p_data, IBIS_FUNC_LST(VS_PerformanceHistogramInfo)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


int Ibis::VSPerformanceHistogramBufferControlGet(u_int16_t lid,
        phys_port_t port_number, uint8_t vl, uint8_t dir,
        bool port_global, bool vl_global,
        struct VS_PerformanceHistogramBufferControl *p_data,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_data);
    u_int32_t attribute_modifier =
                (port_global << 31) | (vl_global << 30) | ((dir & 0x01) << 29) | ((vl & 0x0F) << 20) | port_number;


    IBIS_LOG(TT_LOG_LEVEL_MAD,"Sending VS_PerformanceHistogramBufferControl Get MAD " \
                               "lid = %u, port = %u, port_global = %d, vl_global = %d, vl = %d \n",
                               lid, port_number, port_global, vl_global, vl);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PERFORMANCE_HISTOGRAM_BUFFER_CONTROL,
            attribute_modifier,
            data_func_set_t(p_data, IBIS_FUNC_LST(VS_PerformanceHistogramBufferControl)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


int Ibis::VSPerformanceHistogramBufferControlSet(u_int16_t lid,
        phys_port_t port_number, uint8_t vl, uint8_t dir,
        bool port_global, bool vl_global,
        struct VS_PerformanceHistogramBufferControl *p_data,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_data);
    u_int32_t attribute_modifier =
                (port_global << 31) | (vl_global << 30) | ((dir & 0x01) << 29) | ((vl & 0x0F) << 20) | port_number;


    IBIS_LOG(TT_LOG_LEVEL_MAD,"Sending VS_PerformanceHistogramBufferControl Set MAD " \
                               "lid = %u, port = %u, port_global = %d, vl_global = %d, vl = %d \n",
                               lid, port_number, port_global, vl_global, vl);


    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_SET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PERFORMANCE_HISTOGRAM_BUFFER_CONTROL,
            attribute_modifier,
            data_func_set_t(p_data, IBIS_FUNC_LST(VS_PerformanceHistogramBufferControl)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}


int Ibis::VSPerformanceHistogramBufferDataGet(u_int16_t lid,
        phys_port_t port_number, uint8_t vl, uint8_t dir, bool clr,
        struct VS_PerformanceHistogramBufferData *p_data,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_data);
    u_int32_t attribute_modifier = (clr << 31) | ((dir & 0x01) << 29) | ((vl & 0x0F) << 20) | port_number;


    IBIS_LOG(TT_LOG_LEVEL_MAD,"Sending VS_PerformanceHistogramBufferData Get MAD " \
                               "lid = %u, port = %u, vl = %d, clr = %d \n",
                               lid, port_number, vl, clr);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PERFORMANCE_HISTOGRAM_BUFFER_DATA,
            attribute_modifier,
            data_func_set_t(p_data, IBIS_FUNC_LST(VS_PerformanceHistogramBufferData)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSPerformanceHistogramPortsControlGet(u_int16_t lid,
        phys_port_t port_number, u_int8_t hist_id,
        struct VS_PerformanceHistogramPortsControl *p_data,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_data);
    u_int32_t attribute_modifier = ((hist_id & 0x0F) << 16) | port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,"Sending VS_PerformanceHistogramPortControl Get MAD " \
                               "lid = %u, port = %u, hist_id = %d \n",
                               lid, port_number, hist_id);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PERFORMANCE_HISTOGRAM_PORTS_CONTROL,
            attribute_modifier,
            data_func_set_t(p_data, IBIS_FUNC_LST(VS_PerformanceHistogramPortsControl)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

int Ibis::VSPerformanceHistogramPortsDataGet(u_int16_t lid,
        phys_port_t port_number, u_int8_t hist_id, bool clr,
        struct VS_PerformanceHistogramPortsData *p_data,
        const clbck_data_t *p_clbck_data)
{
    IBIS_ENTER;

    CLEAR_STRUCT(*p_data);
    u_int32_t attribute_modifier = ((clr & 0x01) << 31) | ((hist_id & 0x0F) << 16) | port_number;

    IBIS_LOG(TT_LOG_LEVEL_MAD,"Sending VS_PerformanceHistogramPortData Get MAD " \
                               "lid = %u, port = %u, hist_id = %d \n",
                               lid, port_number, hist_id);

    int rc = this->VSMadGetSet(lid,
            IBIS_IB_MAD_METHOD_GET,
            IBIS_IB_ATTR_VENDOR_SPEC_MELLANOX_PERFORMANCE_HISTOGRAM_PORTS_DATA,
            attribute_modifier,
            data_func_set_t(p_data, IBIS_FUNC_LST(VS_PerformanceHistogramPortsData)),
            p_clbck_data);

    IBIS_RETURN(rc & 0x00ff);   //bits [8-15] are class specific - don't need them here
}

