/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <typeinfo>
#include <math.h>
#include <assert.h>
#include "phy_diag.h"
#include "acc_reg.h"
#include "phy_diag_errs.h"
#include "infiniband/ibdm/cable/PrtlRecord.h"
#include "phy_diag_types.h"
#include "infiniband/ibutils/output_stream_primitives.h"

string PDDRModuleInfoRegister::transmitter_technology_arr[] = {
    "850 nm VCSEL"                                             ,
    "1310 nm VCSEL"                                            ,
    "1550 nm VCSEL"                                            ,
    "1310 nm FP"                                               ,
    "1310 nm DFB"                                              ,
    "1550 nm DFB"                                              ,
    "1310 nm EML"                                              ,
    "1550 nm EML"                                              ,
    "others"                                                   ,
    "1490 nm DFB"                                              ,
    "Copper cable- unequalized"                                ,
    "Copper cable- passive equalized"                          ,
    "Copper cable- near and far end limiting active equalizers",
    "Copper cable- far end limiting active equalizers"         ,
    "Copper cable- near end limiting active equalizers"        ,
    "Copper cable- linear active equalizers"                   ,
    NO_DATA_VAL };

string PDDRModuleInfoRegister::ib_compliance_code_arr[] = {
    "SDR/",
    "DDR/",
    "QDR/",
    "FDR/",
    "EDR/",
    "HDR/",
    "NDR/" };


template<typename _Type, int (_Type::*_Method)(const clbck_data_t &, int , void*)>
    void AccRegHandler_ForwardClbck(const clbck_data_t &clbck_data,
                      int rec_status,
                      void *p_attribute_data)
{
    //wW cast to a reference in order to avoid having unchecked NULL ptr
    //which dynamic_cast my return.
    //In case of casting to the reference an exception is thrown.
    (dynamic_cast<_Type&>
        (*reinterpret_cast<AccRegHandler*>(clbck_data.m_p_obj)->GetPReg()).*_Method)
              (clbck_data, rec_status, p_attribute_data);
}

static inline bool IsSupportAccRegFailed(IBNode* p_node,
                                         u_int8_t acc_reg_type,
                                         u_int8_t not_support_bit)
{
    return NSB_BITSET(p_node)[acc_reg_type] || NSB_BITSET(p_node)[not_support_bit];
}

bool keycomp(AccRegKey *k1, AccRegKey *k2)
{
    return ((*k1) < (*k2));
};

Register::Register(PhyDiag * phy_diag,
                   uint32_t reg_id,
                   const unpack_data_func_t p_unpack_func,
                   string sn,
                   string name,
                   uint32_t fn,
                   uint64_t nsb,
                   string header,
                   PhyPluginSupportedNodesType support_nodes,
                   bool dump_enabled,
                   bool retrive_disconnected,
                   AccRegVia_t acc_reg_type,
                   AccRegVia_t acc_type_priority)

    : m_phy_diag(phy_diag), m_register_id(reg_id), m_fields_num(fn), m_not_supported_bit(nsb),
      m_section_name(sn), m_header(header), m_name(name),
      m_support_nodes(support_nodes), m_retrive_disconnected(retrive_disconnected),
      m_dump_enabled(dump_enabled), m_acc_reg_type(acc_reg_type), unpack_data_func(p_unpack_func)

{
    IBDIAGNET_ENTER;

    m_acc_type_priority = PhyDiag::acc_reg_priority == NOT_SUP_ACC_REG ?
            acc_type_priority : PhyDiag::acc_reg_priority;

    IBDIAGNET_RETURN_VOID;
}

bool Register::IsRegSupportNodeType(const IBNode* node)
{
    IBDIAGNET_ENTER;
    bool rc = IsPhyPluginSupportNodeType(m_support_nodes, node);
    IBDIAGNET_RETURN(rc);
}


void Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    sstream << handler_header;
    if (this->GetHeader().empty())
        for (u_int32_t cnt = 0; cnt < this->GetFieldsNum(); cnt++)
            sstream << ",field" <<  cnt;
    else
        sstream << this->GetHeader();

    IBDIAGNET_RETURN_VOID;
}

void Register::GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices)
{
    IBDIAGNET_ENTER;
    assert(true); // make sure that this function called only from
                  // derived registers that overrun this function.

    IBDIAGNET_RETURN_VOID;
}

int Register::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors)
{
    IBDIAGNET_ENTER;
    assert(true); // make sure that this function called only from
                  // derived classes that overrun this function.

    IBDIAGNET_RETURN(IBDIAG_ERR_CODE_ILLEGAL_FUNCTION_CALL);
}

int Register::SensorsBitsToList(u_int64_t sensors,
                                u_int8_t max_sensors,
                                list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    u_int64_t mask = 1;

    for (u_int8_t i = 0; i < max_sensors; ++i, mask <<= 1) {
        if (mask & sensors)
            sensors_list.push_back(i);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int Register::SensorsCountToList(u_int8_t sensor_count, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;

    for (u_int8_t i = 0; i < sensor_count; ++i)
        sensors_list.push_back(i);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int Register::BuildDB(AccRegHandler* p_handler,
                      list_p_fabric_general_err &phy_errors,
                      ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    assert(false); // make sure that this function called only from
                   // derived classes that overrun this function.

    IBDIAGNET_RETURN(IBDIAG_ERR_CODE_ILLEGAL_FUNCTION_CALL);
}

void Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                  export_data_phy_node_t *p_export_data_phy_node,
                                  struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

bool Register::CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                export_data_phy_node_t *p_export_data_phy_node,
                                AccRegKey *p_key)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(false);
}

int Register::FillLocalPortAndPlaneInd(const AccRegKey* p_key, uint8_t * local_port, uint8_t * plane_ind) {
    const AccRegKeyPort* key = dynamic_cast<const AccRegKeyPort*>(p_key);

    if (!key)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    IBPort *port = this->m_phy_diag->GetPort(key->node_guid, key->port_num);
    if (!port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISABLED);
    
    return FillLocalPortAndPlaneInd(port, local_port, plane_ind);
}


int Register::FillLocalPortAndPlaneInd(IBPort *p_port, uint8_t * local_port, uint8_t * plane_ind) {
    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    IBNode *p_node = p_port->p_node;

    if (!p_node)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    if (!local_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    // Fill legacy mode
    *local_port = p_port->num;
    if (plane_ind)
        *plane_ind = 0;

    if (p_node->type != IB_CA_NODE)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    if (p_node->isSpecialNode())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    if (p_node->ext_type != IB_MULTI_PLANE_HCA_NODE)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    if (!p_port->p_port_hierarchy_info)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    if (p_port->p_port_hierarchy_info->m_aport == -1)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    if (!plane_ind) {
        PHYNodeData * phy_data = dynamic_cast<PHYNodeData *>(p_node->p_phy_data);

        if (!phy_data)
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

        if (phy_data->planarized_hca_per_reg_sent.count(std::make_pair(this, p_port->p_aport)))
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISABLED);

        phy_data->planarized_hca_per_reg_sent.insert(std::make_pair(this, p_port->p_aport));
    }
    else
    {
        if (p_port->p_port_hierarchy_info->m_plane == -1)
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

        *plane_ind  = p_port->p_port_hierarchy_info->m_plane;
    }

    *local_port = p_port->p_port_hierarchy_info->m_aport;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//*****************************************
//***     PRTL Register functions       ***
//*****************************************
PRTLRegister::PRTLRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
            ACCESS_REGISTER_ID_PRTL,
            (const unpack_data_func_t)prtl_reg_unpack,
            ACC_REG_PRTL_INTERNAL_SECTION_NAME,
            ACC_REG_PRTL_NAME,
            ACC_REG_PRTL_FIELDS_NUM,//how many output columns of the register are going to print
            NSB::get(this),
            ACC_REG_PRTL_FIELDS_HEADER,
            SUPPORT_SW_CA, //should support HCA and Switches
            true,
            false,
            VIA_SMP,
            VIA_GMP)
{
}

int PRTLRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    AccRegKeyPort * key = dynamic_cast<AccRegKeyPort*>(p_key);
    if (!key || !p_node)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    if (p_node->serdes_version < SerDes_7nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    //On switches CAP mask should be taken from "0 port"
    const IBPort *p_port = p_node->getPort(p_node->type == IB_SW_NODE ? 0 : key->port_num);
    if (!p_port)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    const SMP_PortInfo * portInfo =
           this->m_phy_diag->GetIBDiag()->GetIBDMExtendedInfoPtr()->getSMPPortInfo(p_port->createIndex);
    if (!portInfo || !IS_SUPPORT_LINK_ROUND_TRIP_LATENCY(portInfo->CapMsk))
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    //sent any time cause it used for cable's length calculation even
    //if a cable is fetched by the "Cable" plugin

    struct prtl_reg prtl;
    CLEAR_STRUCT(prtl);
    prtl.pnat =  ACC_REG_PNAT_IB_PORT;
    IBPort* local_port = p_node->getPort(key->port_num);

    int rc = FillLocalPortAndPlaneInd(local_port, &prtl.local_port, nullptr);

    if (rc)
        return rc;

    prtl_reg_pack(&prtl, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

#ifdef COPY
    #error "The PRTL COPY macros has been already defined"
#else
    #define COPY(value) p_port->p_prtl->value = p_reg->value;
#endif

int PRTLRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer)
{
    IBDIAGNET_ENTER;
    int rc = Register::UnpackData(p_key, data_to_unpack, unpacked_buffer);
    if (rc)
        IBDIAG_RETURN(rc);

    const AccRegKeyPort *p_regKey = dynamic_cast<const AccRegKeyPort*>(p_key);
    if (!p_regKey)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    IBPort *p_port = this->m_phy_diag->GetPort(p_regKey->node_guid, p_regKey->port_num);
    if (!p_port || p_port->p_prtl)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    const prtl_reg *p_reg = reinterpret_cast<prtl_reg*>(data_to_unpack);
    p_port->p_prtl = new PrtlRecord();
    COPY(lp_msb);
    COPY(local_port);
    COPY(rtt_support);
    COPY(latency_accuracy);
    COPY(latency_res);
    COPY(local_phy_latency);
    COPY(local_mod_dp_latency);
    COPY(round_trip_latency);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

#undef COPY

void PRTLRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                   << '#'
            << handler_header         << ','
            << "lp_msb"               << ','
            << "pnat"                 << ','
            << "local_port"           << ','
            << "rtt_support"          << ','
            << "latency_accuracy"     << ','
            << "latency_res"          << ','
            << "local_phy_latency"    << ','
            << "local_mod_dp_latency" << ','
            << "round_trip_latency";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PRTLRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct prtl_reg &prtl = areg.regs.prtl;
    sstream << +prtl.lp_msb              << ','
            << +prtl.pnat                << ','
            << +prtl.local_port          << ','
            << +prtl.rtt_support         << ','
            << +prtl.latency_accuracy    << ','
            << +prtl.latency_res         << ','
            << prtl.local_phy_latency    << ','
            << prtl.local_mod_dp_latency << ','
            << prtl.round_trip_latency   << endl;

    IBDIAGNET_RETURN_VOID;
}

void PRTLRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node,
                                    struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_prtl_reg = &areg.regs.prtl;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***     PCAM Register functions       ***
//*****************************************
PCAMRegister::PCAMRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PCAM,
                (const unpack_data_func_t)pcam_reg_unpack,
                ACC_REG_PCAM_INTERNAL_SECTION_NAME,
                ACC_REG_PCAM_NAME,
                ACC_REG_PCAM_FIELDS_NUM,//how many output columns of the register are going to print
                NSB::get(this),
                ACC_REG_PCAM_FIELDS_HEADER,
                SUPPORT_CA,
                true,
                false,
                VIA_GMP,
                VIA_GMP),
      m_access_reg_group(0), m_feature_group(0)
{
}

AccRegVia_t PCAMRegister::Validation(IBNode *p_node, int &rc)
{
    //TODO take care of
    //HCA_CAP.PCAM == 1

    IBDIAG_RETURN(Register::Validation(p_node, rc));
}

int PCAMRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pcam_reg pcam;
    CLEAR_STRUCT(pcam);

    pcam.access_reg_group   = m_access_reg_group;
    pcam.feature_group      = m_feature_group;

    pcam_reg_pack(&pcam, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PCAMRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer)
{
    IBDIAGNET_ENTER;
    int rc = Register::UnpackData(p_key, data_to_unpack, unpacked_buffer);

    if (!rc) {
        PHYNodeData* data = this->m_phy_diag->GetPHYNodeData(p_key->node_guid);
        if (data && !data->pcam)
            data->pcam = new pcam_reg(*(pcam_reg*)data_to_unpack);
    }

    IBDIAG_RETURN(rc);
}

void PCAMRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                        << '#'
            << handler_header              << ","
            << "access_reg_group"          << ","
            << "feature_group"             << "," 
            << "port_access_reg_cap_mask"  << ","
            << "feature_cap_mask";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PCAMRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct pcam_reg &pcam = areg.regs.pcam;
    sstream << +pcam.access_reg_group      << ','
            << +pcam.feature_group         << ',';

    for (size_t i = 0; i < sizeof(pcam.port_access_reg_cap_mask); i++) {
        if (!(i % 8))
            sstream << ' ';

        sstream << +pcam.port_access_reg_cap_mask[i];
    }

    sstream << ',';

    for (size_t i = 0; i < sizeof(pcam.feature_cap_mask); i++) {
        if (!(i % 8))
            sstream << ' ';

        sstream << +pcam.feature_cap_mask[i];
    }

    sstream << endl;

    IBDIAG_RETURN_VOID;
}

void PCAMRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node)
        p_export_data_phy_node->p_pcam_reg = &areg.regs.pcam;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* PPLL Register functions ***
//*****************************************
PPLLRegister::PPLLRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PPLL,
                (const unpack_data_func_t)ppll_reg_unpack,
                ACC_REG_PPLL_INTERNAL_SECTION_NAME,
                ACC_REG_PPLL_NAME,
                ACC_REG_PPLL_MAX_FIELDS_NUM,
                NSB::get(this))
{
}

int PPLLRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct ppll_reg ppll;
    CLEAR_STRUCT(ppll);
    ppll.pll_group = ((AccRegKeyGroup *)p_key)->group_num;
    ppll_reg_pack(&ppll, data);
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPLLRegister::Header_Dump_28nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "version_28nm=1"    << ','
            << "num_pll_groups"    << ','
            << "pll_group"         << ','
            << "pci_oob_pll"       << ','
            << "num_plls"          << ','
            << "ae";

    for (size_t i = 0; i < ARRAY_SIZE(ppll_28nm::pll_status); ++i)
        sstream << ',' << "pll_28[" << i << "].lock_cal"
                << ',' << "pll_28[" << i << "].lock_status"
                << ',' << "pll_28[" << i << "].algo_f_ctrl"
                << ',' << "pll_28[" << i << "].analog_algo_num_var"
                << ',' << "pll_28[" << i << "].f_ctrl_measure"
                << ',' << "pll_28[" << i << "].analog_var"
                << ',' << "pll_28[" << i << "].high_var"
                << ',' << "pll_28[" << i << "].low_var"
                << ',' << "pll_28[" << i << "].mid_var";

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Header_Dump_16nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "version_16nm=3"    << ','
            << "num_pll_groups"    << ','
            << "pll_group"         << ','
            << "pci_oob_pll"       << ','
            << "num_plls";

    for (size_t i = 0; i < ARRAY_SIZE(ppll_reg_16nm::pll_status); ++i)
        sstream << ',' << "pll_16[" << i << "].lock_status"
                << ',' << "pll_16[" << i << "].pll_pwrup"
                << ',' << "pll_16[" << i << "].lock_pulse"
                << ',' << "pll_16[" << i << "].lock_lost_counter"
                << ',' << "pll_16[" << i << "].ae"
                << ',' << "pll_16[" << i << "].clear_cause"
                << ',' << "pll_16[" << i << "].lock_clk_val_cause"
                << ',' << "pll_16[" << i << "].plllock_clk_val"
                << ',' << "pll_16[" << i << "].clock_valid"
                << ',' << "pll_16[" << i << "].pll_speed"
                << ',' << "pll_16[" << i << "].lock"
                << ',' << "pll_16[" << i << "].analog_var"
                << ',' << "pll_16[" << i << "].high_var"
                << ',' << "pll_16[" << i << "].low_var"
                << ',' << "pll_16[" << i << "].mid_var"
                << ',' << "pll_16[" << i << "].fctrl_measure";

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "version_7nm=4"                        << ','
            << "num_pll_groups"                       << ','
            << "pll_group"                            << ','
            << "pci_oob_pll"                          << ','
            << "num_plls"                             << ','
            << "pll_ugl_state"                        << ','
            << "lokcdet_error_count"                  << ','
            << "lockdet_err_measure_done"             << ','
            << "lockdet_err_state"                    << ','
            << "lockdet_cnt_based_lock"               << ','
            << "lockdet_err_cnt_unlocked_sticky"      << ','
            << "lockdet_err_cnt_locked_sticky"        << ','
            << "lockdet_err_cnt_active_lock"          << ','
            << "cal_internal_state"                   << ','
            << "cal_main_state"                       << ','
            << "pll_lockdet_state"                    << ','
            << "pwr_up_state"                         << ','
            << "cal_abort_sticky"                     << ','
            << "cal_abort"                            << ','
            << "cal_valid_sticky"                     << ','
            << "dco_coarse"                           << ','
            << "cal_done"                             << ','
            << "cal_en"                               << ','
            << "dco_coarse_bin"                       << ','
            << "cal_error_counter"                    << ','
            << "lma_rcal_code"                        << ','
            << "lma_rcal_tx_p_code"                   << ','
            << "lma_rcal_tx_n_code"                   << ','
            << "rcal_done_valid_sticky"               << ','
            << "rcal_tx_p_valid_sticky"               << ','
            << "rcal_tx_n_valid_sticky"               << ','
            << "rcal_fsm_done"                        << ','
            << "speedo_valid"                         << ','
            << "cvb_trim_valid"                       << ','
            << "bg_trim_valid"                        << ','
            << "speedo"                               << ','
            << "cvb_trim"                             << ','
            << "bg_trim";

    IBDIAGNET_RETURN_VOID
}

void PPLLRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    IBDIAGNET_ENTER;

    sstream << "version_5nm=5"                        << ','
            << "num_pll_groups"                       << ','
            << "pll_group"                            << ','
            << "pci_oob_pll"                          << ','
            << "num_plls"                             << ','
            << "lockdet_err_measure_done"             << ','
            << "lockdet_err_state"                    << ','
            << "lockdet_cnt_based_lock"               << ','
            << "lockdet_err_cnt_unlocked_sticky"      << ','
            << "lockdet_err_cnt_locked_sticky"        << ','
            << "lockdet_err_cnt_active_lock"          << ','
            << "dlf2_phlock_det_status_sticky"        << ','
            << "lockdet_error_count"                  << ','
            << "speedo_valid"                         << ','
            << "cvb_trim_valid"                       << ','
            << "bg_trim_valid"                        << ','
            << "lockdet_mode_retimer"                 << ','
            << "speedo"                               << ','
            << "cvb_trim"                             << ','
            << "bg_trim";

    IBDIAGNET_RETURN_VOID
}

void PPLLRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_28nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct ppll_reg &ppll = areg.regs.ppll;

    sstream << hex
            << "0x" << +ppll.version           << ','
            << "0x" << +ppll.num_pll_groups    << ','
            << "0x" << +ppll.pll_group         << ','
            << "0x" << +ppll.pci_oob_pll       << ','
            << "0x" << +ppll.num_plls          << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (ppll.version) {

        case 1: // 28nm
            Dump_28nm(ppll, sstream);
            break;

        case 3: // 16nm
            Dump_16nm(ppll, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(ppll, sstream);
            break;

        case 5: // 5nm
            Dump_5nm(ppll, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +ppll.version << '"';
            for (int i = 0; i < ACC_REG_PPLL_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for PPLL: %d, on node: " U64H_FMT ".\n",
                        ppll.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_28nm(const struct ppll_reg &ppll, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    ppll_28nm ppll_28;
    ppll_28nm_unpack(&ppll_28, ppll.page_data.ppll_data_set.data);

    sstream << "0x" << +ppll_28.ae;

    int pll_status_size = ARRAY_SIZE(ppll_28.pll_status);

    for (int i = 0; i < pll_status_size; ++i) {
        sstream << ',';
        Dump_pll_28nm(ppll_28.pll_status[i], sstream);
    }

    for (int i = 0; i < ACC_REG_PPLL_MAX_FIELDS_NUM - ACC_REG_PPLL28NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_16nm(const struct ppll_reg &ppll, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    ppll_reg_16nm ppll_16;
    ppll_reg_16nm_unpack(&ppll_16, ppll.page_data.ppll_data_set.data);

    int pll_status_size = ARRAY_SIZE(ppll_16.pll_status);

    for (int i = 0; i < pll_status_size; ++i) {
        if (i)
            sstream << ',';
        Dump_pll_16nm(ppll_16.pll_status[i], sstream);
    }

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_7nm(const struct ppll_reg &ppll, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    ppll_reg_7nm ppll_7;
    ppll_reg_7nm_unpack(&ppll_7, ppll.page_data.ppll_data_set.data);

    sstream << "0x" << +ppll_7.pll_ugl_state                        << ','
            << "0x" << +ppll_7.lokcdet_error_count                  << ','
            << "0x" << +ppll_7.lockdet_err_measure_done             << ','
            << "0x" << +ppll_7.lockdet_err_state                    << ','
            << "0x" << +ppll_7.lockdet_cnt_based_lock               << ','
            << "0x" << +ppll_7.lockdet_err_cnt_unlocked_sticky      << ','
            << "0x" << +ppll_7.lockdet_err_cnt_locked_sticky        << ','
            << "0x" << +ppll_7.lockdet_err_cnt_active_lock          << ','
            << "0x" << +ppll_7.cal_internal_state                   << ','
            << "0x" << +ppll_7.cal_main_state                       << ','
            << "0x" << +ppll_7.pll_lockdet_state                    << ','
            << "0x" << +ppll_7.pwr_up_state                         << ','
            << "0x" << +ppll_7.cal_abort_sticky                     << ','
            << "0x" << +ppll_7.cal_abort                            << ','
            << "0x" << +ppll_7.cal_valid_sticky                     << ','
            << "0x" << +ppll_7.dco_coarse                           << ','
            << "0x" << +ppll_7.cal_done                             << ','
            << "0x" << +ppll_7.cal_en                               << ','
            << "0x" << +ppll_7.dco_coarse_bin                       << ','
            << "0x" << +ppll_7.cal_error_counter                    << ','
            << "0x" << +ppll_7.lma_rcal_code                        << ','
            << "0x" << +ppll_7.lma_rcal_tx_p_code                   << ','
            << "0x" << +ppll_7.lma_rcal_tx_n_code                   << ','
            << "0x" << +ppll_7.rcal_done_valid_sticky               << ','
            << "0x" << +ppll_7.rcal_tx_p_valid_sticky               << ','
            << "0x" << +ppll_7.rcal_tx_n_valid_sticky               << ','
            << "0x" << +ppll_7.rcal_fsm_done                        << ','
            << "0x" << +ppll_7.speedo_valid                         << ','
            << "0x" << +ppll_7.cvb_trim_valid                       << ','
            << "0x" << +ppll_7.bg_trim_valid                        << ','
            << "0x" << +ppll_7.speedo                               << ','
            << "0x" << +ppll_7.cvb_trim                             << ','
            << "0x" << +ppll_7.bg_trim                              ;

    for (int i = 0; i < ACC_REG_PPLL_MAX_FIELDS_NUM - ACC_REG_PPLL7NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_5nm(const struct ppll_reg &ppll, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    ppll_reg_5nm ppll_5;
    ppll_reg_5nm_unpack(&ppll_5, ppll.page_data.ppll_data_set.data);

    sstream << "0x" << +ppll_5.lockdet_err_measure_done                    << ','
            << "0x" << +ppll_5.lockdet_err_state                           << ','
            << "0x" << +ppll_5.lockdet_cnt_based_lock                      << ','
            << "0x" << +ppll_5.lockdet_err_cnt_unlocked_sticky             << ','
            << "0x" << +ppll_5.lockdet_err_cnt_locked_sticky               << ','
            << "0x" << +ppll_5.lockdet_err_cnt_active_lock                 << ','
            << "0x" << +ppll_5.dlf2_phlock_det_status_sticky               << ','
            << "0x" << +ppll_5.lockdet_error_count                         << ','
            << "0x" << +ppll_5.speedo_valid                                << ','
            << "0x" << +ppll_5.cvb_trim_valid                              << ','
            << "0x" << +ppll_5.bg_trim_valid                               << ','
            << "0x" << +ppll_5.lockdet_mode_retimer                        << ','
            << "0x" << +ppll_5.speedo                                      << ','
            << "0x" << +ppll_5.cvb_trim                                    << ','
            << "0x" << +ppll_5.bg_trim;

    for (int i = 0; i < ACC_REG_PPLL_MAX_FIELDS_NUM - ACC_REG_PPLL5NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_pll_28nm(const struct pll_28nm &pll_28, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "0x" << +pll_28.lock_cal               << ','
            << "0x" << +pll_28.lock_status            << ','
            << "0x" << +pll_28.algo_f_ctrl            << ','
            << "0x" << +pll_28.analog_algo_num_var    << ','
            << "0x" << +pll_28.f_ctrl_measure         << ','
            << "0x" << +pll_28.analog_var             << ','
            << "0x" << +pll_28.high_var               << ','
            << "0x" << +pll_28.low_var                << ','
            << "0x" << +pll_28.mid_var                ;

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::Dump_pll_16nm(const struct pll_status_data &pll_16, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "0x" << +pll_16.lock_status             << ','
            << "0x" << +pll_16.pll_pwrup               << ','
            << "0x" << +pll_16.lock_pulse              << ','
            << "0x" << +pll_16.lock_lost_counter       << ','
            << "0x" << +pll_16.ae                      << ','
            << "0x" << +pll_16.clear_cause             << ','
            << "0x" << +pll_16.lock_clk_val_cause      << ','
            << "0x" << +pll_16.plllock_clk_val         << ','
            << "0x" << +pll_16.clock_valid             << ','
            << "0x" << +pll_16.pll_speed               << ','
            << "0x" << +pll_16.lock                    << ','
            << "0x" << +pll_16.analog_var              << ','
            << "0x" << +pll_16.high_var                << ','
            << "0x" << +pll_16.low_var                 << ','
            << "0x" << +pll_16.mid_var                 << ','
            << "0x" << +pll_16.fctrl_measure           ;

    IBDIAGNET_RETURN_VOID;
}

void PPLLRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node) {
        struct ppll_reg *p_ppll_reg = new struct ppll_reg(areg.regs.ppll);

        switch (p_ppll_reg->version) {
            case 0: // 28nm
                ppll_28nm_unpack(&p_ppll_reg->page_data.ppll_28nm, areg.regs.ppll.page_data.ppll_data_set.data);
                break;
            case 1: // 16nm
                ppll_reg_16nm_unpack(&p_ppll_reg->page_data.ppll_reg_16nm, areg.regs.ppll.page_data.ppll_data_set.data);
                break;
            case 4: // 7nm
                ppll_reg_7nm_unpack(&p_ppll_reg->page_data.ppll_reg_7nm, areg.regs.ppll.page_data.ppll_data_set.data);
                break;
            case 5: // 5nm
                ppll_reg_5nm_unpack(&p_ppll_reg->page_data.ppll_reg_5nm, areg.regs.ppll.page_data.ppll_data_set.data);
                break;
            default:
                break;
        }

        p_export_data_phy_node->p_ppll_reg[((AccRegKeyGroup *)p_key)->group_num] = p_ppll_reg;
    }

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* PTAS Register functions ***
//*****************************************
PTASRegister::PTASRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PTAS,
                (const unpack_data_func_t)ptas_reg_unpack,
                ACC_REG_PTAS_INTERNAL_SECTION_NAME,
                ACC_REG_PTAS_NAME,
                ACC_REG_PTAS_FIELDS_NUM,
                NSB::get(this))
{
}

int PTASRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PTASRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                             <<  '#'
            << handler_header                   <<  ','
            << "algorithm_options"              <<  ','
            << "repetitions_mode"               <<  ','
            << "num_of_repetitions"             <<  ','
            << "grade_version"                  <<  ','
            << "height_grade_type"              <<  ','
            << "phase_grade_type"               <<  ','
            << "height_grade_weight"            <<  ','
            << "phase_grade_weight"             <<  ','
            << "gisim_measure_bits"             <<  ','
            << "adaptive_tap_measure_bits"      <<  ','
            << "ber_bath_high_error_threshold"  <<  ','
            << "ber_bath_mid_error_threshold"   <<  ','
            << "ber_bath_low_error_threshold"   <<  ','
            << "one_ratio_high_threshold"       <<  ','
            << "one_ratio_high_mid_threshold"   <<  ','
            << "one_ratio_low_mid_threshold"    <<  ','
            << "one_ratio_low_threshold"        <<  ','
            << "ndeo_error_threshold"           <<  ','
            << "mixer_offset_step_size"         <<  ','
            << "mix90_phase_for_voltage_bath"   <<  ','
            << "mixer_offset_start"             <<  ','
            << "mixer_offset_end"               <<  ','
            << "ber_test_time";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PTASRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    char buffer[1024] = {0};
    const struct ptas_reg & p_ptas = areg.regs.ptas;
    //register fields
    snprintf(buffer, sizeof(buffer),
             UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
             UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
             UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
             UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT "," UH_FMT ","
             UH_FMT "," UH_FMT "," UH_FMT ,
             p_ptas.algorithm_options,
             p_ptas.repetitions_mode,
             p_ptas.num_of_repetitions,
             p_ptas.grade_version,
             p_ptas.height_grade_type,
             p_ptas.phase_grade_type,
             p_ptas.height_grade_weight,
             p_ptas.phase_grade_weight,
             p_ptas.gisim_measure_bits,
             p_ptas.adaptive_tap_measure_bits,
             p_ptas.ber_bath_high_error_threshold,
             p_ptas.ber_bath_mid_error_threshold,
             p_ptas.ber_bath_low_error_threshold,
             p_ptas.one_ratio_high_threshold,
             p_ptas.one_ratio_high_mid_threshold,
             p_ptas.one_ratio_low_mid_threshold,
             p_ptas.one_ratio_low_threshold,
             p_ptas.ndeo_error_threshold,
             p_ptas.mixer_offset_step_size,
             p_ptas.mix90_phase_for_voltage_bath,
             p_ptas.mixer_offset_start,
             p_ptas.mixer_offset_end,
             p_ptas.ber_test_time);
    sstream << buffer << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* PMDR Register functions ***
//*****************************************
PMDRRegister::PMDRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PMDR,
                (const unpack_data_func_t)pmdr_reg_unpack,
                ACC_REG_PMDR_INTERNAL_SECTION_NAME,
                ACC_REG_PMDR_NAME,
                ACC_REG_PMDR_FIELDS_NUM,
                NSB::get(this),
                ACC_REG_PMDR_FIELDS_HEADER,
                SUPPORT_SW_CA,
                true,
                false,
                VIA_GMP,
                VIA_GMP),
       m_pnat(ACC_REG_PNAT_IB_PORT)
{
}

int PMDRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pmdr_reg pmdr;
    CLEAR_STRUCT(pmdr);
    pmdr.pnat = this->m_pnat;

    int rc = FillLocalPortAndPlaneInd(p_key, &pmdr.local_port, &pmdr.plane_ind);

    if (rc)
        return rc;

    pmdr_reg_pack(&pmdr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PMDRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS    
    sstream << endl                    << '#'
            << handler_header          << ','
            << "status"                << ","
            << "version"               << ","
            << "gb_valid"              << ","
            << "mcm_tile_valid"        << ","
            << "scc_valid"             << ","
            << "plane_ind"             << ","
            << "pport"                 << ","
            << "module"                << ","
            << "cluster"               << ","
            << "pport_msb"             << ","
            << "clp_4x"                << ","
            << "clp_1x"                << ","
            << "gb_dp_num"             << ","
            << "split"                 << ","
            << "swid"                  << ","
            << "module_lane_mask"      << ","
            << "ib_port"               << ","
            << "label_port_query"      << ","
            << "local_port_query"      << ","
            << "usr_sd_4x_tx"          << ","
            << "usr_sd_4x_rx"          << ","
            << "tile_pport_msb"        << ","
            << "lbp_query_msb"         << ","
            << "lp_query_msb"          << ","
            << "slot_index"            << ","
            << "tile_cluster"          << ","
            << "mcm_tile_num"          << ","
            << "pll_cnt_rx"            << ","
            << "tile_pport"            << ","
            << "gearbox_die_num"       << ","
            << "vl_num"                << ","
            << "pll_cnt_tx"            << ","
            << "lane7_physical_rx"     << ","
            << "lane6_physical_rx"     << ","
            << "lane5_physical_rx"     << ","
            << "lane4_physical_rx"     << ","
            << "lane3_physical_rx"     << ","
            << "lane2_physical_rx"     << ","
            << "lane1_physical_rx"     << ","
            << "lane0_physical_rx"     << ","
            << "pll_index"             << ","
            << "lane7_physical_tx"     << ","
            << "lane6_physical_tx"     << ","
            << "lane5_physical_tx"     << ","
            << "lane4_physical_tx"     << ","
            << "lane3_physical_tx"     << ","
            << "lane2_physical_tx"     << ","
            << "lane1_physical_tx"     << ","
            << "lane0_physical_tx"     << ","
            << "VL7_lane_map"          << ","
            << "VL6_lane_map"          << ","
            << "VL5_lane_map"          << ","
            << "VL4_lane_map"          << ","
            << "VL3_lane_map"          << ","
            << "VL2_lane_map"          << ","
            << "VL1_lane_map"          << ","
            << "VL0_lane_map"          << ","
            << "VL15_lane_map"         << ","
            << "VL14_lane_map"         << ","
            << "VL13_lane_map"         << ","
            << "VL12_lane_map"         << ","
            << "VL11_lane_map"         << ","
            << "VL10_lane_map"         << ","
            << "VL9_lane_map"          << ","
            << "VL8_lane_map"          << ","
            << "VL23_lane_map"         << ","
            << "VL22_lane_map"         << ","
            << "VL21_lane_map"         << ","
            << "VL20_lane_map"         << ","
            << "VL19_lane_map"         << ","
            << "VL18_lane_map"         << ","
            << "VL17_lane_map"         << ","
            << "VL16_lane_map"         << ","
            << "VL31_lane_map"         << ","
            << "VL30_lane_map"         << ","
            << "VL29_lane_map"         << ","
            << "VL28_lane_map"         << ","
            << "VL27_lane_map"         << ","
            << "VL26_lane_map"         << ","
            << "VL25_lane_map"         << ","
            << "VL24_lane_map"         << ","
            << "physical_tx_cluster"   << ","
            << "physical_rx_cluster"   << ","
            << "scc_hs_lanes"          << ","
            << "lane7_ph_tx_msb"       << ","
            << "lane6_ph_tx_msb"       << ","
            << "lane5_ph_tx_msb"       << ","
            << "lane4_ph_tx_msb"       << ","
            << "lane3_ph_tx_msb"       << ","
            << "lane2_ph_tx_msb"       << ","
            << "lane1_ph_tx_msb"       << ","
            << "lane0_ph_tx_msb"       << ","
            << "lane7_ph_rx_msb"       << ","
            << "lane6_ph_rx_msb"       << ","
            << "lane5_ph_rx_msb"       << ","
            << "lane4_ph_rx_msb"       << ","
            << "lane3_ph_rx_msb"       << ","
            << "lane2_ph_rx_msb"       << ","
            << "lane1_ph_rx_msb"       << ","
            << "lane0_ph_rx_msb"       << ","
            << "ph_cluster_tx_cnt"     << ","
            << "ph_cluster_rx_cnt"     << ","
            << "lane7_cluster_tx"      << ","
            << "lane6_cluster_tx"      << ","
            << "lane5_cluster_tx"      << ","
            << "lane4_cluster_tx"      << ","
            << "lane3_cluster_tx"      << ","
            << "lane2_cluster_tx"      << ","
            << "lane1_cluster_tx"      << ","
            << "lane0_cluster_tx"      << ","
            << "pll_index_rx"          << ","
            << "lane7_cluster_rx"      << ","
            << "lane6_cluster_rx"      << ","
            << "lane5_cluster_rx"      << ","
            << "lane4_cluster_rx"      << ","
            << "lane3_cluster_rx"      << ","
            << "lane2_cluster_rx"      << ","
            << "lane1_cluster_rx"      << ","
            << "lane0_cluster_rx";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PMDRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct pmdr_reg &pmdr = areg.regs.pmdr;

    //register fields
    sstream << +pmdr.status                << ","
            << +pmdr.version               << ","
            << +pmdr.gb_valid              << ","
            << +pmdr.mcm_tile_valid        << ","
            << +pmdr.scc_valid             << ","
            << +pmdr.plane_ind             << ","
            << +pmdr.pport                 << ","
            << +pmdr.module                << ","
            << +pmdr.cluster               << ","
            << +pmdr.pport_msb             << ","
            << +pmdr.clp_4x                << ","
            << +pmdr.clp_1x                << ","
            << +pmdr.gb_dp_num             << ","
            << +pmdr.split                 << ","
            << +pmdr.swid                  << ","
            << +pmdr.module_lane_mask      << ","
            << +pmdr.ib_port               << ","
            << +pmdr.label_port_query      << ","
            << +pmdr.local_port_query      << ","
            << +pmdr.usr_sd_4x_tx          << ","
            << +pmdr.usr_sd_4x_rx          << ","
            << +pmdr.tile_pport_msb        << ","
            << +pmdr.lbp_query_msb         << ","
            << +pmdr.lp_query_msb          << ","
            << +pmdr.slot_index            << ","
            << +pmdr.tile_cluster          << ","
            << +pmdr.mcm_tile_num          << ","
            << +pmdr.pll_cnt_rx            << ","
            << +pmdr.tile_pport            << ","
            << +pmdr.gearbox_die_num       << ","
            << +pmdr.vl_num                << ","
            << +pmdr.pll_cnt_tx            << ","
            << +pmdr.lane7_physical_rx     << ","
            << +pmdr.lane6_physical_rx     << ","
            << +pmdr.lane5_physical_rx     << ","
            << +pmdr.lane4_physical_rx     << ","
            << +pmdr.lane3_physical_rx     << ","
            << +pmdr.lane2_physical_rx     << ","
            << +pmdr.lane1_physical_rx     << ","
            << +pmdr.lane0_physical_rx     << ","
            << +pmdr.pll_index             << ","
            << +pmdr.lane7_physical_tx     << ","
            << +pmdr.lane6_physical_tx     << ","
            << +pmdr.lane5_physical_tx     << ","
            << +pmdr.lane4_physical_tx     << ","
            << +pmdr.lane3_physical_tx     << ","
            << +pmdr.lane2_physical_tx     << ","
            << +pmdr.lane1_physical_tx     << ","
            << +pmdr.lane0_physical_tx     << ","
            << +pmdr.VL7_lane_map          << ","
            << +pmdr.VL6_lane_map          << ","
            << +pmdr.VL5_lane_map          << ","
            << +pmdr.VL4_lane_map          << ","
            << +pmdr.VL3_lane_map          << ","
            << +pmdr.VL2_lane_map          << ","
            << +pmdr.VL1_lane_map          << ","
            << +pmdr.VL0_lane_map          << ","
            << +pmdr.VL15_lane_map         << ","
            << +pmdr.VL14_lane_map         << ","
            << +pmdr.VL13_lane_map         << ","
            << +pmdr.VL12_lane_map         << ","
            << +pmdr.VL11_lane_map         << ","
            << +pmdr.VL10_lane_map         << ","
            << +pmdr.VL9_lane_map          << ","
            << +pmdr.VL8_lane_map          << ","
            << +pmdr.VL23_lane_map         << ","
            << +pmdr.VL22_lane_map         << ","
            << +pmdr.VL21_lane_map         << ","
            << +pmdr.VL20_lane_map         << ","
            << +pmdr.VL19_lane_map         << ","
            << +pmdr.VL18_lane_map         << ","
            << +pmdr.VL17_lane_map         << ","
            << +pmdr.VL16_lane_map         << ","
            << +pmdr.VL31_lane_map         << ","
            << +pmdr.VL30_lane_map         << ","
            << +pmdr.VL29_lane_map         << ","
            << +pmdr.VL28_lane_map         << ","
            << +pmdr.VL27_lane_map         << ","
            << +pmdr.VL26_lane_map         << ","
            << +pmdr.VL25_lane_map         << ","
            << +pmdr.VL24_lane_map         << ","
            << +pmdr.physical_tx_cluster   << ","
            << +pmdr.physical_rx_cluster   << ","
            << +pmdr.scc_hs_lanes          << ","
            << +pmdr.lane7_ph_tx_msb       << ","
            << +pmdr.lane6_ph_tx_msb       << ","
            << +pmdr.lane5_ph_tx_msb       << ","
            << +pmdr.lane4_ph_tx_msb       << ","
            << +pmdr.lane3_ph_tx_msb       << ","
            << +pmdr.lane2_ph_tx_msb       << ","
            << +pmdr.lane1_ph_tx_msb       << ","
            << +pmdr.lane0_ph_tx_msb       << ","
            << +pmdr.lane7_ph_rx_msb       << ","
            << +pmdr.lane6_ph_rx_msb       << ","
            << +pmdr.lane5_ph_rx_msb       << ","
            << +pmdr.lane4_ph_rx_msb       << ","
            << +pmdr.lane3_ph_rx_msb       << ","
            << +pmdr.lane2_ph_rx_msb       << ","
            << +pmdr.lane1_ph_rx_msb       << ","
            << +pmdr.lane0_ph_rx_msb       << ","
            << +pmdr.ph_cluster_tx_cnt     << ","
            << +pmdr.ph_cluster_rx_cnt     << ","
            << +pmdr.lane7_cluster_tx      << ","
            << +pmdr.lane6_cluster_tx      << ","
            << +pmdr.lane5_cluster_tx      << ","
            << +pmdr.lane4_cluster_tx      << ","
            << +pmdr.lane3_cluster_tx      << ","
            << +pmdr.lane2_cluster_tx      << ","
            << +pmdr.lane1_cluster_tx      << ","
            << +pmdr.lane0_cluster_tx      << ","
            << +pmdr.pll_index_rx          << ","
            << +pmdr.lane7_cluster_rx      << ","
            << +pmdr.lane6_cluster_rx      << ","
            << +pmdr.lane5_cluster_rx      << ","
            << +pmdr.lane4_cluster_rx      << ","
            << +pmdr.lane3_cluster_rx      << ","
            << +pmdr.lane2_cluster_rx      << ","
            << +pmdr.lane1_cluster_rx      << ","
            << +pmdr.lane0_cluster_rx
            << endl;

    IBDIAGNET_RETURN_VOID;
}

void PMDRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pmdr_reg = &areg.regs.pmdr;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* PMCR Register functions ***
//*****************************************
PMCRRegister::PMCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PMCR,
                (const unpack_data_func_t)pmcr_reg_unpack,
                ACC_REG_PMCR_INTERNAL_SECTION_NAME,
                ACC_REG_PMCR_NAME,
                ACC_REG_PMCR_FIELDS_NUM,
                NSB::get(this))
{
}

int PMCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pmcr_reg pmcr;
    CLEAR_STRUCT(pmcr);

    int rc = FillLocalPortAndPlaneInd(p_key, &pmcr.local_port, &pmcr.plane_ind);

    if (rc)
        return rc;

    pmcr_reg_pack(&pmcr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PMCRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                             << '#'
            << handler_header                   << ','
            << "cs_sel"                         << ','
            << "local_port"                     << ','
            << "lp_msb"                         << ','
            << "error_code_response"            << ','
            << "plane_ind"                      << ','
            << "db"                             << ','
            << "dir_dp_ctrl"                    << ','
            << "dir_tx_dis_ctrl"                << ','
            << "dir_dp_ctrl_cap"                << ','
            << "dir_tx_dis_ctrl_cap"            << ','
            << "tx_disable_override_value"      << ','
            << "tx_disable_override_cap"        << ','
            << "tx_disable_override_cntl"       << ','
            << "cdr_override_value"             << ','
            << "cdr_override_cap"               << ','
            << "cdr_override_cntl"              << ','
            << "rx_los_override_admin"          << ','
            << "rx_los_override_cap"            << ','
            << "rx_los_override_cntl"           << ','
            << "rx_amp_override_value_cap"      << ','
            << "rx_amp_override_value"          << ','
            << "rx_amp_override_cap"            << ','
            << "rx_amp_override_cntl"           << ','
            << "rx_post_emp_override_value_cap" << ','
            << "rx_post_emp_override_value"     << ','
            << "rx_post_emp_override_cap"       << ','
            << "rx_post_emp_override_cntl"      << ','
            << "rx_emp_override_value_cap"      << ','
            << "rx_emp_override_value"          << ','
            << "rx_emp_override_cap"            << ','
            << "rx_emp_override_cntl"           << ','
            << "tx_adaptive_override_cap"       << ','
            << "tx_adaptive_override_cntrl"     << ','
            << "tx_adaptive_override_value"     << ','
            << "tx_equ_override_value_cap"      << ','
            << "tx_equ_override_value"          << ','
            << "tx_equ_override_cap"            << ','
            << "tx_equ_override_cntl"           << ','
            << "rs_override_value"              << ','
            << "rs_override_cntl"               << ','
            << "agc_override_cap"               << ','
            << "agc_override_cntl"              << ','
            << "ap_sel_override_value"          << ','
            << "ap_sel_override_cntrl"          << ','
            << "agc_target_value"               ;
#endif

    IBDIAGNET_RETURN_VOID;

}

void PMCRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct pmcr_reg &pmcr = areg.regs.pmcr;

    //register fields
    sstream << +pmcr.cs_sel                         << ','
            << +pmcr.local_port                     << ','
            << +pmcr.lp_msb                         << ','
            << +pmcr.error_code_response            << ','
            << +pmcr.plane_ind                      << ','
            << +pmcr.db                             << ','
            << +pmcr.dir_dp_ctrl                    << ','
            << +pmcr.dir_tx_dis_ctrl                << ','
            << +pmcr.dir_dp_ctrl_cap                << ','
            << +pmcr.dir_tx_dis_ctrl_cap            << ','
            << +pmcr.tx_disable_override_value      << ','
            << +pmcr.tx_disable_override_cap        << ','
            << +pmcr.tx_disable_override_cntl       << ','
            << +pmcr.cdr_override_value             << ','
            << +pmcr.cdr_override_cap               << ','
            << +pmcr.cdr_override_cntl              << ','
            << +pmcr.rx_los_override_admin          << ','
            << +pmcr.rx_los_override_cap            << ','
            << +pmcr.rx_los_override_cntl           << ','
            << +pmcr.rx_amp_override_value_cap      << ','
            << +pmcr.rx_amp_override_value          << ','
            << +pmcr.rx_amp_override_cap            << ','
            << +pmcr.rx_amp_override_cntl           << ','
            << +pmcr.rx_post_emp_override_value_cap << ','
            << +pmcr.rx_post_emp_override_value     << ','
            << +pmcr.rx_post_emp_override_cap       << ','
            << +pmcr.rx_post_emp_override_cntl      << ','
            << +pmcr.rx_emp_override_value_cap      << ','
            << +pmcr.rx_emp_override_value          << ','
            << +pmcr.rx_emp_override_cap            << ','
            << +pmcr.rx_emp_override_cntl           << ','
            << +pmcr.tx_adaptive_override_cap       << ','
            << +pmcr.tx_adaptive_override_cntrl     << ','
            << +pmcr.tx_adaptive_override_value     << ','
            << +pmcr.tx_equ_override_value_cap      << ','
            << +pmcr.tx_equ_override_value          << ','
            << +pmcr.tx_equ_override_cap            << ','
            << +pmcr.tx_equ_override_cntl           << ','
            << +pmcr.rs_override_value              << ','
            << +pmcr.rs_override_cntl               << ','
            << +pmcr.agc_override_cap               << ','
            << +pmcr.agc_override_cntl              << ','
            << +pmcr.ap_sel_override_value          << ','
            << +pmcr.ap_sel_override_cntrl          << ','
            << +pmcr.agc_target_value               << endl;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* PTYS Register functions ***
//*****************************************
PTYSRegister::PTYSRegister(PhyDiag * phy_diag)
    : Register(phy_diag, ACCESS_REGISTER_ID_PTYS,
                (const unpack_data_func_t)ptys_reg_unpack,
                ACC_REG_PTYS_INTERNAL_SECTION_NAME,
                ACC_REG_PTYS_NAME,
                ACC_REG_PTYS_FIELDS_NUM,
                NSB::get(this),
                "",
                SUPPORT_SW_CA,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP),
    m_pnat(ACC_REG_PNAT_IB_PORT)
{
}

int PTYSRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct ptys_reg ptys;
    CLEAR_STRUCT(ptys);

    ptys.pnat = this->m_pnat;
    int rc = FillLocalPortAndPlaneInd(p_key, &ptys.local_port, &ptys.plane_ind);

    if (rc)
        return rc;

    ptys_reg_pack(&ptys, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PTYSRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                       << '#'
            << handler_header             << ','
            << "reserved_high"            << ','
            << "an_disable_admin"         << ','
            << "an_disable_cap"           << ','
            << "ee_tx_ready"              << ','
            << "tx_ready_e"               << ','
            << "dis_optic_neg_admin"      << ','
            << "force_tx_aba_param"       << ','
            << "local_port"               << ','
            << "pnat"                     << ','
            << "lp_msb"                   << ','
            << "port_type"                << ','
            << "plane_ind"                << ','
            << "transmit_allowed"         << ','
            << "proto_mask"               << ','
            << "an_status"                << ','
            << "max_port_rate"            << ','
            << "data_rate_oper"           << ','
            << "ext_eth_proto_capability" << ','
            << "eth_proto_capability"     << ','
            << "ib_link_width_capability" << ','
            << "ib_proto_capability"      << ','
            << "ext_eth_proto_admin"      << ','
            << "eth_proto_admin"          << ','
            << "ib_link_width_admin"      << ','
            << "ib_proto_admin"           << ','
            << "ext_eth_proto_oper"       << ','
            << "eth_proto_oper"           << ','
            << "ib_link_width_oper"       << ','
            << "ib_proto_oper"            << ','
            << "xdr_2x_slow_cap"          << ','
            << "force_lt_frames_cap"      << ','
            << "force_lt_frames_admin"    << ','
            << "ext_eth_proto_lp_valid"   << ','
            << "eth_proto_lp_valid"       << ','
            << "xdr_2x_slow_admin"        << ','
            << "xdr_2x_slow_active"       << ','
            << "lane_rate_oper"           << ','
            << "connector_type"           << ','
            << "lp_advertised_proto_ext"  << ','
            << "eth_proto_lp_advertise"   << ','
            << "ext_ib_proto_cap"         << ','
            << "ext_ib_proto_admin"       << ','
            << "ext_ib_proto_oper"        ;
#endif

    IBDIAGNET_RETURN_VOID;

}


void PTYSRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct ptys_reg &ptys = areg.regs.ptys;

    //register fields
    sstream << +ptys.reserved_high                 << ','
            << +ptys.an_disable_admin              << ','
            << +ptys.an_disable_cap                << ','
            << +ptys.ee_tx_ready                   << ','
            << +ptys.tx_ready_e                    << ','
            << +ptys.dis_optic_neg_admin           << ','
            << +ptys.force_tx_aba_param            << ','
            << +ptys.local_port                    << ','
            << +ptys.pnat                          << ','
            << +ptys.lp_msb                        << ','
            << +ptys.port_type                     << ','
            << +ptys.plane_ind                     << ','
            << +ptys.transmit_allowed              << ','
            << +ptys.proto_mask                    << ','
            << +ptys.an_status                     << ','
            << +ptys.max_port_rate                 << ','
            << +ptys.data_rate_oper                << ','
            << +ptys.ext_eth_proto_capability      << ','
            << +ptys.eth_proto_capability          << ','
            << +ptys.ib_link_width_capability      << ','
            << +ptys.ib_proto_capability           << ','
            << +ptys.ext_eth_proto_admin           << ','
            << +ptys.eth_proto_admin               << ','
            << +ptys.ib_link_width_admin           << ','
            << +ptys.ib_proto_admin                << ','
            << +ptys.ext_eth_proto_oper            << ','
            << +ptys.eth_proto_oper                << ','
            << +ptys.ib_link_width_oper            << ','
            << +ptys.ib_proto_oper                 << ','
            << +ptys.xdr_2x_slow_cap               << ','
            << +ptys.force_lt_frames_cap           << ','
            << +ptys.force_lt_frames_admin         << ','
            << +ptys.ext_eth_proto_lp_valid        << ','
            << +ptys.eth_proto_lp_valid            << ','
            << +ptys.xdr_2x_slow_admin             << ','
            << +ptys.xdr_2x_slow_active            << ','
            << +ptys.lane_rate_oper                << ','
            << +ptys.connector_type                << ','
            << +ptys.lp_advertised_proto_ext.
                ptys_lp_ext_ib.
                link_partner_protocol_extended_ib  << ','
            << +ptys.eth_proto_lp_advertise        << ','
            << +ptys.ext_ib_proto_cap              << ','
            << +ptys.ext_ib_proto_admin            << ','
            << +ptys.ext_ib_proto_oper             << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void PTYSRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ptys_reg = &areg.regs.ptys;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* PPSLS Register functions ***
//*****************************************
PPSLSRegister::PPSLSRegister(PhyDiag * phy_diag)
    : Register(phy_diag, ACCESS_REGISTER_ID_PPSLS,
                (const unpack_data_func_t)ppsls_reg_unpack,
                ACC_REG_PPSLS_INTERNAL_SECTION_NAME,
                ACC_REG_PPSLS_NAME,
                ACC_REG_PPSLS_FIELDS_NUM,
                NSB::get(this),
                "",
                SUPPORT_SW_CA,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP),
    m_pnat(ACC_REG_PNAT_IB_PORT)
{
}

int PPSLSRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->type == IB_SW_NODE && p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);
    
    if (p_node->type == IB_CA_NODE &&  p_node->ext_type != IB_GPU_NODE)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct ppsls_reg reg;
    CLEAR_STRUCT(reg);

    reg.pnat        = this->m_pnat;

    int rc = FillLocalPortAndPlaneInd(p_key, &reg.local_port, nullptr);

    if (rc)
        return rc;

    ppsls_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPSLSRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                             << '#'
            << handler_header                   << ','
            << "l1_fw_mode_cap"                 << ','
            << "l1_cap"                         << ','
            << "fw_mode_remote"                 << ','
            << "fw_mode_act"                    << ','
            << "fw_mode_neg_status"             << ','
            << "l1_neg_status";
#endif

    IBDIAGNET_RETURN_VOID;

}

void PPSLSRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct ppsls_reg &reg = areg.regs.ppsls;

    //register fields
    sstream << +reg.l1_fw_mode_cap                 << ','
            << +reg.l1_cap                         << ','
            << +reg.fw_mode_remote                 << ','
            << +reg.fw_mode_act                    << ','
            << +reg.fw_mode_neg_status             << ','
            << +reg.l1_neg_status                  << endl;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* MPSCR Register functions ***
//*****************************************
MPSCRRegister::MPSCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag, ACCESS_REGISTER_ID_MPSCR,
                (const unpack_data_func_t)mpscr_reg_unpack,
                ACC_REG_MPSCR_INTERNAL_SECTION_NAME,
                ACC_REG_MPSCR_NAME,
                ACC_REG_MPSCR_FIELDS_NUM,
                NSB::get(this),
                "",
                SUPPORT_SW_CA,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP)
{
}

int MPSCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->type == IB_SW_NODE && p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);
    
    if (p_node->type == IB_CA_NODE &&  p_node->ext_type != IB_GPU_NODE)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct mpscr_reg reg;
    CLEAR_STRUCT(reg);

    //This would clear counters in MTSR Register
    reg.cc = this->m_phy_diag->ShouldResetCounters();

    mpscr_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MPSCRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                             << '#'
            << handler_header                   << ','
            << "warning_inactive_time"          << ','
            << "warning_active_time"            << ','
            << "critical_inactive_time"         << ','
            << "critical_active_time"           << ','
            << "cc"                             << ','
            << "queue_depth_th";
#endif

    IBDIAGNET_RETURN_VOID;

}

void MPSCRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct mpscr_reg &reg = areg.regs.mpscr;

    //register fields
    sstream << +reg.warning_inactive_time     << ','
            << +reg.warning_active_time       << ','
            << +reg.critical_inactive_time    << ','
            << +reg.critical_active_time      << ','
            << +reg.cc                        << ','
            << +reg.queue_depth_th            << endl;


    IBDIAGNET_RETURN_VOID;
}



//*****************************************
//************* MTSR Register functions ***
//*****************************************
MTSRRegister::MTSRRegister(PhyDiag * phy_diag)
    : Register(phy_diag, ACCESS_REGISTER_ID_MTSR,
                (const unpack_data_func_t)mtsr_reg_unpack,
                ACC_REG_MTSR_INTERNAL_SECTION_NAME,
                ACC_REG_MTSR_NAME,
                ACC_REG_MTSR_FIELDS_NUM,
                NSB::get(this),
                "",
                SUPPORT_SW_CA,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP)
{
}

int MTSRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->type == IB_SW_NODE && p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);
    
    if (p_node->type == IB_CA_NODE &&  p_node->ext_type != IB_GPU_NODE)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct mtsr_reg reg;
    CLEAR_STRUCT(reg);

    mtsr_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MTSRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                             << '#'
            << handler_header                   << ','
            << "mode"                           << ','
            << "normal_events_cnt"              << ','
            << "warning_events_cnt"             << ','
            << "critical_events_cnt";
#endif

    IBDIAGNET_RETURN_VOID;

}

void MTSRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct mtsr_reg &reg = areg.regs.mtsr;

    //register fields
    sstream << +reg.mode                      << ','
            << +reg.normal_events_cnt         << ','
            << +reg.warning_events_cnt        << ','
            << +reg.critical_events_cnt       << endl;


    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* PLTC Register functions ***
//*****************************************
PLTCRegister::PLTCRegister(PhyDiag * phy_diag)
    : Register(phy_diag, ACCESS_REGISTER_ID_PLTC,
                (const unpack_data_func_t)pltc_reg_unpack,
                ACC_REG_PLTC_INTERNAL_SECTION_NAME,
                ACC_REG_PLTC_NAME,
                ACC_REG_PLTC_FIELDS_NUM,
                NSB::get(this),
                "",
                SUPPORT_SW,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP),
    m_pnat(ACC_REG_PNAT_IB_PORT)
{
}

int PLTCRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pltc_reg reg;
    CLEAR_STRUCT(reg);

    AccRegKeyPortLane * k = dynamic_cast<AccRegKeyPortLane*>(p_key);

    if (!k)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISABLED);

    reg.local_port  = k->port_num;
    reg.lane_mask   = 1 << k->lane;
    reg.pnat        = this->m_pnat;

    pltc_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PLTCRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                         << '#'
            << handler_header               << ','
            << "precoding_cap_mask"         << ','
            << "ph_pol_tx_cap"              << ','
            << "ph_pol_rx_cap"              << ','
            << "gray_tx_cap"                << ','
            << "gray_rx_cap"                << ','
            << "local_tx_precoding_admin"   << ','
            << "local_rx_precoding_admin"   << ','
            << "ph_pol_tx_admin"            << ','
            << "ph_pol_rx_admin"            << ','
            << "gray_tx_admin"              << ','
            << "gray_rx_admin"              << ','
            << "local_tx_precoding_oper"    << ','
            << "local_rx_precoding_oper"    << ','
            << "gray_tx_oper"               << ','
            << "gray_rx_oper";
#endif

    IBDIAGNET_RETURN_VOID;

}

void PLTCRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct pltc_reg &reg = areg.regs.pltc;

    //register fields
    sstream << +reg.precoding_cap_mask         << ','
            << +reg.ph_pol_tx_cap              << ','
            << +reg.ph_pol_rx_cap              << ','
            << +reg.gray_tx_cap                << ','
            << +reg.gray_rx_cap                << ','
            << +reg.local_tx_precoding_admin   << ','
            << +reg.local_rx_precoding_admin   << ','
            << +reg.ph_pol_tx_admin            << ','
            << +reg.ph_pol_rx_admin            << ','
            << +reg.gray_tx_admin              << ','
            << +reg.gray_rx_admin              << ','
            << +reg.local_tx_precoding_oper    << ','
            << +reg.local_rx_precoding_oper    << ','
            << +reg.gray_tx_oper               << ','
            << +reg.gray_rx_oper               << endl;


    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* PAOS Register functions ***
//*****************************************
PAOSRegister::PAOSRegister(PhyDiag * phy_diag)
    : Register(phy_diag, ACCESS_REGISTER_ID_PAOS,
                (const unpack_data_func_t)paos_reg_unpack,
                ACC_REG_PAOS_INTERNAL_SECTION_NAME,
                ACC_REG_PAOS_NAME,
                ACC_REG_PAOS_FIELDS_NUM,
                NSB::get(this),
                "",
                SUPPORT_SW_CA,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP),
    m_pnat(ACC_REG_PNAT_IB_PORT)
{
}

int PAOSRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    paos_reg reg;
    CLEAR_STRUCT(reg);

    reg.pnat        = this->m_pnat;

    int rc = FillLocalPortAndPlaneInd(p_key, &reg.local_port, &reg.plane_ind);

    if (rc)
        return rc;

    paos_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PAOSRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                     << '#'
            << handler_header           << ','
            << "admin_status"           << ','
            << "plane_ind"              << ','
            << "oper_status"            << ','
            << "ase"                    << ','
            << "ee"                     << ','
            << "ee_ls"                  << ','
            << "ee_ps"                  << ','
            << "ls_e"                   << ','
            << "logical_state_status"   << ','
            << "ps_e"                   << ','
            << "lock_en"                << ','
            << "sleep_cap"              << ','
            << "fd"                     << ','
            << "physical_state_status"  << ','
            << "e"                      << ','
            << "lock_mode";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PAOSRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct paos_reg &reg = areg.regs.paos;
    sstream << +reg.admin_status            << ','
            << +reg.plane_ind               << ','
            << +reg.oper_status             << ','
            << +reg.ase                     << ','
            << +reg.ee                      << ','
            << +reg.ee_ls                   << ','
            << +reg.ee_ps                   << ','
            << +reg.ls_e                    << ','
            << +reg.logical_state_status    << ','
            << +reg.ps_e                    << ','
            << +reg.lock_en                 << ','
            << +reg.sleep_cap               << ','
            << +reg.fd                      << ','
            << +reg.physical_state_status   << ','
            << +reg.e                       << ','
            << +reg.lock_mode               << endl;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************ PPHCR Register functions ***
//*****************************************
PPHCRRegister::PPHCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PPHCR,
                (const unpack_data_func_t)pphcr_reg_unpack,
                ACC_REG_PPHCR_INTERNAL_SECTION_NAME,
                ACC_REG_PPHCR_NAME,
                ACC_REG_PPHCR_FIELDS_NUM,
                NSB::get(this),
                "",
                SUPPORT_SW_CA,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP),
        m_pnat(ACC_REG_PNAT_IB_PORT)
{
}

int PPHCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pphcr_reg pphcr;
    CLEAR_STRUCT(pphcr);

    pphcr.pnat = this->m_pnat;

    int rc = FillLocalPortAndPlaneInd(p_key, &pphcr.local_port, &pphcr.plane_ind);

    if (rc)
        return rc;

    pphcr_reg_pack(&pphcr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPHCRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                    << '#'
            << handler_header          << ','
            << "we"                    << ','
            << "plane_ind"             << ','
            << "port_type"             << ','
            << "lp_msb"                << ','
            << "pnat"                  << ','
            << "local_port"            << ','
            << "active_hist_type"      << ','
            << "hist_type"             << ','
            << "num_of_bins"           << ','
            << "hist_min_measurement"  << ','
            << "hist_max_measurement"  << ','
            << "bin_range_write_mask";

    for (size_t i = 0; i < ARRAY_SIZE(pphcr_reg::bin_range); i++)
         sstream << ',' << "bin_range[" << i << "].low_val"
                 << ',' << "bin_range[" << i << "].high_val";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPHCRRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct pphcr_reg &pphcr = areg.regs.pphcr;

    //register fields
    sstream << +pphcr.we                    << ","
            << +pphcr.plane_ind             << ","
            << +pphcr.port_type             << ","
            << +pphcr.lp_msb                << ","
            << +pphcr.pnat                  << ","
            << +pphcr.local_port            << ","
            << +pphcr.active_hist_type      << ","
            << +pphcr.hist_type             << ","
            << +pphcr.num_of_bins           << ","
            << +pphcr.hist_min_measurement  << ","
            << +pphcr.hist_max_measurement  << ","
            << DEC(pphcr.bin_range_write_mask);

    int bin_range_size = ARRAY_SIZE(pphcr.bin_range);
    for (int i = 0; i < bin_range_size; i++) {

         sstream << "," << +pphcr.bin_range[i].low_val
                 << "," << +pphcr.bin_range[i].high_val;
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void PPHCRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node,
                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pphcr_reg = &areg.regs.pphcr;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************ PEUCG Register functions ***
//*****************************************
PEUCGRegister::PEUCGRegister(PhyDiag * phy_diag, string section_name)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_PEUCG,
               (const unpack_data_func_t)peucg_reg_unpack,
               section_name,
               ACC_REG_PEUCG_NAME,
               ACC_REG_PEUCG_FIELDS_NUM,
               NSB::get(this),
               "",
               SUPPORT_SW_CA,
               true,  // dump enabled
               false, // disconnected ports
               VIA_GMP),
      m_pnat(ACC_REG_PNAT_IB_PORT),
      m_cur_mad_builder(NULL)
{
}

void PEUCGRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "unit"             << ","
            << "local_port"       << ","
            << "pnat"             << ","
            << "lp_msb"           << ","
            << "lane"             << ","
            << "status"           << ","
            << "payload_size"     << ","
            << "db"               << ","
            << "clr"              << ","
            << "enum_init"        << ","
            << "num_of_entries"   << ","
            << "db_index";

    for (size_t i = 0; i < ARRAY_SIZE(peucg_reg::page_data); i++)
        sstream << "," << "page_data[" << i << "].address"
                << "," << "page_data[" << i << "].rxtx"
                << "," << "page_data[" << i << "].payload_data";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PEUCGRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct peucg_reg &peucg = areg.regs.peucg;

    //register fields
    sstream << +peucg.unit               << ","
            << +peucg.local_port         << ","
            << +peucg.pnat               << ","
            << +peucg.lp_msb             << ","
            << +peucg.lane               << ","
            << +peucg.status             << ","
            << +peucg.payload_size       << ","
            << +peucg.db                 << ","
            << +peucg.clr                << ","
            << +peucg.enum_init          << ","
            << peucg.num_of_entries      << ","
            << peucg.db_index;

    u_int8_t page_data_size = ARRAY_SIZE(peucg.page_data);
    u_int8_t i = 0;

    for ( ; i < peucg.payload_size; ++i) {
        sstream << "," << peucg.page_data[i].address
                << "," << +peucg.page_data[i].rxtx
                << "," << peucg.page_data[i].payload_data;
    }

    for ( ; i < page_data_size; ++i) {
        sstream << "," << "N/A"
                << "," << "N/A"
                << "," << "N/A";
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//**** PEUCG Version Register functions ***
//*****************************************
PEUCG_Ver_Register::PEUCG_Ver_Register(PhyDiag * phy_diag)
    : PEUCGRegister(phy_diag, ACC_REG_PEUCG_VER_INTERNAL_SECTION_NAME)
{
}

int PEUCG_Ver_Register::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct peucg_reg peucg;
    CLEAR_STRUCT(peucg);

    // for version - all fields are 0, except:
    peucg.pnat = this->m_pnat;
    peucg.payload_size = 1;
    peucg.local_port = 1;

    peucg_reg_pack(&peucg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PEUCG_Ver_Register::HandleData(IBNode *p_node, AccRegKey* p_key, acc_reg_data& reg_data) const
{
    IBDIAGNET_ENTER;

    if (!p_node || !p_node->p_phy_data)
        IBDIAGNET_RETURN_VOID;

    PHYNodeData* phy_data = (PHYNodeData*)p_node->p_phy_data;

    // if we already get uphy version we'll use existing data.
    if (phy_data->uphy)
        IBDIAGNET_RETURN_VOID;

    phy_data->uphy = new PHYNodeData::UPHYData();

    phy_data->uphy->cln.version = new UPHY::Version(reg_data.regs.peucg.page_data[0].payload_data & 0x0f,
            (reg_data.regs.peucg.page_data[0].payload_data & 0xf0) >> 4);

    phy_data->uphy->dln.version = new UPHY::Version(reg_data.regs.peucg.page_data[0].payload_data & 0x0f,
            (reg_data.regs.peucg.page_data[0].payload_data & 0xf0) >> 4);

    phy_data->uphy->cln.dataset = UPHY::DB::instance().get_compatible(
            UPHY::DataSetType_CLN, *phy_data->uphy->cln.version);

    phy_data->uphy->dln.dataset = UPHY::DB::instance().get_compatible(
            UPHY::DataSetType_DLN, *phy_data->uphy->dln.version);

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//****** PEUCG DLN Register functions *****
//*****************************************
PEUCG_DLN_Register::PEUCG_DLN_Register(PhyDiag * phy_diag) :
        PEUCGRegister(phy_diag, ACC_REG_PEUCG_DLN_INTERNAL_SECTION_NAME)
{
}

int PEUCG_DLN_Register::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct peucg_reg peucg;
    CLEAR_STRUCT(peucg);

    UPHY::DataKey *p_cur_key = (UPHY::DataKey*)p_key;

    AccRegKeyPortLane* p_plkey = (AccRegKeyPortLane*)(p_cur_key->base_key());

    peucg.local_port = p_plkey->port_num;
    peucg.lane = p_plkey->lane;
    peucg.pnat = this->m_pnat;

    if (m_cur_mad_builder)
        m_cur_mad_builder->next(&peucg);

    peucg_reg_pack(&peucg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PEUCG_DLN_Register::BuildDB(AccRegHandler* p_handler,
                                list_p_fabric_general_err &phy_errors,
                                ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    for (map_str_pnode::iterator nI = m_phy_diag->GetFabric()->NodeByName.begin();
         nI != m_phy_diag->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            m_phy_diag->SetLastError("DB error - found null node in"  \
                                           " NodeByName map for key = %s",
                                           nI->first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            IBDIAGNET_RETURN(rc);
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        PHYNodeData::UPHYData* cur_uphy_data = ((PHYNodeData*)(p_curr_node->p_phy_data))->uphy;
        if (!cur_uphy_data || !cur_uphy_data->dln.dataset)
            continue;

        // go over all ports of this node:
        for (phys_port_t i = 1; i < p_curr_node->numPorts + 1; ++i) {

            IBPort *p_curr_port = p_curr_node->getPort(i);
            //port doesn't exist or down or have undefined state
            //or are not in the scope of the fabric.
            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;

            if (p_curr_port->isSpecialPort())
                continue;

            for (u_int32_t ln = 0; ln < LANE_NUM; ln++) {

                AccessRegister acc_reg;
                CLEAR_STRUCT(acc_reg);

                uint8_t idx = 0; // for the dln key to be unique.

                this->m_cur_mad_builder = new UPHY::MadBuilder(*cur_uphy_data->dln.dataset, m_phy_diag->GetUphyDumps());

                while (!this->m_cur_mad_builder->empty()) {

                    AccRegKeyPortLane *p_plkey = new AccRegKeyPortLane(
                            p_curr_node->guid_get(),
                            p_curr_port->guid_get(),
                            i,
                            ln,
                            0); // idx in lane - not relevant

                    UPHY::DataKey* p_cur_key = new UPHY::DataKey(p_plkey, idx++);

                    rc = p_handler->SendAccReg(acc_reg_via,
                                               p_curr_node,
                                               i,
                                               p_curr_port->base_lid,
                                               acc_reg,
                                               p_cur_key,
                                               p_progress_bar);

                    if (rc == IBDIAG_ERR_CODE_DB_ERR) {
                        delete this->m_cur_mad_builder;
                        this->m_cur_mad_builder = NULL;
                        IBDIAGNET_RETURN(rc);
                    }
                } // elements blocks

                delete this->m_cur_mad_builder;
                this->m_cur_mad_builder = NULL;
            } // lane
        } // port
    } // node

    IBDIAGNET_RETURN(rc);
}


//*****************************************
//****** PEUCG CLN Register functions *****
//*****************************************
PEUCG_CLN_Register::PEUCG_CLN_Register(PhyDiag * phy_diag, map_akey_areg *pmdr_map) :
        PEUCGRegister(phy_diag, ACC_REG_PEUCG_CLN_INTERNAL_SECTION_NAME),
        p_pmdr_map(pmdr_map)
{
}

int PEUCG_CLN_Register::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct peucg_reg peucg;
    CLEAR_STRUCT(peucg);

    UPHY::DataKey *p_cur_key = (UPHY::DataKey*)p_key;
    AccRegKeyGroup* p_gkey = (AccRegKeyGroup*)(p_cur_key->base_key());

    peucg.unit = 1; // 1 - common pll (CLN)
    peucg.local_port = p_gkey->group_num; // for CLN local_port is pll_index

    if (m_cur_mad_builder)
        m_cur_mad_builder->next(&peucg);

    peucg_reg_pack(&peucg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PEUCG_CLN_Register::BuildDB(AccRegHandler* p_handler,
                                list_p_fabric_general_err &phy_errors,
                                ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    // 1. go over all PMDRs and make a map of node and set of PLLs
    map < u_int64_t , set<u_int8_t> > node2plls;

    for (map_akey_areg::iterator pmdr_iter = p_pmdr_map->begin();
            pmdr_iter != p_pmdr_map->end(); pmdr_iter++) {

        AccRegKeyPort *p_port_key = (AccRegKeyPort*)pmdr_iter->first;
        pmdr_reg pmdr = pmdr_iter->second.regs.pmdr;

        node2plls[p_port_key->node_guid].insert(pmdr.pll_index);
    }

    // 2. go over all nodes and send CLN to all PLLs
    for (map < u_int64_t , set<u_int8_t> >::iterator nodes_iter = node2plls.begin();
            nodes_iter != node2plls.end(); nodes_iter++) {

        IBNode *p_curr_node =
                m_phy_diag->GetFabric()->getNodeByGuid(nodes_iter->first);

        if (!p_curr_node) {
            m_phy_diag->SetLastError("DB error - found null node in "
                    "NodeByName map for key = " U64H_FMT, nodes_iter->first);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        PHYNodeData::UPHYData* cur_uphy_data = ((PHYNodeData*)(p_curr_node->p_phy_data))->uphy;
        if (!cur_uphy_data || !cur_uphy_data->cln.dataset)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        // go over all PLLs of this node:
        for (set<u_int8_t>::iterator plls_iter = nodes_iter->second.begin();
                plls_iter != nodes_iter->second.end(); plls_iter++) {

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            uint8_t idx = 0; // for the cln key to be unique.

            this->m_cur_mad_builder = new UPHY::MadBuilder(*cur_uphy_data->cln.dataset, m_phy_diag->GetUphyDumps());

            while (!this->m_cur_mad_builder->empty()) {

                AccRegKeyGroup *p_gkey = new AccRegKeyGroup(
                        p_curr_node->guid_get(),
                        *plls_iter);

                UPHY::DataKey* p_cur_key = new UPHY::DataKey(p_gkey, idx++);

                rc = p_handler->SendAccReg(acc_reg_via,
                        p_curr_node,
                        0, // port 0
                        lid,
                        acc_reg,
                        p_cur_key,
                        p_progress_bar);

                if (rc == IBDIAG_ERR_CODE_DB_ERR) {
                    delete this->m_cur_mad_builder;
                    this->m_cur_mad_builder = NULL;
                    IBDIAGNET_RETURN(rc);
                }
            }

            delete this->m_cur_mad_builder;
            this->m_cur_mad_builder = NULL;
        }
    }

    IBDIAGNET_RETURN(rc);
}


//*****************************************
//********   SLRegister functions    ******
//*****************************************
SLRegister::SLRegister(PhyDiag * phy_diag,
                       uint32_t reg_id,
                       const unpack_data_func_t p_unpack_func,
                       string sn,
                       string name,
                       uint32_t fn,
                       uint64_t nsb,
                       map_akey_areg *mpein_map,
                       map_akey_areg *mpir_map)
    : Register(phy_diag,
               reg_id,
               p_unpack_func,
               sn,
               name,
               fn,
               nsb),
      p_mpein_map(mpein_map),
      p_mpir_map(mpir_map){}

/*
 * this register is for slrg slrp sltp registers in pci mode.
 * it goes over all valid MPIR queries and get the above registers.
 * p_handler - contain the MPIR map.
 * */
int SLRegister::BuildDB(AccRegHandler* p_handler,
                        list_p_fabric_general_err &phy_errors,
                        ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    p_handler->SetHeader(PCI_PORT_LANE_HEADER);

    for(map_akey_areg::iterator mpir_iter = p_mpir_map->begin();
            mpir_iter != p_mpir_map->end(); mpir_iter++) {

        AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)mpir_iter->first;
        mpir_reg mpir = mpir_iter->second.regs.mpir;
        mpein_reg mpein = p_mpein_map->find(p_dpnkey)->second.regs.mpein;

        IBNode *p_curr_node =
                m_phy_diag->GetFabric()->getNodeByGuid(p_dpnkey->node_guid);

        if (!p_curr_node) {
            m_phy_diag->SetLastError("DB error - found null node in "
                    "NodeByName map for key = " U64H_FMT, p_dpnkey->node_guid);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        AccRegVia_t acc_reg_via = this->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        IBPort *p_curr_port = NULL;
        phys_port_t port_num = 1;
        for ( ; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
            p_curr_port = p_curr_node->getPort(port_num);

            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;
            break; // use the first valid port.
        }

        if (!p_curr_port) {
            m_phy_diag->SetLastError("DB error - failed to find valid port for node %s",
                                                p_curr_node->getName().c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        for (u_int8_t lane = 0; lane < mpein.link_width_active; lane++) {

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            AccRegKeyPCIPortLane *p_pciplkey = new AccRegKeyPCIPortLane(
                    p_dpnkey->node_guid,
                    mpir.local_port,
                    lane);

            rc = p_handler->SendAccReg(acc_reg_via,
                                       p_curr_node,
                                       port_num,
                                       p_curr_port->base_lid,
                                       acc_reg,
                                       p_pciplkey,
                                       p_progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR)
                IBDIAGNET_RETURN(rc);

        } // lane loop
    } // mpir loop

    IBDIAGNET_RETURN(rc);
}

bool SLRegister::CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                  export_data_phy_node_t *p_export_data_phy_node,
                                  AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node &&
            p_key->node_guid == p_export_data_phy_node->node_guid)
        IBDIAGNET_RETURN(true);

    IBDIAGNET_RETURN(false);
}


//*****************************************
//***     SLSIR Register functions      ***
//*****************************************
SLSIRRegister::SLSIRRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(phy_diag,
               ACCESS_REGISTER_ID_SLSIR,
               (const unpack_data_func_t)slsir_reg_unpack,
               sn,
               ACC_REG_SLSIR_NAME,
               ACC_REG_SLSIR_MAX_FIELDS_NUM,
               NSB::get(this),
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLSIRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slsir_reg slsir;
    CLEAR_STRUCT(slsir);

    slsir.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slsir.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slsir.pnat = m_pnat;

    slsir_reg_pack(&slsir, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLSIRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;

}

void SLSIRRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct slsir_reg &slsir = areg.regs.slsir;

    sstream << +slsir.version                << ','
            << +slsir.local_port             << ','
            << +slsir.pnat                   << ','
            << +slsir.lp_msb                 << ','
            << +slsir.lane                   << ','
            << +slsir.port_type              << ',';

    switch(slsir.version)
    {
        case 4: // 7nm
            Dump_7nm(slsir, sstream);
            break;
        case 5: // 5nm
            Dump_5nm(slsir, sstream);
            break;
        default:
            static bool should_print = true;
            sstream << '"' << "Unknown version " << +slsir.version << '"';
            for (int i = 0; i < ACC_REG_SLSIR_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLSIR: %d, on node: " U64H_FMT ".\n",
                        slsir.version, key.node_guid);
                should_print = false;
            }
    }
    sstream << endl;
    IBDIAGNET_RETURN_VOID;
}

AccRegVia_t SLSIRRegister::Validation(IBNode *p_node, int &rc)
{
    if (p_node->serdes_version < SerDes_7nm)
        return NOT_SUP_ACC_REG;

    return Register::Validation(p_node, rc);
}

void SLSIRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node,
                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port && this->m_pnat == ACC_REG_PNAT_IB_PORT)
        p_export_data_phy_port->p_slsir_reg[((AccRegKeyPortLane *)p_key)->lane] = &areg.regs.slsir;

    IBDIAGNET_RETURN_VOID;
}

void SLSIRRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;
    sstream << "version_7nm=4"          << ','
            << "local_port"             << ','
            << "pnat"                   << ','
            << "lp_msb"                 << ','
            << "lane"                   << ','
            << "port_type"              << ','
            << "nop_rsunf_error"        <<','
            << "nop_rsovf_error"        <<','
            << "nop_dsunf_error"        <<','
            << "nop_dsovf_error"        <<','
            << "peq_adc_overload"       <<','
            << "feq_adc_overload"       <<','
            << "cdr_error"              <<','
            << "imem_chksm_error"       <<','
            << "rx_ugl_state"           <<','
            << "rx_eom_ugl_state"       <<','
            << "rx_cal_ugl_state"       <<','
            << "rx_eq_ugl_state"        <<','
            << "tx_ugl_state"           <<','
            << "recovery_retries_cnt"   <<','
            << "imem_loading_retries"   <<','
            << "sd_hits_cnt"            <<','
            << "sd_iter_cnt"            <<','
            << "rd_iter_cnt"            <<','
            << "ae_state"               <<','
            << "rx_init_abort_cnt"      <<','
            << "rx_init_done_cnt"       <<','
            << "cdr_abort_cnt"          <<','
            << "cdr_done_cnt"           <<','
            << "cal_abort_cnt"          <<','
            << "cal_done_cnt"           <<','
            << "eq_abort_cnt"           <<','
            << "eq_done_cnt"            <<','
            << "eom_abort_cnt"          <<','
            << "eom_done_cnt"           <<','
            << "uphy_rev_subminor"      <<','
            << "bkv_revision_cln"       <<','
            << "bkv_revision_rx"        <<','
            << "uphy_rev_major"         <<','
            << "uphy_rev_minor"         <<','
            << "err_ind_it_3"           <<','
            << "err_ind_it_2"           <<','
            << "err_ind_it_1"           <<','
            << "err_ind_it_0";
    IBDIAGNET_RETURN_VOID;
}

void SLSIRRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;
    sstream << "version_5nm=5"          << ','
            << "local_port"             << ','
            << "pnat"                   << ','
            << "lp_msb"                 << ','
            << "lane"                   << ','
            << "port_type"              << ','
            << "clear_causes"           <<','
            << "clear_counters"         <<','
            << "sfec_align_ok"          <<','
            << "sfec_align_fail"        <<','
            << "dfd_trigger"            <<','
            << "nop_rsunf_error"        <<','
            << "nop_rsovf_error"        <<','
            << "nop_dsunf_error"        <<','
            << "nop_dsovf_error"        <<','
            << "peq_adc_overload"       <<','
            << "feq_adc_overload"       <<','
            << "cdr_error"              <<','
            << "imem_chksm_error"       <<','
            << "cdr_error_cnt"          <<','
            << "cal_error_cnt"          <<','
            << "bist_err_cnt"           <<','
            << "cdr_seq_done_cnt"       <<','
            << "cdr_seq_abort_cnt"      <<','
            << "rx_init_done_cnt"       <<','
            << "rx_init_abort_cnt"      <<','
            << "eq_done_cnt"            <<','
            << "eq_abort_cnt"           <<','
            << "cal_done_cnt"           <<','
            << "cal_abort_cnt"          <<','
            << "margin_done_cnt"        <<','
            << "margin_yield_cnt"       <<','
            << "eom_done_cnt"           <<','
            << "eom_abort_cnt"          <<','
            << "ae_state"               <<','
            << "pll_ui_shift_cnt"       <<','
            << "bin_cnt"                <<','
            << "event_cnt"              <<','
            << "bkv_major_cln"          <<','
            << "bkv_minor_cln"          <<','
            << "bkv_major_dln"          <<','
            << "bkv_minor_dln"          <<','
            << "uphy_ver_major"         <<','
            << "uphy_ver_minor";
    IBDIAGNET_RETURN_VOID;
}

void SLSIRRegister::Dump_7nm(const struct slsir_reg &slsir, stringstream &sstream) const
{
    IBDIAGNET_ENTER;
    slsir_7nm slsir_7;
    slsir_7nm_unpack(&slsir_7, slsir.page_data.slsir_data_set.data);

    sstream << +slsir_7.nop_rsunf_error <<','
            << +slsir_7.nop_rsovf_error <<','
            << +slsir_7.nop_dsunf_error <<','
            << +slsir_7.nop_dsovf_error <<','
            << +slsir_7.peq_adc_overload <<','
            << +slsir_7.feq_adc_overload <<','
            << +slsir_7.cdr_error <<','
            << +slsir_7.imem_chksm_error <<','
            << +slsir_7.rx_ugl_state <<','
            << +slsir_7.rx_eom_ugl_state <<','
            << +slsir_7.rx_cal_ugl_state <<','
            << +slsir_7.rx_eq_ugl_state <<','
            << +slsir_7.tx_ugl_state <<','
            << +slsir_7.recovery_retries_cnt <<','
            << +slsir_7.imem_loading_retries <<','
            << +slsir_7.sd_hits_cnt <<','
            << +slsir_7.sd_iter_cnt <<','
            << +slsir_7.rd_iter_cnt <<','
            << +slsir_7.ae_state <<','
            << +slsir_7.rx_init_abort_cnt <<','
            << +slsir_7.rx_init_done_cnt <<','
            << +slsir_7.cdr_abort_cnt <<','
            << +slsir_7.cdr_done_cnt <<','
            << +slsir_7.cal_abort_cnt <<','
            << +slsir_7.cal_done_cnt <<','
            << +slsir_7.eq_abort_cnt <<','
            << +slsir_7.eq_done_cnt <<','
            << +slsir_7.eom_abort_cnt <<','
            << +slsir_7.eom_done_cnt <<','
            << +slsir_7.uphy_rev_subminor <<','
            << +slsir_7.bkv_revision_cln <<','
            << +slsir_7.bkv_revision_rx <<','
            << +slsir_7.uphy_rev_major <<','
            << +slsir_7.uphy_rev_minor <<','
            << +slsir_7.err_ind_it_3 <<','
            << +slsir_7.err_ind_it_2 <<','
            << +slsir_7.err_ind_it_1 <<','
            << +slsir_7.err_ind_it_0;

    IBDIAGNET_RETURN_VOID;
}

void SLSIRRegister::Dump_5nm(const struct slsir_reg &slsir, stringstream &sstream) const
{
    IBDIAGNET_ENTER;
    slsir_5nm slsir_5;
    slsir_5nm_unpack(&slsir_5, slsir.page_data.slsir_data_set.data);

    sstream << +slsir_5.clear_causes <<','
            << +slsir_5.clear_counters <<','
            << +slsir_5.sfec_align_ok <<','
            << +slsir_5.sfec_align_fail <<','
            << +slsir_5.dfd_trigger <<','
            << +slsir_5.nop_rsunf_error <<','
            << +slsir_5.nop_rsovf_error <<','
            << +slsir_5.nop_dsunf_error <<','
            << +slsir_5.nop_dsovf_error <<','
            << +slsir_5.peq_adc_overload <<','
            << +slsir_5.feq_adc_overload <<','
            << +slsir_5.cdr_error <<','
            << +slsir_5.imem_chksm_error <<','
            << +slsir_5.cdr_error_cnt <<','
            << +slsir_5.cal_error_cnt <<','
            << +slsir_5.bist_err_cnt <<','
            << +slsir_5.cdr_seq_done_cnt <<','
            << +slsir_5.cdr_seq_abort_cnt <<','
            << +slsir_5.rx_init_done_cnt <<','
            << +slsir_5.rx_init_abort_cnt <<','
            << +slsir_5.eq_done_cnt <<','
            << +slsir_5.eq_abort_cnt <<','
            << +slsir_5.cal_done_cnt <<','
            << +slsir_5.cal_abort_cnt <<','
            << +slsir_5.margin_done_cnt <<','
            << +slsir_5.margin_yield_cnt <<','
            << +slsir_5.eom_done_cnt <<','
            << +slsir_5.eom_abort_cnt <<','
            << +slsir_5.ae_state <<','
            << +slsir_5.pll_ui_shift_cnt <<','
            << +slsir_5.bin_cnt <<','
            << +slsir_5.event_cnt <<','
            << +slsir_5.bkv_major_cln <<','
            << +slsir_5.bkv_minor_cln <<','
            << +slsir_5.bkv_major_dln <<','
            << +slsir_5.bkv_minor_dln <<','
            << +slsir_5.uphy_ver_major <<','
            << +slsir_5.uphy_ver_minor;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* SLRP Register functions ***
//*****************************************
SLRPRegister::SLRPRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(phy_diag,
               ACCESS_REGISTER_ID_SLRP,
               (const unpack_data_func_t)slrp_reg_unpack,
               sn,
               ACC_REG_SLRP_NAME,
               ACC_REG_SLRP_MAX_FIELDS_NUM,
               NSB::get(this),
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLRPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slrp_reg slrp;
    CLEAR_STRUCT(slrp);
    slrp.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slrp.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slrp.pnat = m_pnat;
    slrp_reg_pack(&slrp, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLRPRegister::Header_Dump_40nm_28nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                    << ','
            << "version_40nm=0_28nm=1"     << ','
            << "local_port"                << ','
            << "pnat"                      << ','
            << "lp_msb"                    << ','
            << "lane"                      << ','
            << "port_type"                 << ','
            << "ib_sel"                    << ','
            << "dp_sel"                    << ','
            << "dp90sel"                   << ','
            << "mix90phase"                << ','
            << "ffe_tap0"                  << ','
            << "ffe_tap1"                  << ','
            << "ffe_tap2"                  << ','
            << "ffe_tap3"                  << ','
            << "ffe_tap4"                  << ','
            << "ffe_tap5"                  << ','
            << "ffe_tap6"                  << ','
            << "ffe_tap7"                  << ','
            << "ffe_tap8"                  << ','
            << "mixerbias_tap_amp"         << ','
            << "ffe_tap_en"                << ','
            << "ffe_tap_offset0"           << ','
            << "ffe_tap_offset1"           << ','
            << "slicer_offset0"            << ','
            << "mixer_offset0"             << ','
            << "mixer_offset1"             << ','
            << "mixerbgn_inp"              << ','
            << "mixerbgn_inn"              << ','
            << "mixerbgn_refp"             << ','
            << "mixerbgn_refn"             << ','
            << "sel_slicer_lctrl_h"        << ','
            << "sel_slicer_lctrl_l"        << ','
            << "ref_mixer_vreg"            << ','
            << "slicer_gctrl"              << ','
            << "lctrl_input"               << ','
            << "mixer_offset_cm1"          << ','
            << "common_mode"               << ','
            << "mixer_offset_cm0"          << ','
            << "slicer_offset_cm"          ;

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Header_Dump_16nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                                   << ','
            << "version_16nm=3"                           << ','
            << "local_port"                               << ','
            << "pnat"                                     << ','
            << "lp_msb"                                   << ','
            << "lane"                                     << ','
            << "port_type"                                << ','
            << "mixer_offset_up"                          << ','
            << "mixer_offset_down"                        << ','
            << "sel_enc"                                  << ','
            << "dp_sel"                                   << ','
            << "sel_ref_in_mixer_biasgen_up"              << ','
            << "sel_ref_in_mixer_biasgen_mid"             << ','
            << "sel_ref_in_mixer_biasgen_down"            << ','
            << "sel_ref_in_mixer_biasgen_eo"              << ','
            << "sel_ref_out_mixer_biasgen_up"             << ','
            << "sel_ref_out_mixer_biasgen_mid"            << ','
            << "sel_ref_out_mixer_biasgen_down"           << ','
            << "sel_ref_out_mixer_biasgen_eo"             << ','
            << "sel_amp_mixer_biasgen_up"                 << ','
            << "sel_amp_mixer_biasgen_mid"                << ','
            << "sel_amp_mixer_biasgen_down"               << ','
            << "sel_amp_mixer_biasgen_eo"                 << ','
            << "mixer_offset_eye_scan"                    << ','
            << "gctrl_bin_bgn0_n"                         << ','
            << "gctrl_bin_bgn0_p"                         << ','
            << "gctrl_bin_bgn1_n"                         << ','
            << "gctrl_bin_bgn1_p"                         << ','
            << "sel_digital_gctrl"                        << ','
            << "sel_bias2ctle"                            << ','
            << "selgc_ref_op0_alev_ctle"                  << ','
            << "selgc_ref_op1_alev_ctle"                  << ','
            << "sel_bias_en_0"                            << ','
            << "sel_bias_en_1"                            << ','
            << "sel_bias_en_2"                            << ','
            << "sel_bias_en_3"                            << ','
            << "sel_bias_en_4"                            << ','
            << "sel_bias_en_5"                            << ','
            << "sel_bias_en_6"                            << ','
            << "sel_gctrln_en_0"                          << ','
            << "sel_gctrln_en_1"                          << ','
            << "sel_gctrln_en_2"                          << ','
            << "sel_gctrln_en_3"                          << ','
            << "sel_gctrln_en_4"                          << ','
            << "sel_gctrln_en_5"                          << ','
            << "sel_gctrln_en_6"                          << ','
            << "sel_gctrlp_en_0"                          << ','
            << "sel_gctrlp_en_1"                          << ','
            << "sel_gctrlp_en_2"                          << ','
            << "sel_gctrlp_en_3"                          << ','
            << "sel_gctrlp_en_4"                          << ','
            << "sel_gctrlp_en_5"                          << ','
            << "sel_gctrlp_en_6"                          << ','
            << "crnt_bgn_offset_p1"                       << ','
            << "crnt_bgn_offset_n1"                       << ','
            << "crnt_bgn_offset_p0"                       << ','
            << "crnt_bgn_offset_n0"                       ;

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                       << ','
            << "version_7nm=4"                << ','
            << "local_port"                   << ','
            << "pnat"                         << ','
            << "lp_msb"                       << ','
            << "lane"                         << ','
            << "port_type"                    << ','
            << "adc_recording_admin"          << ','
            << "adc_recording_status"         << ','
            << "adc_rocording_lanes"          << ','
            << "edge_vos_ccal_en"             << ','
            << "adc_gain_shift_auto"          << ','
            << "ccal_mode"                    << ','
            << "ctle_override_ctrl"           << ','
            << "vga_override_ctrl"            << ','
            << "adc_vos_override_ctrl"        << ','
            << "adc_gain_override_ctrl"       << ','
            << "phos_override_ctrl"           << ','
            << "cal_error_cnt"                << ','
            << "ccal_state"                   << ','
            << "ccal_op"                      << ','
            << "phos"                         << ','
            << "vga_vos1"                     << ','
            << "vga_vos0"                     << ','
            << "ctle_vos0"                    ;
    for (size_t i = 0; i < ARRAY_SIZE(slrp_7nm::adc_vos); ++i)
        sstream << ',' << "adc_vos[" << i << "]";
    for (size_t i = 0; i < ARRAY_SIZE(slrp_7nm::adc_gos); ++i)
        sstream << ',' << "adc_gos[" << i << "]";
    
    sstream << ',' 
            << "phos3"                        << ','
            << "phos2"                        << ','
            << "phos1"                        << ','
            << "phos0"                        << ','
            << "phos7"                        << ','
            << "phos6"                        << ','
            << "phos5"                        << ','
            << "phos4"                        ;
            
    IBDIAG_RETURN_VOID;
}

void SLRPRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                       << ','
            << "version_5nm=5"                << ','
            << "local_port"                   << ','
            << "pnat"                         << ','
            << "lp_msb"                       << ','
            << "lane"                         << ','
            << "port_type"                    << ','
            << "ctle_vos_val"                 << ','
            << "ctle_vos_stg1"                << ','
            << "ctle_vos_aux"                 << ','
            << "adc_recording_admin"          << ','
            << "adc_recording_status"         << ','
            << "fixed_val"                    << ','
            << "fine_val"                     << ','
            << "coarse_val";

    for (int i = 0; i < 16; ++i)
        sstream << ',' << "tah_vos[" << i << ']';

    for (int i = 0; i < 16; ++i)
        sstream << ',' << "adc_vos[" << i << ']';

    for (int i = 0; i < 16; ++i)
        sstream << ',' << "adc_gos[" << i << ']';

    IBDIAG_RETURN_VOID;
}

void SLRPRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_40nm_28nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct slrp_reg & slrp = areg.regs.slrp;

    sstream << +slrp.status           << ','
            << +slrp.version          << ','
            << +slrp.local_port       << ','
            << +slrp.pnat             << ','
            << +slrp.lp_msb           << ','
            << +slrp.lane             << ','
            << +slrp.port_type        << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (slrp.version) {

        case 0: // 40nm
        case 1: // 28nm
            Dump_40nm_28nm(slrp, sstream);
            break;

        case 3: // 16nm
            Dump_16nm(slrp, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(slrp, sstream);
            break;

        case 5: // 5nm_scc
            Dump_5nm_scc(slrp, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +slrp.version << '"';
            for (int i = 0; i < ACC_REG_SLRP_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLRP: %d, on node: " U64H_FMT ".\n",
                        slrp.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << endl;
    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Dump_40nm_28nm(const struct slrp_reg &slrp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrp_40nm_28nm slrp_40_28;
    slrp_40nm_28nm_unpack(&slrp_40_28, slrp.page_data.slrp_data_set.data);

    sstream << +slrp_40_28.ib_sel                    << ','
            << +slrp_40_28.dp_sel                    << ','
            << +slrp_40_28.dp90sel                   << ','
            << +slrp_40_28.mix90phase                << ','
            << +slrp_40_28.ffe_tap0                  << ','
            << +slrp_40_28.ffe_tap1                  << ','
            << +slrp_40_28.ffe_tap2                  << ','
            << +slrp_40_28.ffe_tap3                  << ','
            << +slrp_40_28.ffe_tap4                  << ','
            << +slrp_40_28.ffe_tap5                  << ','
            << +slrp_40_28.ffe_tap6                  << ','
            << +slrp_40_28.ffe_tap7                  << ','
            << +slrp_40_28.ffe_tap8                  << ','
            << +slrp_40_28.mixerbias_tap_amp         << ','
            << +slrp_40_28.ffe_tap_en                << ','
            << +slrp_40_28.ffe_tap_offset0           << ','
            << +slrp_40_28.ffe_tap_offset1           << ','
            << +slrp_40_28.slicer_offset0            << ','
            << +slrp_40_28.mixer_offset0             << ','
            << +slrp_40_28.mixer_offset1             << ','
            << +slrp_40_28.mixerbgn_inp              << ','
            << +slrp_40_28.mixerbgn_inn              << ','
            << +slrp_40_28.mixerbgn_refp             << ','
            << +slrp_40_28.mixerbgn_refn             << ','
            << +slrp_40_28.sel_slicer_lctrl_h        << ','
            << +slrp_40_28.sel_slicer_lctrl_l        << ','
            << +slrp_40_28.ref_mixer_vreg            << ','
            << +slrp_40_28.slicer_gctrl              << ','
            << +slrp_40_28.lctrl_input               << ','
            << +slrp_40_28.mixer_offset_cm1          << ','
            << +slrp_40_28.common_mode               << ','
            << +slrp_40_28.mixer_offset_cm0          << ','
            << +slrp_40_28.slicer_offset_cm          ;

    for (int i = 0; i < ACC_REG_SLRP_MAX_FIELDS_NUM - ACC_REG_SLRP28NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Dump_16nm(const struct slrp_reg &slrp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;
    slrp_16nm slrp_16;
    slrp_16nm_unpack(&slrp_16, slrp.page_data.slrp_data_set.data);

    sstream << +slrp_16.mixer_offset_up                          << ','
            << +slrp_16.mixer_offset_down                        << ','
            << +slrp_16.sel_enc                                  << ','
            << +slrp_16.dp_sel                                   << ','
            << +slrp_16.sel_ref_in_mixer_biasgen_up              << ','
            << +slrp_16.sel_ref_in_mixer_biasgen_mid             << ','
            << +slrp_16.sel_ref_in_mixer_biasgen_down            << ','
            << +slrp_16.sel_ref_in_mixer_biasgen_eo              << ','
            << +slrp_16.sel_ref_out_mixer_biasgen_up             << ','
            << +slrp_16.sel_ref_out_mixer_biasgen_mid            << ','
            << +slrp_16.sel_ref_out_mixer_biasgen_down           << ','
            << +slrp_16.sel_ref_out_mixer_biasgen_eo             << ','
            << +slrp_16.sel_amp_mixer_biasgen_up                 << ','
            << +slrp_16.sel_amp_mixer_biasgen_mid                << ','
            << +slrp_16.sel_amp_mixer_biasgen_down               << ','
            << +slrp_16.sel_amp_mixer_biasgen_eo                 << ','
            << +slrp_16.mixer_offset_eye_scan                    << ','
            << +slrp_16.gctrl_bin_bgn0_n                         << ','
            << +slrp_16.gctrl_bin_bgn0_p                         << ','
            << +slrp_16.gctrl_bin_bgn1_n                         << ','
            << +slrp_16.gctrl_bin_bgn1_p                         << ','
            << +slrp_16.sel_digital_gctrl                        << ','
            << +slrp_16.sel_bias2ctle                            << ','
            << +slrp_16.selgc_ref_op0_alev_ctle                  << ','
            << +slrp_16.selgc_ref_op1_alev_ctle                  << ','
            << +slrp_16.sel_bias_en_0                            << ','
            << +slrp_16.sel_bias_en_1                            << ','
            << +slrp_16.sel_bias_en_2                            << ','
            << +slrp_16.sel_bias_en_3                            << ','
            << +slrp_16.sel_bias_en_4                            << ','
            << +slrp_16.sel_bias_en_5                            << ','
            << +slrp_16.sel_bias_en_6                            << ','
            << +slrp_16.sel_gctrln_en_0                          << ','
            << +slrp_16.sel_gctrln_en_1                          << ','
            << +slrp_16.sel_gctrln_en_2                          << ','
            << +slrp_16.sel_gctrln_en_3                          << ','
            << +slrp_16.sel_gctrln_en_4                          << ','
            << +slrp_16.sel_gctrln_en_5                          << ','
            << +slrp_16.sel_gctrln_en_6                          << ','
            << +slrp_16.sel_gctrlp_en_0                          << ','
            << +slrp_16.sel_gctrlp_en_1                          << ','
            << +slrp_16.sel_gctrlp_en_2                          << ','
            << +slrp_16.sel_gctrlp_en_3                          << ','
            << +slrp_16.sel_gctrlp_en_4                          << ','
            << +slrp_16.sel_gctrlp_en_5                          << ','
            << +slrp_16.sel_gctrlp_en_6                          << ','
            << +slrp_16.crnt_bgn_offset_p1                       << ','
            << +slrp_16.crnt_bgn_offset_n1                       << ','
            << +slrp_16.crnt_bgn_offset_p0                       << ','
            << +slrp_16.crnt_bgn_offset_n0                       ;

    for (int i = 0; i < ACC_REG_SLRP_MAX_FIELDS_NUM - ACC_REG_SLRP16NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Dump_7nm(const struct slrp_reg &slrp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;
    slrp_7nm slrp_7;
    slrp_7nm_unpack(&slrp_7, slrp.page_data.slrp_data_set.data);

    sstream << +slrp_7.adc_recording_admin          << ','
            << +slrp_7.adc_recording_status         << ','
            << +slrp_7.adc_rocording_lanes          << ','
            << +slrp_7.edge_vos_ccal_en             << ','
            << +slrp_7.adc_gain_shift_auto          << ','
            << +slrp_7.ccal_mode                    << ','
            << +slrp_7.ctle_override_ctrl           << ','
            << +slrp_7.vga_override_ctrl            << ','
            << +slrp_7.adc_vos_override_ctrl        << ','
            << +slrp_7.adc_gain_override_ctrl       << ','
            << +slrp_7.phos_override_ctrl           << ','
            << +slrp_7.cal_error_cnt                << ','
            << +slrp_7.ccal_state                   << ','
            << +slrp_7.ccal_op                      << ','
            << +slrp_7.phos                         << ','
            << +slrp_7.vga_vos1                     << ','
            << +slrp_7.vga_vos0                     << ','
            << +slrp_7.ctle_vos0                    ;
    for (size_t i = 0; i < ARRAY_SIZE(slrp_7.adc_vos); ++i) {
        sstream << ',' << +slrp_7.adc_vos[i];
    }
    for (size_t i = 0; i < ARRAY_SIZE(slrp_7.adc_gos); ++i) {
        sstream << ',' << +slrp_7.adc_gos[i];
    }
    sstream << ','
            << +slrp_7.phos3                        << ','
            << +slrp_7.phos2                        << ','
            << +slrp_7.phos1                        << ','
            << +slrp_7.phos0                        << ','
            << +slrp_7.phos7                        << ','
            << +slrp_7.phos6                        << ','
            << +slrp_7.phos5                        << ','
            << +slrp_7.phos4                        ;

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::Dump_5nm_scc(const struct slrp_reg &slrp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;
    slrp_5nm slrp_5;
    slrp_5nm_unpack(&slrp_5, slrp.page_data.slrp_data_set.data);

    sstream << +slrp_5.ctle_vos_val                  << ','
            << +slrp_5.ctle_vos_stg1                 << ','
            << +slrp_5.ctle_vos_aux                  << ','
            << +slrp_5.adc_recording_admin           << ','
            << +slrp_5.adc_recording_status          << ','
            << +slrp_5.fixed_val                     << ','
            << +slrp_5.fine_val                      << ','
            << +slrp_5.coarse_val                    ;

    for (int i = 0; i < 16; ++i) {
        sstream << ',' << +slrp_5.tah_vos[i];
    }
    for (int i = 0; i < 16; ++i) {
        sstream << ',' << +slrp_5.adc_vos[i];
    }
    for (int i = 0; i < 16; ++i) {
        sstream << ',' << +slrp_5.adc_gos[i];
    }

    for (int i = 0; i < ACC_REG_SLRP_MAX_FIELDS_NUM - ACC_REG_SLRP5NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRPRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if ((p_export_data_phy_port && this->m_pnat == ACC_REG_PNAT_IB_PORT) ||
            (p_export_data_phy_node && this->m_pnat == ACC_REG_PNAT_OOB_PORT)) {
        struct slrp_reg *p_slrp_reg = new struct slrp_reg(areg.regs.slrp);

        switch (p_slrp_reg->version) {
            case 0: // 40nm
            case 1: // 28nm
                slrp_40nm_28nm_unpack(&p_slrp_reg->page_data.slrp_40nm_28nm, areg.regs.slrp.page_data.slrp_data_set.data);
                break;
            case 3: // 16nm
                slrp_16nm_unpack(&p_slrp_reg->page_data.slrp_16nm, areg.regs.slrp.page_data.slrp_data_set.data);
                break;
            case 4: // 7nm
                slrp_7nm_unpack(&p_slrp_reg->page_data.slrp_7nm, areg.regs.slrp.page_data.slrp_data_set.data);
                break;
            case 5: // 5nm
                slrp_5nm_unpack(&p_slrp_reg->page_data.slrp_5nm, areg.regs.slrp.page_data.slrp_data_set.data);
                break;
            default:
                break;
        }

        if (this->m_pnat == ACC_REG_PNAT_IB_PORT) {
            p_export_data_phy_port->p_slrp_reg[((AccRegKeyPortLane *)p_key)->lane] = p_slrp_reg;
        } else if (this->m_pnat == ACC_REG_PNAT_OOB_PORT) {
            AccRegKeyPCIPortLane *p_pcipl_key = (AccRegKeyPCIPortLane *)p_key;
            p_export_data_phy_node->p_slrp_reg[p_pcipl_key->port_num] = p_slrp_reg;
        }
    }

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* SLTP Register functions ***
//*****************************************
SLTPRegister::SLTPRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(phy_diag,
               ACCESS_REGISTER_ID_SLTP,
               (const unpack_data_func_t)sltp_reg_unpack,
               sn,
               ACC_REG_SLTP_NAME,
               ACC_REG_SLTP_MAX_FIELDS_NUM,
               NSB::get(this),
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLTPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct sltp_reg sltp;
    CLEAR_STRUCT(sltp);
    sltp.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    sltp.lane = ((AccRegKeyPortLane *)p_key)->lane;
    sltp.pnat = m_pnat;
    sltp_reg_pack(&sltp, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLTPRegister::Header_Dump_40nm_28nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                  << ','
            << "version_40nm=0_28nm=1"   << ','
            << "local_port"              << ','
            << "pnat"                    << ','
            << "tx_policy"               << ','
            << "conf_mod"                << ','
            << "lane"                    << ','
            << "lane_speed"              << ','
            << "port_type"               << ','
            << "c_db"                    << ','            
            << "polarity"                << ','
            << "ob_tap0"                 << ','
            << "ob_tap1"                 << ','
            << "ob_tap2"                 << ','
            << "ob_leva"                 << ','
            << "ob_preemp_mode"          << ','
            << "ob_reg"                  << ','
            << "ob_bias"                 << ','
            << "ob_norm"                 << ','
            << "ob_bad_stat";

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Header_Dump_16nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"             << ','
            << "version_16nm=3"     << ','
            << "local_port"         << ','
            << "pnat"               << ','
            << "tx_policy"          << ','
            << "conf_mod"           << ','
            << "lane"               << ','
            << "lane_speed"         << ','
            << "port_type"          << ','
            << "c_db"               << ','
            << "pre_2_tap"          << ','
            << "pre_tap"            << ','
            << "main_tap"           << ','
            << "post_tap"           << ','
            << "ob_m2lp"            << ','
            << "ob_amp"             << ','
            << "ob_alev_out"        << ','
            << "ob_bad_stat"        << ','
            << "obplev"             << ','
            << "obnlev"             << ','
            << "regn_bfm1p"         << ','
            << "regp_bfm1n"         << ','
            << "blev"               << ','
            << "tx_alev"            << ','
            << "alev_plus_bfm2"     << ','
            << "alev_minus_bfm2";

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"             << ','
            << "version_7nm=4"      << ','
            << "local_port"         << ','
            << "pnat"               << ','
            << "tx_policy"          << ','
            << "conf_mod"           << ','
            << "lane"               << ','
            << "lane_speed"         << ','
            << "port_type"          << ','
            << "c_db"               << ','
            << "vs_peer_db"         << ','
            << "drv_amp"            << ','
            << "ob_bad_stat"        << ','
            << "fir_pre3"           << ','
            << "fir_pre2"           << ','
            << "fir_pre1"           << ','
            << "fir_main"           << ','
            << "fir_post1";

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"             << ','
            << "version_5nm=5"      << ','
            << "local_port"         << ','
            << "pnat"               << ','
            << "tx_policy"          << ','
            << "conf_mod"           << ','
            << "lane"               << ','
            << "lane_speed"         << ','
            << "port_type"          << ','
            << "c_db"               << ','
            << "ob_bad_stat"        << ','
            << "tap3"               << ','
            << "tap2"               << ','
            << "tap1"               << ','
            << "tap0"               << ','
            << "tap7"               << ','
            << "tap6"               << ','
            << "tap5"               << ','
            << "tap4"               << ','
            << "tap11"              << ','
            << "tap10"              << ','
            << "tap9"               << ','
            << "tap8"               << ','
            << "upper_eye_amp"      << ','
            << "middle_eye_amp"     << ','
            << "lower_eye_amp";

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_40nm_28nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct sltp_reg &sltp = areg.regs.sltp;

    sstream << hex
            << "0x" << +sltp.status        << ','
            << "0x" << +sltp.version       << ','
            << "0x" << +sltp.local_port    << ','
            << "0x" << +sltp.pnat          << ','
            << "0x" << +sltp.tx_policy     << ','
            << "0x" << +sltp.conf_mod      << ','
            << "0x" << +sltp.lane          << ','
            << "0x" << +sltp.lane_speed    << ','
            << "0x" << +sltp.port_type     << ','
            << "0x" << +sltp.c_db          << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (sltp.version) {

        case 0: // 40nm
        case 1: // 28nm
            Dump_40nm_28nm(sltp, sstream);
            break;

        case 3: // 16nm
            Dump_16nm(sltp, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(sltp, sstream);
            break;
        case 5: // 5nm
            Dump_5nm(sltp, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +sltp.version << '"';
            for (int i = 0; i < ACC_REG_SLTP_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLTP: %d, on node: " U64H_FMT ".\n",
                        sltp.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Dump_40nm_28nm(const struct sltp_reg &sltp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sltp_28nm_40nm sltp_40_28;
    sltp_28nm_40nm_unpack(&sltp_40_28, sltp.page_data.sltp_data_set.data);

    sstream << "0x" << +sltp_40_28.polarity           << ','
            << "0x" << +sltp_40_28.ob_tap0            << ','
            << "0x" << +sltp_40_28.ob_tap1            << ','
            << "0x" << +sltp_40_28.ob_tap2            << ','
            << "0x" << +sltp_40_28.ob_leva            << ','
            << "0x" << +sltp_40_28.ob_preemp_mode     << ','
            << "0x" << +sltp_40_28.ob_reg             << ','
            << "0x" << +sltp_40_28.ob_bias            << ','
            << "0x" << +sltp_40_28.ob_norm            << ','
            << "0x" << +sltp_40_28.ob_bad_stat        ;

    for (int i = 0; i < ACC_REG_SLTP_MAX_FIELDS_NUM - ACC_REG_SLTP28NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Dump_16nm(const struct sltp_reg &sltp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sltp_16nm sltp_16;
    sltp_16nm_unpack(&sltp_16, sltp.page_data.sltp_data_set.data);

    sstream << "0x" << +sltp_16.pre_2_tap          << ','
            << "0x" << +sltp_16.pre_tap            << ','
            << "0x" << +sltp_16.main_tap           << ','
            << "0x" << +sltp_16.post_tap           << ','
            << "0x" << +sltp_16.ob_m2lp            << ','
            << "0x" << +sltp_16.ob_amp             << ','
            << "0x" << +sltp_16.ob_alev_out        << ','
            << "0x" << +sltp_16.ob_bad_stat        << ','
            << "0x" << +sltp_16.obplev             << ','
            << "0x" << +sltp_16.obnlev             << ','
            << "0x" << +sltp_16.regn_bfm1p         << ','
            << "0x" << +sltp_16.regp_bfm1n         << ','
            << "0x" << +sltp_16.blev               << ','
            << "0x" << +sltp_16.tx_alev            << ','
            << "0x" << +sltp_16.alev_plus_bfm2     << ','
            << "0x" << +sltp_16.alev_minus_bfm2    ;

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Dump_7nm(const struct sltp_reg &sltp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sltp_7nm sltp_7;
    sltp_7nm_unpack(&sltp_7, sltp.page_data.sltp_data_set.data);

    sstream << "0x" << +sltp_7.vs_peer_db  << ','
            << "0x" << +sltp_7.drv_amp     << ','
            << "0x" << +sltp_7.ob_bad_stat << ','
            << "0x" << +sltp_7.fir_pre3    << ','
            << "0x" << +sltp_7.fir_pre2    << ','
            << "0x" << +sltp_7.fir_pre1    << ','
            << "0x" << +sltp_7.fir_main    << ','
            << "0x" << +sltp_7.fir_post1   ;

    for (int i = 0; i < ACC_REG_SLTP_MAX_FIELDS_NUM - ACC_REG_SLTP7NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::Dump_5nm(const struct sltp_reg &sltp, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sltp_5nm sltp_5;
    sltp_5nm_unpack(&sltp_5, sltp.page_data.sltp_data_set.data);

    sstream << "0x" << +sltp_5.ob_bad_stat         << ','
            << "0x" << +sltp_5.tap3                << ','
            << "0x" << +sltp_5.tap2                << ','
            << "0x" << +sltp_5.tap1                << ','
            << "0x" << +sltp_5.tap0                << ','
            << "0x" << +sltp_5.tap7                << ','
            << "0x" << +sltp_5.tap6                << ','
            << "0x" << +sltp_5.tap5                << ','
            << "0x" << +sltp_5.tap4                << ','
            << "0x" << +sltp_5.tap11               << ','
            << "0x" << +sltp_5.tap10               << ','
            << "0x" << +sltp_5.tap9                << ','
            << "0x" << +sltp_5.tap8                << ','
            << "0x" << +sltp_5.upper_eye_amp       << ','
            << "0x" << +sltp_5.middle_eye_amp      << ','
            << "0x" << +sltp_5.lower_eye_amp;

#if ACC_REG_SLTP_MAX_FIELDS_NUM < ACC_REG_SLTP5NM_FIELDS_NUM
#error "Number of field of SLTP 5nm more of max fields of SLTP access register"
#endif

#if ACC_REG_SLTP_MAX_FIELDS_NUM - ACC_REG_SLTP5NM_FIELDS_NUM
    for (int i = 0; i < ACC_REG_SLTP_MAX_FIELDS_NUM - ACC_REG_SLTP5NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLTPRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if ((p_export_data_phy_port && this->m_pnat == ACC_REG_PNAT_IB_PORT) ||
            (p_export_data_phy_node && this->m_pnat == ACC_REG_PNAT_OOB_PORT)) {
        struct sltp_reg *p_sltp_reg = new struct sltp_reg(areg.regs.sltp);

        switch (p_sltp_reg->version) {
            case 0: // 40nm
            case 1: // 28nm
                sltp_28nm_40nm_unpack(&p_sltp_reg->page_data.sltp_28nm_40nm, areg.regs.sltp.page_data.sltp_data_set.data);
                break;
            case 3: // 16nm
                sltp_16nm_unpack(&p_sltp_reg->page_data.sltp_16nm, areg.regs.sltp.page_data.sltp_data_set.data);
                break;
            case 4: // 7nm
                sltp_7nm_unpack(&p_sltp_reg->page_data.sltp_7nm, areg.regs.sltp.page_data.sltp_data_set.data);
                break;
            case 5: // 5nm
                sltp_5nm_unpack(&p_sltp_reg->page_data.sltp_5nm, areg.regs.sltp.page_data.sltp_data_set.data);
                break;
            default:
                break;
        }

        if (this->m_pnat == ACC_REG_PNAT_IB_PORT) {
            p_export_data_phy_port->p_sltp_reg[((AccRegKeyPortLane *)p_key)->lane] = p_sltp_reg;
        } else if (this->m_pnat == ACC_REG_PNAT_OOB_PORT) {
            AccRegKeyPCIPortLane *p_pcipl_key = (AccRegKeyPCIPortLane *)p_key;
            p_export_data_phy_node->p_sltp_reg[p_pcipl_key->port_num] = p_sltp_reg;
        }
    }

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//************* SLRG Register functions ***
//*****************************************
SLRGRegister::SLRGRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map):
    SLRegister(phy_diag,
               ACCESS_REGISTER_ID_SLRG,
               (const unpack_data_func_t)slrg_reg_unpack,
               sn,
               ACC_REG_SLRG_NAME,
               ACC_REG_SLRG_MAX_FIELDS_NUM,
               NSB::get(this),
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLRGRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slrg_reg slrg;
    CLEAR_STRUCT(slrg);

    slrg.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slrg.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slrg.pnat = m_pnat;
    slrg_reg_pack(&slrg, data);

    if (p_node &&
            (p_node->serdes_version == SerDes_7nm || p_node->serdes_version == SerDes_5nm)) {
        slrg.page_data.slrg_7nm.fom_measurment = 0xF;
        slrg_7nm_pack(&slrg.page_data.slrg_7nm, data + 4); // +4 - packed SLRG header size
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLRGRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
     IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_40nm_28nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Header_Dump_16nm(stringstream &sstream) const
{
    IBDIAG_ENTER;

    sstream << "status"                << ','
            << "version_16nm=3"        << ','
            << "local_port"            << ','
            << "pnat"                  << ','
            << "lp_msb"                << ','
            << "lane"                  << ','
            << "port_type"             << ','
            << "test_mode"             << ','
            << "grade_lane_speed"      << ','
            << "grade_version"         << ','
            << "grade"                 << ','
            << "height_eo_pos_up"      << ','
            << "height_eo_neg_up"      << ','
            << "phase_eo_pos_up"       << ','
            << "phase_eo_neg_up"       << ','
            << "height_eo_pos_mid"     << ','
            << "height_eo_neg_mid"     << ','
            << "phase_eo_pos_mid"      << ','
            << "phase_eo_neg_mid"      << ','
            << "height_eo_pos_low"     << ','
            << "height_eo_neg_low"     << ','
            << "phase_eo_pos_low"      << ','
            << "phase_eo_neg_low"      << ','
            << "offset_units"          << ','
            << "phase_units"           << ','
            << "height_grade_type"     << ','
            << "phase_grade_type"      << ','
            << "mid_eye_grade"         << ','
            << "up_eye_grade"          << ','
            << "dn_eye_grade";

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Header_Dump_40nm_28nm(stringstream &sstream) const
{
    IBDIAG_ENTER;

    sstream << "status"                   << ','
            << "version_40nm=0_28nm=1"    << ','
            << "local_port"               << ','
            << "pnat"                     << ','
            << "lp_msb"                   << ','
            << "lane"                     << ','
            << "port_type"                << ','
            << "test_mode"                << ','
            << "grade_lane_speed"         << ','
            << "grade_version"            << ','
            << "grade"                    << ','
            << "offset_units"             << ','
            << "phase_units"              << ','
            << "height_grade_type"        << ','
            << "height_grade"             << ','
            << "height_dz"                << ','
            << "height_dv"                << ','
            << "height_sigma"             << ','
            << "height_eo_pos"            << ','
            << "height_eo_neg"            << ','
            << "phase_grade_type"         << ','
            << "phase_grade"              << ','
            << "phase_eo_pos"             << ','
            << "phase_eo_neg"             << ','
            << "ffe_set_tested"           << ','
            << "test_errors_per_lane";

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                   << ','
            << "version_7nm=4"            << ','
            << "local_port"               << ','
            << "pnat"                     << ','
            << "lp_msb"                   << ','
            << "lane"                     << ','
            << "port_type"                << ','
            << "test_mode"                << ','
            << "fom_measurment"           << ','
            << "fom_mode"                 << ','
            << "initial_fom"              << ','
            << "last_fom"                 << ','
            << "upper_eye"                << ','
            << "mid_eye"                  << ','
            << "lower_eye"                << ','
            << "fom_ext_conf"             << ','
            << "fom_ext_conf_cap"         << ','
            << "nblks_max"                << ','
            << "nerrs_min";

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "status"                   << ','
            << "version_5nm=5"            << ','
            << "local_port"               << ','
            << "pnat"                     << ','
            << "lp_msb"                   << ','
            << "lane"                     << ','
            << "port_type"                << ','
            << "test_mode"                << ','
            << "fom_measurment"           << ','
            << "fom_mode"                 << ','
            << "initial_fom"              << ','
            << "last_fom"                 << ','
            << "upper_eye"                << ','
            << "mid_eye"                  << ','
            << "lower_eye"                << ','
            << "fom_ext_conf"             << ','
            << "fom_ext_conf_cap"         << ','
            << "nblks_max"                << ','
            << "nerrs_min";

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct slrg_reg & slrg = areg.regs.slrg;

    sstream << +slrg.status         << ','
            << +slrg.version        << ','
            << +slrg.local_port     << ','
            << +slrg.pnat           << ','
            << +slrg.lp_msb         << ','
            << +slrg.lane           << ','
            << +slrg.port_type      << ','
            << +slrg.test_mode      << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (slrg.version) {

    case 0: // 40nm
    case 1: // 28nm
        Dump_40nm_28nm(slrg, sstream);
        break;

    case 3: // 16nm
        Dump_16nm(slrg, sstream);
        break;

    case 4: // 7nm
        Dump_7nm(slrg, sstream);
        break;

    case 5: // 5nm
        Dump_5nm(slrg, sstream);
        break;

    default:
        static bool should_print = true;

        sstream << '"' << "Unknown version " << +slrg.version << '"';
        for (int i = 0; i < ACC_REG_SLRG_MAX_FIELDS_NUM - 1; ++i)
            sstream << ",NA"; // add NAs to fill the empty fields

        if (should_print) {
            WARN_PRINT("Unknown version for SLRG: %d, on node: " U64H_FMT ".\n",
                    slrg.version, key.node_guid);
            should_print = false;
        }
    }

    sstream << endl;
    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Dump_40nm_28nm(const struct slrg_reg &slrg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrg_40nm_28nm slrg_40_28;
    slrg_40nm_28nm_unpack(&slrg_40_28, slrg.page_data.slrg_data_set.data);

    sstream << +slrg_40_28.grade_lane_speed         << ','
            << +slrg_40_28.grade_version            << ','
            << +slrg_40_28.grade                    << ','
            << +slrg_40_28.offset_units             << ','
            << +slrg_40_28.phase_units              << ','
            << +slrg_40_28.height_grade_type        << ','
            << +slrg_40_28.height_grade             << ','
            << +slrg_40_28.height_dz                << ','
            << +slrg_40_28.height_dv                << ','
            << +slrg_40_28.height_sigma             << ','
            << +slrg_40_28.height_eo_pos            << ','
            << +slrg_40_28.height_eo_neg            << ','
            << +slrg_40_28.phase_grade_type         << ','
            << +slrg_40_28.phase_grade              << ','
            << +slrg_40_28.phase_eo_pos             << ','
            << +slrg_40_28.phase_eo_neg             << ','
            << +slrg_40_28.ffe_set_tested           << ','
            << +slrg_40_28.test_errors_per_lane     ;

    for (int i = 0; i < ACC_REG_SLRG_MAX_FIELDS_NUM - ACC_REG_SLRG28NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Dump_16nm(const struct slrg_reg &slrg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrg_16nm slrg_16;
    slrg_16nm_unpack(&slrg_16, slrg.page_data.slrg_data_set.data);

    sstream << +slrg_16.grade_lane_speed      << ','
            << +slrg_16.grade_version         << ','
            << +slrg_16.grade                 << ','
            << +slrg_16.height_eo_pos_up      << ','
            << +slrg_16.height_eo_neg_up      << ','
            << +slrg_16.phase_eo_pos_up       << ','
            << +slrg_16.phase_eo_neg_up       << ','
            << +slrg_16.height_eo_pos_mid     << ','
            << +slrg_16.height_eo_neg_mid     << ','
            << +slrg_16.phase_eo_pos_mid      << ','
            << +slrg_16.phase_eo_neg_mid      << ','
            << +slrg_16.height_eo_pos_low     << ','
            << +slrg_16.height_eo_neg_low     << ','
            << +slrg_16.phase_eo_pos_low      << ','
            << +slrg_16.phase_eo_neg_low      << ','
            << +slrg_16.offset_units          << ','
            << +slrg_16.phase_units           << ','
            << +slrg_16.height_grade_type     << ','
            << +slrg_16.phase_grade_type      << ','
            << +slrg_16.mid_eye_grade         << ','
            << +slrg_16.up_eye_grade          << ','
            << +slrg_16.dn_eye_grade          ;

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Dump_7nm(const struct slrg_reg &slrg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrg_7nm slrg_7;
    slrg_7nm_unpack(&slrg_7, slrg.page_data.slrg_data_set.data);

    sstream << +slrg_7.fom_measurment    << ','
            << +slrg_7.fom_mode          << ','
            << +slrg_7.initial_fom       << ','
            << +slrg_7.last_fom          << ','
            << +slrg_7.upper_eye         << ','
            << +slrg_7.mid_eye           << ','
            << +slrg_7.lower_eye         << ','
            << +slrg_7.fom_ext_conf      << ','
            << +slrg_7.fom_ext_conf_cap  << ','
            << +slrg_7.nblks_max         << ','
            << +slrg_7.nerrs_min;

    for (int i = 0; i < ACC_REG_SLRG_MAX_FIELDS_NUM - ACC_REG_SLRG7NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::Dump_5nm(const struct slrg_reg &slrg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrg_5nm slrg_5;
    slrg_5nm_unpack(&slrg_5, slrg.page_data.slrg_data_set.data);

    sstream << +slrg_5.fom_measurment       << ','
            << +slrg_5.fom_mode             << ','
            << +slrg_5.initial_fom          << ','
            << +slrg_5.last_fom             << ','
            << +slrg_5.upper_eye            << ','
            << +slrg_5.mid_eye              << ','
            << +slrg_5.lower_eye            << ','
            << +slrg_5.fom_ext_conf         << ','
            << +slrg_5.fom_ext_conf_cap     << ','
            << +slrg_5.nblks_max            << ','
            << +slrg_5.nerrs_min;

    for (int i = 0; i < ACC_REG_SLRG_MAX_FIELDS_NUM - ACC_REG_SLRG5NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRGRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if ((p_export_data_phy_port && this->m_pnat == ACC_REG_PNAT_IB_PORT) ||
            (p_export_data_phy_node && this->m_pnat == ACC_REG_PNAT_OOB_PORT)) {
        struct slrg_reg *p_slrg_reg = new struct slrg_reg(areg.regs.slrg);

        switch (p_slrg_reg->version) {
            case 0: // 40nm
            case 1: // 28nm
                slrg_40nm_28nm_unpack(&p_slrg_reg->page_data.slrg_40nm_28nm, areg.regs.slrg.page_data.slrg_data_set.data);
                break;
            case 3: // 16nm
                slrg_16nm_unpack(&p_slrg_reg->page_data.slrg_16nm, areg.regs.slrg.page_data.slrg_data_set.data);
                break;
            case 4: // 7nm
            case 5: // 5nm
                slrg_7nm_unpack(&p_slrg_reg->page_data.slrg_7nm, areg.regs.slrg.page_data.slrg_data_set.data);
                break;
            default:
                break;
        }

        if (this->m_pnat == ACC_REG_PNAT_IB_PORT) {
            p_export_data_phy_port->p_slrg_reg[((AccRegKeyPortLane *)p_key)->lane] = p_slrg_reg;
        } else if (this->m_pnat == ACC_REG_PNAT_OOB_PORT) {
            AccRegKeyPCIPortLane *p_pcipl_key = (AccRegKeyPCIPortLane *)p_key;
            p_export_data_phy_node->p_slrg_reg[p_pcipl_key->port_num] = p_slrg_reg;
        }
    }

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLLM Register functions      ***
//*****************************************
SLLMRegister::SLLMRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map,
                           map_akey_areg *mpir_map):
    SLRegister(phy_diag,
               ACCESS_REGISTER_ID_SLLM,
               (const unpack_data_func_t)sllm_reg_unpack,
               sn,
               ACC_REG_SLLM_NAME,
               ACC_REG_SLLM7NM_FIELDS_NUM,
               NSB::get(this),
               mpein_map,
               mpir_map),
    m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLLMRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct sllm_reg sllm;
    CLEAR_STRUCT(sllm);
    sllm.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    sllm.lane = ((AccRegKeyPortLane *)p_key)->lane;
    sllm.pnat = this->m_pnat;
    sllm.port_type = 0;
    sllm_reg_pack(&sllm, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLLMRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLLMRegister::Header_Dump_7nm(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "status"                 << ','
            << "version_7nm=4"          << ','
            << "local_port"             << ','
            << "pnat"                   << ','
            << "lp_msb"                 << ','
            << "lane"                   << ','
            << "port_type"              << ','
            << "c_db"                   << ','
            << "ctle_peq_en"            << ','
            << "peq_tsense_en"          << ','
            << "peq_f1_adapt_skip"      << ','
            << "vref_peq_en"            << ','
            << "dffe_peq_scout_skip"    << ','
            << "peq_train_mode"         << ','
            << "peq_vref_iters"         << ','
            << "peq_adc_vref_step"      << ','
            << "dffe_peq_en"            << ','
            << "peq_dffe_iters"         << ','
            << "peq_dffe_delay"         << ','
            << "ctle_peq_cnt"           << ','
            << "ber_mon_mantissa"       << ','
            << "ber_mon_exp";

    IBDIAGNET_RETURN_VOID;
}

void SLLMRegister::Header_Dump_16nm(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "status"               << ','
            << "version_16nm=3"       << ','
            << "local_port"           << ','
            << "pnat"                 << ','
            << "lp_msb"               << ','
            << "lane"                 << ','
            << "port_type"            << ','
            << "c_db"                 << ','
            << "lm_active"            << ','
            << "lm_was_active"        << ','
            << "pib_gw_lock"          << ','
            << "lm_en"                << ','
            << "lm_clk90_fl_err_max"  << ','
            << "lm_clk90_fl_err_acc"  << ','
            << "ib3_max_lm_90_tries"  << ','
            << "lm_counter_up"        << ','
            << "ib1_max_lm_90_tries"  << ','
            << "lm_counter_mid"       << ','
            << "lm_counter_dn";

    IBDIAG_RETURN_VOID;
}

void SLLMRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct sllm_reg &sllm = areg.regs.sllm;
    sstream << +sllm.status         << ','
            << +sllm.version        << ','
            << +sllm.local_port     << ','
            << +sllm.pnat           << ','
            << +sllm.lp_msb         << ','
            << +sllm.lane           << ','
            << +sllm.port_type      << ','
            << +sllm.c_db           << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (sllm.version) {

        case 3: // 16nm
            Dump_16nm(sllm, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(sllm, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +sllm.version << '"';
            for (int i = 0; i < ACC_REG_SLLM_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLLM: %d, on node: " U64H_FMT ".\n",
                        sllm.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void SLLMRegister::Dump_16nm(const struct sllm_reg &sllm, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sllm_16nm sllm_16;
    sllm_16nm_unpack(&sllm_16, sllm.page_data.sllm_data_set.data);

    sstream << +sllm_16.lm_active                   << ','
            << +sllm_16.lm_was_active               << ','
            << +sllm_16.pib_gw_lock                 << ','
            << +sllm_16.lm_en                       << ','
            << +sllm_16.lm_clk90_fl_err_max         << ','
            << +sllm_16.lm_clk90_fl_err_acc         << ','
            << +sllm_16.ib3_max_lm_90_tries         << ','
            << +sllm_16.lm_counter_up               << ','
            << +sllm_16.ib1_max_lm_90_tries         << ','
            << +sllm_16.lm_counter_mid              << ','
            << +sllm_16.lm_counter_dn               << ','
            << "NA,NA,NA";  // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLLMRegister::Dump_7nm(const struct sllm_reg &sllm, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sllm_7nm sllm_7;
    sllm_7nm_unpack(&sllm_7, sllm.page_data.sllm_data_set.data);

    sstream << +sllm_7.ctle_peq_en                 << ','
            << +sllm_7.peq_tsense_en               << ','
            << +sllm_7.peq_f1_adapt_skip           << ','
            << +sllm_7.vref_peq_en                 << ','
            << +sllm_7.dffe_peq_scout_skip         << ','
            << +sllm_7.peq_train_mode              << ','
            << +sllm_7.peq_vref_iters              << ','
            << +sllm_7.peq_adc_vref_step           << ','
            << +sllm_7.dffe_peq_en                 << ','
            << +sllm_7.peq_dffe_iters              << ','
            << +sllm_7.peq_dffe_delay              << ','
            << +sllm_7.ctle_peq_cnt                << ','
            << +sllm_7.ber_mon_mantissa            << ','
            << +sllm_7.ber_mon_exp                 ;

    IBDIAGNET_RETURN_VOID;
}

void SLLMRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port) {
        struct sllm_reg *p_sllm_reg = new struct sllm_reg(areg.regs.sllm);

        switch (p_sllm_reg->version) {
            case 3: // 16nm
                sllm_16nm_unpack(&p_sllm_reg->page_data.sllm_16nm, areg.regs.sllm.page_data.sllm_data_set.data);
                break;
            case 4: // 7nm
                sllm_7nm_unpack(&p_sllm_reg->page_data.sllm_7nm, areg.regs.sllm.page_data.sllm_data_set.data);
                break;
            default:
                break;
        }

        p_export_data_phy_port->p_sllm_reg[((AccRegKeyPortLane *)p_key)->lane] = p_sllm_reg;
    }

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* MPEIN Register functions ***
//*****************************************
MPEINRegister::MPEINRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MPEIN,
                (const unpack_data_func_t)mpein_reg_unpack,
                ACC_REG_MPEIN_INTERNAL_SECTION_NAME,
                ACC_REG_MPEIN_NAME,
                ACC_REG_MPEIN_FIELDS_NUM,
                NSB::get(this))
{
}


int MPEINRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)p_key;
    struct mpein_reg mpein;
    CLEAR_STRUCT(mpein);

    mpein.depth = p_dpnkey->depth;
    mpein.pcie_index = p_dpnkey->pci_idx;
    mpein.node = p_dpnkey->pci_node;
    mpein_reg_pack(&mpein, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MPEINRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                           << '#'
            << handler_header                 << ','
            << "capability_mask"              << ','
            << "link_width_enabled"           << ','
            << "link_speed_enabled"           << ','
            << "lane0_physical_position"      << ','
            << "link_width_active"            << ','
            << "link_speed_active"            << ','
            << "num_of_pfs"                   << ','
            << "num_of_vfs"                   << ','
            << "bdf0"                         << ','
            << "max_read_request_size"        << ','
            << "max_payload_size"             << ','
            << "pwr_status"                   << ','
            << "port_type"                    << ','
            << "lane_reversal"                << ','
            << "link_peer_max_speed"          << ','
            << "pci_power"                    << ','
            << "device_status"                << ','
            << "port_state"                   << ','
            << "receiver_detect_result";
#endif
    IBDIAGNET_RETURN_VOID;
}

void MPEINRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct mpein_reg &p_mpein = areg.regs.mpein;

    // HEADER
    // NodeGuid, PCIIndex, Depth, PCINode,
    // capability_mask, link_width_enabled, link_speed_enabled, lane0_physical_position,
    // link_width_active, link_speed_active, num_of_pfs, num_of_vfs, bdf0, max_read_request_size,
    // max_payload_size, pwr_status, port_type, lane_reversal, link_peer_max_speed,
    // pci_power, device_status, port_state, receiver_detect_result

    sstream << hex
            << "0x" << +p_mpein.capability_mask              << ','
            << "0x" << +p_mpein.link_width_enabled           << ','
            << "0x" << +p_mpein.link_speed_enabled           << ','
            << "0x" << +p_mpein.lane0_physical_position      << ','
            << "0x" << +p_mpein.link_width_active            << ','
            << "0x" << +p_mpein.link_speed_active            << ','
            << "0x" << +p_mpein.num_of_pfs                   << ','
            << "0x" << +p_mpein.num_of_vfs                   << ','
            << "0x" << +p_mpein.bdf0                         << ','
            << "0x" << +p_mpein.max_read_request_size        << ','
            << "0x" << +p_mpein.max_payload_size             << ','
            << "0x" << +p_mpein.pwr_status                   << ','
            << "0x" << +p_mpein.port_type                    << ','
            << "0x" << +p_mpein.lane_reversal                << ','
            << "0x" << +p_mpein.link_peer_max_speed          << ','
            << "0x" << +p_mpein.pci_power                    << ','
            << "0x" << +p_mpein.device_status                << ','
            << "0x" << +p_mpein.port_state                   << ','
            << "0x" << +p_mpein.receiver_detect_result
            << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

/*
 * Build DB in the following order:
 * For each CA
 *   For each PCI index
 *     For each PCI depth
 *       For each PCI node
*/
int MPEINRegister::BuildDB(AccRegHandler* p_handler,
                           list_p_fabric_general_err &phy_errors,
                           ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    // clbck_data.m_p_obj            - handler
    // clbck_data.m_data1            - node
    // clbck_data.m_data2            - key
    // clbck_data.m_data3            - port
    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func = &AccRegHandler_ForwardClbck<MPEINRegister,
                &MPEINRegister::LoopPCIDepth>;

    clbck_data.m_p_obj = p_handler;

    p_handler->SetHeader(PCI_DPN_HEADER);

    for (map_str_psys::iterator sys_iter = m_phy_diag->GetFabric()->SystemByName.begin();
            sys_iter != m_phy_diag->GetFabric()->SystemByName.end(); sys_iter++) {

        IBSystem *p_curr_sys = sys_iter->second;
        if (!p_curr_sys) {
            m_phy_diag->SetLastError("DB error - found null node in"  \
                    " SystemByName map for key = %s",
                    sys_iter->first.c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        set<uint64_t> sys_guids;

        for (map_str_pnode::iterator it = p_curr_sys->NodeByName.begin();
             it != p_curr_sys->NodeByName.end(); it++) {

            IBNode *p_curr_node = it->second;
            if (!p_curr_node) {
                m_phy_diag->SetLastError("DB error - found null node in"  \
                        " NodeByName map in IBSystem for key = %s",
                        p_curr_sys->NodeByName.begin()->first.c_str());
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            if (p_curr_node->type == IB_SW_NODE)
                continue;

            // skip ANs
            if (p_curr_node->isSpecialNode())
                continue;

            // todo - filter node not in scope already in the SystemByName map
            //        as in NodeByName map.
            // skip not in scope
            if (!p_curr_node->getInSubFabric())
                continue;

            AccRegVia_t acc_reg_via = Validation(p_curr_node, rc);

            if (acc_reg_via == NOT_SUP_ACC_REG)
                continue;

            pair<set<uint64_t>::iterator,bool> ret = sys_guids.insert(p_curr_node->system_guid_get());
            if (!ret.second)
                continue;

            clbck_data.m_data1 = p_curr_node;
            clbck_data.m_data4 = (void*)(u_int64_t)acc_reg_via;

            for (u_int32_t port_num = 1; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
                IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)port_num);

                if (!p_curr_port || !p_curr_port->is_data_worthy())
                   continue;

                clbck_data.m_data3 = p_curr_port;

                for (u_int8_t idx = 0; idx < MAX_PCI_IDX; ++idx) {

                    AccRegKeyDPN *p_dpnkey = new AccRegKeyDPN(
                            p_curr_port, // port
                            0,        // depth
                            idx,      // pci index
                            0);       // pci node

                    clbck_data.m_data2 = p_dpnkey;

                    AccessRegister acc_reg;
                    CLEAR_STRUCT(acc_reg);

                    rc = p_handler->SendAccReg(acc_reg_via,
                                          p_curr_node,
                                          p_curr_port->num,
                                          p_curr_port->base_lid,
                                          acc_reg,
                                          p_dpnkey,
                                          p_progress_bar,
                                          &clbck_data);

                    if (rc == IBDIAG_ERR_CODE_DB_ERR)
                        IBDIAGNET_RETURN(rc);

                } // pci index loop

                break; // send only for one port per node
            } // port loop
        } // nodes loop
    } // system loop

    IBDIAGNET_RETURN(rc);
}

int MPEINRegister::LoopPCIDepth(const clbck_data_t &clbck_data,
                                int rec_status,
                                void *p_attribute_data)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    AccRegHandler *p_handler = (AccRegHandler*)clbck_data.m_p_obj;
    IBNode *p_curr_node = (IBNode*)clbck_data.m_data1;
    AccRegKeyDPN *p_dpnkey_old = (AccRegKeyDPN *)clbck_data.m_data2;
    IBPort *p_curr_port = (IBPort*)clbck_data.m_data3;
    AccRegVia_t acc_reg_via = (AccRegVia_t)(u_int64_t)clbck_data.m_data4;

    direct_route_t *p_curr_direct_route = m_phy_diag->GetIBDiag()->GetDR(p_curr_node);
    if (!p_curr_direct_route) {
        m_phy_diag->SetLastError("DB error - current direct route is NULL. Node GUID="
                U64H_FMT ", node name %s.", p_curr_node->guid_get(), p_curr_node->getName().c_str());

        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    mpein_reg mpein;
    CLEAR_STRUCT(mpein);

    // invoke handler callback in order to handle return status and insert reply to the map
    if (acc_reg_via == VIA_SMP) {
        rc = p_handler->SMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);
        if (rc)
            IBDIAGNET_RETURN(rc);

        SMP_AccessRegister *p_acc_reg_smp = (SMP_AccessRegister*)p_attribute_data;
        mpein_reg_unpack(&mpein, p_acc_reg_smp->reg.data);
    }

    if (acc_reg_via == VIA_GMP){
        rc = p_handler->GMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);
        if (rc)
            IBDIAGNET_RETURN(rc);

        struct GMP_AccessRegister *p_acc_reg_gmp = (struct GMP_AccessRegister *)p_attribute_data;
        mpein_reg_unpack(&mpein, p_acc_reg_gmp->reg.data);
    }

    clbck_data_t clbck_data_new = clbck_data;
    clbck_data_new.m_handle_data_func = &AccRegHandler_ForwardClbck<MPEINRegister,
                   &MPEINRegister::LoopPCINode>;

    if (mpein.port_type == PCI_PORT_TYPE_US) {

        for (u_int8_t depth = 0; depth < MAX_PCI_DEPTH; ++depth) {

            AccRegKeyDPN *p_dpnkey = new AccRegKeyDPN(
                    p_curr_port,             // port
                    depth,                   // depth
                    p_dpnkey_old->pci_idx,   // pci index
                    0);                      // pci node

            clbck_data_new.m_data2 = p_dpnkey;

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            rc = p_handler->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  p_curr_port->num,
                                  p_curr_port->base_lid,
                                  acc_reg,
                                  p_dpnkey,
                                  (ProgressBar *)clbck_data.m_p_progress_bar,
                                  &clbck_data_new);

            if (rc == IBDIAG_ERR_CODE_DB_ERR)
                IBDIAGNET_RETURN(rc);
        }
    }

    IBDIAGNET_RETURN(rc);
}

int MPEINRegister::LoopPCINode(const clbck_data_t &clbck_data,
                                int rec_status,
                                void *p_attribute_data)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    AccRegHandler *p_handler = (AccRegHandler*)clbck_data.m_p_obj;
    IBNode *p_curr_node = (IBNode*)clbck_data.m_data1;
    AccRegKeyDPN *p_dpnkey_old = (AccRegKeyDPN *)clbck_data.m_data2;
    IBPort *p_curr_port = (IBPort*)clbck_data.m_data3;
    AccRegVia_t acc_reg_via = (AccRegVia_t)(u_int64_t)clbck_data.m_data4;

    direct_route_t *p_curr_direct_route = m_phy_diag->GetIBDiag()->GetDR(p_curr_node);
    if (!p_curr_direct_route) {
        m_phy_diag->SetLastError("DB error - current direct route is NULL. Node GUID="
                U64H_FMT ", node name %s.", p_curr_node->guid_get(), p_curr_node->getName().c_str());

        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    mpein_reg mpein;
    CLEAR_STRUCT(mpein);

    // invoke handler callback in order to handle return status and insert reply to the map
    if (acc_reg_via == VIA_SMP) {
        rc = p_handler->SMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);
        if (rc)
            IBDIAGNET_RETURN(rc);

        SMP_AccessRegister *p_acc_reg_smp = (SMP_AccessRegister*)p_attribute_data;
        mpein_reg_unpack(&mpein, p_acc_reg_smp->reg.data);
    }

    if (acc_reg_via == VIA_GMP){
        rc = p_handler->GMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);
        if (rc)
            IBDIAGNET_RETURN(rc);

        struct GMP_AccessRegister *p_acc_reg_gmp = (struct GMP_AccessRegister *)p_attribute_data;
        mpein_reg_unpack(&mpein, p_acc_reg_gmp->reg.data);
    }

    if (mpein.port_type == PCI_PORT_TYPE_DS) {

        for (u_int8_t node = 0; node < MAX_PCI_NODES; ++node) {

            AccRegKeyDPN *p_dpnkey = new AccRegKeyDPN(
                    p_curr_port,                // port
                    p_dpnkey_old->depth,        // depth
                    p_dpnkey_old->pci_idx,      // pci index
                    node);                      // pci node

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            rc = p_handler->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  p_curr_port->num,
                                  p_curr_port->base_lid,
                                  acc_reg,
                                  p_dpnkey,
                                  (ProgressBar *)clbck_data.m_p_progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR)
                IBDIAGNET_RETURN(rc);
        }
    }

    IBDIAGNET_RETURN(rc);
}

bool MPEINRegister::IsValidPakcet(const struct acc_reg_data & reg_data) const
{
    u_int8_t port_type = reg_data.regs.mpein.port_type;

    if (port_type == PCI_PORT_TYPE_EP ||
        port_type == PCI_PORT_TYPE_US ||
        port_type == PCI_PORT_TYPE_DS)
        return true;

    return false;
}

void MPEINRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node,
                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node) {
        AccRegKeyDPN *p_dpn_key = (AccRegKeyDPN *)p_key;
        p_export_data_phy_node->p_mpein_reg[p_dpn_key->pci_idx][p_dpn_key->depth][p_dpn_key->pci_node] = &areg.regs.mpein;
    }

    IBDIAGNET_RETURN_VOID;
}

bool MPEINRegister::CheckRegisterKey(export_data_phy_port_t *p_export_data_phy_port,
                                     export_data_phy_node_t *p_export_data_phy_node,
                                     AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node &&
            p_key->node_guid == p_export_data_phy_node->node_guid)
        IBDIAGNET_RETURN(true);

    IBDIAGNET_RETURN(false);
}

//*****************************************
//***      MPIR Register functions      ***
//*****************************************
MPIRRegister::MPIRRegister(PhyDiag * phy_diag, map_akey_areg &mpein_map) :
    Register(phy_diag,
             ACCESS_REGISTER_ID_MPIR,
             (const unpack_data_func_t)mpir_reg_unpack,
             ACC_REG_MPIR_INTERNAL_SECTION_NAME,
             ACC_REG_MPIR_NAME,
             ACC_REG_MPIR_FIELDS_NUM,
             NSB::get(this)),
    mpein_map(mpein_map) // this register should be sent to
                         // all valid mpein registers.
{
}

int MPIRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    AccRegKeyDPN* p_dpnkey = (AccRegKeyDPN*)p_key;
    mpir_reg mpir;
    CLEAR_STRUCT(mpir);

    mpir.depth = p_dpnkey->depth;
    mpir.pcie_index = p_dpnkey->pci_idx;
    mpir.node = p_dpnkey->pci_node;
    mpir_reg_pack(&mpir, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MPIRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl              << '#'
            << handler_header    << ','
            << "host_buses"      << ','
            << "sdm"             << ','
            << "subordinate_bus" << ','
            << "secondary_bus"   << ','
            << "device"          << ','
            << "bus"             << ','
            << "local_port"      << ','
            << "slot_number"     << ','
            << "num_con_devices" << ','
            << "slot_cap";
#endif
    IBDIAGNET_RETURN_VOID
}

void MPIRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct mpir_reg & mpir = areg.regs.mpir;

    sstream << +mpir.host_buses << ','
            << +mpir.sdm << ','
            << +mpir.subordinate_bus << ','
            << +mpir.secondary_bus << ','
            << +mpir.device << ','
            << +mpir.bus << ','
            << +mpir.local_port << ','
            << +mpir.slot_number << ','
            << +mpir.num_con_devices << ','
            << +mpir.slot_cap << endl;

    IBDIAGNET_RETURN_VOID;
}

int MPIRRegister::BuildDB(AccRegHandler* p_handler,
                          list_p_fabric_general_err &phy_errors,
                          ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    clbck_data_t clbck_data;
    clbck_data.m_handle_data_func =  &forwardClbck<AccRegHandler,
                    &AccRegHandler::PCIAddressCollectGetClbck>;
    clbck_data.m_p_obj = p_handler;

    p_handler->SetHeader(PCI_DPN_HEADER);

    // go over all the mpein MADs and send mpir accordingly
    for (map_akey_areg::iterator mpein_iter = mpein_map.begin();
         mpein_iter != mpein_map.end(); mpein_iter++) {

        AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)mpein_iter->first;

        // get node
        IBNode *p_curr_node =
                m_phy_diag->GetFabric()->getNodeByGuid(p_dpnkey->node_guid);
        if (!p_curr_node) {
            m_phy_diag->SetLastError("DB error - found null node in "
                                     "NodeByName map for key = " U64H_FMT,
                                                p_dpnkey->node_guid);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        AccRegVia_t acc_reg_via = Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        // get port
        IBPort *p_curr_port = NULL;
        for (u_int32_t port_num = 1; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
            p_curr_port = p_curr_node->getPort((phys_port_t)port_num);

            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;
            break; // use the first valid port.
        }

        if (!p_curr_port) {
            m_phy_diag->SetLastError("DB error - failed to find valid port for node %s",
                                     p_curr_node->getName().c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        AccessRegister acc_reg;
        CLEAR_STRUCT(acc_reg);

        AccRegKeyDPN *p_dpnkey_new = new AccRegKeyDPN(p_dpnkey->p_port,
                                                      p_dpnkey->depth,
                                                      p_dpnkey->pci_idx,
                                                      p_dpnkey->pci_node);

        clbck_data.m_data1 = p_curr_node;
        clbck_data.m_data2 = p_dpnkey_new;
        clbck_data.m_data4 = (void*)(uint64_t)acc_reg_via;

        rc = p_handler->SendAccReg(acc_reg_via,
                                   p_curr_node,
                                   0, // port 0
                                   p_curr_port->base_lid,
                                   acc_reg,
                                   p_dpnkey_new,
                                   p_progress_bar,
                                   &clbck_data);

        if (rc == IBDIAG_ERR_CODE_DB_ERR)
            IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(rc);
}


//*****************************************
//***      MFCR Register functions      ***
//*****************************************
MFCRRegister::MFCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MFCR,
                (const unpack_data_func_t)mfcr_reg_unpack,
                ACC_REG_MFCR_INTERNAL_SECTION_NAME,
                ACC_REG_MFCR_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                AVAILABLE_FANS,
                SUPPORT_SW,
                false)
{
}

int MFCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MFCRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct mfcr_reg & p_mfcr = areg.regs.mfcr;
    sstream << hex << p_mfcr.tacho_active << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

int MFCRRegister::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    struct mfcr_reg *p_mfcr = &areg.regs.mfcr;
    int rc  = SensorsBitsToList(p_mfcr->tacho_active, MAX_FANS, sensors_list);
    IBDIAGNET_RETURN(rc);
}

//*****************************************
//***      MFSM Register functions      ***
//*****************************************
MFSMRegister::MFSMRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MFSM,
                (const unpack_data_func_t)mfsm_reg_unpack,
                ACC_REG_MFSM_INTERNAL_SECTION_NAME,
                ACC_REG_MFSM_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                FANS_SPEED_HEADER,
                SUPPORT_SW)
{
}

int MFSMRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct mfsm_reg mfsm;
    CLEAR_STRUCT(mfsm);
    mfsm.tacho = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mfsm_reg_pack(&mfsm, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MFSMRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    char buffer[1024] = {0};
    const struct mfsm_reg & p_mfsm = areg.regs.mfsm;

    snprintf(buffer, sizeof(buffer), U32D_FMT, p_mfsm.rpm);

    sstream << buffer << endl;
    IBDIAGNET_RETURN_VOID;
}

void MFSMRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node)
        p_export_data_phy_node->p_mfsm_reg[((AccRegKeyNodeSensor *)p_key)->sensor_id] = &areg.regs.mfsm;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MFSL Register functions      ***
//*****************************************
MFSLRegister::MFSLRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MFSL,
                (const unpack_data_func_t)mfsl_reg_unpack,
                ACC_REG_MFSL_INTERNAL_SECTION_NAME,
                ACC_REG_MFSL_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                FANS_THRESHOLDS_HEADER,
                SUPPORT_SW)
{
}

int MFSLRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct mfsl_reg mfsl;
    CLEAR_STRUCT(mfsl);
    mfsl.tacho = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mfsl_reg_pack(&mfsl, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MFSLRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    char buffer[1024] = {0};
    const struct mfsl_reg & p_mfsl = areg.regs.mfsl;

    snprintf(buffer, sizeof(buffer), U32D_FMT "," U32D_FMT, p_mfsl.tach_min, p_mfsl.tach_max);

    sstream << buffer << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      FORE Register functions      ***
//*****************************************
FORERegister::FORERegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_FORE,
                (const unpack_data_func_t)fore_reg_unpack,
                ACC_REG_FORE_INTERNAL_SECTION_NAME,
                ACC_REG_FORE_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                FANS_ALERT_HEADER,
                SUPPORT_SW)
{
}

int FORERegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void FORERegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct fore_reg & p_fore = areg.regs.fore;
    stringstream under_limit;
    stringstream over_limit;
    u_int16_t mask = 1;

    if (!p_fore.fan_under_limit && !p_fore.fan_over_limit) {
        sstream << "-1,-1" << endl;
        IBDIAGNET_RETURN_VOID;
    }

    for (int i = 0; i < MAX_FANS; ++i, mask <<= 1) {
        if (mask & p_fore.fan_under_limit)
            under_limit << i << "|";
        if (mask & p_fore.fan_over_limit)
            over_limit << i << "|";
    }

    string under;
    string over;

    if (!p_fore.fan_under_limit)
        under = "-1";
    else {
        under = under_limit.str();
        under = under.substr(0, under.size() - 1);

    }

    if (!p_fore.fan_over_limit)
        over = "-1";
    else {
        over = over_limit.str();
        over = over.substr(0, over.size() - 1);

    }

    sstream << under << "," << over << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***     MFNR Register functions       ***
//*****************************************
MFNRRegister::MFNRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MFNR,
                (const unpack_data_func_t)mfnr_reg_unpack,
                ACC_REG_MFNR_INTERNAL_SECTION_NAME,
                ACC_REG_MFNR_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                FANS_NUMBERS_HEADER,
                SUPPORT_SW)
{
}

int MFNRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct mfnr_reg mfnr;
    CLEAR_STRUCT(mfnr);
    mfnr.fan_index = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mfnr_reg_pack(&mfnr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MFNRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct mfnr_reg & p_mfnr = areg.regs.mfnr;
    sstream << p_mfnr.serial_number << endl;
    IBDIAGNET_RETURN_VOID;
}

//******************************************
//***      MTCAP Register functions      ***
//******************************************
MTCAPRegister::MTCAPRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MTCAP,
                (const unpack_data_func_t)mtcap_reg_unpack,
                ACC_REG_MTCAP_INTERNAL_SECTION_NAME,
                ACC_REG_MTCAP_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                TEMPERATURE_SENSORS_COUNT,
                SUPPORT_SW,
                false) // dump enabled
{
}

int MTCAPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MTCAPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    char buffer[32] = {0};
    const struct mtcap_reg & p_mtcap = areg.regs.mtcap;
    snprintf(buffer, sizeof(buffer), U32D_FMT, p_mtcap.sensor_count);
    sstream << buffer << endl;

    IBDIAGNET_RETURN_VOID;
}

int MTCAPRegister::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    struct mtcap_reg *p_mtcap = &areg.regs.mtcap;
    IBDIAGNET_RETURN(SensorsCountToList(p_mtcap->sensor_count, sensors_list));
}

//*****************************************
//***      MTMP Register functions      ***
//*****************************************
MTMPRegister::MTMPRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MTMP,
                (const unpack_data_func_t)mtmp_reg_unpack,
                ACC_REG_MTMP_INTERNAL_SECTION_NAME,
                ACC_REG_MTMP_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                TEMPERATURE_SENSORS_HEADER,
                SUPPORT_SW)
{
}

int MTMPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct mtmp_reg mtmp;
    CLEAR_STRUCT(mtmp);
    mtmp.sensor_index = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mtmp_reg_pack(&mtmp, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MTMPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    char buffer[1024] = {0};
    const struct mtmp_reg & p_mtmp = areg.regs.mtmp;

    int16_t temperature = p_mtmp.temperature;
    int16_t max_temperature = p_mtmp.max_temperature;
    int16_t threshold_lo = p_mtmp.temperature_threshold_lo;
    int16_t threshold_hi = p_mtmp.temperature_threshold_hi;

    snprintf(buffer, sizeof(buffer), FLT_FMT "," FLT_FMT "," FLT_FMT "," FLT_FMT,
            (float)temperature / 8, (float)max_temperature / 8,
            (float)threshold_lo / 8, (float)threshold_hi / 8);

    sstream << p_mtmp.sensor_name_hi << p_mtmp.sensor_name_lo << "," << buffer << endl;

    IBDIAGNET_RETURN_VOID;
}

void MTMPRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node)
        p_export_data_phy_node->p_mtmp_reg[((AccRegKeyNodeSensor *)p_key)->sensor_id] = &areg.regs.mtmp;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MTWE Register functions      ***
//*****************************************
MTWERegister::MTWERegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MTWE,
                (const unpack_data_func_t)mtwe_reg_unpack,
                ACC_REG_MTWE_INTERNAL_SECTION_NAME,
                ACC_REG_MTWE_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                TEMPERATURE_SENSORS_ALERT_HEADER,
                SUPPORT_SW)
{
}

int MTWERegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MTWERegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    const struct mtwe_reg & p_mtwe = areg.regs.mtwe;
    stringstream sensor_warning;
    const int num_of_warnings = 4;
    const int bits_per_warning = 32;
    u_int32_t mask = 1;
    u_int32_t sensor_warnings[num_of_warnings];
    string warning;

    if (!p_mtwe.sensor_warning0 && !p_mtwe.sensor_warning1 &&
        !p_mtwe.sensor_warning2 && !p_mtwe.sensor_warning3 ) {
        sstream << "-1" << endl;
        IBDIAGNET_RETURN_VOID;
    }

    sensor_warnings[0] = p_mtwe.sensor_warning0;
    sensor_warnings[1] = p_mtwe.sensor_warning1;
    sensor_warnings[2] = p_mtwe.sensor_warning2;
    sensor_warnings[3] = p_mtwe.sensor_warning3;

    for (int i = 0; i < num_of_warnings; ++i) {
        for (int j = 0; j < bits_per_warning; ++j, mask <<= 1) {
            if (mask & sensor_warnings[i])
                sensor_warning << (i * bits_per_warning + j) << "|";
        }
        mask = 1;
    }

    warning = sensor_warning.str();
    warning = warning.substr(0, warning.size() - 1);
    sstream << warning << endl;
    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***     MVCAP Register functions     ***
//*****************************************
MVCAPRegister::MVCAPRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MVCAP,
                (const unpack_data_func_t)mvcap_reg_unpack,
                ACC_REG_MVCAP_INTERNAL_SECTION_NAME,
                ACC_REG_MVCAP_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                AVAILABLE_POWER_SENSORS,
                SUPPORT_SW,
                false)
{
}

int MVCAPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MVCAPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct mvcap_reg & p_mvcap = areg.regs.mvcap;
    sstream << hex << p_mvcap.sensor_map << dec << endl;

    IBDIAGNET_RETURN_VOID;
}

int MVCAPRegister::AvailableSensors(struct acc_reg_data& areg, list_uint8& sensors_list)
{
    IBDIAGNET_ENTER;
    struct mvcap_reg *p_mvcap = &areg.regs.mvcap;
    int rc = SensorsBitsToList(p_mvcap->sensor_map, MAX_VOLTAGE_CURRENT_SENSORS, sensors_list);
    IBDIAGNET_RETURN(rc);
}

//*****************************************
//***      MVCR Register functions      ***
//*****************************************
MVCRRegister::MVCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MVCR,
                (const unpack_data_func_t)mvcr_reg_unpack,
                ACC_REG_MVCR_INTERNAL_SECTION_NAME,
                ACC_REG_MVCR_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                POWER_SENSORS_HEADER,
                SUPPORT_SW)
{
}

int MVCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct mvcr_reg mvcr;
    CLEAR_STRUCT(mvcr);
    mvcr.sensor_index = ((AccRegKeyNodeSensor *)p_key)->sensor_id;
    mvcr_reg_pack(&mvcr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MVCRRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    char buffer[128] = {0};
    char name[128] = {0};
    const struct mvcr_reg & p_mvcr = areg.regs.mvcr;

    memcpy(name, &p_mvcr.sensor_name, sizeof(u_int64_t));

    // todo change current from 'N/A' to real value when supported.
    snprintf(buffer, sizeof(buffer), FLT_FMT "," "N/A",
             p_mvcr.voltage_sensor_value / 100.0);

    sstream << name << "," << buffer << endl;
    IBDIAGNET_RETURN_VOID;
}

void MVCRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node)
        p_export_data_phy_node->p_mvcr_reg[((AccRegKeyNodeSensor *)p_key)->sensor_id] = &areg.regs.mvcr;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MSPS Register functions      ***
//*****************************************
MSPSRegister::MSPSRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MSPS,
                (const unpack_data_func_t)msps_reg_unpack,
                ACC_REG_MSPS_INTERNAL_SECTION_NAME,
                ACC_REG_MSPS_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                POWER_SUPPLIES_HEADER,
                SUPPORT_SW,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP)
{
}

int MSPSRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

static string DumpPSU(const struct psu_status & psu)
{
    string psu_status;

    switch(psu.present) {
        case 0 :
            psu_status += ",N/A,N/A";
            break;
        case 1 :
            psu_status += ",Yes,No";
            break;
        case 2 :
            psu_status += ",No,No";
            break;
        case 5 :
            psu_status += ",Yes,Yes";
            break;
        case 6 :
            psu_status += ",No,Yes";
            break;
        default:
            psu_status += ",,";
    }

    switch(psu.ac_status) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",Yes";
            break;
        case 2 :
            psu_status += ",No";
            break;
        default:
            psu_status += ",";
    }

    switch(psu.dc_status) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",OK";
            break;
        case 2 :
            psu_status += ",Error";
            break;
        default:
            psu_status += ",";
    }

    switch(psu.alert) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",No";
            break;
        case 2 :
            psu_status += ",Yes";
            break;
        default:
            psu_status += ",";
    }

    switch(psu.fan) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",No Fan";
            break;
        case 2 :
            psu_status += ",OK";
            break;
        case 3 :
            psu_status += ",Error";
            break;
        default:
            psu_status += ",";
    }

    switch(psu.temp) {
        case 0 :
            psu_status += ",N/A";
            break;
        case 1 :
            psu_status += ",OK";
            break;
        case 2 :
            psu_status += ",Overtemp";
            break;
        default:
            psu_status += ",";
    }

    psu_status += ',';
    psu_status += to_string(psu.power_cap);

    if (psu.power_cap == 2) {
        psu_status += ',';
        psu_status += to_string(psu.power_consumption);
    } else
        psu_status += ",N/A";

    if (psu.snp) {
        psu_status += ',';
        psu_status += psu.serial_number;
    } else
        psu_status += ",N/A";

    return psu_status;
}

void MSPSRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct msps_reg & p_msps = areg.regs.msps;
    string node_guid = sstream.str(); // get the node guid from the key

    sstream << "1" << DumpPSU(p_msps.psu0) << endl
            << node_guid << "2" << DumpPSU(p_msps.psu1) << endl;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//***      MSGI Register functions      ***
//*****************************************
MSGIRegister::MSGIRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_MSGI,
                (const unpack_data_func_t)msgi_reg_unpack,
                ACC_REG_MSGI_INTERNAL_SECTION_NAME,
                ACC_REG_MSGI_NAME,
                UNUSED_FIELDS_NUM,
                NSB::get(this),
                SYSTEM_GENERAL_INFORMATION_HEADER,
                SUPPORT_SW,
                true,  // dump enabled
                false, // disconnected ports
                VIA_GMP)
{
}

int MSGIRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MSGIRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct msgi_reg & p_msgi = areg.regs.msgi;

    sstream << p_msgi.serial_number << ','
            << p_msgi.part_number << ','
            << p_msgi.revision << ','
            << '\"' << p_msgi.product_name << '\"' << endl;

    IBDIAGNET_RETURN_VOID;
}

void MSGIRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                      export_data_phy_node_t *p_export_data_phy_node,
                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_node)
        p_export_data_phy_node->p_msgi_reg = &areg.regs.msgi;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLRIP Register functions     ***
//*****************************************
SLRIPRegister::SLRIPRegister(PhyDiag * phy_diag, u_int8_t pnat, string sn, map_akey_areg *mpein_map, map_akey_areg *mpir_map)
    : SLRegister(phy_diag,
                ACCESS_REGISTER_ID_SLRIP,
                (const unpack_data_func_t)slrip_reg_unpack,
                sn,
                ACC_REG_SLRIP_NAME,
                ACC_REG_SLRIP_MAX_FIELDS_NUM,
                NSB::get(this),
                mpein_map,
                mpir_map),
     m_pnat(pnat)
{
    if (pnat == ACC_REG_PNAT_OOB_PORT)
        m_support_nodes = SUPPORT_CA;
}

int SLRIPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slrip_reg slrip;
    CLEAR_STRUCT(slrip);

    slrip.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slrip.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slrip.pnat = m_pnat;

    // ib_sel is just for 16nm version
    struct slrip_16nm slrip_16;
    CLEAR_STRUCT(slrip_16);
    slrip_16.ib_sel = ((AccRegKeyPortLane *)p_key)->idx_in_lane;
    slrip_16nm_pack(&slrip_16, slrip.page_data.slrip_data_set.data);

    slrip_reg_pack(&slrip, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLRIPRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_5nm(sstream);

#endif

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Header_Dump_16nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "stauts"               << ','
            << "version_16nm=0"       << ','
            << "local_port"           << ','
            << "p_nat"                << ','
            << "lp_msb"               << ','
            << "lane"                 << ','
            << "port_type"            << ','
            << "ib_sel"               << ','
            << "ffe_tap0"             << ','
            << "ffe_tap1"             << ','
            << "ffe_tap2"             << ','
            << "ffe_tap3"             << ','
            << "ffe_tap4"             << ','
            << "ffe_tap5"             << ','
            << "ffe_tap6"             << ','
            << "ffe_tap7"             << ','
            << "ffe_tap8"             << ','
            << "sel_enc0"             << ','
            << "ffe_tap_en"           << ','
            << "sel_enc1"             << ','
            << "sel_enc2"             << ','
            << "mixer_offset0"        << ','
            << "mixer_offset1"        << ','
            << "saved0_sel_enc0"      << ','
            << "saved0_mixer_offset0" << ','
            << "saved1_sel_enc0"      << ','
            << "saved1_mixer_offset0";

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Header_Dump_7nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "stauts"                      << ','
            << "version_7nm=4"               << ','
            << "local_port"                  << ','
            << "p_nat"                       << ','
            << "lp_msb"                      << ','
            << "lane"                        << ','
            << "port_type"                   << ','
            << "feq_train_mode"              << ','
            << "term_lpf_override_ctrl"      << ','
            << "vref_val_mode_ctrl"          << ','
            << "dffe_override_ctrl"          << ','
            << "vga_gain_override_ctrl"      << ','
            << "hf_gain_override_ctrl"       << ','
            << "mf_pole_override_ctrl"       << ','
            << "mf_gain_override_ctrl"       << ','
            << "lf_pole_override_ctrl"       << ','
            << "lf_gain_override_ctrl"       << ','
            << "dc_gain_override_ctrl"       << ','
            << "term_attn_override_ctrl"     << ','
            << "f1_override_ctrl"            << ','
            << "f2_override_ctrl"            << ','
            << "f3_override_ctrl"            << ','
            << "f4_override_ctrl"            << ','
            << "f5_override_ctrl"            << ','
            << "f6_override_ctrl"            << ','
            << "f7_override_ctrl"            << ','
            << "f8_override_ctrl"            << ','
            << "fm1_override_ctrl"           << ','
            << "fm2_override_ctrl"           << ','
            << "fm3_override_ctrl"           << ','
            << "eth_override_ctrl"           << ','
            << "mf_pole"                     << ','
            << "mf_gain"                     << ','
            << "lf_pole"                     << ','
            << "lf_gain"                     << ','
            << "term_attn_ctrl"              << ','
            << "hf_gain"                     << ','
            << "dc_gain"                     << ','
            << "adc_vref_val"                << ','
            << "vga_gain"                    << ','
            << "ffe_fm1"                     << ','
            << "ffe_fm2"                     << ','
            << "ffe_fm3"                     << ','
            << "ffe_f4 "                     << ','
            << "ffe_f3"                      << ','
            << "ffe_f2"                      << ','
            << "ffe_f1"                      << ','
            << "ffe_f8"                      << ','
            << "ffe_f7"                      << ','
            << "ffe_f6"                      << ','
            << "ffe_f5"                      << ','
            << "dffe_dsel3"                  << ','
            << "dffe_coef3"                  << ','
            << "dffe_dsel2"                  << ','
            << "dffe_coef2"                  << ','
            << "dffe_dsel1"                  << ','
            << "dffe_coef1"                  << ','
            << "dffe_dsel0"                  << ','
            << "dffe_coef0"                  << ','
            << "dffe_dsel7"                  << ','
            << "dffe_coef7"                  << ','
            << "dffe_dsel6"                  << ','
            << "dffe_coef6"                  << ','
            << "dffe_dsel5"                  << ','
            << "dffe_coef5"                  << ','
            << "dffe_dsel4"                  << ','
            << "dffe_coef4"                  << ','
            << "dffe_dsel8"                  << ','
            << "dffe_coef8"                  << ','
            << "th1"                         << ','
            << "th0"                         << ','
            << "th3"                         << ','
            << "th2"                         << ','
            << "th5"                         << ','
            << "th4"                         << ','
            << "th7"                         << ','
            << "th6"                         << ','
            << "th9"                         << ','
            << "th8"                         << ','
            << "th11"                        << ','
            << "th10"                        << ','
            << "th13"                        << ','
            << "th12"                        << ','
            << "th15"                        << ','
            << "th14"                        << ','
            << "th17"                        << ','
            << "th16"                        << ','
            << "th19"                        << ','
            << "th18"                        << ','
            << "th21"                        << ','
            << "th20";

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Header_Dump_5nm(stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    sstream << "stauts"                             << ','
            << "version_5nm=5"                      << ','
            << "local_port"                         << ','
            << "p_nat"                              << ','
            << "lp_msb"                             << ','
            << "lane"                               << ','
            << "port_type"                          << ','
            << "tah_amp_gain"                       << ','
            << "lf_gain"                            << ','
            << "lf_pol"                             << ','
            << "mf_gain"                            << ','
            << "mf_pol"                             << ','
            << "dc_gain"                            << ','
            << "hf_gain"                            << ','
            << "ffe_fm3"                            << ','
            << "ffe_fm4"                            << ','
            << "ffe_fm5"                            << ','
            << "ffe_f1"                             << ','
            << "ffe_fm1"                            << ','
            << "ffe_fm2"                            << ','
            << "ffe_f4"                             << ','
            << "ffe_f3"                             << ','
            << "ffe_f2"                             << ','
            << "ffe_f8"                             << ','
            << "ffe_f7"                             << ','
            << "ffe_f6"                             << ','
            << "ffe_f5"                             << ','
            << "ffe_f12"                            << ','
            << "ffe_f11"                            << ','
            << "ffe_f10"                            << ','
            << "ffe_f9"                             << ','
            << "dffe_coef2"                         << ','
            << "dffe_dsel2"                         << ','
            << "dffe_coef1"                         << ','
            << "dffe_dsel1"                         << ','
            << "dffe_coef0"                         << ','
            << "dffe_dsel0"                         << ','
            << "dffe_coef5"                         << ','
            << "dffe_dsel5"                         << ','
            << "dffe_coef4"                         << ','
            << "dffe_dsel4"                         << ','
            << "dffe_coef3"                         << ','
            << "dffe_dsel3"                         << ','
            << "dffe_coef8"                         << ','
            << "dffe_dsel8"                         << ','
            << "dffe_coef7"                         << ','
            << "dffe_dsel7"                         << ','
            << "dffe_coef6"                         << ','
            << "dffe_dsel6"                         << ','
            << "dffe_coef11"                        << ','
            << "dffe_dsel11"                        << ','
            << "dffe_coef10"                        << ','
            << "dffe_dsel10"                        << ','
            << "dffe_coef9"                         << ','
            << "dffe_dsel9"                         << ','
            << "dffe_coef14"                        << ','
            << "dffe_dsel14"                        << ','
            << "dffe_coef13"                        << ','
            << "dffe_dsel13"                        << ','
            << "dffe_coef12"                        << ','
            << "dffe_dsel12"                        << ','
            << "fb_slicer_th1"                      << ','
            << "fb_slicer_th0"                      << ','
            << "dffe_coef15"                        << ','
            << "dffe_dsel15"                        << ','
            << "fb_slicer_th4"                      << ','
            << "fb_slicer_th3"                      << ','
            << "fb_slicer_th2"                      << ','
            << "fb_slicer_th7"                      << ','
            << "fb_slicer_th6"                      << ','
            << "fb_slicer_th5"                      << ','
            << "fb_slicer_th10"                     << ','
            << "fb_slicer_th9"                      << ','
            << "fb_slicer_th8"                      << ','
            << "ff_slicer_th0"                      << ','
            << "fb_slicer_th12"                     << ','
            << "fb_slicer_th11"                     << ','
            << "ff_slicer_th3"                      << ','
            << "ff_slicer_th2"                      << ','
            << "ff_slicer_th1"                      << ','
            << "ff_slicer_th6"                      << ','
            << "ff_slicer_th5"                      << ','
            << "ff_slicer_th4"                      << ','
            << "ff_slicer_th9"                      << ','
            << "ff_slicer_th8"                      << ','
            << "ff_slicer_th7"                      << ','
            << "ff_slicer_th12"                     << ','
            << "ff_slicer_th11"                     << ','
            << "ff_slicer_th10";

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct slrip_reg & slrip = areg.regs.slrip;

    sstream << +slrip.status          << ','
            << +slrip.version         << ','
            << +slrip.local_port      << ','
            << +slrip.pnat            << ','
            << +slrip.lp_msb          << ','
            << +slrip.lane            << ','
            << +slrip.port_type       << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (slrip.version) {

        case 0: // 16nm
            Dump_16nm(slrip, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(slrip, sstream);
            break;

        case 5: // 5nm
            Dump_5nm(slrip, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +slrip.version << '"';
            for (int i = 0; i < ACC_REG_SLRIP_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLRIP: %d, on node: " U64H_FMT ".\n",
                        slrip.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Dump_16nm(const struct slrip_reg &slrip, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrip_16nm slrip_16;
    slrip_16nm_unpack(&slrip_16, slrip.page_data.slrip_data_set.data);

    sstream << +slrip_16.ib_sel                         << ','
            << +slrip_16.ffe_tap0                       << ','
            << +slrip_16.ffe_tap1                       << ','
            << +slrip_16.ffe_tap2                       << ','
            << +slrip_16.ffe_tap3                       << ','
            << +slrip_16.ffe_tap4                       << ','
            << +slrip_16.ffe_tap5                       << ','
            << +slrip_16.ffe_tap6                       << ','
            << +slrip_16.ffe_tap7                       << ','
            << +slrip_16.ffe_tap8                       << ','
            << +slrip_16.sel_enc0                       << ','
            << +slrip_16.ffe_tap_en                     << ','
            << +slrip_16.sel_enc1                       << ','
            << +slrip_16.sel_enc2                       << ','
            << +slrip_16.mixer_offset0                  << ','
            << +slrip_16.mixer_offset1                  << ','
            << +slrip_16.saved0_sel_enc0                << ','
            << +slrip_16.saved0_mixer_offset0           << ','
            << +slrip_16.saved1_sel_enc0                << ','
            << +slrip_16.saved1_mixer_offset0           ;

    for (int i = 0; i < ACC_REG_SLRIP_MAX_FIELDS_NUM - ACC_REG_SLRIP16NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Dump_7nm(const struct slrip_reg &slrip, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrip_7nm slrip_7;
    slrip_7nm_unpack(&slrip_7, slrip.page_data.slrip_data_set.data);

    sstream << +slrip_7.feq_train_mode              << ','
            << +slrip_7.term_lpf_override_ctrl      << ','
            << +slrip_7.vref_val_mode_ctrl          << ','
            << +slrip_7.dffe_override_ctrl          << ','
            << +slrip_7.vga_gain_override_ctrl      << ','
            << +slrip_7.hf_gain_override_ctrl       << ','
            << +slrip_7.mf_pole_override_ctrl       << ','
            << +slrip_7.mf_gain_override_ctrl       << ','
            << +slrip_7.lf_pole_override_ctrl       << ','
            << +slrip_7.lf_gain_override_ctrl       << ','
            << +slrip_7.dc_gain_override_ctrl       << ','
            << +slrip_7.term_attn_override_ctrl     << ','
            << +slrip_7.f1_override_ctrl            << ','
            << +slrip_7.f2_override_ctrl            << ','
            << +slrip_7.f3_override_ctrl            << ','
            << +slrip_7.f4_override_ctrl            << ','
            << +slrip_7.f5_override_ctrl            << ','
            << +slrip_7.f6_override_ctrl            << ','
            << +slrip_7.f7_override_ctrl            << ','
            << +slrip_7.f8_override_ctrl            << ','
            << +slrip_7.fm1_override_ctrl           << ','
            << +slrip_7.fm2_override_ctrl           << ','
            << +slrip_7.fm3_override_ctrl           << ','
            << +slrip_7.eth_override_ctrl           << ','
            << +slrip_7.mf_pole                     << ','
            << +slrip_7.mf_gain                     << ','
            << +slrip_7.lf_pole                     << ','
            << +slrip_7.lf_gain                     << ','
            << +slrip_7.term_attn_ctrl              << ','
            << +slrip_7.hf_gain                     << ','
            << +slrip_7.dc_gain                     << ','
            << +slrip_7.adc_vref_val                << ','
            << +slrip_7.vga_gain                    << ','
            << +slrip_7.ffe_fm1                     << ','
            << +slrip_7.ffe_fm2                     << ','
            << +slrip_7.ffe_fm3                     << ','
            << +slrip_7.ffe_f4                      << ','
            << +slrip_7.ffe_f3                      << ','
            << +slrip_7.ffe_f2                      << ','
            << +slrip_7.ffe_f1                      << ','
            << +slrip_7.ffe_f8                      << ','
            << +slrip_7.ffe_f7                      << ','
            << +slrip_7.ffe_f6                      << ','
            << +slrip_7.ffe_f5                      << ','
            << +slrip_7.dffe_dsel3                  << ','
            << +slrip_7.dffe_coef3                  << ','
            << +slrip_7.dffe_dsel2                  << ','
            << +slrip_7.dffe_coef2                  << ','
            << +slrip_7.dffe_dsel1                  << ','
            << +slrip_7.dffe_coef1                  << ','
            << +slrip_7.dffe_dsel0                  << ','
            << +slrip_7.dffe_coef0                  << ','
            << +slrip_7.dffe_dsel7                  << ','
            << +slrip_7.dffe_coef7                  << ','
            << +slrip_7.dffe_dsel6                  << ','
            << +slrip_7.dffe_coef6                  << ','
            << +slrip_7.dffe_dsel5                  << ','
            << +slrip_7.dffe_coef5                  << ','
            << +slrip_7.dffe_dsel4                  << ','
            << +slrip_7.dffe_coef4                  << ','
            << +slrip_7.dffe_dsel8                  << ','
            << +slrip_7.dffe_coef8                  << ','
            << +slrip_7.th1                         << ','
            << +slrip_7.th0                         << ','
            << +slrip_7.th3                         << ','
            << +slrip_7.th2                         << ','
            << +slrip_7.th5                         << ','
            << +slrip_7.th4                         << ','
            << +slrip_7.th7                         << ','
            << +slrip_7.th6                         << ','
            << +slrip_7.th9                         << ','
            << +slrip_7.th8                         << ','
            << +slrip_7.th11                        << ','
            << +slrip_7.th10                        << ','
            << +slrip_7.th13                        << ','
            << +slrip_7.th12                        << ','
            << +slrip_7.th15                        << ','
            << +slrip_7.th14                        << ','
            << +slrip_7.th17                        << ','
            << +slrip_7.th16                        << ','
            << +slrip_7.th19                        << ','
            << +slrip_7.th18                        << ','
            << +slrip_7.th21                        << ','
            << +slrip_7.th20                        ;

    IBDIAGNET_RETURN_VOID;
}

void SLRIPRegister::Dump_5nm(const struct slrip_reg &slrip, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slrip_5nm slrip_5;
    slrip_5nm_unpack(&slrip_5, slrip.page_data.slrip_data_set.data);
    sstream << +slrip_5.tah_amp_gain                       << ','
            << +slrip_5.lf_gain                            << ','
            << +slrip_5.lf_pol                             << ','
            << +slrip_5.mf_gain                            << ','
            << +slrip_5.mf_pol                             << ','
            << +slrip_5.dc_gain                            << ','
            << +slrip_5.hf_gain                            << ','
            << +slrip_5.ffe_fm3                            << ','
            << +slrip_5.ffe_fm4                            << ','
            << +slrip_5.ffe_fm5                            << ','
            << +slrip_5.ffe_f1                             << ','
            << +slrip_5.ffe_fm1                            << ','
            << +slrip_5.ffe_fm2                            << ','
            << +slrip_5.ffe_f4                             << ','
            << +slrip_5.ffe_f3                             << ','
            << +slrip_5.ffe_f2                             << ','
            << +slrip_5.ffe_f8                             << ','
            << +slrip_5.ffe_f7                             << ','
            << +slrip_5.ffe_f6                             << ','
            << +slrip_5.ffe_f5                             << ','
            << +slrip_5.ffe_f12                            << ','
            << +slrip_5.ffe_f11                            << ','
            << +slrip_5.ffe_f10                            << ','
            << +slrip_5.ffe_f9                             << ','
            << +slrip_5.dffe_coef2                         << ','
            << +slrip_5.dffe_dsel2                         << ','
            << +slrip_5.dffe_coef1                         << ','
            << +slrip_5.dffe_dsel1                         << ','
            << +slrip_5.dffe_coef0                         << ','
            << +slrip_5.dffe_dsel0                         << ','
            << +slrip_5.dffe_coef5                         << ','
            << +slrip_5.dffe_dsel5                         << ','
            << +slrip_5.dffe_coef4                         << ','
            << +slrip_5.dffe_dsel4                         << ','
            << +slrip_5.dffe_coef3                         << ','
            << +slrip_5.dffe_dsel3                         << ','
            << +slrip_5.dffe_coef8                         << ','
            << +slrip_5.dffe_dsel8                         << ','
            << +slrip_5.dffe_coef7                         << ','
            << +slrip_5.dffe_dsel7                         << ','
            << +slrip_5.dffe_coef6                         << ','
            << +slrip_5.dffe_dsel6                         << ','
            << +slrip_5.dffe_coef11                        << ','
            << +slrip_5.dffe_dsel11                        << ','
            << +slrip_5.dffe_coef10                        << ','
            << +slrip_5.dffe_dsel10                        << ','
            << +slrip_5.dffe_coef9                         << ','
            << +slrip_5.dffe_dsel9                         << ','
            << +slrip_5.dffe_coef14                        << ','
            << +slrip_5.dffe_dsel14                        << ','
            << +slrip_5.dffe_coef13                        << ','
            << +slrip_5.dffe_dsel13                        << ','
            << +slrip_5.dffe_coef12                        << ','
            << +slrip_5.dffe_dsel12                        << ','
            << +slrip_5.fb_slicer_th1                      << ','
            << +slrip_5.fb_slicer_th0                      << ','
            << +slrip_5.dffe_coef15                        << ','
            << +slrip_5.dffe_dsel15                        << ','
            << +slrip_5.fb_slicer_th4                      << ','
            << +slrip_5.fb_slicer_th3                      << ','
            << +slrip_5.fb_slicer_th2                      << ','
            << +slrip_5.fb_slicer_th7                      << ','
            << +slrip_5.fb_slicer_th6                      << ','
            << +slrip_5.fb_slicer_th5                      << ','
            << +slrip_5.fb_slicer_th10                     << ','
            << +slrip_5.fb_slicer_th9                      << ','
            << +slrip_5.fb_slicer_th8                      << ','
            << +slrip_5.ff_slicer_th0                      << ','
            << +slrip_5.fb_slicer_th12                     << ','
            << +slrip_5.fb_slicer_th11                     << ','
            << +slrip_5.ff_slicer_th3                      << ','
            << +slrip_5.ff_slicer_th2                      << ','
            << +slrip_5.ff_slicer_th1                      << ','
            << +slrip_5.ff_slicer_th6                      << ','
            << +slrip_5.ff_slicer_th5                      << ','
            << +slrip_5.ff_slicer_th4                      << ','
            << +slrip_5.ff_slicer_th9                      << ','
            << +slrip_5.ff_slicer_th8                      << ','
            << +slrip_5.ff_slicer_th7                      << ','
            << +slrip_5.ff_slicer_th12                     << ','
            << +slrip_5.ff_slicer_th11                     << ','
            << +slrip_5.ff_slicer_th10                     ;

    for (int i = 0; i < ACC_REG_SLRIP_MAX_FIELDS_NUM - ACC_REG_SLRIP5NM_FIELDS_NUM; ++i)
        sstream << ",NA"; // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}
void SLRIPRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node,
                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port) {
        struct slrip_reg *p_slrip_reg = new struct slrip_reg(areg.regs.slrip);

        switch (p_slrip_reg->version) {
            case 0: // 16nm
                slrip_16nm_unpack(&p_slrip_reg->page_data.slrip_16nm, areg.regs.slrip.page_data.slrip_data_set.data);
                break;
            case 4: // 7nm
                slrip_7nm_unpack(&p_slrip_reg->page_data.slrip_7nm, areg.regs.slrip.page_data.slrip_data_set.data);
                break;
            case 5: // 5nm
                slrip_5nm_unpack(&p_slrip_reg->page_data.slrip_5nm, areg.regs.slrip.page_data.slrip_data_set.data);
                break;
            default:
                break;
        }

        AccRegKeyPortLane *p_pl_key = (AccRegKeyPortLane *)p_key;
        p_export_data_phy_port->p_slrip_reg[p_pl_key->lane][p_pl_key->idx_in_lane] = p_slrip_reg;
    }

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLREG Register functions     ***
//*****************************************
SLREGRegister::SLREGRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_SLREG,
                (const unpack_data_func_t)slreg_reg_unpack,
                ACC_REG_SLREG_INTERNAL_SECTION_NAME,
                ACC_REG_SLREG_NAME,
                ACC_REG_SLREG7NM_FIELDS_NUM,
                NSB::get(this))
{
}

int SLREGRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slreg_reg slreg;
    CLEAR_STRUCT(slreg);
    slreg.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slreg.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slreg.pnat = ACC_REG_PNAT_IB_PORT;
    slreg.eidx = ((AccRegKeyPortLane *)p_key)->idx_in_lane;
    slreg_reg_pack(&slreg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLREGRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) 
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_16nm(sstream);

    sstream << endl                      << '#'
            << handler_header            << ',';
    Header_Dump_7nm(sstream);

#endif

    IBDIAGNET_RETURN_VOID;
}

void SLREGRegister::Header_Dump_16nm(stringstream& sstream) const
{
    IBDIAGNET_ENTER;
    sstream << "eidx"                                   << ','
            << "status"                                 << ','
            << "version_16nm=0"                         << ','
            << "local_port"                             << ','
            << "pnat"                                   << ','
            << "lp_msb"                                 << ','
            << "lane"                                   << ','
            << "port_type"                              << ','
            << "initial_error"                          << ','
            << "eye_grade"                              << ','
            << "post_error"                             << ','
            << "initial_area"                           << ','
            << "post_area"                              << ','
            << "post_area_after_centering"              << ','
            << "gd_iter_num_errors"                     << ','
            << "gd_iter_num_area"                       << ','
            << "post_sel_enc_slave_centering"           << ','
            << "post_sel_enc_slave"                     << ','
            << "initial_sel_enc_slave"                  << ','
            << "ffe_gain"                               << ','
            << "cur_phase_ib"                           << ','
            << "rect_size"                              << ','
            << "sel_enc_tune_area_failure"              << ','
            << "area_zero_event_counter"                << ','
            << "monotonic_reached_low_limit"            << ','
            << "monotonic_reached_high_limit"           << ','
            << "monotonic_flow_height_inconsistent"     << ','
            << "update_single_gain_bad_value"           << ','
            << "init_all_gains_bad_value";

    IBDIAGNET_RETURN_VOID;
}

void SLREGRegister::Header_Dump_7nm(stringstream& sstream) const 
{
    IBDIAGNET_ENTER;
    sstream << "eidx"                           << ','
            << "status"                         << ','
            << "version_7nm=4"                  << ','
            << "local_port"                     << ','
            << "pnat"                           << ','
            << "lp_msb"                         << ','
            << "lane"                           << ','
            << "port_type"                      << ','
            << "eom_en"                         << ','
            << "eom_status"                     << ','
            << "sth_start_mode"                 << ','
            << "eom_vsign"                      << ','
            << "fom_timeout"                    << ','
            << "ber_grad_inv"                   << ','
            << "eom_mode"                       << ','
            << "vos_adc_test_en"                << ','
            << "gos_adc_test_en"                << ','
            << "vos_vga_test_en"                << ','
            << "vos_ctle_test_en"               << ','
            << "latch_phos_test_en"             << ','
            << "latch_adc_vos_test_en"          << ','
            << "latch_adc_gos_test_en"          << ','
            << "fom_retest_mode"                << ','
            << "phos_test_en"                   << ','
            << "latch_test_mode"                << ','
            << "vos_test_mode"                  << ','
            << "fom_test_mode"                  << ','
            << "phos_threshold"                 << ','
            << "fom_threshold"                  << ','
            << "adc_vos_threshold"              << ','
            << "adc_gos_threshold"              << ','
            << "vga_vos_threshold"              << ','
            << "ctle_vos_threshold"             << ','
            << "fom_mode"                       << ','
            << "eom_fmr_sel"                    << ','
            << "eom_fmr_ovrd"                   << ','
            << "fom_norm_max_en"                << ','
            << "eom_grad"                       << ','
            << "ber_mode"                       << ','
            << "nblks_max"                      << ','
            << "nerrs_min"                      << ','
            << "eyeo_range"                     << ','
            << "ber_eye_sel"                    << ','
            << "ber_t"                          << ','
            << "ber_v"                          << ','
            << "margin_en"                      << ','
            << "margin_v"                       << ','
            << "margin_t";

    IBDIAGNET_RETURN_VOID;
}


void SLREGRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct slreg_reg & slreg = areg.regs.slreg;

    sstream << "eidx_" << +slreg.eidx << ','
            << +slreg.status          << ','
            << +slreg.version         << ','
            << +slreg.local_port      << ','
            << +slreg.pnat            << ','
            << +slreg.lp_msb          << ','
            << +slreg.lane            << ','
            << +slreg.port_type       << ',';

    // Note: 'enum SerDesVersion' can not be used here due to mismatches
    // between register version and GeneralInfo.HW_Info.technology

    switch (slreg.version) {

        case 0: // 16nm
            Dump_16nm(slreg, sstream);
            break;

        case 4: // 7nm
            Dump_7nm(slreg, sstream);
            break;

        default:
            static bool should_print = true;

            sstream << '"' << "Unknown version " << +slreg.version << '"';
            for (int i = 0; i < ACC_REG_SLREG_MAX_FIELDS_NUM - 1; ++i)
                sstream << ",NA"; // add NAs to fill the empty fields

            if (should_print) {
                WARN_PRINT("Unknown version for SLREG: %d, on node: " U64H_FMT ".\n",
                        slreg.version, key.node_guid);
                should_print = false;
            }
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void SLREGRegister::Dump_16nm(const struct slreg_reg &slreg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slreg_16nm slreg_16;
    slreg_16nm_unpack(&slreg_16, slreg.page_data.slreg_data_set.data);

    sstream << +slreg_16.initial_error                           << ','
            << +slreg_16.eye_grade                               << ','
            << +slreg_16.post_error                              << ','
            << +slreg_16.initial_area                            << ','
            << +slreg_16.post_area                               << ','
            << +slreg_16.post_area_after_centering               << ','
            << +slreg_16.gd_iter_num_errors                      << ','
            << +slreg_16.gd_iter_num_area                        << ','
            << +slreg_16.post_sel_enc_slave_centering            << ','
            << +slreg_16.post_sel_enc_slave                      << ','
            << +slreg_16.initial_sel_enc_slave                   << ','
            << +slreg_16.ffe_gain                                << ','
            << +slreg_16.cur_phase_ib                            << ','
            << +slreg_16.rect_size                               << ','
            << +slreg_16.sel_enc_tune_area_failure               << ','
            << +slreg_16.area_zero_event_counter                 << ','
            << +slreg_16.monotonic_reached_low_limit             << ','
            << +slreg_16.monotonic_reached_high_limit            << ','
            << +slreg_16.monotonic_flow_height_inconsistent      << ','
            << +slreg_16.update_single_gain_bad_value            << ','
            << +slreg_16.init_all_gains_bad_value                << ','
            << "NA,NA,NA,NA,NA,NA,NA,NA,NA,NA"                   << ','
            << "NA,NA,NA,NA,NA,NA,NA,NA,NA"                      ;  // add NAs to fill the empty fields

    IBDIAGNET_RETURN_VOID;
}

void SLREGRegister::Dump_7nm(const struct slreg_reg &slreg, stringstream &sstream) const
{
    IBDIAGNET_ENTER;

    slreg_7nm slreg_7;
    slreg_7nm_unpack(&slreg_7, slreg.page_data.slreg_data_set.data);

    sstream << +slreg_7.eom_en                          << ','
            << +slreg_7.eom_status                      << ','
            << +slreg_7.sth_start_mode                  << ','
            << +slreg_7.eom_vsign                       << ','
            << +slreg_7.fom_timeout                     << ','
            << +slreg_7.ber_grad_inv                    << ','
            << +slreg_7.eom_mode                        << ','
            << +slreg_7.vos_adc_test_en                 << ','
            << +slreg_7.gos_adc_test_en                 << ','
            << +slreg_7.vos_vga_test_en                 << ','
            << +slreg_7.vos_ctle_test_en                << ','
            << +slreg_7.latch_phos_test_en              << ','
            << +slreg_7.latch_adc_vos_test_en           << ','
            << +slreg_7.latch_adc_gos_test_en           << ','
            << +slreg_7.fom_retest_mode                 << ','
            << +slreg_7.phos_test_en                    << ','
            << +slreg_7.latch_test_mode                 << ','
            << +slreg_7.vos_test_mode                   << ','
            << +slreg_7.fom_test_mode                   << ','
            << +slreg_7.phos_threshold                  << ','
            << +slreg_7.fom_threshold                   << ','
            << +slreg_7.adc_vos_threshold               << ','
            << +slreg_7.adc_gos_threshold               << ','
            << +slreg_7.vga_vos_threshold               << ','
            << +slreg_7.ctle_vos_threshold              << ','
            << +slreg_7.fom_mode                        << ','
            << +slreg_7.eom_fmr_sel                     << ','
            << +slreg_7.eom_fmr_ovrd                    << ','
            << +slreg_7.fom_norm_max_en                 << ','
            << +slreg_7.eom_grad                        << ','
            << +slreg_7.ber_mode                        << ','
            << +slreg_7.nblks_max                       << ','
            << +slreg_7.nerrs_min                       << ','
            << +slreg_7.eyeo_range                      << ','
            << +slreg_7.ber_eye_sel                     << ','
            << +slreg_7.ber_t                           << ','
            << +slreg_7.ber_v                           << ','
            << +slreg_7.margin_en                       << ','
            << +slreg_7.margin_v                        << ','
            << +slreg_7.margin_t                        ;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      SLCCT Register functions     ***
//*****************************************
SLCCTRegister::SLCCTRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_SLCCT,
                (const unpack_data_func_t)slcct_reg_unpack,
                ACC_REG_SLCCT_INTERNAL_SECTION_NAME,
                ACC_REG_SLCCT_NAME,
                ACC_REG_SLCCT_FIELDS_NUM,
                NSB::get(this))
{
}

int SLCCTRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct slcct_reg slcct;
    CLEAR_STRUCT(slcct);
    slcct.local_port = ((AccRegKeyPortLane *)p_key)->port_num;
    slcct.lane = ((AccRegKeyPortLane *)p_key)->lane;
    slcct.pnat = ACC_REG_PNAT_IB_PORT;
    slcct.conf_index = ((AccRegKeyPortLane *)p_key)->idx_in_lane? SLCCT_CTLE_ARR_SIZE : 0;
    slcct.num_of_idx = SLCCT_CTLE_ARR_SIZE;

    slcct_reg_pack(&slcct, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void SLCCTRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "pnat"             << ','
            << "version"          << ','
            << "status"           << ','
            << "conf_index"       << ','
            << "num_of_idx";

    for (size_t i = 0; i < ARRAY_SIZE(slcct_reg::ctle_calib_set); ++i)
        sstream << ',' << "ctle_calib_set[" << i << "].bgn_offset_n0"
                << ',' << "ctle_calib_set[" << i << "].bgn_offset_p0"
                << ',' << "ctle_calib_set[" << i << "].bgn_offset_n1"
                << ',' << "ctle_calib_set[" << i << "].bgn_offset_p1";
#endif

    IBDIAGNET_RETURN_VOID;
}

void SLCCTRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct slcct_reg & p_slcct = areg.regs.slcct;

    sstream << +p_slcct.pnat << ','
            << +p_slcct.version << ','
            << +p_slcct.status << ','
            << +p_slcct.conf_index << ','
            << +p_slcct.num_of_idx;

    for (int i = 0; i < p_slcct.num_of_idx; ++i) {
        sstream << ',' << +p_slcct.ctle_calib_set[i].bgn_offset_n0
                << ',' << +p_slcct.ctle_calib_set[i].bgn_offset_p0
                << ',' << +p_slcct.ctle_calib_set[i].bgn_offset_n1
                << ',' << +p_slcct.ctle_calib_set[i].bgn_offset_p1;
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***      PPAMP Register functions     ***
//*****************************************
PPAMPRegister::PPAMPRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
                ACCESS_REGISTER_ID_PPAMP,
                (const unpack_data_func_t)ppamp_reg_unpack,
                ACC_REG_PPAMP_INTERNAL_SECTION_NAME,
                ACC_REG_PPAMP_NAME,
                ACC_REG_PPAMP_FIELDS_NUM,
                NSB::get(this))
{
}

int PPAMPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct ppamp_reg ppamp;
    CLEAR_STRUCT(ppamp);
    ppamp.opamp_group_type = ((AccRegKeyTypeGroup *)p_key)->group_type;
    ppamp.opamp_group = ((AccRegKeyTypeGroup *)p_key)->group_num;
    ppamp.start_index = ((AccRegKeyTypeGroup *)p_key)->start_index;
    ppamp.num_of_indices = ((AccRegKeyTypeGroup *)p_key)->num_of_indices;
    ppamp_reg_pack(&ppamp, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPAMPRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "max_opamp_group"    << ','
            << "max_num_of_indices";
    
    for (int i = 0; i < PPAMP_IDX_DATA_SIZE; ++i)
        sstream << ',' << "index_data[" << i << ']';

#endif

    IBDIAGNET_RETURN_VOID;
}

void PPAMPRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct ppamp_reg & p_ppamp = areg.regs.ppamp;

    sstream << +p_ppamp.max_opamp_group << ','
            << p_ppamp.max_num_of_indices;

    for (int i = 0; i < PPAMP_IDX_DATA_SIZE; ++i) {
        sstream << ',' << p_ppamp.index_data[i];
    }

    sstream << endl;

    IBDIAGNET_RETURN_VOID;
}

void PPAMPRegister::GetGroupData(struct acc_reg_data& areg, u_int8_t& group, u_int16_t& indices)
{
    IBDIAGNET_ENTER;

    group = areg.regs.ppamp.max_opamp_group;
    indices = areg.regs.ppamp.max_num_of_indices;

    IBDIAGNET_RETURN_VOID;
}

//*****************************************
//************* PDDR Register functions ***
//*****************************************
int PDDRRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer)
{
    IBDIAGNET_ENTER;

    pddr_reg tmp;
    pddr_reg_unpack(&tmp, unpacked_buffer);

    unpack_data_func(data_to_unpack, tmp.page_data.raw.data);

    reg_raw_data *p_raw_data = reinterpret_cast<reg_raw_data *>(data_to_unpack);
    p_raw_data->module_info_ext = tmp.module_info_ext;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PDDRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct pddr_reg pddr;
    CLEAR_STRUCT(pddr);

    pddr.port_type          = this->m_port_type;
    pddr.lp_msb             = this->m_lp_msb;
    pddr.pnat               = this->m_pnat;
    pddr.page_select        = this->m_page_select;
    pddr.module_info_ext    = this->m_module_info_ext;
    pddr.clr                = this->m_phy_diag->ShouldResetCounters();

    int rc = FillLocalPortAndPlaneInd(p_key, &pddr.local_port, &pddr.plane_ind);

    if (rc)
        return rc;

    pddr_reg_pack(&pddr, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


PDDRModuleInfoRegister::PDDRModuleInfoRegister(PhyDiag * phy_diag)
    : PDDRRegister(phy_diag,
                   0x03,
                   (const unpack_data_func_t)pddr_module_info_unpack,
                   ACC_REG_PDDR_MODULE_INTERNAL_SECTION_NAME,
                   ACC_REG_PDDR_MODULE_NAME,
                   ACC_REG_PDDR_MODULE_FIELDS_NUM,
                   ACC_REG_PDDR_MODULE_HEADER,
                   SUPPORT_SW_CA,
                   false,
                   phy_diag->GetToGetCableDisconnectedPorts())
{
    this->m_module_info_ext = 1;
}

int PDDRModuleInfoRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    AccRegKeyPort * key = dynamic_cast<AccRegKeyPort*>(p_key);

    if (!key || !p_node)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    IBPort * p_port = p_node->getPort(key->port_num);

    if (!p_port)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    if (p_port->p_combined_cable && !m_phy_diag->GetPhyCableFullData())
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    IBDIAGNET_RETURN(PDDRRegister::PackData(p_key, data, p_node));
}

void PDDRModuleInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ',';
    PDDRModuleInfoRegister::DumpRawDataHeader(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PDDRModuleInfoRegister::DumpRawDataHeader(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "cable_technology"                  << ','
            << "cable_breakout"                    << ','
            << "ext_ethernet_compliance_code"      << ','
            << "ethernet_compliance_code"          << ','
            << "cable_type"                        << ','
            << "cable_vendor"                      << ','
            << "cable_length"                      << ','
            << "cable_identifier"                  << ','
            << "cable_power_class"                 << ','
            << "max_power"                         << ','
            << "cable_rx_amp"                      << ','
            << "cable_rx_emphasis"                 << ','
            << "cable_tx_equalization"             << ','
            << "cable_attenuation_25g"             << ','
            << "cable_attenuation_12g"             << ','
            << "cable_attenuation_7g"              << ','
            << "cable_attenuation_5g"              << ','
            << "cable_rx_post_emphasis"            << ','
            << "rx_cdr_cap"                        << ','
            << "tx_cdr_cap"                        << ','
            << "rx_cdr_state"                      << ','
            << "tx_cdr_state"                      << ','
            << "vendor_name"                       << ','
            << "vendor_pn"                         << ','
            << "revision"                          << ','
            << "fw_version"                        << ','
            << "vendor_sn"                         << ','
            << "temperature"                       << ','
            << "voltage"                           << ','
            << "rx_power_lane0"                    << ','
            << "rx_power_lane1"                    << ','
            << "rx_power_lane2"                    << ','
            << "rx_power_lane3"                    << ','
            << "rx_power_lane4"                    << ','
            << "rx_power_lane5"                    << ','
            << "rx_power_lane6"                    << ','
            << "rx_power_lane7"                    << ','
            << "tx_power_lane0"                    << ','
            << "tx_power_lane1"                    << ','
            << "tx_power_lane2"                    << ','
            << "tx_power_lane3"                    << ','
            << "tx_power_lane4"                    << ','
            << "tx_power_lane5"                    << ','
            << "tx_power_lane6"                    << ','
            << "tx_power_lane7"                    << ','
            << "tx_bias_lane0"                     << ','
            << "tx_bias_lane1"                     << ','
            << "tx_bias_lane2"                     << ','
            << "tx_bias_lane3"                     << ','
            << "tx_bias_lane4"                     << ','
            << "tx_bias_lane5"                     << ','
            << "tx_bias_lane6"                     << ','
            << "tx_bias_lane7"                     << ','
            << "temperature_high_th"               << ','
            << "temperature_low_th"                << ','
            << "voltage_high_th"                   << ','
            << "voltage_low_th"                    << ','
            << "rx_power_high_th"                  << ','
            << "rx_power_low_th"                   << ','
            << "tx_power_high_th"                  << ','
            << "tx_power_low_th"                   << ','
            << "tx_bias_high_th"                   << ','
            << "tx_bias_low_th"                    << ','
            << "module_st"                         << ','
            << "rx_power_type"                     << ','
            << "did_cap"                           << ','
            << "smf_length"                        << ','
            << "wavelength"                        << ','
            << "active_set_host_compliance_code"   << ','
            << "active_set_media_compliance_code"  << ','
            << "tx_bias_scaling_factor"            << ','
            << "ib_compliance_code"                << ','
            << "nbr250"                            << ','
            << "nbr100"                            << ','
            << "monitor_cap_mask"                  << ','
            << "ib_width"                          << ','
            << "dp_st_lane0"                       << ','
            << "dp_st_lane1"                       << ','
            << "dp_st_lane2"                       << ','
            << "dp_st_lane3"                       << ','
            << "dp_st_lane4"                       << ','
            << "dp_st_lane5"                       << ','
            << "dp_st_lane6"                       << ','
            << "dp_st_lane7"                       << ','
            << "length_om2"                        << ','
            << "length_om3"                        << ','
            << "length_om4"                        << ','
            << "length_om5"                        << ','
            << "length_om1"                        << ','
            << "wavelength_tolerance"              << ','
            << "memory_map_rev"                    << ','
            << "memory_map_compliance"             << ','
            << "date_code"                         << ','
            << "connector_type"                    << ','
            << "vendor_oui"                        << ','
            << "cable_attenuation_53g"             << ','
            << "rx_output_valid"                   << ','
            << "tx_input_freq_sync"                << ','
            << "error_code"                        << ','
            << "cdr_vendor"                        << ','
            << "max_fiber_length";

    IBDIAGNET_RETURN_VOID;
}

void PDDRModuleInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    PDDRModuleInfoRegister::DumpLayout(sstream, areg.regs.pddr_module);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void PDDRModuleInfoRegister::DumpLayout(stringstream &sstream, const pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    sstream << +module_info.cable_technology                  << ','
            << +module_info.cable_breakout                    << ','
            << +module_info.ext_ethernet_compliance_code      << ','
            << +module_info.ethernet_compliance_code          << ','
            << +module_info.cable_type                        << ','
            << +module_info.cable_vendor                      << ','
            << +module_info.cable_length                      << ','
            << +module_info.cable_identifier                  << ','
            << +module_info.cable_power_class                 << ','
            << +module_info.max_power                         << ','
            << +module_info.cable_rx_amp                      << ','
            << +module_info.cable_rx_emphasis                 << ','
            << +module_info.cable_tx_equalization             << ','
            << +module_info.cable_attenuation_25g             << ','
            << +module_info.cable_attenuation_12g             << ','
            << +module_info.cable_attenuation_7g              << ','
            << +module_info.cable_attenuation_5g              << ','
            << +module_info.cable_rx_post_emphasis            << ','
            << +module_info.rx_cdr_cap                        << ','
            << +module_info.tx_cdr_cap                        << ','
            << +module_info.rx_cdr_state                      << ','
            << +module_info.tx_cdr_state                      << ','
            << QUOTED(DescToCsvDesc(module_info.vendor_name)) << ','
            << QUOTED(DescToCsvDesc(module_info.vendor_pn))   << ','
            << QUOTED(DescToCsvDesc(ConvertRevisionToStr(module_info))) << ','
            << module_info.fw_version                         << ','
            << QUOTED(DescToCsvDesc(module_info.vendor_sn))   << ','
            << +module_info.temperature                       << ','
            << +module_info.voltage                           << ','
            << +module_info.rx_power_lane0                    << ','
            << +module_info.rx_power_lane1                    << ','
            << +module_info.rx_power_lane2                    << ','
            << +module_info.rx_power_lane3                    << ','
            << +module_info.rx_power_lane4                    << ','
            << +module_info.rx_power_lane5                    << ','
            << +module_info.rx_power_lane6                    << ','
            << +module_info.rx_power_lane7                    << ','
            << +module_info.tx_power_lane0                    << ','
            << +module_info.tx_power_lane1                    << ','
            << +module_info.tx_power_lane2                    << ','
            << +module_info.tx_power_lane3                    << ','
            << +module_info.tx_power_lane4                    << ','
            << +module_info.tx_power_lane5                    << ','
            << +module_info.tx_power_lane6                    << ','
            << +module_info.tx_power_lane7                    << ','
            << +module_info.tx_bias_lane0                     << ','
            << +module_info.tx_bias_lane1                     << ','
            << +module_info.tx_bias_lane2                     << ','
            << +module_info.tx_bias_lane3                     << ','
            << +module_info.tx_bias_lane4                     << ','
            << +module_info.tx_bias_lane5                     << ','
            << +module_info.tx_bias_lane6                     << ','
            << +module_info.tx_bias_lane7                     << ','
            << +module_info.temperature_high_th               << ','
            << +module_info.temperature_low_th                << ','
            << +module_info.voltage_high_th                   << ','
            << +module_info.voltage_low_th                    << ','
            << +module_info.rx_power_high_th                  << ','
            << +module_info.rx_power_low_th                   << ','
            << +module_info.tx_power_high_th                  << ','
            << +module_info.tx_power_low_th                   << ','
            << +module_info.tx_bias_high_th                   << ','
            << +module_info.tx_bias_low_th                    << ','
            << +module_info.module_st                         << ','
            << +module_info.rx_power_type                     << ','
            << +module_info.did_cap                           << ','
            << +module_info.smf_length                        << ','
            << +module_info.wavelength                        << ','
            << +module_info.active_set_host_compliance_code   << ','
            << +module_info.active_set_media_compliance_code  << ','
            << +module_info.tx_bias_scaling_factor            << ','
            << +module_info.ib_compliance_code                << ','
            << +module_info.nbr250                            << ','
            << +module_info.nbr100                            << ','
            << +module_info.monitor_cap_mask                  << ','
            << +module_info.ib_width                          << ','
            << +module_info.dp_st_lane0                       << ','
            << +module_info.dp_st_lane1                       << ','
            << +module_info.dp_st_lane2                       << ','
            << +module_info.dp_st_lane3                       << ','
            << +module_info.dp_st_lane4                       << ','
            << +module_info.dp_st_lane5                       << ','
            << +module_info.dp_st_lane6                       << ','
            << +module_info.dp_st_lane7                       << ','
            << +module_info.length_om2                        << ','
            << +module_info.length_om3                        << ','
            << +module_info.length_om4                        << ','
            << +module_info.length_om5                        << ','
            << +module_info.length_om1                        << ','
            << +module_info.wavelength_tolerance              << ','
            << +module_info.memory_map_rev                    << ','
            << +module_info.memory_map_compliance             << ','
            << "0x" << HEX(module_info.date_code)             << ','
            << +module_info.connector_type                    << ','
            << +module_info.vendor_oui                        << ','
            << +module_info.cable_attenuation_53g             << ','
            << +module_info.rx_output_valid                   << ','
            << +module_info.tx_input_freq_sync                << ','
            << +module_info.error_code                        << ','
            << +module_info.cdr_vendor                        << ','
            << +module_info.max_fiber_length                  ;

    IBDIAGNET_RETURN_VOID;
}

string PDDRModuleInfoRegister::ConvertCableTemperatureToStr(const struct pddr_module_info &module_info, u_int16_t temp)
{
    IBDIAGNET_ENTER;

    stringstream temperature_ss;
    u_int8_t cable_technology = module_info.cable_technology >> 4;
    int8_t temperature = (int8_t)(temp >> 8);
    if (cable_technology == CABLE_MODULE_INFO_TYPE_COPPER_UNEQUALIZED ||
            cable_technology == CABLE_MODULE_INFO_TYPE_COPPER_PASSIVE_EQUALIZED ||
            temperature < CABLE_MODULE_INFO_MIN_SUPPORT_TEMPERATURE ||
            temperature > CABLE_MODULE_INFO_MAX_SUPPORT_TEMPERATURE) {
        temperature_ss << NO_DATA_VAL;
        IBDIAGNET_RETURN(temperature_ss.str());
    }

    temperature_ss << +temperature << 'C';

    IBDIAGNET_RETURN(temperature_ss.str());
}

string PDDRModuleInfoRegister::ConvertVoltageToStr(u_int16_t vcc)
{
    IBDIAGNET_ENTER;

    stringstream voltage_ss;
    voltage_ss << vcc * 100 << "uV";

    IBDIAGNET_RETURN(voltage_ss.str());
}

string PDDRModuleInfoRegister::ConvertMaxPowerToStr(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    if (!module_info.max_power)
        IBDIAGNET_RETURN(NO_DATA_VAL);

    stringstream max_power_ss;
    max_power_ss << (double)module_info.max_power / 4.0 << " W";
    IBDIAGNET_RETURN(max_power_ss.str());
}
const string& PDDRModuleInfoRegister::ConvertTransmitterTechnologyToStr(u_int8_t transmitter_technology)
{
    IBDIAGNET_ENTER;

    u_int8_t size = (u_int8_t)ARRAY_SIZE(PDDRModuleInfoRegister::transmitter_technology_arr);
    if (transmitter_technology >= size)
        IBDIAGNET_RETURN(PDDRModuleInfoRegister::transmitter_technology_arr[size - 1]);

    IBDIAGNET_RETURN(PDDRModuleInfoRegister::transmitter_technology_arr[transmitter_technology]);
}

#ifndef BIT_TO_STR
    #define BIT_TO_STR(val, index) (val & (1 << index) ? "1" : "0")
#else
    #error "the BIT_TO_STR has already been defined"
#endif

string PDDRModuleInfoRegister::ConvertCableTechnologyBitsToStr(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    stringstream cable_technology_ss;
    if (PDDRModuleInfoRegister::IsCMISCable(module_info.cable_identifier))
        cable_technology_ss << NO_DATA_VAL << ',' << NO_DATA_VAL << ','
                            << NO_DATA_VAL << ',' << NO_DATA_VAL;
    else
        cable_technology_ss << BIT_TO_STR(module_info.cable_technology, 0) << ','
                            << BIT_TO_STR(module_info.cable_technology, 1) << ','
                            << BIT_TO_STR(module_info.cable_technology, 2) << ','
                            << BIT_TO_STR(module_info.cable_technology, 3);

    IBDIAGNET_RETURN(cable_technology_ss.str());
}

string PDDRModuleInfoRegister::ConvertCableIdentifierToStr(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    string result;

    switch (module_info.cable_identifier) {
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP28:
        result = "QSFP28";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_PLUS:
        result = "QSFP+";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus:
        result = "SFP28/SFP+";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSA:
        result = "QSA (QSFP->SFP)";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_Backplane:
        result = "Backplane";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_SFP_DD:
        result = "SFP-DD";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_DD:
        result = "QSFP-DD";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_CMIS:
        result = "QSFP_CMIS";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_OSFP:
        result = "OSFP";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_C2C:
        result = "C2C";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_DSFP:
        result = "DSFP";
        break;
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_SPLIT_CABLE:
        result = "QSFP_Split_Cable";
        break;
    default:
        result = NO_DATA_VAL;
        break;
    }

    IBDIAGNET_RETURN(result);
}
string PDDRModuleInfoRegister::ConvertCableLengthToStr(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    if (!module_info.cable_length)
        IBDIAGNET_RETURN(NO_DATA_VAL);

    stringstream cable_length_ss;
    u_int64_t cable_length = (u_int64_t)module_info.cable_length;
    if (PDDRModuleInfoRegister::IsCMISCable(module_info.cable_identifier)) {
        u_int8_t cable_length_value = module_info.cable_length & 0x3f;
        if (!cable_length_value)
            IBDIAGNET_RETURN(NO_DATA_VAL);

        u_int8_t cable_length_mul = module_info.cable_length >> 6;
        u_int64_t mul = 0;
        if (cable_length_mul == CABLE_MODULE_INFO_CMIS_LENGTH_MUL_0_1) {
            double cable_length_float = 0.1 * (double)cable_length_value;
            cable_length_ss << cable_length_float << " m";
            IBDIAGNET_RETURN(cable_length_ss.str());
        } else {
            if (cable_length_mul == CABLE_MODULE_INFO_CMIS_LENGTH_MUL_1)
                mul = 1;
            else if (cable_length_mul == CABLE_MODULE_INFO_CMIS_LENGTH_MUL_10)
                mul = 10;
            else if (cable_length_mul == CABLE_MODULE_INFO_CMIS_LENGTH_MUL_100)
                mul = 100;
        }

        cable_length = mul * (u_int64_t)cable_length_value;
    }

    cable_length_ss << cable_length << " m";

    IBDIAGNET_RETURN(cable_length_ss.str());
}
string PDDRModuleInfoRegister::ConvertCableLengthSMFiberToStr(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    u_int16_t length_base = module_info.smf_length & 0xff;
    u_int16_t unit_bit = (module_info.smf_length >> 8) & 0x3;
    stringstream smf_length_ss;

    if (!length_base)
        IBDIAGNET_RETURN(NO_DATA_VAL);

    switch (unit_bit) {
    case 0:
        smf_length_ss << length_base << " km";
        break;
    case 0x1:
        smf_length_ss << (double)(length_base * 100) / 1000.0 << " km";
        break;
    default:
        smf_length_ss << NO_DATA_VAL;
    }

    IBDIAGNET_RETURN(smf_length_ss.str());
}
string PDDRModuleInfoRegister::ConvertCableLengthOMXToStr(const struct pddr_module_info &module_info, u_int8_t om_index)
{
    IBDIAGNET_ENTER;

    u_int8_t unit;
    u_int8_t length_omx;
    switch (om_index) {
    case 1:
        length_omx = module_info.length_om1;
        unit = 10;
        break;
    case 2:
        length_omx = module_info.length_om2;
        if (module_info.cable_identifier == CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus)
            unit = 10;
        else
            unit = 1;
        break;
    case 3:
        length_omx = module_info.length_om3;
        if (module_info.cable_identifier == CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus)
           unit = 10;
        else
           unit = 2;
        break;
    case 4:
        length_omx = module_info.length_om4;
        if (module_info.cable_identifier == CABLE_MODULE_INFO_IDENTIFIER_SFP28_SFP_plus)
           unit = 10;
        else
           unit = 2;
        break;
    case 5:
        length_omx = module_info.length_om5;
        unit = 2;
        break;
    default:
        IBDIAGNET_RETURN(NO_DATA_VAL);
    }

    if (!length_omx)
        IBDIAGNET_RETURN(NO_DATA_VAL);

    stringstream length_omx_ss;
    length_omx_ss << (u_int64_t)length_omx * unit << " m";
    IBDIAGNET_RETURN(length_omx_ss.str());
}

bool PDDRModuleInfoRegister::IsCMISCable(uint8_t identifier)
{
    IBDIAGNET_ENTER;

    switch (identifier)
    {
    case CABLE_MODULE_INFO_IDENTIFIER_SFP_DD:
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_DD:
    case CABLE_MODULE_INFO_IDENTIFIER_QSFP_CMIS:
    case CABLE_MODULE_INFO_IDENTIFIER_OSFP:
    case CABLE_MODULE_INFO_IDENTIFIER_DSFP:
          IBDIAGNET_RETURN(true);

    default:
          IBDIAGNET_RETURN(false);
    }

}

string PDDRModuleInfoRegister::ConvertIBComplianceCodeToStr(u_int8_t ib_compliance_code)
{
    IBDIAGNET_ENTER;
    stringstream ib_compliance_code_ss;

    u_int8_t size = ARRAY_SIZE(PDDRModuleInfoRegister::ib_compliance_code_arr);
    for (u_int8_t i = 0; i < size; i++) {
         if (ib_compliance_code & (1 << i))
             ib_compliance_code_ss << PDDRModuleInfoRegister::ib_compliance_code_arr[i];
    }

    string ib_compliance_code_str = ib_compliance_code_ss.str();
    if (ib_compliance_code_str.empty()) {
        ib_compliance_code_str = NO_DATA_VAL;
    } else {
        ib_compliance_code_str.erase(ib_compliance_code_str.size() - 1, 1);
    }

    IBDIAGNET_RETURN(ib_compliance_code_str);

}

u_int8_t PDDRModuleInfoRegister::SelectNominalBR(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    if (module_info.nbr100 == 0xff)
        IBDIAGNET_RETURN(module_info.nbr250 / 4);

    IBDIAGNET_RETURN(module_info.nbr100 / 10);

}

u_int8_t PDDRModuleInfoRegister::SelectTransmitterTechnology(const struct pddr_module_info &module_info)
{
     IBDIAGNET_ENTER;

     if (PDDRModuleInfoRegister::IsCMISCable(module_info.cable_identifier))
         IBDIAGNET_RETURN(module_info.cable_technology);

     IBDIAGNET_RETURN(module_info.cable_technology >> 4);

}

bool PDDRModuleInfoRegister::IsModule(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    if (((module_info.cable_technology >> 4) != CABLE_MODULE_INFO_TYPE_COPPER_UNEQUALIZED) &&
        (module_info.connector_type != CABLE_MODULE_INFO_NO_SEPARABLE_CONNECTOR))
        IBDIAGNET_RETURN (true);
    IBDIAGNET_RETURN(false);

}

bool PDDRModuleInfoRegister::IsActiveCable(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    if (((module_info.cable_technology >> 4) != CABLE_MODULE_INFO_TYPE_COPPER_UNEQUALIZED) &&
        (module_info.connector_type == CABLE_MODULE_INFO_NO_SEPARABLE_CONNECTOR))
        IBDIAGNET_RETURN(true);
    IBDIAGNET_RETURN(false);

}

bool PDDRModuleInfoRegister::IsPassiveCable(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    if ((module_info.cable_technology >> 4) == CABLE_MODULE_INFO_TYPE_COPPER_UNEQUALIZED)
        IBDIAGNET_RETURN(true);
    IBDIAGNET_RETURN(false);

}

string PDDRModuleInfoRegister::ConvertCDREnableTxRxToStr(const struct pddr_module_info &module_info, bool is_rx,
                                                         bool is_csv)
{
    IBDIAGNET_ENTER;

    stringstream cdr_state_ss;

    if (PDDRModuleInfoRegister::IsModule(module_info) ||
        PDDRModuleInfoRegister::IsActiveCable(module_info)) {
        u_int8_t cdr_cap;
        u_int8_t cdr_state;
        if (is_rx) {
            cdr_cap = module_info.rx_cdr_cap;
            cdr_state = module_info.rx_cdr_state;
        } else {
            cdr_cap = module_info.tx_cdr_cap;
            cdr_state = module_info.tx_cdr_state;
        }
        if (is_csv || cdr_cap) {
            cdr_state_ss << PTR(cdr_state);
        } else {
            cdr_state_ss << NO_DATA_VAL;
        }
    } else {
        cdr_state_ss << NO_DATA_VAL;
    }

    IBDIAGNET_RETURN(cdr_state_ss.str());

}

string PDDRModuleInfoRegister::ConvertTxEQRxAMPRxEMPToStr(const struct pddr_module_info &module_info, u_int8_t val,
                                                          bool is_csv)
{
    IBDIAGNET_ENTER;

    stringstream val_ss;
    if (PDDRModuleInfoRegister::IsModule(module_info) ||
        PDDRModuleInfoRegister::IsActiveCable(module_info)) {
        if (!is_csv)
            val_ss << +val << " " << +val << " " << +val << " " << +val;
        else
            val_ss << HEX(val) << HEX(val) << HEX(val) << HEX(val);
    } else {
        val_ss << NO_DATA_VAL;
    }

    IBDIAGNET_RETURN(val_ss.str());
}

string PDDRModuleInfoRegister::ConvertAttenuationToStr(const struct pddr_module_info &module_info, bool is_csv)
{
    IBDIAGNET_ENTER;
    stringstream attenuation_ss;

    string separator;
    string na_str;
    if (is_csv) {
        separator = ",";
    } else {
        separator = " ";
    }

    if (PDDRModuleInfoRegister::IsPassiveCable(module_info)) {
        attenuation_ss << +module_info.cable_attenuation_5g << separator
                       << +module_info.cable_attenuation_7g << separator
                       << +module_info.cable_attenuation_12g << separator
                       << +module_info.cable_attenuation_25g;
    } else {
        attenuation_ss << NO_DATA_VAL << separator << NO_DATA_VAL << separator << NO_DATA_VAL
                       << separator << NO_DATA_VAL;
    }

    IBDIAGNET_RETURN(attenuation_ss.str());
}

string PDDRModuleInfoRegister::ConvertFWVersionToStr(const struct pddr_module_info &module_info, bool is_csv)
{
    IBDIAGNET_ENTER;
    stringstream fw_version_ss;

    if ((module_info.fw_version != 0) &&
        (PDDRModuleInfoRegister::IsModule(module_info) ||
         PDDRModuleInfoRegister::IsActiveCable(module_info))) {
        uint8_t fw_hi = (uint8_t)(module_info.fw_version >> 24);
        uint8_t fw_mid = (uint8_t)((module_info.fw_version >> 16) & 0x00ff);
        uint16_t fw_lo = (uint16_t)(module_info.fw_version & 0x0000ffff);
        fw_version_ss << +fw_hi << "." << +fw_mid << "." << fw_lo;
    } else  {
        fw_version_ss << NO_DATA_VAL;
    }

    IBDIAGNET_RETURN(fw_version_ss.str());

}

string PDDRModuleInfoRegister::ConvertDateCodeToStr(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    u_int64_t date_code = module_info.date_code >> 16;
    char *p_date_code = (char *)&date_code;
    for (int i = 0; i < 6; i++) {
         if (!isdigit(p_date_code[i]))
             IBDIAGNET_RETURN(NO_DATA_VAL);
    }

    stringstream date_code_ss;
    date_code_ss << p_date_code[5] << p_date_code[4] << '-'
                 << p_date_code[3] << p_date_code[2] << '-'
                 << p_date_code[1] << p_date_code[0];

    IBDIAGNET_RETURN(date_code_ss.str());

}

string PDDRModuleInfoRegister::ConvertRevisionToStr(const struct pddr_module_info &module_info)
{
    IBDIAGNET_ENTER;

    stringstream rev_ss;
    size_t size = sizeof(module_info.vendor_rev);
    for (size_t i = 0; i < size; i++) {
         if (module_info.vendor_rev[i])
             rev_ss << module_info.vendor_rev[i];
    }

    IBDIAGNET_RETURN(rev_ss.str());

}

double PDDRModuleInfoRegister::MWToDBM(double mW)
{
    if (mW == 0)
        IBDIAGNET_RETURN(-999.999);
    IBDIAGNET_RETURN(10 * log10(mW));

}

void PDDRModuleInfoRegister::DumpModuleInfoHeader(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "Vendor,OUI,PN,SN,REV,Identifier,IdentifierStr,"
            << "LengthCopperOrActive,LengthSMFiber,LengthOM1,LengthOM2,LengthOM3,LengthOM4,"
            << "LengthOM5,TypeDesc,Temperature,PowerClass,MaxPower,Connector,Type,SupportedSpeed,"
            << "SupportedSpeedDesc,NominalBitrate,CDREnableRx,CDREnableTx,"
            << "InputEq,OutputAmp,OutputPreEmp,OutputPostEmp,"
            << "FWVersion,Attenuation5G,Attenuation7G,Attenuation12G,Attenuation25G,RXPowerType,"
            << "module_info_ext,RX1Power,RX2Power,RX3Power,RX4Power,"
            << "RX5Power,RX6Power,RX7Power,RX8Power,"
            << "TX1Bias,TX2Bias,TX3Bias,TX4Bias,"
            << "TX5Bias,TX6Bias,TX7Bias,TX8Bias,"
            << "TX1Power,TX2Power,TX3Power,TX4Power,"
            << "TX5Power,TX6Power,TX7Power,TX8Power,SupplyVoltageReporting,"
            << "DateCode,Lot,TransmitterTechnology,ActiveWavelengthControl,CooledTransmitterDevice,"
            << "ActivePinDetector,TunableTransmitter,ExtendedSpecificationComplianceCodes,"
            << "AlarmTemperatureHighThresh,AlarmTemperatureLowThresh,"
            << "WarnTemperatureHighThresh,WarnTemperatureLowThresh,"
            << "AlarmVoltageHighThresh,AlarmVoltageLowThresh,"
            << "WarnVoltageHighThresh,WarnVoltageLowThresh,"
            << "RXPowerHighThresh,RXPowerLowThresh,TXPowerHighThresh,TXPowerLowThresh,"
            << "TXBiasHighThresh,TXBiasLowThresh";

    IBDIAGNET_RETURN_VOID;

}
void PDDRModuleInfoRegister::DumpModuleInfoData(stringstream &sstream, const struct pddr_module_info *p_module_info,
                                                u_int16_t module_info_ext)
{
    IBDIAGNET_ENTER;

    if (!p_module_info) {
        sstream << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A"  //RX(5-8)Power
                << "N/A,N/A,N/A,N/A"  //TX(5-8)Bias
                << "N/A,N/A,N/A,N/A"  //TX(5-8)Power
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A";
        IBDIAGNET_RETURN_VOID;
    }

    u_int8_t transmitter_technology =
                               PDDRModuleInfoRegister::SelectTransmitterTechnology(*p_module_info);

    sstream << QUOTED(DescToCsvDesc(p_module_info->vendor_name))                                 << ','
            << PTR(p_module_info->vendor_oui)                                                    << ','
            << QUOTED(DescToCsvDesc(p_module_info->vendor_pn))                                   << ','
            << QUOTED(DescToCsvDesc(p_module_info->vendor_sn))                                   << ','
            << QUOTED(DescToCsvDesc(ConvertRevisionToStr(*p_module_info)))                       << ','
            << PTR(p_module_info->cable_identifier)                                              << ','
            << QUOTED(ConvertCableIdentifierToStr(*p_module_info))                               << ','
            << QUOTED(ConvertCableLengthToStr(*p_module_info))                                   << ','
            << QUOTED(ConvertCableLengthSMFiberToStr(*p_module_info))                            << ','
            << QUOTED(ConvertCableLengthOMXToStr(*p_module_info, 1))                             << ','
            << QUOTED(ConvertCableLengthOMXToStr(*p_module_info, 2))                             << ','
            << QUOTED(ConvertCableLengthOMXToStr(*p_module_info, 3))                             << ','
            << QUOTED(ConvertCableLengthOMXToStr(*p_module_info, 4))                             << ','
            << QUOTED(ConvertCableLengthOMXToStr(*p_module_info, 5))                             << ','
            << QUOTED(ConvertTransmitterTechnologyToStr(transmitter_technology))                 << ','
            << QUOTED(ConvertCableTemperatureToStr(*p_module_info, p_module_info->temperature))  << ','
            << +p_module_info->cable_power_class                                                 << ','
            << ConvertMaxPowerToStr(*p_module_info)                                              << ','
            << PTR(p_module_info->connector_type)                                                << ','
            << +transmitter_technology                                                           << ','
            << +p_module_info->ib_compliance_code                                                << ','
            << QUOTED(ConvertIBComplianceCodeToStr(p_module_info->ib_compliance_code))           << ','
            << (SelectNominalBR(*p_module_info) * 1000)                                          << ','
            << ConvertCDREnableTxRxToStr(*p_module_info, true, true)                             << ','
            << ConvertCDREnableTxRxToStr(*p_module_info, false, true)                            << ','
            << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                          p_module_info->cable_tx_equalization, true)            << ','
            << ConvertTxEQRxAMPRxEMPToStr(*p_module_info, p_module_info->cable_rx_amp, true)     << ','
            << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                          p_module_info->cable_rx_emphasis, true)                << ','
            << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                          p_module_info->cable_rx_post_emphasis, true)           << ','
            << ConvertFWVersionToStr(*p_module_info, true)                                       << ','
            << ConvertAttenuationToStr(*p_module_info, true)                                     << ','
            //module_info_ext is not available here
            << +p_module_info->rx_power_type                                                     << ','
            << module_info_ext                                                                   << ','
            << (float)(p_module_info->rx_power_lane0)                                            << ','
            << (float)(p_module_info->rx_power_lane1)                                            << ','
            << (float)(p_module_info->rx_power_lane2)                                            << ','
            << (float)(p_module_info->rx_power_lane3)                                            << ','
            << (float)(p_module_info->rx_power_lane4)                                            << ','
            << (float)(p_module_info->rx_power_lane5)                                            << ','
            << (float)(p_module_info->rx_power_lane6)                                            << ','
            << (float)(p_module_info->rx_power_lane7)                                            << ','
            << (float)(p_module_info->tx_bias_lane0 / 500.0)                                     << ','
            << (float)(p_module_info->tx_bias_lane1 / 500.0)                                     << ','
            << (float)(p_module_info->tx_bias_lane2 / 500.0)                                     << ','
            << (float)(p_module_info->tx_bias_lane3 / 500.0)                                     << ','
            << (float)(p_module_info->tx_bias_lane4 / 500.0)                                     << ','
            << (float)(p_module_info->tx_bias_lane5 / 500.0)                                     << ','
            << (float)(p_module_info->tx_bias_lane6 / 500.0)                                     << ','
            << (float)(p_module_info->tx_bias_lane7 / 500.0)                                     << ','
            << (float)(p_module_info->tx_power_lane0)                                            << ','
            << (float)(p_module_info->tx_power_lane1)                                            << ','
            << (float)(p_module_info->tx_power_lane2)                                            << ','
            << (float)(p_module_info->tx_power_lane3)                                           << ','
            << (float)(p_module_info->tx_power_lane4)                                           << ','
            << (float)(p_module_info->tx_power_lane5)                                           << ','
            << (float)(p_module_info->tx_power_lane6)                                           << ','
            << (float)(p_module_info->tx_power_lane7)                                           << ','
            << QUOTED(ConvertVoltageToStr(p_module_info->voltage))                               << ','
            << ConvertDateCodeToStr(*p_module_info)                                              << ','
            << PTR((u_int16_t)(p_module_info->date_code & 0xffff))                               << ','
            << PTR(transmitter_technology)                                                       << ','
            << ConvertCableTechnologyBitsToStr(*p_module_info)                                   << ','
            << +p_module_info->ext_ethernet_compliance_code                                      << ','
            << QUOTED(ConvertCableTemperatureToStr(*p_module_info,
                                                   p_module_info->temperature_high_th))          << ','
            << QUOTED(ConvertCableTemperatureToStr(*p_module_info,
                                                   p_module_info->temperature_low_th))           << ','
            << "N/A"                                                                             << ','
            << "N/A"                                                                             << ','
            << QUOTED(ConvertVoltageToStr(p_module_info->voltage_high_th))                       << ','
            << QUOTED(ConvertVoltageToStr(p_module_info->voltage_low_th))                        << ','
            << "N/A"                                                                             << ','
            << "N/A"                                                                             << ','
            << +p_module_info->rx_power_high_th                                                  << ','
            << +p_module_info->rx_power_low_th                                                   << ','
            << +p_module_info->tx_power_high_th                                                  << ','
            << +p_module_info->tx_power_low_th                                                   << ','
            << +p_module_info->tx_bias_high_th                                                   << ','
            << +p_module_info->tx_bias_low_th                                                    ;

    IBDIAGNET_RETURN_VOID;

}
void PDDRModuleInfoRegister::DumpModuleInfoData(ofstream &sout, const struct pddr_module_info *p_module_info,
                                                u_int16_t module_info_ext)
{
    IBDIAGNET_ENTER;

    if (!p_module_info) {
         sout << "Identifier: N/A" << endl
              << "Vendor: N/A" << endl
              << "OUI: N/A" << endl
              << "PN: N/A"  << endl
              << "SN: N/A"  << endl
              << "Rev: N/A"   << endl
              << "LengthCopperOrActive: N/A" << endl
              << "LengthSMFiber: N/A" << endl
              << "LengthOM1: N/A" << endl
              << "LengthOM2: N/A" << endl
              << "LengthOM3: N/A" << endl
              << "LengthOM4: N/A" << endl
              << "LengthOM5: N/A" << endl
              << "Type: N/A" << endl
              << "SupportedSpeed: N/A" << endl
              << "Temperature: N/A"  << endl
              << "PowerClass: N/A" << endl
              << "NominalBitrate: N/A"  << endl
              << "CDREnableRx: N/A"  << endl
              << "CDREnableTx: N/A"  << endl
              << "InputEq: N/A"  << endl
              << "OutputAmp: N/A"  << endl
              << "OutputPreEmp: N/A"  << endl
              << "OutputPostEmp: N/A" << endl
              << "FW Version: N/A" <<  endl
              << "Attenuation(5,7,12,25): N/A" << endl
              << "RX power type: N/A" << endl
              << "RX1 Power: N/A" << endl
              << "RX2 Power: N/A" << endl
              << "RX3 Power: N/A" << endl
              << "RX4 Power: N/A" << endl
              << "RX5 Power: N/A" << endl
              << "RX6 Power: N/A" << endl
              << "RX7 Power: N/A" << endl
              << "RX8 Power: N/A" << endl
              << "TX1 Bias: N/A" << endl
              << "TX2 Bias: N/A" << endl
              << "TX3 Bias: N/A" << endl
              << "TX4 Bias: N/A" << endl
              << "TX5 Bias: N/A" << endl
              << "TX6 Bias: N/A" << endl
              << "TX7 Bias: N/A" << endl
              << "TX8 Bias: N/A" << endl
              << "TX1 Power: N/A" << endl
              << "TX2 Power: N/A" << endl
              << "TX3 Power: N/A" << endl
              << "TX4 Power: N/A" << endl
              << "TX5 Power: N/A" << endl
              << "TX6 Power: N/A" << endl
              << "TX7 Power: N/A" << endl
              << "TX8 Power: N/A" << endl
              << "Supply Voltage Reporting: N/A" << endl
              << "DateCode: N/A" << endl
              << "Lot: N/A" << endl
              << "Transmitter Technology: N/A" << endl
              << "Extended Specification Compliance Codes: N/A" << endl
              << "AlarmTemperatureHighThresh: N/A" << endl
              << "AlarmTemperatureLowThresh: N/A" << endl
              << "WarnTemperatureHighThresh: N/A" << endl
              << "WarnTemperatureLowThresh: N/A" << endl
              << "AlarmVoltageHighThresh: N/A" << endl
              << "AlarmVoltageLowThresh: N/A" << endl
              << "WarnVoltageHighThresh: N/A" << endl
              << "WarnVoltageLowThresh: N/A";
         IBDIAGNET_RETURN_VOID;
    }

    u_int8_t transmitter_technology =
                              PDDRModuleInfoRegister::SelectTransmitterTechnology(*p_module_info);
    sout << "Identifier: " << ConvertCableIdentifierToStr(*p_module_info) << " ("
            << PTR(p_module_info->cable_identifier) << ')' << endl
         << "Vendor: " << DescToCsvDesc(p_module_info->vendor_name) << endl
         << "OUI: " << PTR(p_module_info->vendor_oui) << endl
         << "PN: " << DescToCsvDesc(p_module_info->vendor_pn) << endl
         << "SN: " << DescToCsvDesc(p_module_info->vendor_sn) << endl
         << "Rev: " << DescToCsvDesc(ConvertRevisionToStr(*p_module_info)) << endl
         << "LengthCopperOrActive: " << ConvertCableLengthToStr(*p_module_info) << endl
         << "LengthSMFiber: " << ConvertCableLengthSMFiberToStr(*p_module_info) << endl
         << "LengthOM1: " << ConvertCableLengthOMXToStr(*p_module_info, 1) << endl
         << "LengthOM2: " << ConvertCableLengthOMXToStr(*p_module_info, 2) << endl
         << "LengthOM3: " << ConvertCableLengthOMXToStr(*p_module_info, 3) << endl
         << "LengthOM4: " << ConvertCableLengthOMXToStr(*p_module_info, 4) << endl
         << "LengthOM5: " << ConvertCableLengthOMXToStr(*p_module_info, 5) << endl
         << "Type: "
         << ConvertTransmitterTechnologyToStr(transmitter_technology) << endl
         << "SupportedSpeed: "
         << ConvertIBComplianceCodeToStr(p_module_info->ib_compliance_code) << endl
         << "Temperature: " << ConvertCableTemperatureToStr(*p_module_info,
                                                            p_module_info->temperature) << endl
         << "PowerClass: " << +p_module_info->cable_power_class << endl
         << "NominalBitrate: " << +SelectNominalBR(*p_module_info) << endl
         << "CDREnableRx: " << ConvertCDREnableTxRxToStr(*p_module_info, true, false) << endl
         << "CDREnableTx: " << ConvertCDREnableTxRxToStr(*p_module_info, false, false) << endl
         << "InputEq: " << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                                      p_module_info->cable_tx_equalization,
                                                      false) << endl
         << "OutputAmp: " << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                                        p_module_info->cable_rx_amp,
                                                        false) << endl
         << "OutputPreEmp: " << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                                           p_module_info->cable_rx_emphasis,
                                                           false) << endl
         << "OutputPostEmp: " << ConvertTxEQRxAMPRxEMPToStr(*p_module_info,
                                                            p_module_info->cable_rx_post_emphasis,
                                                            false) << endl
         << "FW Version: " << ConvertFWVersionToStr(*p_module_info, false) << endl
         << "Attenuation(5,7,12,25): " << ConvertAttenuationToStr(*p_module_info, false) << endl
         << "RX power type: " << (p_module_info->rx_power_type ? "Average Power" : "OMA") << endl
         << "RX1 Power: " << p_module_info->rx_power_lane0 << (module_info_ext ? " uW" : " dBm") << endl
         << "RX2 Power: " << p_module_info->rx_power_lane1 << (module_info_ext ? " uW" : " dBm") << endl
         << "RX3 Power: " << p_module_info->rx_power_lane2 << (module_info_ext ? " uW" : " dBm") << endl
         << "RX4 Power: " << p_module_info->rx_power_lane3 << (module_info_ext ? " uW" : " dBm") << endl
         << "RX5 Power: " << p_module_info->rx_power_lane4 << (module_info_ext ? " uW" : " dBm") << endl
         << "RX6 Power: " << p_module_info->rx_power_lane5 << (module_info_ext ? " uW" : " dBm") << endl
         << "RX7 Power: " << p_module_info->rx_power_lane6 << (module_info_ext ? " uW" : " dBm") << endl
         << "RX8 Power: " << p_module_info->rx_power_lane7 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX1 Bias: " << (float)(p_module_info->tx_bias_lane0 / 500.0) << " mA" << endl
         << "TX2 Bias: " << (float)(p_module_info->tx_bias_lane1 / 500.0) << " mA" << endl
         << "TX3 Bias: " << (float)(p_module_info->tx_bias_lane2 / 500.0) << " mA" << endl
         << "TX4 Bias: " << (float)(p_module_info->tx_bias_lane3 / 500.0) << " mA" << endl
         << "TX5 Bias: " << (float)(p_module_info->tx_bias_lane4 / 500.0) << " mA" << endl
         << "TX6 Bias: " << (float)(p_module_info->tx_bias_lane5 / 500.0) << " mA" << endl
         << "TX7 Bias: " << (float)(p_module_info->tx_bias_lane6 / 500.0) << " mA" << endl
         << "TX8 Bias: " << (float)(p_module_info->tx_bias_lane7 / 500.0) << " mA" << endl
         << "TX1 Power: " << p_module_info->tx_power_lane0 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX2 Power: " << p_module_info->tx_power_lane1 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX3 Power: " << p_module_info->tx_power_lane2 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX4 Power: " << p_module_info->tx_power_lane3 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX5 Power: " << p_module_info->tx_power_lane4 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX6 Power: " << p_module_info->tx_power_lane5 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX7 Power: " << p_module_info->tx_power_lane6 << (module_info_ext ? " uW" : " dBm") << endl
         << "TX8 Power: " << p_module_info->tx_power_lane7 << (module_info_ext ? " uW" : " dBm") << endl
         << "Supply Voltage Reporting: " << ConvertVoltageToStr(p_module_info->voltage) << endl
         << "DateCode: "<< ConvertDateCodeToStr(*p_module_info) << endl
         << "Lot: " << PTR((u_int16_t)(p_module_info->date_code & 0xffff))<< endl
         << "Transmitter Technology: " << PTR(p_module_info->cable_technology) << endl
         << "Extended Specification Compliance Codes: "
            << +p_module_info->ext_ethernet_compliance_code << endl
         << "AlarmTemperatureHighThresh: "
            << ConvertCableTemperatureToStr(*p_module_info, p_module_info->temperature_high_th) << endl
         << "AlarmTemperatureLowThresh: "
            << ConvertCableTemperatureToStr(*p_module_info, p_module_info->temperature_low_th) << endl
         << "WarnTemperatureHighThresh: " << "N/A" << endl
         << "WarnTemperatureLowThresh: " << "N/A" << endl
         << "AlarmVoltageHighThresh: " << ConvertVoltageToStr(p_module_info->voltage_high_th) << endl
         << "AlarmVoltageLowThresh: " << ConvertVoltageToStr(p_module_info->voltage_low_th) << endl
         << "WarnVoltageHighThresh: N/A" << endl
         << "WarnVoltageLowThresh: N/A";

    IBDIAGNET_RETURN_VOID;

}
PhyCableRecord::ModuleRecord * PDDRModuleInfoRegister::CreateRecord(const struct pddr_module_info *p_module_info,
                                                                    u_int16_t module_info_ext)
{
    IBDIAGNET_ENTER;

    if (!p_module_info)
        IBDIAGNET_RETURN(NULL);

    PhyCableRecord::ModuleRecord *p_record = new PhyCableRecord::ModuleRecord();
    PDDRModuleInfoRegister::CopyTo(*p_module_info, *p_record);
    p_record->module_info_ext = module_info_ext;

    IBDIAGNET_RETURN(p_record);

}

#ifdef COPY
    #error "The COPY macros has been already defined"
#else
    #define COPY(X) to.X = from.X
#endif

#ifdef COPY_ARRAY
    #error "The COPY_ARRAY macros has been already defined"
#else
    #define COPY_ARRAY(X) memcpy(to.X, from.X, sizeof(to.X))
#endif

void PDDRModuleInfoRegister::CopyTo(const pddr_module_info &from, PhyCableRecord::ModuleRecord &to)
{
    IBDIAGNET_ENTER;

    COPY(ethernet_compliance_code);
    COPY(ext_ethernet_compliance_code);
    COPY(cable_breakout);
    COPY(cable_technology);
    COPY(cable_power_class);
    COPY(cable_identifier);
    COPY(cable_length);
    COPY(cable_vendor);
    COPY(cable_type);
    COPY(cable_tx_equalization);
    COPY(cable_rx_emphasis);
    COPY(cable_rx_amp);
    COPY(max_power);
    COPY(cable_attenuation_5g);
    COPY(cable_attenuation_7g);
    COPY(cable_attenuation_12g);
    COPY(cable_attenuation_25g);
    COPY(tx_cdr_state);
    COPY(rx_cdr_state);
    COPY(tx_cdr_cap);
    COPY(rx_cdr_cap);
    COPY(cable_rx_post_emphasis);
    COPY_ARRAY(vendor_name);
    COPY_ARRAY(vendor_pn);
    COPY_ARRAY(vendor_rev);
    COPY(fw_version);
    COPY_ARRAY(vendor_sn);
    COPY(voltage);
    COPY(temperature);
    COPY(rx_power_lane1);
    COPY(rx_power_lane0);
    COPY(rx_power_lane3);
    COPY(rx_power_lane2);
    COPY(rx_power_lane5);
    COPY(rx_power_lane4);
    COPY(rx_power_lane7);
    COPY(rx_power_lane6);
    COPY(tx_power_lane1);
    COPY(tx_power_lane0);
    COPY(tx_power_lane3);
    COPY(tx_power_lane2);
    COPY(tx_power_lane5);
    COPY(tx_power_lane4);
    COPY(tx_power_lane7);
    COPY(tx_power_lane6);
    COPY(tx_bias_lane1);
    COPY(tx_bias_lane0);
    COPY(tx_bias_lane3);
    COPY(tx_bias_lane2);
    COPY(tx_bias_lane5);
    COPY(tx_bias_lane4);
    COPY(tx_bias_lane7);
    COPY(tx_bias_lane6);
    COPY(temperature_low_th);
    COPY(temperature_high_th);
    COPY(voltage_low_th);
    COPY(voltage_high_th);
    COPY(rx_power_low_th);
    COPY(rx_power_high_th);
    COPY(tx_power_low_th);
    COPY(tx_power_high_th);
    COPY(tx_bias_low_th);
    COPY(tx_bias_high_th);
    COPY(wavelength);
    COPY(smf_length);
    COPY(did_cap);
    COPY(rx_power_type);
    COPY(module_st);
    COPY(ib_compliance_code);
    COPY(active_set_media_compliance_code);
    COPY(active_set_host_compliance_code);
    COPY(ib_width);
    COPY(monitor_cap_mask);
    COPY(nbr100);
    COPY(nbr250);
    COPY(dp_st_lane7);
    COPY(dp_st_lane6);
    COPY(dp_st_lane5);
    COPY(dp_st_lane4);
    COPY(dp_st_lane3);
    COPY(dp_st_lane2);
    COPY(dp_st_lane1);
    COPY(dp_st_lane0);
    COPY(length_om5);
    COPY(length_om4);
    COPY(length_om3);
    COPY(length_om2);
    COPY(memory_map_rev);
    COPY(wavelength_tolerance);
    COPY(length_om1);
    COPY(memory_map_compliance);
    COPY(date_code);
    COPY(vendor_oui);
    COPY(connector_type);
    COPY(rx_output_valid);
    COPY(cable_attenuation_53g);
    COPY(tx_input_freq_sync);
    COPY(error_code);
    COPY(cdr_vendor);
    COPY(max_fiber_length);

    IBDIAGNET_RETURN_VOID;
}

#undef COPY
#undef COPY_ARRAY

PDDRLinkUpInfoRegister::PDDRLinkUpInfoRegister(PhyDiag * phy_diag)
    : PDDRRegister(phy_diag, 
                   0x08, 
                   (const unpack_data_func_t)pddr_link_up_info_page_unpack,
                   ACC_REG_PDDR_LINKUP_INTERNAL_SECTION_NAME,
                   ACC_REG_PDDR_LINKUP_NAME,
                   ACC_REG_PDDR_LINKUP_FIELDS_NUM,
                   ACC_REG_PDDR_LINKUP_HEADER,
                   SUPPORT_SW_CA)
{
}

void PDDRLinkUpInfoRegister::DumpLayout(stringstream &sstream, const struct pddr_link_up_info_page &linkup_info)
{
    IBDIAGNET_ENTER;
    sstream << +linkup_info.up_reason_mng << ','
            << +linkup_info.up_reason_drv << ','
            << +linkup_info.up_reason_pwr << ','
            << +linkup_info.time_to_link_up << ','
            << +linkup_info.lt_cnt << ','
            << +linkup_info.fast_link_up_status << ','
            << +linkup_info.time_to_link_up_phy_up_to_active << ','
            << +linkup_info.time_to_link_up_sd_to_phy_up << ','
            << +linkup_info.time_to_link_up_disable_to_sd << ','
            << +linkup_info.time_to_link_up_disable_to_pd;
    IBDIAGNET_RETURN_VOID;
}

void PDDRLinkUpInfoRegister::DumpLinkUpInfoRealHeader(stringstream &sstream, const string& header)
{
    IBDIAGNET_ENTER;
#ifdef SHOW_PHY_HEADERS
    sstream << '#' << header                         << ','
            << "up_reason_mng"                       << ','
            << "up_reason_drv"                       << ','
            << "up_reason_pwr"                       << ','
            << "time_to_link_up"                     << ','
            << "lt_cnt"                              << ','
            << "fast_link_up_status"                 << ','
            << "time_to_link_up_phy_up_to_active"    << ','
            << "time_to_link_up_sd_to_phy_up"        << ','
            << "time_to_link_up_disable_to_sd"       << ','
            << "time_to_link_up_disable_to_pd";

#endif
    IBDIAGNET_RETURN_VOID;
}

void PDDRLinkUpInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
    sstream << std::endl;
    DumpLinkUpInfoRealHeader(sstream, handler_header);
    IBDIAGNET_RETURN_VOID;
}

void PDDRLinkUpInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    DumpLayout(sstream, areg.regs.pddr_link_up);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

PDDRLatchedFlagInfoRegister::PDDRLatchedFlagInfoRegister(PhyDiag * phy_diag)
    : PDDRRegister(phy_diag,
                   0x09,
                   (const unpack_data_func_t)module_latched_flag_info_unpack,
                   ACC_REG_PDDR_LATCHED_INTERNAL_SECTION_NAME,
                   ACC_REG_PDDR_LATCHED_NAME,
                   ACC_REG_PDDR_LATCHED_FIELDS_NUM,
                   ACC_REG_PDDR_LATCHED_HEADER,
                   SUPPORT_SW_CA,
                   false,
                   phy_diag->GetToGetCableDisconnectedPorts())
{
}

int PDDRLatchedFlagInfoRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    AccRegKeyPort * key = dynamic_cast<AccRegKeyPort*>(p_key);

    if (!key || !p_node)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    IBPort * p_port = p_node->getPort(key->port_num);

    if (!p_port)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    if (p_port->p_combined_cable && !m_phy_diag->GetPhyCableFullData())
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    IBDIAGNET_RETURN(PDDRRegister::PackData(p_key, data, p_node));
}

void PDDRLatchedFlagInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                       << '#'
            << handler_header             << ',';
    PDDRLatchedFlagInfoRegister::DumpRawDataHeader(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PDDRLatchedFlagInfoRegister::DumpRawDataHeader(stringstream &sstream) {
    IBDIAGNET_ENTER;

    sstream << "dp_fw_fault"              << ','
            << "mod_fw_fault"             << ','
            << "vcc_flags"                << ','
            << "temp_flags"               << ','
            << "tx_ad_eq_fault"           << ','
            << "tx_cdr_lol"               << ','
            << "tx_los"                   << ','
            << "tx_fault"                 << ','
            << "tx_power_lo_war"          << ','
            << "tx_power_hi_war"          << ','
            << "tx_power_lo_al"           << ','
            << "tx_power_hi_al"           << ','
            << "tx_bias_lo_war"           << ','
            << "tx_bias_hi_war"           << ','
            << "tx_bias_lo_al"            << ','
            << "tx_bias_hi_al"            << ','
            << "rx_cdr_lol"               << ','
            << "rx_los"                   << ','
            << "rx_power_lo_war"          << ','
            << "rx_power_hi_war"          << ','
            << "rx_power_lo_al"           << ','
            << "rx_power_hi_al"           << ','
            << "rx_output_valid_change"   << ','
            << "flag_in_use";

    IBDIAGNET_RETURN_VOID;
}

void PDDRLatchedFlagInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream& sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    PDDRLatchedFlagInfoRegister::DumpLayout(sstream, areg.regs.pddr_latched);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void PDDRLatchedFlagInfoRegister::DumpLayout(stringstream &sstream, const module_latched_flag_info &latched_flag_info)
{
    IBDIAGNET_ENTER;

    sstream << +latched_flag_info.dp_fw_fault << ','
            << +latched_flag_info.mod_fw_fault << ','
            << +latched_flag_info.vcc_flags << ','
            << +latched_flag_info.temp_flags << ','
            << +latched_flag_info.tx_ad_eq_fault << ','
            << +latched_flag_info.tx_cdr_lol << ','
            << +latched_flag_info.tx_los << ','
            << +latched_flag_info.tx_fault << ','
            << +latched_flag_info.tx_power_lo_war << ','
            << +latched_flag_info.tx_power_hi_war << ','
            << +latched_flag_info.tx_power_lo_al << ','
            << +latched_flag_info.tx_power_hi_al << ','
            << +latched_flag_info.tx_bias_lo_war << ','
            << +latched_flag_info.tx_bias_hi_war << ','
            << +latched_flag_info.tx_bias_lo_al << ','
            << +latched_flag_info.tx_bias_hi_al << ','
            << +latched_flag_info.rx_cdr_lol << ','
            << +latched_flag_info.rx_los << ','
            << +latched_flag_info.rx_power_lo_war << ','
            << +latched_flag_info.rx_power_hi_war << ','
            << +latched_flag_info.rx_power_lo_al << ','
            << +latched_flag_info.rx_power_hi_al << ','
            << +latched_flag_info.rx_output_valid_change << ','
            << PTR(latched_flag_info.flag_in_use);

    IBDIAGNET_RETURN_VOID;
}

void PDDRLatchedFlagInfoRegister::DumpLatchedFlagInfoHeader(stringstream &sstream) 
{
    IBDIAGNET_ENTER;

    sstream << "RX1LatchedLossIndicator,RX2LatchedLossIndicator,RX3LatchedLossIndicator,"
            << "RX4LatchedLossIndicator,TX1LatchedLossIndicator,TX2LatchedLossIndicator,"
            << "TX3LatchedLossIndicator,TX4LatchedLossIndicator,"
            << "TX1AdaptiveEqualizationFaultIndicator,"
            << "TX2AdaptiveEqualizationFaultIndicator,TX3AdaptiveEqualizationFaultIndicator,"
            << "TX4AdaptiveEqualizationFaultIndicator,RX1CDRLOL,RX2CDRLOL,RX3CDRLOL,RX4CDRLOL,"
            << "TX1CDRLOL,TX2CDRLOL,TX3CDRLOL,TX4CDRLOL,"
            << "HighTemperatureAlarm,LowTemperatureAlarm,"
            << "HighTemperatureWarning,LowTemperatureWarning,"
            << "HighSupplyVoltageAlarm,LowSupplyVoltageAlarm,HighSupplyVoltageWarning,"
            << "LowSupplyVoltageWarning,"
            << "HighRX1PowerAlarm,LowRX1PowerAlarm,HighRX1PowerWarning,LowRX1PowerWarning,"
            << "HighRX2PowerAlarm,LowRX2PowerAlarm,HighRX2PowerWarning,LowRX2PowerWarning,"
            << "HighRX3PowerAlarm,LowRX3PowerAlarm,HighRX3PowerWarning,LowRX3PowerWarning,"
            << "HighRX4PowerAlarm,LowRX4PowerAlarm,HighRX4PowerWarning,LowRX4PowerWarning,"
            << "HighTX1BiasAlarm,LowTX1BiasAlarm,HighTX1BiasWarning,LowTX1BiasWarning,"
            << "HighTX2BiasAlarm,LowTX2BiasAlarm,HighTX2BiasWarning,LowTX2BiasWarning,"
            << "HighTX3BiasAlarm,LowTX3BiasAlarm,HighTX3BiasWarning,LowTX3BiasWarning,"
            << "HighTX4BiasAlarm,LowTX4BiasAlarm,HighTX4BiasWarning,LowTX4BiasWarning,"
            << "HighTX1PowerAlarm,LowTX1PowerAlarm,HighTX1PowerWarning,LowTX1PowerWarning,"
            << "HighTX2PowerAlarm,LowTX2PowerAlarm,HighTX2PowerWarning,LowTX2PowerWarning,"
            << "HighTX3PowerAlarm,LowTX3PowerAlarm,HighTX3PowerWarning,LowTX3PowerWarning,"
            << "HighTX4PowerAlarm,LowTX4PowerAlarm,HighTX4PowerWarning,LowTX4PowerWarning";

    IBDIAGNET_RETURN_VOID;
}

void PDDRLatchedFlagInfoRegister::DumpLatchedFlagInfoData(stringstream &sstream, const struct module_latched_flag_info *p_latched_flag_info)
{
    IBDIAGNET_ENTER;

    if (!p_latched_flag_info) {
        sstream << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A";
        IBDIAGNET_RETURN_VOID;
    }


    sstream << +(p_latched_flag_info->rx_los & 0x1)                                              << ','
            << +((p_latched_flag_info->rx_los >> 1) & 0x1)                                       << ','
            << +((p_latched_flag_info->rx_los >> 2) & 0x1)                                       << ','
            << +((p_latched_flag_info->rx_los >> 3) & 0x1)                                       << ','
            << +(p_latched_flag_info->tx_los & 0x1)                                              << ','
            << +((p_latched_flag_info->tx_los >> 1) & 0x1)                                       << ','
            << +((p_latched_flag_info->tx_los >> 2) & 0x1)                                       << ','
            << +((p_latched_flag_info->tx_los >> 3) & 0x1)                                       << ','
            << +(p_latched_flag_info->tx_ad_eq_fault & 0x1)                                      << ','
            << +((p_latched_flag_info->tx_ad_eq_fault >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_ad_eq_fault >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_ad_eq_fault >> 3) & 0x1)                               << ','
            << +(p_latched_flag_info->rx_cdr_lol & 0x1)                                          << ','
            << +((p_latched_flag_info->rx_cdr_lol >> 1) & 0x1)                                   << ','
            << +((p_latched_flag_info->rx_cdr_lol >> 2) & 0x1)                                   << ','
            << +((p_latched_flag_info->rx_cdr_lol >> 3) & 0x1)                                   << ','
            << +(p_latched_flag_info->tx_cdr_lol & 0x1)                                          << ','
            << +((p_latched_flag_info->tx_cdr_lol >> 1) & 0x1)                                   << ','
            << +((p_latched_flag_info->tx_cdr_lol >> 2) & 0x1)                                   << ','
            << +((p_latched_flag_info->tx_cdr_lol >> 3) & 0x1)                                   << ','
            << +(p_latched_flag_info->temp_flags & 0x1)                                          << ','
            << +((p_latched_flag_info->temp_flags >> 1) & 0x1)                                   << ','
            << +((p_latched_flag_info->temp_flags >> 2) & 0x1)                                   << ','
            << +((p_latched_flag_info->temp_flags >> 3) & 0x1)                                   << ','
            << +(p_latched_flag_info->vcc_flags & 0x1)                                           << ','
            << +((p_latched_flag_info->vcc_flags >> 1) & 0x1)                                    << ','
            << +((p_latched_flag_info->vcc_flags >> 2) & 0x1)                                    << ','
            << +((p_latched_flag_info->vcc_flags >> 3) & 0x1)                                    << ','
            << +(p_latched_flag_info->rx_power_hi_al & 0x1)                                      << ','
            << +(p_latched_flag_info->rx_power_lo_al & 0x1)                                      << ','
            << +(p_latched_flag_info->rx_power_hi_war & 0x1)                                     << ','
            << +(p_latched_flag_info->rx_power_lo_war & 0x1)                                     << ','
            << +((p_latched_flag_info->rx_power_hi_al >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_lo_al >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_hi_war >> 1) & 0x1)                              << ','
            << +((p_latched_flag_info->rx_power_lo_war >> 1) & 0x1)                              << ','
            << +((p_latched_flag_info->rx_power_hi_al >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_lo_al >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_hi_war >> 2) & 0x1)                              << ','
            << +((p_latched_flag_info->rx_power_lo_war >> 2) & 0x1)                              << ','
            << +((p_latched_flag_info->rx_power_hi_al >> 3) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_lo_al >> 3) & 0x1)                               << ','
            << +((p_latched_flag_info->rx_power_hi_war >> 3) & 0x1)                              << ','
            << +((p_latched_flag_info->rx_power_lo_war >> 3) & 0x1)                              << ','
            << +(p_latched_flag_info->tx_bias_hi_al & 0x1)                                       << ','
            << +(p_latched_flag_info->tx_bias_lo_al & 0x1)                                       << ','
            << +(p_latched_flag_info->tx_bias_hi_war & 0x1)                                      << ','
            << +(p_latched_flag_info->tx_bias_lo_war & 0x1)                                      << ','
            << +((p_latched_flag_info->tx_bias_hi_al >> 1) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_lo_al >> 1) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_hi_war >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_bias_lo_war >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_bias_hi_al >> 2) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_lo_al >> 2) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_hi_war >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_bias_lo_war >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_bias_hi_al >> 3) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_lo_al >> 3) & 0x1)                                << ','
            << +((p_latched_flag_info->tx_bias_hi_war >> 3) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_bias_lo_war >> 3) & 0x1)                               << ','
            << +(p_latched_flag_info->tx_power_hi_al & 0x1)                                      << ','
            << +(p_latched_flag_info->tx_power_lo_al & 0x1)                                      << ','
            << +(p_latched_flag_info->tx_power_hi_war & 0x1)                                     << ','
            << +(p_latched_flag_info->tx_power_lo_war & 0x1)                                     << ','
            << +((p_latched_flag_info->tx_power_hi_al >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_lo_al >> 1) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_hi_war >> 1) & 0x1)                              << ','
            << +((p_latched_flag_info->tx_power_lo_war >> 1) & 0x1)                              << ','
            << +((p_latched_flag_info->tx_power_hi_al >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_lo_al >> 2) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_hi_war >> 2) & 0x1)                              << ','
            << +((p_latched_flag_info->tx_power_lo_war >> 2) & 0x1)                              << ','
            << +((p_latched_flag_info->tx_power_hi_al >> 3) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_lo_al >> 3) & 0x1)                               << ','
            << +((p_latched_flag_info->tx_power_hi_war >> 3) & 0x1)                              << ','
            << +((p_latched_flag_info->tx_power_lo_war >> 3) & 0x1);

    IBDIAGNET_RETURN_VOID;
}

void PDDRLatchedFlagInfoRegister::DumpLatchedFlagInfoData(ofstream &sout, const struct module_latched_flag_info *p_latched_flag_info)
{
    IBDIAGNET_ENTER;

    if (!p_latched_flag_info) {
         sout << "CDR Latched RX Loss Indicator: N/A" << endl
              << "CDR Latched TX Loss Indicator: N/A" << endl
              << "Latched Adaptive Equalization Fault: N/A" << endl
              << "Latched  RX LOL Indicator: N/A" << endl
              << "Latched  TX LOL Indicator: N/A" << endl
              << "Latched Temperature Alarm and Warning: N/A" << endl
              << "Latched Voltage Alarm and Warning: N/A" << endl
              << "RX Power High Alarm: N/A" << endl
              << "RX Power Low Alarm: N/A" << endl
              << "RX Power High Warning: N/A" << endl
              << "RX Power Low Warning: N/A" << endl
              << "TX Bias High Alarm: N/A" << endl
              << "TX Bias Low Alarm: N/A" << endl
              << "TX Bias High Warning: N/A" << endl
              << "TX Bias Low Warning: N/A" << endl
              << "TX Power High Alarm: N/A" << endl
              << "TX Power Low Alarm: N/A" << endl
              << "TX Power High Warning: N/A" << endl
              << "TX Power Low Warning: N/A";

         IBDIAGNET_RETURN_VOID;
    }

    sout << "CDR Latched RX Loss Indicator: " << +(p_latched_flag_info->rx_los & 0xf) << endl
         << "CDR Latched TX Loss Indicator: " << +(p_latched_flag_info->tx_los & 0xf) << endl
         << "Latched Adaptive Equalization Fault: "
             << +(p_latched_flag_info->tx_ad_eq_fault & 0xf) << endl
         << "Latched  RX LOL Indicator: " << +(p_latched_flag_info->rx_cdr_lol & 0xf) << endl
         << "Latched  TX LOL Indicator: " << +(p_latched_flag_info->tx_cdr_lol & 0xf) << endl
         << "Latched Temperature Alarm and Warning: "
             << +(p_latched_flag_info->temp_flags & 0xf) << endl
         << "Latched Voltage Alarm and Warning: " << +(p_latched_flag_info->vcc_flags & 0xf) << endl
         << "RX Power High Alarm: " << +(p_latched_flag_info->rx_power_hi_al & 0xf) << endl
         << "RX Power Low Alarm: " << +(p_latched_flag_info->rx_power_lo_al & 0xf) << endl
         << "RX Power High Warning: " << +(p_latched_flag_info->rx_power_hi_war & 0xf) << endl
         << "RX Power Low Warning: " << +(p_latched_flag_info->rx_power_lo_war & 0xf) << endl
         << "TX Bias High Alarm: " << +(p_latched_flag_info->tx_bias_hi_al & 0xf) << endl
         << "TX Bias Low Alarm: " << +(p_latched_flag_info->tx_bias_lo_al & 0xf) << endl
         << "TX Bias High Warning: " << +(p_latched_flag_info->tx_bias_hi_war & 0xf) << endl
         << "TX Bias Low Warning: " << +(p_latched_flag_info->tx_bias_lo_war & 0xf) << endl
         << "TX Power High Alarm: " << +(p_latched_flag_info->tx_power_hi_al & 0xf) << endl
         << "TX Power Low Alarm: " << +(p_latched_flag_info->tx_power_lo_al & 0xf) << endl
         << "TX Power High Warning: " << +(p_latched_flag_info->tx_power_hi_war & 0xf) << endl
         << "TX Power Low Warning: " << +(p_latched_flag_info->tx_power_lo_war & 0xf);

    IBDIAGNET_RETURN_VOID;
}

PhyCableRecord::LatchedRecord
    *PDDRLatchedFlagInfoRegister::CreateRecord(const struct module_latched_flag_info *p_latched_flag_info)
{
    IBDIAGNET_ENTER;

    if (!p_latched_flag_info)
        IBDIAGNET_RETURN(NULL);

    PhyCableRecord::LatchedRecord *p_latched = new PhyCableRecord::LatchedRecord();
    PDDRLatchedFlagInfoRegister::CopyTo(*p_latched_flag_info, *p_latched);

    IBDIAGNET_RETURN(p_latched);
}

#ifdef COPY
    #error "The COPY macros has been already defined"
#else
    #define COPY(X) to.X = from.X
#endif

void PDDRLatchedFlagInfoRegister::CopyTo(const module_latched_flag_info &from , PhyCableRecord::LatchedRecord &to)
{
    IBDIAGNET_ENTER;

    COPY(dp_fw_fault);
    COPY(mod_fw_fault);
    COPY(vcc_flags);
    COPY(temp_flags);
    COPY(tx_ad_eq_fault);
    COPY(tx_cdr_lol);
    COPY(tx_los);
    COPY(tx_fault);
    COPY(tx_power_lo_war);
    COPY(tx_power_hi_war);
    COPY(tx_power_lo_al);
    COPY(tx_power_hi_al);
    COPY(tx_bias_lo_war);
    COPY(tx_bias_hi_war);
    COPY(tx_bias_lo_al);
    COPY(tx_bias_hi_al);
    COPY(rx_cdr_lol);
    COPY(rx_los);
    COPY(rx_power_lo_war);
    COPY(rx_power_hi_war);
    COPY(rx_power_lo_al);
    COPY(rx_power_hi_al);
    COPY(rx_output_valid_change);
    COPY(flag_in_use);

    IBDIAGNET_RETURN_VOID;
}

#undef COPY

PDDRPhyInfoRegister::PDDRPhyInfoRegister(PhyDiag * phy_diag)
    :PDDRRegister(phy_diag, 
                  0x02, 
                  (const unpack_data_func_t)pddr_phy_info_page_unpack, 
                  ACC_REG_PDDR_PHY_INFO_INTERNAL_SECTION_NAME, 
                  ACC_REG_PDDR_PHY_INFO_NAME, 
                  ACC_REG_PDDR_PHY_INFO_FIELDS_NUM,
                  ACC_REG_PDDR_PHY_INFO_HEADER, 
                  SUPPORT_SW_CA)
{

}
void PDDRPhyInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
    sstream << std::endl;
    DumpPhyInfoRealHeader(sstream, handler_header);
    IBDIAGNET_RETURN_VOID;
}

void PDDRPhyInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;
    DumpLayout(sstream, areg.regs.pddr_phyinfo);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PDDRPhyInfoRegister::DumpPhyInfoRealHeader(stringstream &sstream, const string& header)
{
#ifdef SHOW_PHY_HEADERS
    sstream << '#' << header                               << ','
            << "remote_device_type"                        << ','
            << "negotiation_mask"                          << ','
            << "port_notifications"                        << ','
            << "num_of_negotiation_attempts"               << ','
            << "ib_revision"                               << ','
            << "lp_ib_revision"                            << ','
            << "hw_link_phy_state"                         << ','
            << "phy_manger_disable_mask"                   << ','
            << "pcs_phy_state"                             << ','
            << "lp_proto_enabled"                          << ','
            << "lp_fec_mode_support"                       << ','
            << "lp_fec_mode_request"                       << ','
            << "ib_last_link_down_reason"                  << ','
            << "eth_last_link_down_lane"                   << ','
            << "speed_deg_db"                              << ','
            << "degrade_grade_lane_0"                      << ','
            << "degrade_grade_lane_1"                      << ','
            << "degrade_grade_lane_2"                      << ','
            << "degrade_grade_lane_3"                      << ','
            << "kr_startup_debug_indications_4"            << ','
            << "kr_startup_debug_indications_5"            << ','
            << "kr_startup_debug_indications_6"            << ','
            << "kr_startup_debug_indications_7"            << ','
            << "num_of_presets_tested_lane0"               << ','
            << "num_of_presets_tested_lane1"               << ','
            << "num_of_presets_tested_lane2"               << ','
            << "num_of_presets_tested_lane3"               << ','
            << "num_of_presets_tested_lane_4"              << ','
            << "num_of_presets_tested_lane_5"              << ','
            << "num_of_presets_tested_lane_6"              << ','
            << "num_of_presets_tested_lane_7"              << ','
            << "kr_startup_fsm_lane"                       << ','
            << "eth_an_debug_indication"                   << ','
            << "cdr_not_locked_cnt"                        << ','
            << "fall_from_cfg_idle_cnt"                    << ','
            << "fall_from_cfg_idle_cdr_cnt"                << ','
            << "eth_an_watchdog_cnt"                       << ','
            << "rounds_waited_for_peer_to_end_test"        << ','
            << "ib_phy_fsm_state_trace"                    << ','
            << "kr_startup_debug_indications_0"            << ','
            << "kr_startup_debug_indications_1"            << ','
            << "kr_startup_debug_indications_2"            << ','
            << "kr_startup_debug_indications_3"            << ','
            << "tx_tuning_stages_lane_0"                   << ','
            << "tx_tuning_stages_lane_1"                   << ','
            << "tx_tuning_stages_lane_2"                   << ','
            << "tx_tuning_stages_lane_3"                   << ','
            << "tx_tuning_stages_lane_4"                   << ','
            << "tx_tuning_stages_lane_5"                   << ','
            << "tx_tuning_stages_lane_6"                   << ','
            << "tx_tuning_stages_lane_7"                   << ','
            << "plu_rx_polarity"                           << ','
            << "plu_tx_polarity"                           << ','
            << "plu_rx_pwrup"                              << ','
            << "plu_tx_pwrup"                              << ','
            << "signal_detected"                           << ','
            << "plr_outstanding_cells"                     << ','
            << "plu_rx_speed"                              << ','
            << "plu_tx_speed"                              << ','
            << "sd_valid"                                  << ','
            << "ib_cfg_delay_timeout"                      << ','
            << "irisc_status"                              << ','
            << "stamping_reason"                           << ','
            << "kr_frame_lock_tuning_failure_events_count" << ','
            << "kr_full_tuning_failure_count"              << ','
            << "ib_debug_indication"                       << ','
            << "pm_debug_indication"                       << ','
            << "pd_cat_val"                                << ','
            << "pd_catastrophic_enum"                      << ','
            << "hst_cat_val"                               << ','
            << "hst_catastrophic_enum"                     << ','
            << "an_cat_val"                                << ','
            << "an_catastrophic_enum"                      << ','
            << "pm_cat_val"                                << ','
            << "pm_catastrophic_enum"                      << ','
            << "pd_debug_indication"                       << ','
            << "psi_collision1"                            << ','
            << "psi_collision2"                            << ','
            << "hst_mismatch_reason"                       << ','
            << "fp_signal_detect_count"                    << ','
            << "pd_count"                                  << ','
            << "nlpn_debug_ind_mask"                       << ','
            << "phy2mod_speed_req"                         << ','
            << "eeprom_prsnt"                              << ','
            << "no_dme_mod"                                << ','
            << "one_pll_mod"                               << ','
            << "phy2mod_ack_lanes"                         << ','
            << "phy2mod_deactivate_lanes"                  << ','
            << "port_xmit_pkts_inc_s1"                     << ','
            << "port_xmit_pkts_inc_s0"                     << ','
            << "cnt_rx_frame_received_ok_s1"               << ','
            << "cnt_rx_frame_received_ok_s0"               << ','
            << "macsec_en_plt1_s1"                         << ','
            << "macsec_en_plt1_s0"                         << ','
            << "macsec_en_plt0_s1"                         << ','
            << "macsec_en_plt0_s0"                         << ','
            << "reconciliation_mux_plt1"                   << ','
            << "reconciliation_mux_plt0"                   << ','
            << "tx_bypass_mux_plt1"                        << ','
            << "tx_bypass_mux_plt0"                        << ','
            << "rx_bypass_mux_plt1"                        << ','
            << "rx_bypass_mux_plt0"                        << ','
            << "plr_rtt_hdr_threshold"                     << ','
            << "plr_rtt_ndr_threshold"                     << ','
            << "plr_rtt_xdr_threshold";
#endif
}

void PDDRPhyInfoRegister::DumpLayout(stringstream &sstream, const struct pddr_phy_info_page &phy_info)
{
    IBDIAGNET_ENTER;
    #define ARRAY_TO_32_BIT(arr) (u_int32_t)(arr[0] << 24 | (arr[1] << 16) | (arr[2] << 8) | (arr[3]))
    #define ARRAY_TO_64_BIT(arr) (((uint64_t)(((uint8_t *)(arr))[7]) <<  0)+\
                    ((uint64_t)(((uint8_t *)(arr))[6]) <<  8)+\
                    ((uint64_t)(((uint8_t *)(arr))[5]) << 16)+\
                    ((uint64_t)(((uint8_t *)(arr))[4]) << 24)+\
                    ((uint64_t)(((uint8_t *)(arr))[3]) << 32)+\
                    ((uint64_t)(((uint8_t *)(arr))[2]) << 40)+\
                    ((uint64_t)(((uint8_t *)(arr))[1]) << 48)+\
                    ((uint64_t)(((uint8_t *)(arr))[0]) << 56))
    sstream << +phy_info.remote_device_type << ','
            << +phy_info.negotiation_mask << ','
            << +phy_info.port_notifications << ','
            << phy_info.num_of_negotiation_attempts << ','
            << +phy_info.ib_revision << ','
            << +phy_info.lp_ib_revision << ','
            << +phy_info.hw_link_phy_state << ','
            << phy_info.phy_manager_disable_mask << ','
            << phy_info.pcs_phy_state << ','
            << phy_info.lp_proto_enabled << ','
            << phy_info.lp_fec_mode_support << ','
            << phy_info.lp_fec_mode_request << ','
            << phy_info.ib_last_link_down_reason << ','
            << PTR(ARRAY_TO_32_BIT(phy_info.eth_last_link_down_lane)) << ','
            << phy_info.speed_deg_db << ','
            << phy_info.degrade_grade_lane_0 << ','
            << phy_info.degrade_grade_lane_1 << ','
            << phy_info.degrade_grade_lane_2 << ','
            << phy_info.degrade_grade_lane_3 << ','
            << +phy_info.num_of_presets_tested_lane_4 << ','
            << +phy_info.num_of_presets_tested_lane_5 << ','
            << +phy_info.num_of_presets_tested_lane_6 << ','
            << +phy_info.num_of_presets_tested_lane_7 << ','
            << +phy_info.kr_startup_debug_indications_4 << ','
            << +phy_info.kr_startup_debug_indications_5 << ','
            << +phy_info.kr_startup_debug_indications_6 << ','
            << +phy_info.kr_startup_debug_indications_7 << ','
            << +phy_info.num_of_presets_tested_lane0 << ','
            << +phy_info.num_of_presets_tested_lane1 << ','
            << +phy_info.num_of_presets_tested_lane2 << ','
            << +phy_info.num_of_presets_tested_lane3 << ','
            << PTR(ARRAY_TO_64_BIT(phy_info.kr_startup_fsm_lane)) << ','
            << +phy_info.eth_an_debug_indication << ','
            << +phy_info.cdr_not_locked_cnt << ','
            << +phy_info.fall_from_cfg_idle_cnt << ','
            << +phy_info.fall_from_cfg_idle_cdr_cnt << ','
            << +phy_info.eth_an_watchdog_cnt << ','
            << +phy_info.rounds_waited_for_peer_to_end_test << ','
            << phy_info.ib_phy_fsm_state_trace << ','
            << +phy_info.kr_startup_debug_indications_0 << ','
            << +phy_info.kr_startup_debug_indications_1 << ','
            << +phy_info.kr_startup_debug_indications_2 << ','
            << +phy_info.kr_startup_debug_indications_3 << ','
            << +phy_info.tx_tuning_stages_lane[0] << ','
            << +phy_info.tx_tuning_stages_lane[1] << ','
            << +phy_info.tx_tuning_stages_lane[2] << ','
            << +phy_info.tx_tuning_stages_lane[3] << ','
            << +phy_info.tx_tuning_stages_lane[4] << ','
            << +phy_info.tx_tuning_stages_lane[5] << ','
            << +phy_info.tx_tuning_stages_lane[6] << ','
            << +phy_info.tx_tuning_stages_lane[7] << ','
            << +phy_info.plu_rx_polarity << ','
            << +phy_info.plu_tx_polarity << ','
            << +phy_info.plu_rx_pwrup << ','
            << +phy_info.plu_tx_pwrup << ','
            << +phy_info.signal_detected << ','
            << +phy_info.plr_outstanding_cells << ','
            << +phy_info.plu_rx_speed << ','
            << +phy_info.plu_tx_speed << ','
            << +phy_info.sd_valid << ','
            << +phy_info.ib_cfg_delay_timeout << ','
            << +phy_info.irisc_status << ','
            << phy_info.stamping_reason << ','
            << +phy_info.kr_frame_lock_tuning_failure_events_count << ','
            << +phy_info.kr_full_tuning_failure_count << ','
            << +phy_info.ib_debug_indication << ','
            << +phy_info.pm_debug_indication << ','
            << +phy_info.pd_cat_val << ','
            << +phy_info.pd_catastrophic_enum << ','
            << +phy_info.hst_cat_val << ','
            << +phy_info.hst_catastrophic_enum << ','
            << +phy_info.an_cat_val << ','
            << +phy_info.an_catastrophic_enum << ','
            << +phy_info.pm_cat_val << ','
            << +phy_info.pm_catastrophic_enum << ','
            << +phy_info.pd_debug_indication << ','
            << +phy_info.psi_collision1 << ','
            << +phy_info.psi_collision2 << ','
            << +phy_info.hst_mismatch_reason << ','
            << +phy_info.fp_signal_detect_count << ','
            << +phy_info.pd_count << ','
            << +phy_info.nlpn_debug_ind_mask << ','
            << +phy_info.phy2mod_speed_req << ','
            << +phy_info.eeprom_prsnt << ','
            << +phy_info.no_dme_mod << ','
            << +phy_info.one_pll_mod << ','
            << +phy_info.phy2mod_ack_lanes << ','
            << +phy_info.phy2mod_deactivate_lanes << ','
            << +phy_info.port_xmit_pkts_inc_s1 << ','
            << +phy_info.port_xmit_pkts_inc_s0 << ','
            << +phy_info.cnt_rx_frame_received_ok_s1 << ','
            << +phy_info.cnt_rx_frame_received_ok_s0 << ','
            << +phy_info.macsec_en_plt1_s1 << ','
            << +phy_info.macsec_en_plt1_s0 << ','
            << +phy_info.macsec_en_plt0_s1 << ','
            << +phy_info.macsec_en_plt0_s0 << ','
            << +phy_info.reconciliation_mux_plt1 << ','
            << +phy_info.reconciliation_mux_plt0 << ','
            << +phy_info.tx_bypass_mux_plt1 << ','
            << +phy_info.tx_bypass_mux_plt0 << ','
            << +phy_info.rx_bypass_mux_plt1 << ','
            << +phy_info.rx_bypass_mux_plt0 << ','
            << +phy_info.plr_rtt_hdr_threshold << ','
            << +phy_info.plr_rtt_ndr_threshold << ','
            << +phy_info.plr_rtt_xdr_threshold;
    #undef ARRAY_TO_32_BIT
    #undef ARRAY_TO_64_BIT
    IBDIAGNET_RETURN_VOID;
}

// PDDRTroubleshootingInfoRegister
PDDRTroubleshootingInfoRegister::PDDRTroubleshootingInfoRegister(PhyDiag* phy_diag)
    :PDDRRegister(phy_diag,
                  ACC_REG_PDDR_TROUBLESHOOTING_INFO_PAGE_NUM,
                  (const unpack_data_func_t)pddr_troubleshooting_page_unpack,
                  ACC_REG_PDDR_TROUBLESHOOTING_INFO_INTERNAL_SECTION_NAME,
                  ACC_REG_PDDR_TROUBLESHOOTING_INFO_NAME,
                  ACC_REG_PDDR_TROUBLESHOOTING_INFO_FIELDS_NUM,
                  ACC_REG_PDDR_TROUBLESHOOTING_INFO_HEADER,
                  SUPPORT_SW_CA)
{

}

void PDDRTroubleshootingInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
    sstream << std::endl;
    DumpRealHeader(sstream, handler_header);

    IBDIAGNET_RETURN_VOID;
}

void PDDRTroubleshootingInfoRegister::DumpRealHeader(stringstream &sstream, const string& header)
{
#ifdef SHOW_PHY_HEADERS
    sstream << '#' << header          << ','
            << "status_opcode"        << ','
            << "status_message"       << ','
            << "group_opcode"         << ','
            << "user_feedback_data"   << ','
            << "user_feedback_index";
#endif
}

void PDDRTroubleshootingInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    DumpLayout(sstream, areg.regs.pddr_troubleshooting);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void PDDRTroubleshootingInfoRegister::DumpLayout(stringstream &sstream, const pddr_troubleshooting_page &troubleshooting_info)
{
    IBDIAGNET_ENTER;

    u_int16_t status_opcode = troubleshooting_info.group_opcode == 0 ?
                              troubleshooting_info.status_opcode.pddr_monitor_opcode.monitor_opcode :
                              troubleshooting_info.status_opcode.pddr_advanced_opcode.advanced_opcode;

    size_t max_arr_size = ARRAY_SIZE(troubleshooting_info.status_message);
    char fixed_status_message[max_arr_size];
    memset(fixed_status_message, 0, max_arr_size);
    strncpy(fixed_status_message, troubleshooting_info.status_message, max_arr_size);

    sstream << status_opcode << ','
            << QUOTED(std::string(fixed_status_message)) << ','
            << +troubleshooting_info.group_opcode << ','
            << +troubleshooting_info.user_feedback_data << ','
            << +troubleshooting_info.user_feedback_index;

    IBDIAGNET_RETURN_VOID;
}

// PDDROperationInfoRegister
PDDROperationInfoRegister::PDDROperationInfoRegister(PhyDiag* phy_diag)
    :PDDRRegister(phy_diag,
                  ACC_REG_PDDR_OPERATION_INFO_PAGE_NUM,
                  (const unpack_data_func_t)pddr_operation_info_page_unpack,
                  ACC_REG_PDDR_OPERATION_INFO_INTERNAL_SECTION_NAME,
                  ACC_REG_PDDR_OPERATION_INFO_NAME,
                  ACC_REG_PDDR_OPERATION_INFO_FIELDS_NUM,
                  ACC_REG_PDDR_OPERATION_INFO_HEADER,
                  SUPPORT_SW_CA)
{

}

void PDDROperationInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
    sstream << std::endl;
    DumpRealHeader(sstream, handler_header);

    IBDIAGNET_RETURN_VOID;
}

void PDDROperationInfoRegister::DumpRealHeader(stringstream &sstream, const string& header)
{
#ifdef SHOW_PHY_HEADERS
    sstream << '#' << header                          << ','
            << "proto_active"                         << ','
            << "neg_mode_active"                      << ','
            << "pd_fsm_state"                         << ','
            << "phy_mngr_fsm_state"                   << ','
            << "eth_an_fsm_state"                     << ','
            << "ib_phy_fsm_state"                     << ','
            << "phy_hst_fsm_state"                    << ','
            << "phy_manager_link_proto_enabled"       << ','
            << "phy_manager_link_width_enabled"       << ','
            << "core_to_phy_link_proto_enabled"       << ','
            << "core_to_phy_link_width_enabled"       << ','
            << "cable_link_speed_cap"                 << ','
            << "cable_link_width_cap"                 << ','
            << "link_speed_active"                    << ','
            << "link_width_active"                    << ','
            << "retran_mode_active"                   << ','
            << "retran_mode_request"                  << ','
            << "loopback_mode"                        << ','
            << "fec_mode_active"                      << ','
            << "fec_mode_request"                     << ','
            << "profile_fec_in_use"                   << ','
            << "nlpn_fsm_state"                       << ','
            << "eth_25g_50g_fec_support"              << ','
            << "eth_100g_fec_support"                 << ','
            << "pd_link_enabled"                      << ','
            << "phy_hst_link_enabled"                 << ','
            << "eth_an_link_enabled"                  << ','
            << "linkup_retry_iterations"              << ','
            << "core_to_phy_state"                    << ','
            << "psi_fsm_state"                        << ','
            << "cable_proto_cap_ext"                  ;
#endif
}

void PDDROperationInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    DumpLayout(sstream, areg.regs.pddr_operation_info);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void PDDROperationInfoRegister::DumpLayout(stringstream &sstream, const pddr_operation_info_page &operation_info)
{
    IBDIAGNET_ENTER;

    #define PHY_MANAGER(field) operation_info.phy_manager_link_enabled.pddr_phy_manager_link_enabed_ib.field
    #define CORE_TO_PHY_ENABLED(field) operation_info.core_to_phy_link_enabled.pddr_c2p_link_enabed_ib.field
    #define CABLE_PROTO_CAP(field) operation_info.cable_proto_cap.pddr_cable_cap_ib.field
    #define LINK_SPEED_ACTIVE(field) operation_info.link_active.pddr_link_active_ib.field

    sstream << +operation_info.proto_active                             << ','
            << +operation_info.neg_mode_active                          << ','
            << +operation_info.pd_fsm_state                             << ','
            << +operation_info.phy_mngr_fsm_state                       << ','
            << +operation_info.eth_an_fsm_state                         << ','
            << +operation_info.ib_phy_fsm_state                         << ','
            << +operation_info.phy_hst_fsm_state                        << ','
            << PTR(PHY_MANAGER(phy_manager_link_proto_enabled))         << ','
            << PTR(PHY_MANAGER(phy_manager_link_width_enabled))         << ','
            << PTR(CORE_TO_PHY_ENABLED(core_to_phy_link_proto_enabled)) << ','
            << PTR(CORE_TO_PHY_ENABLED(core_to_phy_link_width_enabled)) << ','
            << +CABLE_PROTO_CAP(cable_link_speed_cap)                   << ','
            << +CABLE_PROTO_CAP(cable_link_width_cap)                   << ','
            << PTR(LINK_SPEED_ACTIVE(link_speed_active))                << ','
            << PTR(LINK_SPEED_ACTIVE(link_width_active))                << ','
            << +operation_info.retran_mode_active                       << ','
            << +operation_info.retran_mode_request                      << ','
            << +operation_info.loopback_mode                            << ','
            << +operation_info.fec_mode_active                          << ','
            << +operation_info.fec_mode_request                         << ','
            << +operation_info.profile_fec_in_use                       << ','
            << +operation_info.nlpn_fsm_state                           << ','
            << +operation_info.eth_25g_50g_fec_support                  << ','
            << +operation_info.eth_100g_fec_support                     << ','
            << +operation_info.pd_link_enabled                          << ','
            << +operation_info.phy_hst_link_enabled                     << ','
            << +operation_info.eth_an_link_enabled                      << ','
            << +operation_info.linkup_retry_iterations                  << ','
            << +operation_info.core_to_phy_state                        << ','
            << +operation_info.psi_fsm_state                            << ','
            << +operation_info.cable_proto_cap_ext                      ;

    #undef PHY_MANAGER
    #undef CORE_TO_PHY_ENABLED
    #undef CABLE_PROTO_CAP
    #undef LINK_SPEED_ACTIVE

    IBDIAGNET_RETURN_VOID;
}

// PDDRLinkDownInfoRegister
PDDRLinkDownInfoRegister::PDDRLinkDownInfoRegister(PhyDiag* phy_diag)
    :PDDRRegister(phy_diag,
                  ACC_REG_PDDR_LINK_DOWN_INFO_PAGE_NUM,
                  (const unpack_data_func_t)pddr_link_down_info_page_unpack,
                  ACC_REG_PDDR_LINK_DOWN_INFO_INTERNAL_SECTION_NAME,
                  ACC_REG_PDDR_LINK_DOWN_INFO_NAME,
                  ACC_REG_PDDR_LINK_DOWN_INFO_FIELDS_NUM,
                  ACC_REG_PDDR_LINK_DOWN_INFO_HEADER,
                  SUPPORT_SW_CA)
{

}

void PDDRLinkDownInfoRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
    sstream << std::endl;
    DumpRealHeader(sstream, handler_header);

    IBDIAGNET_RETURN_VOID;
}

void PDDRLinkDownInfoRegister::DumpRealHeader(stringstream &sstream, const string& header)
{
#ifdef SHOW_PHY_HEADERS
    sstream << '#' << header                    << ','
            << "down_blame"                     << ','
            << "local_reason_opcode"            << ','
            << "remote_reason_opcode"           << ','
            << "e2e_reason_opcode"              << ','
            << "cons_raw_norm_ber"              << ','
            << "last_raw_ber_coef"              << ','
            << "last_raw_ber_magnitude"         << ','
            << "num_of_symbol_ber_alarms"       << ','
            << "min_raw_ber_coef"               << ','
            << "min_raw_ber_magnitude"          << ','
            << "max_raw_ber_coef"               << ','
            << "max_raw_ber_magnitude"          << ','
            << "num_of_eff_ber_alarms"          << ','
            << "num_of_raw_ber_alarms"          << ','
            << "time_to_link_down_to_disable"   << ','
            << "time_to_link_down_to_rx_loss"   << ','
            << "cons_eff_norm_ber"              << ','
            << "last_eff_ber_coef"              << ','
            << "last_eff_ber_magnitude"         << ','
            << "min_eff_ber_coef"               << ','
            << "min_eff_ber_magnitude"          << ','
            << "cons_symbol_norm_ber"           << ','
            << "max_symbol_ber_coef"            << ','
            << "max_symbol_ber_magnitude"       << ','
            << "max_eff_ber_coef"               << ','
            << "max_eff_ber_magnitude"          << ','
            << "last_symbol_ber_coef"           << ','
            << "last_symbol_ber_magnitude"      << ','
            << "min_symbol_ber_coef"            << ','
            << "min_symbol_ber_magnitude"       << ','
            << "mean_est_eff_ber_coef"          << ','
            << "mean_est_eff_ber_magnitude"     << ','
            << "min_est_eff_ber_coef"           << ','
            << "min_est_eff_ber_magnitude"      << ','
            << "max_est_eff_ber_coef"           << ','
            << "max_est_eff_ber_magnitude"      << ','
            << "num_of_est_eff_ber_warnings"    << ','
            << "num_of_est_eff_ber_alarms"      ;
#endif
}

void PDDRLinkDownInfoRegister::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    DumpLayout(sstream, areg.regs.pddr_link_down_info);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void PDDRLinkDownInfoRegister::DumpLayout(stringstream &sstream, const pddr_link_down_info_page &link_down_info)
{
    IBDIAGNET_ENTER;

    sstream << +link_down_info.down_blame << ','
            << +link_down_info.local_reason_opcode << ','
            << +link_down_info.remote_reason_opcode << ','
            << +link_down_info.e2e_reason_opcode << ','
            << +link_down_info.cons_raw_norm_ber      << ','
            << +link_down_info.last_raw_ber_coef    << ','
            << +link_down_info.last_raw_ber_magnitude << ','
            << +link_down_info.num_of_symbol_ber_alarms << ','
            << +link_down_info.min_raw_ber_coef << ','
            << +link_down_info.min_raw_ber_magnitude << ','
            << +link_down_info.max_raw_ber_coef << ','
            << +link_down_info.max_raw_ber_magnitude << ','
            << +link_down_info.num_of_eff_ber_alarms << ','
            << +link_down_info.num_of_raw_ber_alarms << ','
            << +link_down_info.time_to_link_down_to_disable << ','
            << +link_down_info.time_to_link_down_to_rx_loss << ','
            << +link_down_info.cons_eff_norm_ber << ','
            << +link_down_info.last_eff_ber_coef << ','
            << +link_down_info.last_eff_ber_magnitude << ','
            << +link_down_info.min_eff_ber_coef << ','
            << +link_down_info.min_eff_ber_magnitude << ','
            << +link_down_info.cons_symbol_norm_ber << ','
            << +link_down_info.max_symbol_ber_coef << ','
            << +link_down_info.max_symbol_ber_magnitude << ','
            << +link_down_info.max_eff_ber_coef << ','
            << +link_down_info.max_eff_ber_magnitude << ','
            << +link_down_info.last_symbol_ber_coef << ','
            << +link_down_info.last_symbol_ber_magnitude << ','
            << +link_down_info.min_symbol_ber_coef << ','
            << +link_down_info.min_symbol_ber_magnitude << ','
            << +link_down_info.mean_est_eff_ber_coef << ','
            << +link_down_info.mean_est_eff_ber_magnitude << ','
            << +link_down_info.min_est_eff_ber_coef << ','
            << +link_down_info.min_est_eff_ber_magnitude << ','
            << +link_down_info.max_est_eff_ber_coef << ','
            << +link_down_info.max_est_eff_ber_magnitude << ','
            << +link_down_info.num_of_est_eff_ber_warnings << ','
            << +link_down_info.num_of_est_eff_ber_alarms;

    IBDIAGNET_RETURN_VOID;
}


int PPCNTRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer) {
    ppcnt_reg *reg = (ppcnt_reg *) data_to_unpack;
    ppcnt_reg_unpack(reg, unpacked_buffer);

    ppcnt_RAW_page_data raw = reg->counter_set.raw;
    unpack_data_func(&reg->counter_set, raw.data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PPCNTRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct ppcnt_reg ppcnt;
    CLEAR_STRUCT(ppcnt);

    ppcnt.grp           = this->m_grp;
    ppcnt.lp_gl         = this->m_lp_gl;
    ppcnt.swid          = this->m_swid;
    ppcnt.clr           = this->m_phy_diag->ShouldResetCounters();
    ppcnt.prio_tc       = this->m_prio_tc;
    ppcnt.pnat          = this->m_pnat;
    ppcnt.lp_msb        = this->m_lp_msb;
    ppcnt.port_type     = this->m_port_type;

    int rc = FillLocalPortAndPlaneInd(p_key, &ppcnt.local_port, &ppcnt.plane_ind);

    if (rc)
        return rc;

    ppcnt_reg_pack(&ppcnt, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

PPCNT_PLR_Counters::PPCNT_PLR_Counters(PhyDiag * phy_diag)
    : PPCNTRegister(phy_diag,
                    0x22,
                    (const unpack_data_func_t)ppcnt_plr_counters_unpack,
                    ACC_REG_PPCNT_PLR_COUNTERS_INTERNAL_SECTION_NAME,
                    ACC_REG_PPCNT_PLR_COUNTERS_NAME,
                    ACC_REG_PPCNT_PLR_COUNTERS_FIELDS_NUM,
                    ACC_REG_PPCNT_PLR_COUNTERS_HEADER,
                    SUPPORT_SW_CA) {

}

void PPCNT_PLR_Counters::DumpFieldNames(stringstream &sstream) {
    sstream << "plr_rcv_codes_high"                          <<','
            << "plr_rcv_codes_low"                           <<','
            << "plr_rcv_code_err_high"                       <<','
            << "plr_rcv_code_err_low"                        <<','
            << "plr_rcv_uncorrectable_code_high"             <<','
            << "plr_rcv_uncorrectable_code_low"              <<','
            << "plr_xmit_codes_high"                         <<','
            << "plr_xmit_codes_low"                          <<','
            << "plr_xmit_retry_codes_high"                   <<','
            << "plr_xmit_retry_codes_low"                    <<','
            << "plr_xmit_retry_events_high"                  <<','
            << "plr_xmit_retry_events_low"                   <<','
            << "plr_sync_events_high"                        <<','
            << "plr_sync_events_low"                         <<','
            << "plr_codes_loss_high"                         <<','
            << "plr_codes_loss_low"                          <<','
            << "plr_xmit_retry_events_within_t_sec_max_high" <<','
            << "plr_xmit_retry_events_within_t_sec_max_low"  <<','
            << "time_since_last_clear_high"                  <<','
            << "time_since_last_clear_low";                   
}

void PPCNT_PLR_Counters::DumpData(const ppcnt_plr_counters& reg, stringstream &sstream, bool has_cap_max_ret_rate) {
    sstream << reg.plr_rcv_codes_high <<','
            << reg.plr_rcv_codes_low <<','
            << reg.plr_rcv_code_err_high <<','
            << reg.plr_rcv_code_err_low <<','
            << reg.plr_rcv_uncorrectable_code_high <<','
            << reg.plr_rcv_uncorrectable_code_low <<','
            << reg.plr_xmit_codes_high <<','
            << reg.plr_xmit_codes_low <<','
            << reg.plr_xmit_retry_codes_high <<','
            << reg.plr_xmit_retry_codes_low <<','
            << reg.plr_xmit_retry_events_high <<','
            << reg.plr_xmit_retry_events_low <<','
            << reg.plr_sync_events_high <<','
            << reg.plr_sync_events_low <<','
            << reg.plr_codes_loss_high <<','
            << reg.plr_codes_loss_low <<',';
            if (has_cap_max_ret_rate) {
                sstream << reg.plr_xmit_retry_events_within_t_sec_max_high <<','
                        << reg.plr_xmit_retry_events_within_t_sec_max_low <<',';
            }
            else {
                sstream <<"N/A,N/A,";
            }
            sstream << reg.time_since_last_clear_high <<','
                    << reg.time_since_last_clear_low;
}

void PPCNT_PLR_Counters::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_PLR_Counters::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    bool has_cap = false;
    const AccRegKeyPort *p_key = dynamic_cast<const AccRegKeyPort*>(&key);
    if (p_key) {
        IBNode* p_node =  this->m_phy_diag->GetNode(p_key->node_guid);
        if (p_node)
            has_cap = this->m_phy_diag->GetCapabilityModule()->
                        IsSupportedGMPCapability(p_node,
                                                EnGMPCAPIsPLRMaxRetransmissionRateSupported);
    }
    DumpData(areg.regs.ppcnt.counter_set.ppcnt_plr_counters, sstream, has_cap);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_PLR_Counters::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppcnt_plr_counters =
                                &areg.regs.ppcnt.counter_set.ppcnt_plr_counters;

    IBDIAGNET_RETURN_VOID;

}

PPCNT_Packet_Type_Counters::PPCNT_Packet_Type_Counters(PhyDiag * phy_diag) 
    : PPCNTRegister(phy_diag, 
                    0x29, 
                    (const unpack_data_func_t)ppcnt_packet_type_counter_unpack,
                    ACC_REG_PPCNT_PACKET_TYPE_COUNTERS_INTERNAL_SECTION_NAME, 
                    ACC_REG_PPCNT_PACKET_TYPE_COUNTERS_NAME, 
                    ACC_REG_PPCNT_PACKET_TYPE_COUNTERS_FIELDS_NUM, 
                    ACC_REG_PPCNT_PACKET_TYPE_COUNTERS_HEADER, 
                    SUPPORT_SW_CA) {

}

void PPCNT_Packet_Type_Counters::DumpFieldNames(stringstream &sstream) {
    sstream << "port_rcv_ibg1_nvl_pkts_high"      << ','
            << "port_rcv_ibg1_nvl_pkts_low"       << ','
            << "port_rcv_ibg1_non_nvl_pkts_high"  << ','
            << "port_rcv_ibg1_non_nvl_pkts_low"   << ','
            << "port_rcv_ibg2_pkts_high"          << ','
            << "port_rcv_ibg2_pkts_low"           << ','
            << "port_xmit_ibg1_nvl_pkts_high"     << ','
            << "port_xmit_ibg1_nvl_pkts_low"      << ','
            << "port_xmit_ibg1_non_nvl_pkts_high" << ','
            << "port_xmit_ibg1_non_nvl_pkts_low"  << ','
            << "port_xmit_ibg2_pkts_high"         << ','
            << "port_xmit_ibg2_pkts_low";
}
void PPCNT_Packet_Type_Counters::DumpData(const struct ppcnt_packet_type_counter& reg, stringstream &sstream) {
    sstream << PTR(reg.port_rcv_ibg1_nvl_pkts_high) << ','
            << PTR(reg.port_rcv_ibg1_nvl_pkts_low) << ','
            << PTR(reg.port_rcv_ibg1_non_nvl_pkts_high) << ','
            << PTR(reg.port_rcv_ibg1_non_nvl_pkts_low) << ','
            << PTR(reg.port_rcv_ibg2_pkts_high) << ','
            << PTR(reg.port_rcv_ibg2_pkts_low) << ','
            << PTR(reg.port_xmit_ibg1_nvl_pkts_high) << ','
            << PTR(reg.port_xmit_ibg1_nvl_pkts_low) << ','
            << PTR(reg.port_xmit_ibg1_non_nvl_pkts_high) << ','
            << PTR(reg.port_xmit_ibg1_non_nvl_pkts_low) << ','
            << PTR(reg.port_xmit_ibg2_pkts_high) << ','
            << PTR(reg.port_xmit_ibg2_pkts_low);
}

void PPCNT_Packet_Type_Counters::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_Packet_Type_Counters::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    DumpData(areg.regs.ppcnt.counter_set.ppcnt_packet_type_counter, sstream);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_Packet_Type_Counters::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                      export_data_phy_node_t *p_export_data_phy_node,
                                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppcnt_packet_type_counter =
                                &areg.regs.ppcnt.counter_set.ppcnt_packet_type_counter;

    IBDIAGNET_RETURN_VOID;
}


PPCNT_InfiniBand_General_Counters::PPCNT_InfiniBand_General_Counters(PhyDiag * phy_diag)
    : PPCNTRegister(phy_diag, 
                    0x26, 
                    (const unpack_data_func_t)ppcnt_infiniband_general_counter_unpack,
                    ACC_REG_PPCNT_INFINIBAND_GENERAL_COUNTERS_INTERNAL_SECTION_NAME, 
                    ACC_REG_PPCNT_INFINIBAND_GENERAL_COUNTERS_NAME, 
                    ACC_REG_PPCNT_INFINIBAND_GENERAL_COUNTERS_FIELDS_NUM, 
                    ACC_REG_PPCNT_INFINIBAND_GENERAL_COUNTERS_HEADER, 
                    SUPPORT_SW_CA) {

}

void PPCNT_InfiniBand_General_Counters::DumpFieldNames(stringstream &sstream) {
    sstream << "port_rcv_data_qword_high"                       <<','
            << "port_rcv_data_qword_low"                        <<','
            << "port_xmit_data_qword_high"                      <<','
            << "port_xmit_data_qword_low"                       <<','
            << "rq_general_error_high"                          <<','
            << "rq_general_error_low"                           <<','
            << "port_xmit_ebp_high"                             <<','
            << "port_xmit_ebp_low"                              <<','
            << "port_sw_hoq_lifetime_limit_discards_high"       <<','
            << "port_sw_hoq_lifetime_limit_discards_low"        <<','
            << "dqs2llu_xmit_wait_arb_global_high"              <<','
            << "dqs2llu_xmit_wait_arb_global_low"               <<','
            << "grxb_fecn_mark_high"                            <<','
            << "grxb_fecn_mark_low"                             <<','
            << "general_transmit_discard_external_contain_high" <<','
            << "general_transmit_discard_external_contain_low"  <<','
            << "general_receive_discard_external_contain_high"  <<','
            << "general_receive_discard_external_contain_low"   <<','
            << "port_receive_poison_high"                       <<','
            << "port_receive_poison_low";
}

void PPCNT_InfiniBand_General_Counters::DumpData(const struct ppcnt_infiniband_general_counter& reg, stringstream &sstream) {
    sstream << reg.port_rcv_data_qword_high <<','
            << reg.port_rcv_data_qword_low <<','
            << reg.port_xmit_data_qword_high <<','
            << reg.port_xmit_data_qword_low <<','
            << reg.rq_general_error_high <<','
            << reg.rq_general_error_low <<','
            << reg.port_xmit_ebp_high <<','
            << reg.port_xmit_ebp_low <<','
            << reg.port_sw_hoq_lifetime_limit_discards_high <<','
            << reg.port_sw_hoq_lifetime_limit_discards_low <<','
            << reg.dqs2llu_xmit_wait_arb_global_high <<','
            << reg.dqs2llu_xmit_wait_arb_global_low <<','
            << reg.grxb_fecn_mark_high <<','
            << reg.grxb_fecn_mark_low <<','
            << reg.general_transmit_discard_external_contain_high <<','
            << reg.general_transmit_discard_external_contain_low <<','
            << reg.general_receive_discard_external_contain_high <<','
            << reg.general_receive_discard_external_contain_low <<','
            << reg.port_receive_poison_high <<','
            << reg.port_receive_poison_low;
}

void PPCNT_InfiniBand_General_Counters::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_InfiniBand_General_Counters::DumpRegisterData(const struct acc_reg_data & areg, 
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    DumpData(areg.regs.ppcnt.counter_set.ppcnt_infiniband_general_counter, sstream);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_InfiniBand_General_Counters::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppcnt_infiniband_general_counter =
                                &areg.regs.ppcnt.counter_set.ppcnt_infiniband_general_counter;

    IBDIAGNET_RETURN_VOID;

}

PPCNT_RS_Fec_Histograms_Counters::PPCNT_RS_Fec_Histograms_Counters(PhyDiag * phy_diag)
    : PPCNTRegister(phy_diag,
                    0x23,
                    (const unpack_data_func_t)ppcnt_rs_fec_histograms_counters_unpack,
                    ACC_REG_PPCNT_RS_FEC_HISTOGRAMS_COUNTERS_INTERNAL_SECTION_NAME,
                    ACC_REG_PPCNT_RS_FEC_HISTOGRAMS_COUNTERS_NAME,
                    ACC_REG_PPCNT_RS_FEC_HISTOGRAMS_COUNTERS_FIELDS_NUM,
                    ACC_REG_PPCNT_RS_FEC_HISTOGRAMS_COUNTERS_HEADER,
                    SUPPORT_SW_CA) {

}

void PPCNT_RS_Fec_Histograms_Counters::DumpFieldNames(stringstream &sstream) {
    for (size_t i = 0; i < ARRAY_SIZE(ppcnt_rs_fec_histograms_counters::hist); ++i)
        sstream << ',' << "hist[" << i << "]";
}

void PPCNT_RS_Fec_Histograms_Counters::DumpData(const ppcnt_rs_fec_histograms_counters& reg, stringstream &sstream) {
    sstream << reg.hist[0];
    for (size_t i = 1; i < ARRAY_SIZE(ppcnt_rs_fec_histograms_counters::hist); ++i) {
        sstream << ',' << reg.hist[i];
    }
}

void PPCNT_RS_Fec_Histograms_Counters::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_RS_Fec_Histograms_Counters::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    DumpData(areg.regs.ppcnt.counter_set.ppcnt_rs_fec_histograms_counters, sstream);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_RS_Fec_Histograms_Counters::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppcnt_rs_fec_histograms_counters =
                                &areg.regs.ppcnt.counter_set.ppcnt_rs_fec_histograms_counters;

    IBDIAGNET_RETURN_VOID;

}

PPCNT_Phys_Layer_Cntrs::PPCNT_Phys_Layer_Cntrs(PhyDiag * phy_diag)
    : PPCNTRegister(phy_diag,
                    0x12,
                    (const unpack_data_func_t)phys_layer_cntrs_unpack,
                    ACC_REG_PPCNT_PHYS_LAYER_CNTRS_INTERNAL_SECTION_NAME,
                    ACC_REG_PPCNT_PHYS_LAYER_CNTRS_NAME,
                    ACC_REG_PPCNT_PHYS_LAYER_CNTRS_FIELDS_NUM,
                    ACC_REG_PPCNT_PHYS_LAYER_CNTRS_HEADER) {

}

void PPCNT_Phys_Layer_Cntrs::DumpFieldNames(stringstream &sstream) {
    sstream << "time_since_last_clear_high"             << ','
            << "time_since_last_clear_low"              << ','
            << "symbol_errors_high"                     << ','
            << "symbol_errors_low"                      << ','
            << "sync_headers_errors_high"               << ','
            << "sync_headers_errors_low"                << ','
            << "edpl_bip_errors_lane0_high"             << ','
            << "edpl_bip_errors_lane0_low"              << ','
            << "edpl_bip_errors_lane1_high"             << ','
            << "edpl_bip_errors_lane1_low"              << ','
            << "edpl_bip_errors_lane2_high"             << ','
            << "edpl_bip_errors_lane2_low"              << ','
            << "edpl_bip_errors_lane3_high"             << ','
            << "edpl_bip_errors_lane3_low"              << ','
            << "fc_fec_corrected_blocks_lane0_high"     << ','
            << "fc_fec_corrected_blocks_lane0_low"      << ','
            << "fc_fec_corrected_blocks_lane1_high"     << ','
            << "fc_fec_corrected_blocks_lane1_low"      << ','
            << "fc_fec_corrected_blocks_lane2_high"     << ','
            << "fc_fec_corrected_blocks_lane2_low"      << ','
            << "fc_fec_corrected_blocks_lane3_high"     << ','
            << "fc_fec_corrected_blocks_lane3_low"      << ','
            << "fc_fec_uncorrectable_blocks_lane0_high" << ','
            << "fc_fec_uncorrectable_blocks_lane0_low"  << ','
            << "fc_fec_uncorrectable_blocks_lane1_high" << ','
            << "fc_fec_uncorrectable_blocks_lane1_low"  << ','
            << "fc_fec_uncorrectable_blocks_lane2_high" << ','
            << "fc_fec_uncorrectable_blocks_lane2_low"  << ','
            << "fc_fec_uncorrectable_blocks_lane3_high" << ','
            << "fc_fec_uncorrectable_blocks_lane3_low"  << ','
            << "rs_fec_corrected_blocks_high"           << ','
            << "rs_fec_corrected_blocks_low"            << ','
            << "rs_fec_uncorrectable_blocks_high"       << ','
            << "rs_fec_uncorrectable_blocks_low"        << ','
            << "rs_fec_no_errors_blocks_high"           << ','
            << "rs_fec_no_errors_blocks_low"            << ','
            << "rs_fec_single_error_blocks_high"        << ','
            << "rs_fec_single_error_blocks_low"         << ','
            << "rs_fec_corrected_symbols_total_high"    << ','
            << "rs_fec_corrected_symbols_total_low"     << ','
            << "rs_fec_corrected_symbols_lane0_high"    << ','
            << "rs_fec_corrected_symbols_lane0_low"     << ','
            << "rs_fec_corrected_symbols_lane1_high"    << ','
            << "rs_fec_corrected_symbols_lane1_low"     << ','
            << "rs_fec_corrected_symbols_lane2_high"    << ','
            << "rs_fec_corrected_symbols_lane2_low"     << ','
            << "rs_fec_corrected_symbols_lane3_high"    << ','
            << "rs_fec_corrected_symbols_lane3_low"     << ','
            << "link_down_events"                       << ','
            << "successful_recovery_events"             ;

}

void PPCNT_Phys_Layer_Cntrs::DumpData(const phys_layer_cntrs& reg, stringstream &sstream) {
    sstream << +reg.time_since_last_clear_high << ','
            << +reg.time_since_last_clear_low << ','
            << +reg.symbol_errors_high << ','
            << +reg.symbol_errors_low << ','
            << +reg.sync_headers_errors_high << ','
            << +reg.sync_headers_errors_low << ','
            << +reg.edpl_bip_errors_lane0_high << ','
            << +reg.edpl_bip_errors_lane0_low << ','
            << +reg.edpl_bip_errors_lane1_high << ','
            << +reg.edpl_bip_errors_lane1_low << ','
            << +reg.edpl_bip_errors_lane2_high << ','
            << +reg.edpl_bip_errors_lane2_low << ','
            << +reg.edpl_bip_errors_lane3_high << ','
            << +reg.edpl_bip_errors_lane3_low << ','
            << +reg.fc_fec_corrected_blocks_lane0_high << ','
            << +reg.fc_fec_corrected_blocks_lane0_low << ','
            << +reg.fc_fec_corrected_blocks_lane1_high << ','
            << +reg.fc_fec_corrected_blocks_lane1_low << ','
            << +reg.fc_fec_corrected_blocks_lane2_high << ','
            << +reg.fc_fec_corrected_blocks_lane2_low << ','
            << +reg.fc_fec_corrected_blocks_lane3_high << ','
            << +reg.fc_fec_corrected_blocks_lane3_low << ','
            << +reg.fc_fec_uncorrectable_blocks_lane0_high << ','
            << +reg.fc_fec_uncorrectable_blocks_lane0_low << ','
            << +reg.fc_fec_uncorrectable_blocks_lane1_high << ','
            << +reg.fc_fec_uncorrectable_blocks_lane1_low << ','
            << +reg.fc_fec_uncorrectable_blocks_lane2_high << ','
            << +reg.fc_fec_uncorrectable_blocks_lane2_low << ','
            << +reg.fc_fec_uncorrectable_blocks_lane3_high << ','
            << +reg.fc_fec_uncorrectable_blocks_lane3_low << ','
            << +reg.rs_fec_corrected_blocks_high << ','
            << +reg.rs_fec_corrected_blocks_low << ','
            << +reg.rs_fec_uncorrectable_blocks_high << ','
            << +reg.rs_fec_uncorrectable_blocks_low << ','
            << +reg.rs_fec_no_errors_blocks_high << ','
            << +reg.rs_fec_no_errors_blocks_low << ','
            << +reg.rs_fec_single_error_blocks_high << ','
            << +reg.rs_fec_single_error_blocks_low << ','
            << +reg.rs_fec_corrected_symbols_total_high << ','
            << +reg.rs_fec_corrected_symbols_total_low << ','
            << +reg.rs_fec_corrected_symbols_lane0_high << ','
            << +reg.rs_fec_corrected_symbols_lane0_low << ','
            << +reg.rs_fec_corrected_symbols_lane1_high << ','
            << +reg.rs_fec_corrected_symbols_lane1_low << ','
            << +reg.rs_fec_corrected_symbols_lane2_high << ','
            << +reg.rs_fec_corrected_symbols_lane2_low << ','
            << +reg.rs_fec_corrected_symbols_lane3_high << ','
            << +reg.rs_fec_corrected_symbols_lane3_low << ','
            << PTR(reg.link_down_events) << ','
            << PTR(reg.successful_recovery_events) ;

}

void PPCNT_Phys_Layer_Cntrs::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_Phys_Layer_Cntrs::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    DumpData(areg.regs.ppcnt.counter_set.phys_layer_cntrs, sstream);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_Phys_Layer_Cntrs::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_phys_layer_cntrs =
                                &areg.regs.ppcnt.counter_set.phys_layer_cntrs;

    IBDIAGNET_RETURN_VOID;

}

PPCNT_Phys_Layer_Stat_Cntrs::PPCNT_Phys_Layer_Stat_Cntrs(PhyDiag * phy_diag)
    : PPCNTRegister(phy_diag,
                    0x16,
                    (const unpack_data_func_t)phys_layer_stat_cntrs_unpack,
                    ACC_REG_PPCNT_PHYS_LAYER_STAT_CNTRS_INTERNAL_SECTION_NAME,
                    ACC_REG_PPCNT_PHYS_LAYER_STAT_CNTRS_NAME,
                    ACC_REG_PPCNT_PHYS_LAYER_STAT_CNTRS_FIELDS_NUM,
                    ACC_REG_PPCNT_PHYS_LAYER_STAT_CNTRS_HEADER,
                    SUPPORT_SW_CA) {

}

void PPCNT_Phys_Layer_Stat_Cntrs::DumpFieldNames(stringstream &sstream) {
    sstream << "time_since_last_clear_high"         <<','
            << "time_since_last_clear_low"          <<','
            << "phy_received_bits_high"             <<','
            << "phy_received_bits_low"              <<','
            << "phy_symbol_errors_high"             <<','
            << "phy_symbol_errors_low"              <<','
            << "phy_corrected_bits_high"            <<','
            << "phy_corrected_bits_low"             <<','
            << "phy_raw_errors_lane0_high"          <<','
            << "phy_raw_errors_lane0_low"           <<','
            << "phy_raw_errors_lane1_high"          <<','
            << "phy_raw_errors_lane1_low"           <<','
            << "phy_raw_errors_lane2_high"          <<','
            << "phy_raw_errors_lane2_low"           <<','
            << "phy_raw_errors_lane3_high"          <<','
            << "phy_raw_errors_lane3_low"           <<','
            << "phy_raw_errors_lane4_high"          <<','
            << "phy_raw_errors_lane4_low"           <<','
            << "phy_raw_errors_lane5_high"          <<','
            << "phy_raw_errors_lane5_low"           <<','
            << "phy_raw_errors_lane6_high"          <<','
            << "phy_raw_errors_lane6_low"           <<','
            << "phy_raw_errors_lane7_high"          <<','
            << "phy_raw_errors_lane7_low"           <<','
            << "fc_zero_hist"                       <<','
            << "raw_ber_magnitude"                  <<','
            << "raw_ber_coef_float"                 <<','
            << "raw_ber_coef"                       <<','
            << "estimated_effective_ber_magnitude"  <<','
            << "estimated_effective_ber_coef_float" <<','
            << "estimated_effective_ber_coef"       <<','
            << "effective_ber_magnitude"            <<','
            << "effective_ber_coef_float"           <<','
            << "effective_ber_coef"                 <<','
            << "symbol_ber_magnitude"               <<','
            << "symbol_ber_coef_float"              <<','
            << "symbol_ber_coef"                    <<','
            << "phy_effective_errors_high"          <<','
            << "phy_effective_errors_low"           <<','
            << "raw_ber_magnitude_lane0"            <<','
            << "raw_ber_coef_float_lane0"           <<','
            << "raw_ber_coef_lane0"                 <<','
            << "raw_ber_magnitude_lane1"            <<','
            << "raw_ber_coef_float_lane1"           <<','
            << "raw_ber_coef_lane1"                 <<','
            << "raw_ber_magnitude_lane2"            <<','
            << "raw_ber_coef_float_lane2"           <<','
            << "raw_ber_coef_lane2"                 <<','
            << "raw_ber_magnitude_lane3"            <<','
            << "raw_ber_coef_float_lane3"           <<','
            << "raw_ber_coef_lane3"                 <<','
            << "raw_ber_magnitude_lane4"            <<','
            << "raw_ber_coef_float_lane4"           <<','
            << "raw_ber_coef_lane4"                 <<','
            << "raw_ber_magnitude_lane5"            <<','
            << "raw_ber_coef_float_lane5"           <<','
            << "raw_ber_coef_lane5"                 <<','
            << "raw_ber_magnitude_lane6"            <<','
            << "raw_ber_coef_float_lane6"           <<','
            << "raw_ber_coef_lane6"                 <<','
            << "raw_ber_magnitude_lane7"            <<','
            << "raw_ber_coef_float_lane7"           <<','
            << "raw_ber_coef_lane7"                 ;

}

void PPCNT_Phys_Layer_Stat_Cntrs::DumpData(const phys_layer_stat_cntrs& reg, stringstream &sstream,
                                           bool is_effective_counter_supported, bool is_raw_ber_per_lane_supported) {
    sstream << reg.time_since_last_clear_high <<','
            << reg.time_since_last_clear_low <<','
            << reg.phy_received_bits_high <<','
            << reg.phy_received_bits_low <<','
            << reg.phy_symbol_errors_high <<','
            << reg.phy_symbol_errors_low <<','
            << reg.phy_corrected_bits_high <<','
            << reg.phy_corrected_bits_low <<','
            << reg.phy_raw_errors_lane0_high <<','
            << reg.phy_raw_errors_lane0_low <<','
            << reg.phy_raw_errors_lane1_high <<','
            << reg.phy_raw_errors_lane1_low <<','
            << reg.phy_raw_errors_lane2_high <<','
            << reg.phy_raw_errors_lane2_low <<','
            << reg.phy_raw_errors_lane3_high <<','
            << reg.phy_raw_errors_lane3_low <<','
            << reg.phy_raw_errors_lane4_high <<','
            << reg.phy_raw_errors_lane4_low <<','
            << reg.phy_raw_errors_lane5_high <<','
            << reg.phy_raw_errors_lane5_low <<','
            << reg.phy_raw_errors_lane6_high <<','
            << reg.phy_raw_errors_lane6_low <<','
            << reg.phy_raw_errors_lane7_high <<','
            << reg.phy_raw_errors_lane7_low <<','
            << reg.fc_zero_hist <<','
            << +reg.raw_ber_magnitude <<','
            << +reg.raw_ber_coef_float <<','
            << +reg.raw_ber_coef <<','
            << +reg.estimated_effective_ber_magnitude <<','
            << +reg.estimated_effective_ber_coef_float <<','
            << +reg.estimated_effective_ber_coef <<','
            << +reg.effective_ber_magnitude <<','
            << +reg.effective_ber_coef_float <<','
            << +reg.effective_ber_coef <<',';
    if (is_effective_counter_supported) {
        sstream << DEC(reg.symbol_ber_magnitude) <<','
                << DEC(reg.symbol_ber_coef_float) <<','
                << DEC(reg.symbol_ber_coef) <<','
                << DEC(reg.phy_effective_errors_high) <<','
                << DEC(reg.phy_effective_errors_low) <<',';
    }
    else {
        sstream << "N/A,N/A,N/A,N/A,N/A,";
    }
    if (is_raw_ber_per_lane_supported) {
        sstream << +reg.raw_ber_magnitude_lane0 <<','
                << +reg.raw_ber_coef_float_lane0 <<','
                << +reg.raw_ber_coef_lane0 <<','
                << +reg.raw_ber_magnitude_lane1 <<','
                << +reg.raw_ber_coef_float_lane1 <<','
                << +reg.raw_ber_coef_lane1 <<','
                << +reg.raw_ber_magnitude_lane2 <<','
                << +reg.raw_ber_coef_float_lane2 <<','
                << +reg.raw_ber_coef_lane2 <<','
                << +reg.raw_ber_magnitude_lane3 <<','
                << +reg.raw_ber_coef_float_lane3 <<','
                << +reg.raw_ber_coef_lane3 <<','
                << +reg.raw_ber_magnitude_lane4 <<','
                << +reg.raw_ber_coef_float_lane4 <<','
                << +reg.raw_ber_coef_lane4 <<','
                << +reg.raw_ber_magnitude_lane5 <<','
                << +reg.raw_ber_coef_float_lane5 <<','
                << +reg.raw_ber_coef_lane5 <<','
                << +reg.raw_ber_magnitude_lane6 <<','
                << +reg.raw_ber_coef_float_lane6 <<','
                << +reg.raw_ber_coef_lane6 <<','
                << +reg.raw_ber_magnitude_lane7 <<','
                << +reg.raw_ber_coef_float_lane7 <<','
                << +reg.raw_ber_coef_lane7;
    }
    else {
        sstream << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A,"
                << "N/A,N/A,N/A,N/A,N/A,N/A,N/A,N/A";
    }

}

void PPCNT_Phys_Layer_Stat_Cntrs::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_Phys_Layer_Stat_Cntrs::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    bool is_effective_counter_supported = false;
    bool is_raw_ber_per_lane_supported = false;
    const AccRegKeyPort *p_key = dynamic_cast<const AccRegKeyPort*>(&key);
    if (p_key) {
        IBNode* p_node =  this->m_phy_diag->GetNode(p_key->node_guid);
        if (p_node) {
            is_effective_counter_supported = this->m_phy_diag->GetCapabilityModule()->
                        IsSupportedGMPCapability(p_node,EnGMPCAPIsEffectiveCounterSupported);
            is_raw_ber_per_lane_supported = this->m_phy_diag->GetCapabilityModule()->
                        IsSupportedGMPCapability(p_node,EnGMPCAPIsRawBerPerLaneSupported);
        }
    }
    DumpData(areg.regs.ppcnt.counter_set.phys_layer_stat_cntrs, sstream, 
             is_effective_counter_supported, is_raw_ber_per_lane_supported);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_Phys_Layer_Stat_Cntrs::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_phys_layer_stat_cntrs =
                                &areg.regs.ppcnt.counter_set.phys_layer_stat_cntrs;

    IBDIAGNET_RETURN_VOID;

}

PPCNT_L1_General_Counters::PPCNT_L1_General_Counters(PhyDiag * phy_diag)
    : PPCNTRegister(phy_diag,
                    0x27,
                    (const unpack_data_func_t)ppcnt_l1_general_counters_unpack,
                    ACC_REG_PPCNT_L1_GENERAL_COUNTERS_INTERNAL_SECTION_NAME,
                    ACC_REG_PPCNT_L1_GENERAL_COUNTERS_NAME,
                    ACC_REG_PPCNT_L1_GENERAL_COUNTERS_FIELDS_NUM,
                    ACC_REG_PPCNT_L1_GENERAL_COUNTERS_HEADER) {

}

void PPCNT_L1_General_Counters::DumpFieldNames(stringstream &sstream) {
    sstream << "l1_entry_quiet_high"                      << ','
            << "l1_entry_quiet_low"                       << ','
            << "l1_force_entry_high"                      << ','
            << "l1_force_entry_low"                       << ','
            << "l1_exit_local_desired_high"               << ','
            << "l1_exit_local_desired_low"                << ','
            << "l1_exit_local_recal_high"                 << ','
            << "l1_exit_local_recal_low"                  << ','
            << "l1_exit_local_recal_with_traffic_high"    << ','
            << "l1_exit_local_recal_with_traffic_low"     << ','
            << "l1_exit_local_recal_without_traffic_high" << ','
            << "l1_exit_local_recal_without_traffic_low"  << ','
            << "l1_exit_remote_high"                      << ','
            << "l1_exit_remote_low"                       << ','
            << "l1_exit_remote_probably_recal_high"       << ','
            << "l1_exit_remote_probably_recal_low"        << ','
            << "local_full_bw_exit_high"                  << ','
            << "local_full_bw_exit_low"                   << ','
            << "local_low_power_enter_high"               << ','
            << "local_low_power_enter_low"                << ','
            << "remote_full_bw_exit_high"                 << ','
            << "remote_full_bw_exit_low"                  << ','
            << "remote_low_power_enter_high"              << ','
            << "remote_low_power_enter_low"               << ','
            << "local_low_power_exit_high"                << ','
            << "local_low_power_exit_low"                 << ','
            << "local_full_bw_enter_high"                 << ','
            << "local_full_bw_enter_low"                  << ','
            << "remote_low_power_exit_high"               << ','
            << "remote_low_power_exit_low"                << ','
            << "remote_full_bw_enter_high"                << ','
            << "remote_full_bw_enter_low"                 << ','
            << "high_speed_steady_state_high"             << ','
            << "high_speed_steady_state_low"              << ','
            << "l1_steady_state_high"                     << ','
            << "l1_steady_state_low"                      << ','
            << "other_state_high"                         << ','
            << "other_state_low"                          << ','
            << "time_since_last_clear_high"               << ','
            << "time_since_last_clear_low"                << ','
            << "l1_total_entries_high"                    << ','
            << "l1_total_entries_low"                     << ','
            << "l1_total_local_exits_high"                << ','
            << "l1_total_local_exits_low"                 << ','
            << "l1_total_exits_high"                      << ','
            << "l1_total_exits_low"                       << ','
            << "relative_high_speed_steady_state_time"    << ','
            << "relative_l1_steady_state_time"            << ','
            << "relative_other_state_time"                ;

}

void PPCNT_L1_General_Counters::DumpData(const ppcnt_l1_general_counters& reg, stringstream &sstream) {
    sstream << +reg.l1_entry_quiet_high << ','
            << +reg.l1_entry_quiet_low << ','
            << +reg.l1_force_entry_high << ','
            << +reg.l1_force_entry_low << ','
            << +reg.l1_exit_local_desired_high << ','
            << +reg.l1_exit_local_desired_low << ','
            << +reg.l1_exit_local_recal_high << ','
            << +reg.l1_exit_local_recal_low << ','
            << +reg.l1_exit_local_recal_with_traffic_high << ','
            << +reg.l1_exit_local_recal_with_traffic_low << ','
            << +reg.l1_exit_local_recal_without_traffic_high << ','
            << +reg.l1_exit_local_recal_without_traffic_low << ','
            << +reg.l1_exit_remote_high << ','
            << +reg.l1_exit_remote_low << ','
            << +reg.l1_exit_remote_probably_recal_high << ','
            << +reg.l1_exit_remote_probably_recal_low << ','
            << +reg.local_full_bw_exit_high << ','
            << +reg.local_full_bw_exit_low << ','
            << +reg.local_low_power_enter_high << ','
            << +reg.local_low_power_enter_low << ','
            << +reg.remote_full_bw_exit_high << ','
            << +reg.remote_full_bw_exit_low << ','
            << +reg.remote_low_power_enter_high << ','
            << +reg.remote_low_power_enter_low << ','
            << +reg.local_low_power_exit_high << ','
            << +reg.local_low_power_exit_low << ','
            << +reg.local_full_bw_enter_high << ','
            << +reg.local_full_bw_enter_low << ','
            << +reg.remote_low_power_exit_high << ','
            << +reg.remote_low_power_exit_low << ','
            << +reg.remote_full_bw_enter_high << ','
            << +reg.remote_full_bw_enter_low << ','
            << +reg.high_speed_steady_state_high << ','
            << +reg.high_speed_steady_state_low << ','
            << +reg.l1_steady_state_high << ','
            << +reg.l1_steady_state_low << ','
            << +reg.other_state_high << ','
            << +reg.other_state_low << ','
            << +reg.time_since_last_clear_high << ','
            << +reg.time_since_last_clear_low << ','
            << +reg.l1_total_entries_high << ','
            << +reg.l1_total_entries_low << ','
            << +reg.l1_total_local_exits_high << ','
            << +reg.l1_total_local_exits_low << ','
            << +reg.l1_total_exits_high << ','
            << +reg.l1_total_exits_low << ','
            << +reg.relative_high_speed_steady_state_time << ','
            << +reg.relative_l1_steady_state_time << ','
            << +reg.relative_other_state_time ;

}

void PPCNT_L1_General_Counters::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl           << '#'
            << handler_header << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_L1_General_Counters::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    DumpData(areg.regs.ppcnt.counter_set.ppcnt_l1_general_counters, sstream);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPCNT_L1_General_Counters::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppcnt_l1_general_counters =
                                &areg.regs.ppcnt.counter_set.ppcnt_l1_general_counters;

    IBDIAGNET_RETURN_VOID;

}

//*****************************************
//************* PEMI Register functions ***
//*****************************************
int PEMIRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer) {
    IBDIAGNET_ENTER;

    pemi_reg *reg = (pemi_reg *) data_to_unpack;
    pemi_reg_unpack(reg, unpacked_buffer);

    const AccRegKeyPort *key = dynamic_cast<const AccRegKeyPort*>(p_key);

    if (key) {
        IBPort *port = this->m_phy_diag->GetPort(key->node_guid, key->port_num);

        if (port && port->p_phy_data) {
            PHYPortData *phy_data = dynamic_cast<PHYPortData*>(port->p_phy_data);

            if (phy_data)
                if (phy_data->pemi_group_cap_mask == std::numeric_limits<PHYPortData::pemi_group_cap_mask_t>::max())
                    phy_data->pemi_group_cap_mask = reg->group_cap_mask;
        }
    }

    if (reg->status)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NOT_SUPPORTED);

    pemi_RAW_page_data raw = reg->page_data.raw;
    unpack_data_func(&reg->page_data, raw.data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PEMIRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    const AccRegKeyPort *key = dynamic_cast<const AccRegKeyPort*>(p_key);

    if (key) {
        IBPort *port = this->m_phy_diag->GetPort(key->node_guid, key->port_num);

        if (port && port->p_phy_data) {
            PHYPortData *phy_data = dynamic_cast<PHYPortData *>(port->p_phy_data);

            if (phy_data) // bit is coded in low 4 bits of page select
                if (phy_data->pemi_group_cap_mask & (1 << (this->m_page_select & 0x0F))) {

                    struct pemi_reg pemi;
                    CLEAR_STRUCT(pemi);

                    pemi.lp_msb             = this->m_lp_msb;
                    pemi.pnat               = this->m_pnat;
                    pemi.page_select        = this->m_page_select;
                    pemi.module_info_ext    = this->m_module_info_ext;

                    int rc = FillLocalPortAndPlaneInd(port, &pemi.local_port, nullptr);

                    if (rc)
                        return rc;

                    pemi_reg_pack(&pemi, data);

                    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
                }
        }
    }

    IBDIAG_RETURN(IBDIAG_ERR_CODE_DISABLED);
}

PEMI_Module_Samples_Register::PEMI_Module_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x00,
                   (const unpack_data_func_t)pemi_Module_Status_Samples_unpack,
                   ACC_REG_PEMI_MODULE_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_MODULE_S_NAME,
                   ACC_REG_PEMI_MODULE_S_FIELDS_NUM,
                   ACC_REG_PEMI_MODULE_S_HEADER)
{
}

void PEMI_Module_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl              << '#'
            << handler_header    << ','
            << "temperature"     << ','
            << "voltage"         << ','

            << "rx_power_lane0"  << ','
            << "rx_power_lane1"  << ','
            << "rx_power_lane2"  << ','
            << "rx_power_lane3"  << ','
            << "rx_power_lane4"  << ','
            << "rx_power_lane5"  << ','
            << "rx_power_lane6"  << ','
            << "rx_power_lane7"  << ','

            << "tx_power_lane0"  << ','
            << "tx_power_lane1"  << ','
            << "tx_power_lane2"  << ','
            << "tx_power_lane3"  << ','
            << "tx_power_lane4"  << ','
            << "tx_power_lane5"  << ','
            << "tx_power_lane6"  << ','
            << "tx_power_lane7"  << ','

            << "tx_bias_lane0"   << ','
            << "tx_bias_lane1"   << ','
            << "tx_bias_lane2"   << ','
            << "tx_bias_lane3"   << ','
            << "tx_bias_lane4"   << ','
            << "tx_bias_lane5"   << ','
            << "tx_bias_lane6"   << ','
            << "tx_bias_lane7"   << ','

            << "module_st"       << ','

            << "dp_st_lane0"     << ','
            << "dp_st_lane1"     << ','
            << "dp_st_lane2"     << ','
            << "dp_st_lane3"     << ','
            << "dp_st_lane4"     << ','
            << "dp_st_lane5"     << ','
            << "dp_st_lane6"     << ','
            << "dp_st_lane7";
#endif
    IBDIAGNET_RETURN_VOID;
}


void PEMI_Module_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Module_Status_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Module_Status_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(temperature)
        << __FMT__(voltage)

        << __FMT__(rx_power_lane0)
        << __FMT__(rx_power_lane1)
        << __FMT__(rx_power_lane2)
        << __FMT__(rx_power_lane3)
        << __FMT__(rx_power_lane4)
        << __FMT__(rx_power_lane5)
        << __FMT__(rx_power_lane6)
        << __FMT__(rx_power_lane7)

        << __FMT__(tx_power_lane0)
        << __FMT__(tx_power_lane1)
        << __FMT__(tx_power_lane2)
        << __FMT__(tx_power_lane3)
        << __FMT__(tx_power_lane4)
        << __FMT__(tx_power_lane5)
        << __FMT__(tx_power_lane6)
        << __FMT__(tx_power_lane7)

        << __FMT__(tx_bias_lane0)
        << __FMT__(tx_bias_lane1)
        << __FMT__(tx_bias_lane2)
        << __FMT__(tx_bias_lane3)
        << __FMT__(tx_bias_lane4)
        << __FMT__(tx_bias_lane5)
        << __FMT__(tx_bias_lane6)
        << __FMT__(tx_bias_lane7)

        << __FMT__(module_st)

        << __FMT__(dp_st_lane0)
        << __FMT__(dp_st_lane1)
        << __FMT__(dp_st_lane2)
        << __FMT__(dp_st_lane3)
        << __FMT__(dp_st_lane4)
        << __FMT__(dp_st_lane5)
        << __FMT__(dp_st_lane6)
        << __FMT_LAST__(dp_st_lane7);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_Module_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                      export_data_phy_node_t *p_export_data_phy_node,
                                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_module_status_samples =
                                &areg.regs.pemi.page_data.pemi_Module_Status_Samples;

    IBDIAGNET_RETURN_VOID;
}

PEMI_SNR_Samples_Register::PEMI_SNR_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x01,
                   (const unpack_data_func_t)pemi_SNR_Samples_unpack,
                   ACC_REG_PEMI_SNR_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_SNR_S_NAME,
                   ACC_REG_PEMI_SNR_S_FIELDS_NUM,
                   ACC_REG_PEMI_SNR_S_HEADER)
{
}

void PEMI_SNR_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "snr_media_lane0"  << ','
            << "snr_media_lane1"  << ','
            << "snr_media_lane2"  << ','
            << "snr_media_lane3"  << ','
            << "snr_media_lane4"  << ','
            << "snr_media_lane5"  << ','
            << "snr_media_lane6"  << ','
            << "snr_media_lane7"  << ','
            << "snr_host_lane0"   << ','
            << "snr_host_lane1"   << ','
            << "snr_host_lane2"   << ','
            << "snr_host_lane3"   << ','
            << "snr_host_lane4"   << ','
            << "snr_host_lane5"   << ','
            << "snr_host_lane6"   << ','
            << "snr_host_lane7";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PEMI_SNR_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_SNR_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_SNR_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(snr_media_lane0)
        << __FMT__(snr_media_lane1)
        << __FMT__(snr_media_lane2)
        << __FMT__(snr_media_lane3)
        << __FMT__(snr_media_lane4)
        << __FMT__(snr_media_lane5)
        << __FMT__(snr_media_lane6)
        << __FMT__(snr_media_lane7)

        << __FMT__(snr_host_lane0)
        << __FMT__(snr_host_lane1)
        << __FMT__(snr_host_lane2)
        << __FMT__(snr_host_lane3)
        << __FMT__(snr_host_lane4)
        << __FMT__(snr_host_lane5)
        << __FMT__(snr_host_lane6)
        << __FMT_LAST__(snr_host_lane7);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_SNR_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                        export_data_phy_node_t *p_export_data_phy_node,
                                        struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_snr_samples = &areg.regs.pemi.page_data.pemi_SNR_Samples;

    IBDIAGNET_RETURN_VOID;
}


PEMI_Laser_Samples_Register::PEMI_Laser_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x02,
                   (const unpack_data_func_t)pemi_Laser_Monitors_Samples_unpack,
                   ACC_REG_PEMI_LASER_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_LASER_S_NAME,
                   ACC_REG_PEMI_LASER_S_FIELDS_NUM,
                   ACC_REG_PEMI_LASER_S_HEADER)
{
}

void PEMI_Laser_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                           << '#'
            << handler_header                 << ','
            << "laser_age"                    << ','
            << "tec_current"                  << ','
            << "laser_frequency_error_lane0"  << ','
            << "laser_frequency_error_lane1"  << ','
            << "laser_frequency_error_lane2"  << ','
            << "laser_frequency_error_lane3"  << ','
            << "laser_frequency_error_lane4"  << ','
            << "laser_frequency_error_lane5"  << ','
            << "laser_frequency_error_lane6"  << ','
            << "laser_frequency_error_lane7"  << ','
            << "laser_temperature_lane0"      << ','
            << "laser_temperature_lane1"      << ','
            << "laser_temperature_lane2"      << ','
            << "laser_temperature_lane3"      << ','
            << "laser_temperature_lane4"      << ','
            << "laser_temperature_lane5"      << ','
            << "laser_temperature_lane6"      << ','
            << "laser_temperature_lane7";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PEMI_Laser_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Laser_Monitors_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Laser_Monitors_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(laser_age)
        << __FMT__(tec_current)

        << __FMT__(laser_frequency_error_lane0)
        << __FMT__(laser_frequency_error_lane1)
        << __FMT__(laser_frequency_error_lane2)
        << __FMT__(laser_frequency_error_lane3)
        << __FMT__(laser_frequency_error_lane4)
        << __FMT__(laser_frequency_error_lane5)
        << __FMT__(laser_frequency_error_lane6)
        << __FMT__(laser_frequency_error_lane7)

        << __FMT__(laser_temperature_lane0)
        << __FMT__(laser_temperature_lane1)
        << __FMT__(laser_temperature_lane2)
        << __FMT__(laser_temperature_lane3)
        << __FMT__(laser_temperature_lane4)
        << __FMT__(laser_temperature_lane5)
        << __FMT__(laser_temperature_lane6)
        << __FMT_LAST__(laser_temperature_lane7);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_Laser_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                     export_data_phy_node_t *p_export_data_phy_node,
                                                     struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_laser_monitors_samples =
                                &areg.regs.pemi.page_data.pemi_Laser_Monitors_Samples;

    IBDIAGNET_RETURN_VOID;
}

PEMI_PAM4_Samples_Register::PEMI_PAM4_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x03,
                   (const unpack_data_func_t)pemi_PAM4_Level_Transition_Samples_unpack,
                   ACC_REG_PEMI_PAM4_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_PAM4_S_NAME,
                   ACC_REG_PEMI_PAM4_S_FIELDS_NUM,
                   ACC_REG_PEMI_PAM4_S_HEADER)
{
}

void PEMI_PAM4_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                                    <<  '#'
            << handler_header                          << ','
            << "pam4_level_transition_media_lane0"     <<  ','
            << "pam4_level_transition_media_lane1"     <<  ','
            << "pam4_level_transition_media_lane2"     <<  ','
            << "pam4_level_transition_media_lane3"     <<  ','
            << "pam4_level_transition_media_lane4"     <<  ','
            << "pam4_level_transition_media_lane5"     <<  ','
            << "pam4_level_transition_media_lane6"     <<  ','
            << "pam4_level_transition_media_lane7"     <<  ','

            << "pam4_level_transition_host_lane0"      <<  ','
            << "pam4_level_transition_host_lane1"      <<  ','
            << "pam4_level_transition_host_lane2"      <<  ','
            << "pam4_level_transition_host_lane3"      <<  ','
            << "pam4_level_transition_host_lane4"      <<  ','
            << "pam4_level_transition_host_lane5"      <<  ','
            << "pam4_level_transition_host_lane6"      <<  ','
            << "pam4_level_transition_host_lane7";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PAM4_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(pam4_level_transition_media_lane0)
        << __FMT__(pam4_level_transition_media_lane1)
        << __FMT__(pam4_level_transition_media_lane2)
        << __FMT__(pam4_level_transition_media_lane3)
        << __FMT__(pam4_level_transition_media_lane4)
        << __FMT__(pam4_level_transition_media_lane5)
        << __FMT__(pam4_level_transition_media_lane6)
        << __FMT__(pam4_level_transition_media_lane7)

        << __FMT__(pam4_level_transition_host_lane0)
        << __FMT__(pam4_level_transition_host_lane1)
        << __FMT__(pam4_level_transition_host_lane2)
        << __FMT__(pam4_level_transition_host_lane3)
        << __FMT__(pam4_level_transition_host_lane4)
        << __FMT__(pam4_level_transition_host_lane5)
        << __FMT__(pam4_level_transition_host_lane6)
        << __FMT_LAST__(pam4_level_transition_host_lane7);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PAM4_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                    export_data_phy_node_t *p_export_data_phy_node,
                                                    struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_pam4_level_transition_samples =
                                    &areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Samples;

    IBDIAGNET_RETURN_VOID;
}

PEMI_PRE_FEC_BER_Samples_Register::PEMI_PRE_FEC_BER_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x04,
                   (const unpack_data_func_t)pemi_Pre_FEC_BER_Samples_unpack,
                   ACC_REG_PEMI_BER_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_BER_S_NAME,
                   ACC_REG_PEMI_BER_S_FIELDS_NUM,
                   ACC_REG_PEMI_BER_S_HEADER)
{
}

void PEMI_PRE_FEC_BER_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                     << '#'
            << handler_header           << ','
            << "pre_fec_ber_min_media"  << ','
            << "pre_fec_ber_min_host"   << ','
            << "pre_fec_ber_max_media"  << ','
            << "pre_fec_ber_max_host"   << ','
            << "pre_fec_ber_avg_media"  << ','
            << "pre_fec_ber_avg_host"   << ','
            << "pre_fec_ber_val_media"  << ','
            << "pre_fec_ber_val_host";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PEMI_PRE_FEC_BER_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(pre_fec_ber_min_media)
        << __FMT__(pre_fec_ber_min_host)
        << __FMT__(pre_fec_ber_max_media)
        << __FMT__(pre_fec_ber_max_host)
        << __FMT__(pre_fec_ber_avg_media)
        << __FMT__(pre_fec_ber_avg_host)
        << __FMT__(pre_fec_ber_val_media)
        << __FMT_LAST__(pre_fec_ber_val_host);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PRE_FEC_BER_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                           export_data_phy_node_t *p_export_data_phy_node,
                                                           struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_pre_fec_ber_samples =
                                    &areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Samples;

    IBDIAGNET_RETURN_VOID;
}


PEMI_FERC_Samples_Register::PEMI_FERC_Samples_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x05,
                   (const unpack_data_func_t)pemi_FERC_Samples_unpack,
                   ACC_REG_PEMI_FERC_S_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_FERC_S_NAME,
                   ACC_REG_PEMI_FERC_S_FIELDS_NUM,
                   ACC_REG_PEMI_FERC_S_HEADER)
{
}

void PEMI_FERC_Samples_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                   << '#'
            << handler_header         << ','
            << "ferc_min_media"       << ','
            << "ferc_min_host"        << ','
            << "ferc_max_media"       << ','
            << "ferc_max_host"        << ','
            << "ferc_avg_media"       << ','
            << "ferc_avg_host"        << ','
            << "ferc_val_media"       << ','
            << "ferc_val_host";

#endif

    IBDIAGNET_RETURN_VOID;
}

void PEMI_FERC_Samples_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_FERC_Samples.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_FERC_Samples.x << std::endl

    sstream
        << std::hex
        << __FMT__(ferc_min_media)
        << __FMT__(ferc_min_host)
        << __FMT__(ferc_max_media)
        << __FMT__(ferc_max_host)
        << __FMT__(ferc_avg_media)
        << __FMT__(ferc_avg_host)
        << __FMT__(ferc_val_media)
        << __FMT_LAST__(ferc_val_host);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_FERC_Samples_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                    export_data_phy_node_t *p_export_data_phy_node,
                                                    struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_ferc_samples = &areg.regs.pemi.page_data.pemi_FERC_Samples;

    IBDIAGNET_RETURN_VOID;
}

PEMI_Module_Properties_Register::PEMI_Module_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x10,
                   (const unpack_data_func_t)pemi_Module_Status_Properties_unpack,
                   ACC_REG_PEMI_MODULE_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_MODULE_P_NAME,
                   ACC_REG_PEMI_MODULE_P_FIELDS_NUM,
                   ACC_REG_PEMI_MODULE_P_HEADER)
{
}

void PEMI_Module_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                   << '#'
            << handler_header         << ','
            << "module_status_cap"     << ','
            << "temp_high_alarm"       << ','
            << "temp_low_alarm"        << ','
            << "temp_high_warning"     << ','
            << "temp_low_warning"      << ','
            << "voltage_high_alarm"    << ','
            << "voltage_low_alarm"     << ','
            << "voltage_high_warning"  << ','
            << "voltage_low_warning"   << ','
            << "rx_power_high_alarm"   << ','
            << "rx_power_low_alarm"    << ','
            << "rx_power_high_warning" << ','
            << "rx_power_low_warning"  << ','
            << "tx_power_high_alarm"   << ','
            << "tx_power_low_alarm"    << ','
            << "tx_power_high_warning" << ','
            << "tx_power_low_warning"  << ','
            << "tx_bias_high_alarm"    << ','
            << "tx_bias_low_alarm"     << ','
            << "tx_bias_high_warning"  << ','
            << "tx_bias_low_warning";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PEMI_Module_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Module_Status_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Module_Status_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT__(module_status_cap)
        << __FMT__(temp_high_alarm)
        << __FMT__(temp_low_alarm)
        << __FMT__(temp_high_warning)
        << __FMT__(temp_low_warning)
        << __FMT__(voltage_high_alarm)
        << __FMT__(voltage_low_alarm)
        << __FMT__(voltage_high_warning)
        << __FMT__(voltage_low_warning)
        << __FMT__(rx_power_high_alarm)
        << __FMT__(rx_power_low_alarm)
        << __FMT__(rx_power_high_warning)
        << __FMT__(rx_power_low_warning)
        << __FMT__(tx_power_high_alarm)
        << __FMT__(tx_power_low_alarm)
        << __FMT__(tx_power_high_warning)
        << __FMT__(tx_power_low_warning)
        << __FMT__(tx_bias_high_alarm)
        << __FMT__(tx_bias_low_alarm)
        << __FMT__(tx_bias_high_warning)
        << __FMT_LAST__(tx_bias_low_warning);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_Module_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                         export_data_phy_node_t *p_export_data_phy_node,
                                                         struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_module_status_properties =
                                        &areg.regs.pemi.page_data.pemi_Module_Status_Properties;

    IBDIAGNET_RETURN_VOID;
}

PEMI_SNR_Properties_Register::PEMI_SNR_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x11,
                   (const unpack_data_func_t)pemi_SNR_Properties_unpack,
                   ACC_REG_PEMI_SNR_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_SNR_P_NAME,
                   ACC_REG_PEMI_SNR_P_FIELDS_NUM,
                   ACC_REG_PEMI_SNR_P_HEADER)
{
}

void PEMI_SNR_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                      << '#'
            << handler_header            << ','
            << "snr_cap"                 << ','
            << "snr_media_low_alarm"     << ','
            << "snr_media_low_warning"   << ','
            << "snr_host_low_alarm"      << ','
            << "snr_host_low_warning";
#endif
    IBDIAGNET_RETURN_VOID;
}

void PEMI_SNR_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_SNR_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_SNR_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT__(snr_cap)
        << __FMT__(snr_media_low_alarm)
        << __FMT__(snr_media_low_warning)
        << __FMT__(snr_host_low_alarm)
        << __FMT_LAST__(snr_host_low_warning);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_SNR_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                      export_data_phy_node_t *p_export_data_phy_node,
                                                      struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_snr_properties = &areg.regs.pemi.page_data.pemi_SNR_Properties;

    IBDIAGNET_RETURN_VOID;
}


PEMI_Laser_Properties_Register::PEMI_Laser_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x12,
                   (const unpack_data_func_t)pemi_Laser_Monitors_Properties_unpack,
                   ACC_REG_PEMI_LASER_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_LASER_P_NAME,
                   ACC_REG_PEMI_LASER_P_FIELDS_NUM,
                   ACC_REG_PEMI_LASER_P_HEADER)
{
}

void PEMI_Laser_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl                                 << '#'
            << handler_header                       << ','
            << "laser_monitor_cap"                  << ','
            << "laser_age_high_alarm"               << ','
            << "laser_age_high_warning"             << ','
            << "tec_current_high_alarm"             << ','
            << "tec_current_low_alarm"              << ','
            << "tec_current_high_warning"           << ','
            << "tec_current_low_warning"            << ','
            << "laser_frequency_error_high_alarm"   << ','
            << "laser_frequency_error_high_warning" << ','
            << "laser_temp_high_alarm"              << ','
            << "laser_temp_low_alarm"               << ','
            << "laser_temp_high_warning"            << ','
            << "laser_temp_low_warning";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PEMI_Laser_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Laser_Monitors_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Laser_Monitors_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT__(laser_monitor_cap)
        << __FMT__(laser_age_high_alarm)
        << __FMT__(laser_age_high_warning)
        << __FMT__(tec_current_high_alarm)
        << __FMT__(tec_current_low_alarm)
        << __FMT__(tec_current_high_warning)
        << __FMT__(tec_current_low_warning)
        << __FMT__(laser_frequency_error_high_alarm)
        << __FMT__(laser_frequency_error_high_warning)
        << __FMT__(laser_temp_high_alarm)
        << __FMT__(laser_temp_low_alarm)
        << __FMT__(laser_temp_high_warning)
        << __FMT_LAST__(laser_temp_low_warning);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_Laser_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                        export_data_phy_node_t *p_export_data_phy_node,
                                                        struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_laser_monitors_properties =
                        &areg.regs.pemi.page_data.pemi_Laser_Monitors_Properties;

    IBDIAGNET_RETURN_VOID;
}

PEMI_PAM4_Properties_Register::PEMI_PAM4_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x13,
                   (const unpack_data_func_t)pemi_PAM4_Level_Transition_Properties_unpack,
                   ACC_REG_PEMI_PAM4_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_PAM4_P_NAME,
                   ACC_REG_PEMI_PAM4_P_FIELDS_NUM,
                   ACC_REG_PEMI_PAM4_P_HEADER)
{
}

void PEMI_PAM4_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
     IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                       << '#'
            <<  handler_header                            << ','
            << "pam4_level_transition_cap"                << ','
            << "pam4_level_transition_media_low_alarm"    << ','
            << "pam4_level_transition_media_low_warning"  << ','
            << "pam4_level_transition_host_low_alarm"     << ','
            << "pam4_level_transition_host_low_warning";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PAM4_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT__(pam4_level_transition_cap)
        << __FMT__(pam4_level_transition_media_low_alarm)
        << __FMT__(pam4_level_transition_media_low_warning)
        << __FMT__(pam4_level_transition_host_low_alarm)
        << __FMT_LAST__(pam4_level_transition_host_low_warning);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PAM4_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                       export_data_phy_node_t *p_export_data_phy_node,
                                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_pam4_level_transition_properties =
                            &areg.regs.pemi.page_data.pemi_PAM4_Level_Transition_Properties;

    IBDIAGNET_RETURN_VOID;
}

PEMI_PRE_FEC_BER_Properties_Register::PEMI_PRE_FEC_BER_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x14,
                   (const unpack_data_func_t)pemi_Pre_FEC_BER_Properties_unpack,
                   ACC_REG_PEMI_BER_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_BER_P_NAME,
                   ACC_REG_PEMI_BER_P_FIELDS_NUM,
                   ACC_REG_PEMI_BER_P_HEADER)
{
}

void PEMI_PRE_FEC_BER_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                   << '#'
            << handler_header                         << ','
            << "pre_fec_ber_cap"                      << ','
            << "pre_fec_ber_min_high_media_alarm"     << ','
            << "pre_fec_ber_min_high_media_warning"   << ','
            << "pre_fec_ber_min_high_host_alarm"      << ','
            << "pre_fec_ber_min_high_host_warning"    << ','
            << "pre_fec_ber_max_high_media_alarm"     << ','
            << "pre_fec_ber_max_high_media_warning"   << ','
            << "pre_fec_ber_max_high_host_alarm"      << ','
            << "pre_fec_ber_max_high_host_warning"    << ','
            << "pre_fec_ber_avg_high_media_alarm"     << ','
            << "pre_fec_ber_avg_high_media_warning"   << ','
            << "pre_fec_ber_avg_high_host_alarm"      << ','
            << "pre_fec_ber_avg_high_host_warning";
#endif
    IBDIAGNET_RETURN_VOID;
}
void PEMI_PRE_FEC_BER_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT__(pre_fec_ber_cap)
        << __FMT__(pre_fec_ber_min_high_media_alarm)
        << __FMT__(pre_fec_ber_min_high_media_warning)
        << __FMT__(pre_fec_ber_min_high_host_alarm)
        << __FMT__(pre_fec_ber_min_high_host_warning)
        << __FMT__(pre_fec_ber_max_high_media_alarm)
        << __FMT__(pre_fec_ber_max_high_media_warning)
        << __FMT__(pre_fec_ber_max_high_host_alarm)
        << __FMT__(pre_fec_ber_max_high_host_warning)
        << __FMT__(pre_fec_ber_avg_high_media_alarm)
        << __FMT__(pre_fec_ber_avg_high_media_warning)
        << __FMT__(pre_fec_ber_avg_high_host_alarm)
        << __FMT_LAST__(pre_fec_ber_avg_high_host_warning);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_PRE_FEC_BER_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                              export_data_phy_node_t *p_export_data_phy_node,
                                                              struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_pre_fec_ber_properties =
                                        &areg.regs.pemi.page_data.pemi_Pre_FEC_BER_Properties;

    IBDIAGNET_RETURN_VOID;
}

PEMI_FERC_Properties_Register::PEMI_FERC_Properties_Register(PhyDiag * phy_diag)
    : PEMIRegister(phy_diag,
                   0x15,
                   (const unpack_data_func_t)pemi_FERC_Properties_unpack,
                   ACC_REG_PEMI_FERC_P_INTERNAL_SECTION_NAME,
                   ACC_REG_PEMI_FERC_P_NAME,
                   ACC_REG_PEMI_FERC_P_FIELDS_NUM,
                   ACC_REG_PEMI_FERC_P_HEADER)
{
}

void PEMI_FERC_Properties_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;

    Register::DumpRegisterHeader(sstream, handler_header);

#ifdef SHOW_PHY_HEADERS
    sstream << endl            << '#'
            << handler_header  << ','
            << "ferc_cap";
#endif

    IBDIAGNET_RETURN_VOID;
}

void PEMI_FERC_Properties_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +areg.regs.pemi.page_data.pemi_FERC_Properties.x << ','
    #define __FMT_LAST__( x )   "0x" << +areg.regs.pemi.page_data.pemi_FERC_Properties.x << std::endl

    sstream
        << std::hex
        << __FMT_LAST__(ferc_cap);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void PEMI_FERC_Properties_Register::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                                                       export_data_phy_node_t *p_export_data_phy_node,
                                                       struct acc_reg_data &areg, AccRegKey *p_key)
{
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_pemi_ferc_properties =
                            &areg.regs.pemi.page_data.pemi_FERC_Properties;

    IBDIAGNET_RETURN_VOID;
}


//*****************************************
//***     MPCNT Register functions      ***
//*****************************************

int MPCNTRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer) {
    IBDIAGNET_ENTER;

    mpcnt_reg *reg = (mpcnt_reg *) data_to_unpack;
    mpcnt_reg_unpack(reg, unpacked_buffer);

    mpcnt_RAW_page_data raw = reg->page_data.raw;
    unpack_data_func(&reg->page_data, raw.data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int MPCNTRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    const AccRegKeyDPN* key = dynamic_cast<const AccRegKeyDPN*>(p_key);

    if (!key)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);

    struct mpcnt_reg mpcnt;
    CLEAR_STRUCT(mpcnt);

    mpcnt.depth         = key->depth;
    mpcnt.pcie_index    = key->pci_idx;
    mpcnt.node          = key->pci_node;
    mpcnt.grp           = this->m_grp;
    mpcnt.clr           = this->m_phy_diag->ShouldResetCounters();

    mpcnt_reg_pack(&mpcnt, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int MPCNTRegister::BuildDB(AccRegHandler* p_handler,
                          list_p_fabric_general_err &phy_errors,
                          ProgressBar *p_progress_bar)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    p_handler->SetHeader(PCI_DPN_HEADER);

    // go over all the mpein MADs and send mpir accordingly
    for (map_akey_areg::iterator mpein_iter = m_mpein_map.begin();
         mpein_iter != m_mpein_map.end(); mpein_iter++) {

        AccRegKeyDPN *p_dpnkey = (AccRegKeyDPN*)mpein_iter->first;

        // get node
        IBNode *p_curr_node =
                m_phy_diag->GetFabric()->getNodeByGuid(p_dpnkey->node_guid);
        if (!p_curr_node) {
            m_phy_diag->SetLastError("DB error - found null node in "
                                     "NodeByName map for key = " U64H_FMT,
                                                p_dpnkey->node_guid);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        AccRegVia_t acc_reg_via = Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        // get port
        IBPort *p_curr_port = NULL;
        for (u_int32_t port_num = 1; port_num < (unsigned)p_curr_node->numPorts + 1; ++port_num) {
            p_curr_port = p_curr_node->getPort((phys_port_t)port_num);

            if (!p_curr_port || !p_curr_port->is_data_worthy())
                continue;
            break; // use the first valid port.
        }

        if (!p_curr_port) {
            m_phy_diag->SetLastError("DB error - failed to find valid port for node %s",
                                     p_curr_node->getName().c_str());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        AccessRegister acc_reg;
        CLEAR_STRUCT(acc_reg);

        AccRegKeyDPN *p_dpnkey_new = new AccRegKeyDPN(p_dpnkey->p_port,
                                                      p_dpnkey->depth,
                                                      p_dpnkey->pci_idx,
                                                      p_dpnkey->pci_node);

        rc = p_handler->SendAccReg(acc_reg_via,
                                   p_curr_node,
                                   0, // port 0
                                   p_curr_port->base_lid,
                                   acc_reg,
                                   p_dpnkey_new,
                                   p_progress_bar);

        if (rc == IBDIAG_ERR_CODE_DB_ERR)
            IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(rc);
}


MPCNT_PCIe_Counters_Register::MPCNT_PCIe_Counters_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map)
    : MPCNTRegister(phy_diag,
                    mpein_map,
                    0x00,
                    (const unpack_data_func_t)pcie_perf_counters_unpack,
                    ACC_REG_MPCNT_PCI_CNT_INTERNAL_SECTION_NAME,
                    ACC_REG_MPCNT_PCI_CNT_NAME,
                    ACC_REG_MPCNT_PCI_CNT_FIELDS_NUM,
                    ACC_REG_MPCNT_PCI_CNT_HEADER)
{
}

void MPCNT_PCIe_Counters_Register::DumpFieldNames(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "life_time_counter_high"             << ','
            << "life_time_counter_low"              << ','
            << "rx_errors"                          << ','
            << "tx_errors"                          << ','
            << "l0_to_recovery_eieos"               << ','
            << "l0_to_recovery_ts"                  << ','
            << "l0_to_recovery_framing"             << ','
            << "l0_to_recovery_retrain"             << ','
            << "crc_error_dllp"                     << ','
            << "crc_error_tlp"                      << ','
            << "tx_overflow_buffer_pkt"             << ','
            << "outbound_stalled_reads"             << ','
            << "outbound_stalled_writes"            << ','
            << "outbound_stalled_reads_events"      << ','
            << "outbound_stalled_writes_events"     << ','
            << "tx_overflow_buffer_marked_pkt"      << ','
            << "effective_ber_magnitude"            << ','
            << "effective_ber_coef"                 << ','
            << "time_since_last_clear_high"         << ','
            << "time_since_last_clear_low";

    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Counters_Register::DumpData(const struct pcie_perf_counters & reg, stringstream &sstream)
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        +reg.x << ','
    #define __FMT_LAST__( x )   +reg.x

    sstream
        << std::dec
        << __FMT__(life_time_counter_high)
        << __FMT__(life_time_counter_low)
        << __FMT__(rx_errors)
        << __FMT__(tx_errors)
        << __FMT__(l0_to_recovery_eieos)
        << __FMT__(l0_to_recovery_ts)
        << __FMT__(l0_to_recovery_framing)
        << __FMT__(l0_to_recovery_retrain)
        << __FMT__(crc_error_dllp)
        << __FMT__(crc_error_tlp)
        << __FMT__(tx_overflow_buffer_pkt)
        << __FMT__(outbound_stalled_reads)
        << __FMT__(outbound_stalled_writes)
        << __FMT__(outbound_stalled_reads_events)
        << __FMT__(outbound_stalled_writes_events)
        << __FMT__(tx_overflow_buffer_marked_pkt)
        << __FMT__(effective_ber_magnitude)
        << __FMT__(effective_ber_coef)
        << __FMT__(time_since_last_clear_high)
        << __FMT_LAST__(time_since_last_clear_low);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Counters_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    DumpData(areg.regs.mpcnt.page_data.pcie_perf_counters, sstream);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Counters_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ',';

    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

MPCNT_PCIe_Lane_Counters_Register::MPCNT_PCIe_Lane_Counters_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map)
    : MPCNTRegister(phy_diag,
                    mpein_map,
                    0x01,
                    (const unpack_data_func_t)pcie_lanes_counters_unpack,
                    ACC_REG_MPCNT_PCI_LCNT_INTERNAL_SECTION_NAME,
                    ACC_REG_MPCNT_PCI_LCNT_NAME,
                    ACC_REG_MPCNT_PCI_LCNT_FIELDS_NUM,
                    ACC_REG_MPCNT_PCI_LCNT_HEADER)
{
}

void MPCNT_PCIe_Lane_Counters_Register::DumpFieldNames(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "life_time_counter_high"             << ','
            << "life_time_counter_low"              << ','
            << "rx_errors"                          << ','
            << "tx_errors"                          << ','
            << "l0_to_recovery_eieos"               << ','
            << "l0_to_recovery_ts"                  << ','
            << "l0_to_recovery_framing"             << ','
            << "l0_to_recovery_retrain"             << ','
            << "crc_error_dllp"                     << ','
            << "crc_error_tlp"                      << ','
            << "tx_overflow_buffer_pkt"             << ','
            << "outbound_stalled_reads"             << ','
            << "outbound_stalled_writes"            << ','
            << "outbound_stalled_reads_events"      << ','
            << "outbound_stalled_writes_events"     << ','
            << "tx_overflow_buffer_marked_pkt"      << ','
            << "effective_ber_magnitude"            << ','
            << "effective_ber_coef"                 << ','
            << "time_since_last_clear_high";


    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Lane_Counters_Register::DumpData(const struct pcie_lanes_counters & reg, stringstream &sstream)
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +reg.x << ','
    #define __FMT_LAST__( x )   "0x" << +reg.x

    sstream
        << std::hex
        << __FMT__(life_time_counter_high)
        << __FMT__(life_time_counter_low)
        << __FMT__(error_counter_lane0)
        << __FMT__(error_counter_lane1)
        << __FMT__(error_counter_lane2)
        << __FMT__(error_counter_lane3)
        << __FMT__(error_counter_lane4)
        << __FMT__(error_counter_lane5)
        << __FMT__(error_counter_lane6)
        << __FMT__(error_counter_lane7)
        << __FMT__(error_counter_lane8)
        << __FMT__(error_counter_lane9)
        << __FMT__(error_counter_lane10)
        << __FMT__(error_counter_lane11)
        << __FMT__(error_counter_lane12)
        << __FMT__(error_counter_lane13)
        << __FMT__(error_counter_lane14)
        << __FMT_LAST__(error_counter_lane15);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Lane_Counters_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    DumpData(areg.regs.mpcnt.page_data.pcie_lanes_counters, sstream);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Lane_Counters_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ',';

    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

MPCNT_PCIe_Timers_Register::MPCNT_PCIe_Timers_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map)
    : MPCNTRegister(phy_diag,
                    mpein_map,
                    0x02,
                    (const unpack_data_func_t)pcie_timers_states_unpack,
                    ACC_REG_MPCNT_PCI_TIMERS_INTERNAL_SECTION_NAME,
                    ACC_REG_MPCNT_PCI_TIMERS_NAME,
                    ACC_REG_MPCNT_PCI_TIMERS_FIELDS_NUM,
                    ACC_REG_MPCNT_PCI_TIMERS_HEADER)
{
}

void MPCNT_PCIe_Timers_Register::DumpFieldNames(stringstream &sstream)
{
    IBDIAGNET_ENTER;

    sstream << "life_time_counter_high"            << ','
            << "life_time_counter_low"             << ','
            << "time_to_boot_image_start"          << ','
            << "time_to_link_image"                << ','
            << "calibration_time"                  << ','
            << "time_to_first_perst"               << ','
            << "time_to_detect_state"              << ','
            << "time_to_l0"                        << ','
            << "time_to_crs_en"                    << ','
            << "time_to_plastic_image_start"       << ','
            << "time_to_iron_image_start"          << ','
            << "perst_handler"                     << ','
            << "times_in_l1"                       << ','
            << "times_in_l23"                      << ','
            << "dl_down"                           << ','
            << "config_cycle1usec"                 << ','
            << "config_cycle2to7usec"              << ','
            << "config_cycle8to15usec"             << ','
            << "config_cycle16to63usec"            << ','
            << "config_cycle64usec"                << ','
            << "correctable_err_msg_sent"          << ','
            << "non_fatal_err_msg_sent"            << ','
            << "fatal_err_msg_sent";

    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Timers_Register::DumpData(const struct pcie_timers_states & reg, stringstream &sstream)
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +reg.x << ','
    #define __FMT_LAST__( x )   "0x" << +reg.x

    sstream
        << std::hex
        << __FMT__(life_time_counter_high)
        << __FMT__(life_time_counter_low)
        << __FMT__(time_to_boot_image_start)
        << __FMT__(time_to_link_image)
        << __FMT__(calibration_time)
        << __FMT__(time_to_first_perst)
        << __FMT__(time_to_detect_state)
        << __FMT__(time_to_l0)
        << __FMT__(time_to_crs_en)
        << __FMT__(time_to_plastic_image_start)
        << __FMT__(time_to_iron_image_start)
        << __FMT__(perst_handler)
        << __FMT__(times_in_l1)
        << __FMT__(times_in_l23)
        << __FMT__(dl_down)
        << __FMT__(config_cycle1usec)
        << __FMT__(config_cycle2to7usec)
        << __FMT__(config_cycle8to15usec)
        << __FMT__(config_cycle16to63usec)
        << __FMT__(config_cycle64usec)
        << __FMT__(correctable_err_msg_sent)
        << __FMT__(non_fatal_err_msg_sent)
        << __FMT_LAST__(fatal_err_msg_sent);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Timers_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    DumpData(areg.regs.mpcnt.page_data.pcie_timers_states, sstream);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Timers_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ',';

    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

MPCNT_PCIe_Phys_Layer_Register::MPCNT_PCIe_Phys_Layer_Register(PhyDiag * phy_diag, map_akey_areg &mpein_map)
    : MPCNTRegister(phy_diag,
                    mpein_map,
                    0x03,
                    (const unpack_data_func_t)pcie_lanes_counters_unpack,
                    ACC_REG_MPCNT_PCI_PHYS_INTERNAL_SECTION_NAME,
                    ACC_REG_MPCNT_PCI_PHYS_NAME,
                    ACC_REG_MPCNT_PCI_PHYS_FIELDS_NUM,
                    ACC_REG_MPCNT_PCI_PHYS_HEADER)
{
}

void MPCNT_PCIe_Phys_Layer_Register::DumpData(const struct pcie_physical_layer & reg, stringstream &sstream)
{
    IBDIAGNET_ENTER;

    ostream_flags_saver saver(sstream);

    #define __FMT__( x )        "0x" << +reg.x << ','
    #define __FMT_LAST__( x )   "0x" << +reg.x

    sstream
        << std::hex
        << __FMT__(eqlz_timeout_ph3)
        << __FMT__(eqlz_timeout_ph2)
        << __FMT__(eqlz_timeout_ph1)
        << __FMT_LAST__(eqlz_timeout_ph0);

    #undef __FMT__
    #undef __FMT_LAST__

    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Phys_Layer_Register::DumpRegisterData(const struct acc_reg_data & areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    DumpData(areg.regs.mpcnt.page_data.pcie_physical_layer, sstream);
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void MPCNT_PCIe_Phys_Layer_Register::DumpRegisterHeader(stringstream &sstream, const string& handler_header)
{
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "eqlz_timeout_ph3" << ','
            << "eqlz_timeout_ph2" << ','
            << "eqlz_timeout_ph1" << ','
            << "eqlz_timeout_ph0";
#endif
    IBDIAGNET_RETURN_VOID;
}




//*****************************************
//***     PPBMP Register functions      ***
//*****************************************
PPBMPRegister::PPBMPRegister(PhyDiag* phy_diag) : Register(phy_diag,
                                                           ACCESS_REGISTER_ID_PPBMP,
                                                           (const unpack_data_func_t)ppbmp_reg_unpack,
                                                           ACC_REG_PPBMP_INTERNAL_SECTION_NAME,
                                                           ACC_REG_PPBMP_NAME,
                                                           ACC_REG_PPBMP_FIELDS_NUM,
                                                           NSB::get(this),
                                                           ACC_REG_PPBMP_FIELDS_HEADER,
                                                           SUPPORT_SW_CA,
                                                           true,
                                                           false,
                                                           VIA_SMP,
                                                           VIA_GMP),
                                                  m_pnat(ACC_REG_PNAT_IB_PORT)
{



}

int PPBMPRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    struct ppbmp_reg ppbmp;
    CLEAR_STRUCT(ppbmp);

    ppbmp.local_port = ((AccRegKeyPortIndex *)p_key)->port_num;
    ppbmp.pnat = this->m_pnat;
    ppbmp.monitor_group = ((AccRegKeyPortIndex *)p_key)->index;

    ppbmp_reg_pack(&ppbmp, data);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPBMPRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct ppbmp_reg &ppbmp = areg.regs.ppbmp;

    double normal_th = ppbmp.monitor_parameters.normal_th_mantissa *
            (1 / pow(10, ppbmp.monitor_parameters.normal_th_exp));

    double warning_th = ppbmp.monitor_parameters.warning_th_mantissa *
            (1 / pow(10, ppbmp.monitor_parameters.warning_th_exp));

    double alarm_th = ppbmp.monitor_parameters.alarm_th_mantissa *
            (1 / pow(10, ppbmp.monitor_parameters.alarm_th_exp));


    ios_base::fmtflags f(sstream.flags());

    sstream << scientific
            << normal_th             << ','
            << warning_th            << ','
            << alarm_th              << ',';

    sstream.flags(f);

    sstream << +ppbmp.time_window    << ','
            << +ppbmp.sampling_rate  << endl;

    IBDIAGNET_RETURN_VOID;
}

void PPBMPRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl             << '#'
            << handler_header   << ","
            << "scientific"     << ','
            << "normal_th"      << ','
            << "warning_th"     << ','
            << "alarm_th"       << ','
            << "time_window"    << ','
            << "sampling_rate";
#endif
    IBDIAGNET_RETURN_VOID;
}




//*****************************************
//***     MRCS Register functions      ***
//*****************************************
MRCSRegister::MRCSRegister(PhyDiag* phy_diag) : Register(phy_diag,
                                                         ACCESS_REGISTER_ID_MRCS,
                                                         (const unpack_data_func_t)mrcs_reg_unpack,
                                                         ACC_REG_MRCS_INTERNAL_SECTION_NAME,
                                                         ACC_REG_MRCS_NAME,
                                                         ACC_REG_MRCS_FIELDS_NUM,
                                                         NSB::get(this),
                                                         ACC_REG_MRCS_FIELDS_HEADER,
                                                         SUPPORT_SW_CA,
                                                         true,
                                                         false,
                                                         VIA_SMP,
                                                         VIA_GMP)
{
}

int MRCSRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->serdes_version != SerDes_7nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void MRCSRegister::DumpRegisterData(const struct acc_reg_data &areg, stringstream &sstream, const AccRegKey &key) const
{
    IBDIAGNET_ENTER;

    const struct mrcs_reg &mrcs = areg.regs.mrcs;

    sstream << +mrcs.ref_clock_stat[0].measured_freq    << ','
            << +mrcs.ref_clock_stat[0].min_freq         << ','
            << +mrcs.ref_clock_stat[0].max_delta_freq   << ','
            << +mrcs.ref_clock_stat[0].max_freq         << ','
            << +mrcs.ref_clock_stat[1].measured_freq    << ','
            << +mrcs.ref_clock_stat[1].min_freq         << ','
            << +mrcs.ref_clock_stat[1].max_delta_freq   << ','
            << +mrcs.ref_clock_stat[1].max_freq         << endl;

    IBDIAGNET_RETURN_VOID;
}

void MRCSRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                       << '#'
            << handler_header             << ","
            << "mrcs[0].measured_freq"    << ','
            << "mrcs[0].min_freq"         << ','
            << "mrcs[0].max_delta_freq"   << ','
            << "mrcs[0].max_freq"         << ','
            << "mrcs[1].measured_freq"    << ','
            << "mrcs[1].min_freq"         << ','
            << "mrcs[1].max_delta_freq"   << ','
            << "mrcs[1].max_freq";
#endif
    IBDIAGNET_RETURN_VOID;
}

PPSLDRegister::PPSLDRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_PPSLD,
               (const unpack_data_func_t)ppsld_reg_unpack,
               ACC_REG_PPSLD_INTERNAL_SECTION_NAME,
               ACC_REG_PPSLD_NAME,
               ACC_REG_PPSLD_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_PPSLD_HEADER,
               SUPPORT_SW,
               true,
               false,
               VIA_GMP),
            m_pnat(ACC_REG_PNAT_IB_PORT)
{

}

int PPSLDRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct ppsld_reg reg;
    CLEAR_STRUCT(reg);

    reg.pnat        = this->m_pnat;

    int rc = FillLocalPortAndPlaneInd(p_key, &reg.local_port, nullptr);

    if (rc)
        return rc;

    ppsld_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPSLDRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                    << '#'
            << handler_header          << ','
            << "fw_mode"               << ','
            << "trigger"               << ','
            << "ehld"                  << ','
            << "fw_desire_window_time" << ','
            << "fw_idle_window_time"   ;

#endif
    IBDIAGNET_RETURN_VOID;
}

void PPSLDRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const ppsld_reg& reg = areg.regs.ppsld;
    sstream << +reg.fw_mode               << ','
            << +reg.trigger               << ','
            << +reg.ehld                  << ','
            << +reg.fw_desire_window_time << ','
            << +reg.fw_idle_window_time   << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void PPSLDRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppsld =
                                &areg.regs.ppsld;

    IBDIAGNET_RETURN_VOID;

}

PPSLCRegister::PPSLCRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_PPSLC, 
               (const unpack_data_func_t)ppslc_reg_unpack,
               ACC_REG_PPSLC_INTERNAL_SECTION_NAME,
               ACC_REG_PPSLC_NAME,
               ACC_REG_PPSLC_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_PPSLC_HEADER,
               SUPPORT_SW_CA,
               true,
               false,
               VIA_GMP),
            m_pnat(ACC_REG_PNAT_IB_PORT)
{

}

int PPSLCRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->type == IB_SW_NODE && p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);
    
    if (p_node->type == IB_CA_NODE &&  p_node->ext_type != IB_GPU_NODE)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct ppslc_reg reg;
    CLEAR_STRUCT(reg);

    reg.pnat        = this->m_pnat;

    int rc = FillLocalPortAndPlaneInd(p_key, &reg.local_port, nullptr);

    if (rc)
        return rc;

    ppslc_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPSLCRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                  << '#'
            << handler_header        << ','
            << "l1_req_en"           << ','
            << "l1_fw_cap_adv"       << ','
            << "l1_cap_adv"          << ','
            << "l1_fw_req_en"        << ','
            << "pred_algo_en"        << ','
            << "ignore_pred_pm"      << ','
            << "hp_queues_bitmap"    << ','
            << "l1_hw_active_time"   << ','
            << "l1_hw_inactive_time" << ','
            << "qem[0]"              << ','
            << "qem[1]"              << ','
            << "qem[2]"              << ','
            << "qem[3]"              << ','
            << "qem[4]"              << ','
            << "qem[5]"              << ','
            << "qem[6]"              << ','
            << "qem[7]"              << ','
            << "qem[8]"              << ','
            << "qem[9]"              << ','
            << "qem[10]"             << ','
            << "qem[11]"             << ','
            << "qem[12]"             << ','
            << "qem[13]"             << ','
            << "qem[14]"             << ','
            << "qem[15]"             << ','
            << "qem[16]"             << ','
            << "qem[17]"             << ','
            << "qem[18]"             << ','
            << "qem[19]"             ;

#endif
    IBDIAGNET_RETURN_VOID;
}

void PPSLCRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const ppslc_reg& reg = areg.regs.ppslc;
    sstream << +reg.l1_req_en           << ','
            << +reg.l1_fw_cap_adv       << ','
            << +reg.l1_cap_adv          << ','
            << +reg.l1_fw_req_en        << ','
            << +reg.pred_algo_en        << ','
            << +reg.ignore_pred_pm      << ','
            << +reg.hp_queues_bitmap    << ','
            << +reg.l1_hw_active_time   << ','
            << +reg.l1_hw_inactive_time << ',';
    sstream << +reg.qem[0];
    for (size_t i = 1; i < ARRAY_SIZE(ppslc_reg::qem); ++i) {
        sstream << ',' << +reg.qem[i];
    }
    sstream << endl;
    IBDIAGNET_RETURN_VOID;
}

void PPSLCRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppslc =
                                &areg.regs.ppslc;

    IBDIAGNET_RETURN_VOID;

}

MLPCRegister::MLPCRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_MLPC,
               (const unpack_data_func_t)mlpc_reg_unpack,
               ACC_REG_MLPC_INTERNAL_SECTION_NAME,
               ACC_REG_MLPC_NAME,
               ACC_REG_MLPC_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_MLPC_HEADER,
               SUPPORT_SW_CA,
               true,
               false,
               VIA_GMP),
            m_pnat(ACC_REG_PNAT_IB_PORT)
{

}

int MLPCRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
   IBDIAGNET_ENTER;

    if (p_node->type == IB_SW_NODE && p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);
    
    if (p_node->type == IB_CA_NODE &&  p_node->ext_type != IB_GPU_NODE)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct mlpc_reg reg;
    CLEAR_STRUCT(reg);

    reg.pnat        = this->m_pnat;
    reg.counter_rst = this->m_phy_diag->ShouldResetCounters();

    int rc = FillLocalPortAndPlaneInd(p_key, &reg.local_port, nullptr);

    if (rc)
        return rc;

    mlpc_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MLPCRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl               << '#'
            << handler_header     << ','
            << "counter_en"       << ','
            << "counter_rst"      << ','
            << "stop_at_ff"       << ','
            << "cnt_64bit"        << ','
            << "force_count_mask" << ','
            << "cnt_type[0]"      << ','
            << "cnt_type[1]"      << ','
            << "cnt_type[2]"      << ','
            << "cnt_type[3]"      << ','
            << "cnt_type[4]"      << ','
            << "cnt_type[5]"      << ','
            << "cnt_type[6]"      << ','
            << "cnt_type[7]"      << ','
            << "cnt_val[0]"       << ','
            << "cnt_val[1]"       << ','
            << "cnt_val[2]"       << ','
            << "cnt_val[3]"       << ','
            << "cnt_val[4]"       << ','
            << "cnt_val[5]"       << ','
            << "cnt_val[6]"       << ','
            << "cnt_val[7]"       ;

#endif
    IBDIAGNET_RETURN_VOID;
}

void MLPCRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const mlpc_reg& reg = areg.regs.mlpc;
    sstream << +reg.counter_en       << ','
            << +reg.counter_rst      << ','
            << +reg.stop_at_ff       << ','
            << +reg.cnt_64bit        << ','
            << +reg.force_count_mask << ',';
    for (size_t i = 0; i < ARRAY_SIZE(mlpc_reg::cnt_type); ++i) {
        sstream << +reg.cnt_type[i] << ',';
    }
    sstream << +reg.cnt_val[0];
    for (size_t i = 1; i < ARRAY_SIZE(mlpc_reg::cnt_val); ++i) {
        sstream << ',' << +reg.cnt_val[i];
    }
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void MLPCRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_mlpc =
                                &areg.regs.mlpc;

    IBDIAGNET_RETURN_VOID;

}

MPPGSRegister::MPPGSRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_MPPGS,
               (const unpack_data_func_t)mppgs_reg_unpack,
               ACC_REG_MPPGS_INTERNAL_SECTION_NAME,
               ACC_REG_MPPGS_NAME,
               ACC_REG_MPPGS_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_MPPGS_HEADER,
               SUPPORT_SW,
               true,
               false,
               VIA_GMP)
{

}

int MPPGSRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct mppgs_reg reg;
    CLEAR_STRUCT(reg);

    reg.op = this->m_phy_diag->ShouldResetCounters();

    mppgs_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MPPGSRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                  << '#'
            << handler_header        << ','
            << "op"                  << ','
            << "critical_power_cnt"  << ','
            << "short_win_power_avg" << ','
            << "long_win_power_avg"  ;

#endif
    IBDIAGNET_RETURN_VOID;
}

void MPPGSRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const mppgs_reg& reg = areg.regs.mppgs;
    sstream << +reg.op                  << ','
            << +reg.critical_power_cnt  << ','
            << +reg.short_win_power_avg << ','
            << +reg.long_win_power_avg  << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void MPPGSRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_mppgs =
                                &areg.regs.mppgs;

    IBDIAGNET_RETURN_VOID;

}

MPPDRRegister::MPPDRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_MPPDR,
               (const unpack_data_func_t)mppdr_reg_unpack,
               ACC_REG_MPPDR_INTERNAL_SECTION_NAME,
               ACC_REG_MPPDR_NAME,
               ACC_REG_MPPDR_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_MPPDR_HEADER,
               SUPPORT_SW,
               true,
               false,
               VIA_GMP) {

}

int MPPDRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MPPDRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ',';
    sstream << "vr_power[0]";
    for (size_t i = 1; i < ARRAY_SIZE(mppdr_reg::vr_power); ++i) {
        sstream << ',' << "vr_power["<< i << "]";
    }
#endif
    IBDIAGNET_RETURN_VOID;
}

void MPPDRRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const mppdr_reg& reg = areg.regs.mppdr;
    sstream << +reg.vr_power[0];
    for (size_t i = 1; i < ARRAY_SIZE(mppdr_reg::vr_power); ++i) {
        sstream << ',' << +reg.vr_power[i];
    }
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void MPPDRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_mppdr =
                                &areg.regs.mppdr;

    IBDIAGNET_RETURN_VOID;

}


MPPGHRegister::MPPGHRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_MPPGH,
               (const unpack_data_func_t)mppgh_reg_unpack,
               ACC_REG_MPPGH_INTERNAL_SECTION_NAME,
               ACC_REG_MPPGH_NAME,
               ACC_REG_MPPGH_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_MPPGH_HEADER,
               SUPPORT_SW,
               true,
               false,
               VIA_GMP)
{

}

int MPPGHRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct mppgh_reg reg;
    CLEAR_STRUCT(reg);

    reg.op = this->m_phy_diag->ShouldResetCounters();

    mppgh_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MPPGHRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl             << '#'
            << handler_header   << ','
            << "op"             << ',';
    sstream << "shaper_bin[0]";
    for (size_t i = 1; i < ARRAY_SIZE(mppgh_reg::shaper_bin); ++i) {
        sstream << ',' << "shaper_bin[" << i << "]";
    }

#endif
    IBDIAGNET_RETURN_VOID;
}

void MPPGHRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const mppgh_reg& reg = areg.regs.mppgh;
    sstream << +reg.op << ',';
    sstream << +reg.shaper_bin[0];
    for (size_t i = 1; i < ARRAY_SIZE(mppgh_reg::shaper_bin); ++i) {
        sstream << ',' << +reg.shaper_bin[i];
    }
    sstream << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void MPPGHRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_mppgh =
                                &areg.regs.mppgh;

    IBDIAGNET_RETURN_VOID;

}

PPSPCRegister::PPSPCRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_PPSPC,
               (const unpack_data_func_t)ppspc_reg_unpack,
               ACC_REG_PPSPC_INTERNAL_SECTION_NAME,
               ACC_REG_PPSPC_NAME,
               ACC_REG_PPSPC_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_PPSPC_HEADER,
               SUPPORT_SW,
               true,
               false,
               VIA_GMP)
            ,m_pnat(ACC_REG_PNAT_IB_PORT)
{

}

int PPSPCRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct ppspc_reg reg;
    CLEAR_STRUCT(reg);

    reg.pnat        = this->m_pnat;

    int rc = FillLocalPortAndPlaneInd(p_key, &reg.local_port, nullptr);

    if (rc)
        return rc;

    ppspc_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPSPCRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                        << '#'
            << handler_header              << ','
            << "predictor_inactive_window" << ','
            << "windows_num"               << ',';
    sstream << "win_thr[0]";
    for (size_t i = 1; i < ARRAY_SIZE(ppspc_reg::win_thr); ++i) {
        sstream << ',' << "win_thr[" << i << "]";
    }

#endif
    IBDIAGNET_RETURN_VOID;
}

void PPSPCRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const ppspc_reg& reg = areg.regs.ppspc;
    sstream << +reg.predictor_inactive_window << ','
            << +reg.windows_num               << ',';
    sstream << +reg.win_thr[0];
    for (size_t i = 1; i < ARRAY_SIZE(ppspc_reg::win_thr); ++i) {
        sstream << ',' << +reg.win_thr[i];
    }
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPSPCRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppspc =
                                &areg.regs.ppspc;

    IBDIAGNET_RETURN_VOID;
}

PPSPIRegister::PPSPIRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_PPSPI,
               (const unpack_data_func_t)ppspi_reg_unpack,
               ACC_REG_PPSPI_INTERNAL_SECTION_NAME,
               ACC_REG_PPSPI_NAME,
               ACC_REG_PPSPI_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_PPSPI_HEADER,
               SUPPORT_SW,
               true,
               false,
               VIA_GMP)
            ,m_pnat(ACC_REG_PNAT_IB_PORT)
{

}

int PPSPIRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
   IBDIAGNET_ENTER;

    if (p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct ppspi_reg reg;
    CLEAR_STRUCT(reg);

    reg.pnat        = this->m_pnat;
    reg.dbg_cc      = this->m_phy_diag->ShouldResetCounters();

    int rc = FillLocalPortAndPlaneInd(p_key, &reg.local_port, nullptr);

    if (rc)
        return rc;

    ppspi_reg_pack(&reg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPSPIRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ','
            << "dbg_cnt_en"             << ','
            << "dbg_cc"                 << ','
            << "pred_win"               << ','
            << "pred_mode"              << ','
            << "too_long_pred_counter"  << ','
            << "too_short_pred_counter" << ','
            << "min_win_too_high"       << ','
            << "max_win_too_low"        << ','
            << "rcnt_idle_win_time_0"   << ','
            << "rcnt_idle_win_time_1"   << ','
            << "rcnt_idle_win_time_2"   << ','
            << "rcnt_idle_win_time_3"   << ','
            << "rcnt_idle_win_time_4"   << ','
            << "pred_on_counter"        << ','
            << "pred_off_counter"       << ','
            << "rcnt_idle_win_idx_0"    << ','
            << "rcnt_idle_win_idx_1"    << ','
            << "rcnt_idle_win_idx_2"    << ','
            << "rcnt_idle_win_idx_3"    << ','
            << "rcnt_idle_win_idx_4"    << ',';
    for (size_t i = 0; i < ARRAY_SIZE(ppspi_reg::st_scr); ++i) {
        sstream << "st_scr[" << i << "]" << ',';
    }
    sstream << "cyc_scr_2"              << ','
            << "cyc_scr_3"              << ','
            << "cyc_scr_4"              << ','
            << "cyc_scr_5"              ;

#endif
    IBDIAGNET_RETURN_VOID;
}

void PPSPIRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const ppspi_reg& reg = areg.regs.ppspi;
    sstream << +reg.dbg_cnt_en             << ','
            << +reg.dbg_cc                 << ','
            << +reg.pred_win               << ','
            << +reg.pred_mode              << ','
            << +reg.too_long_pred_counter  << ','
            << +reg.too_short_pred_counter << ','
            << +reg.min_win_too_high       << ','
            << +reg.max_win_too_low        << ','
            << +reg.rcnt_idle_win_time_0   << ','
            << +reg.rcnt_idle_win_time_1   << ','
            << +reg.rcnt_idle_win_time_2   << ','
            << +reg.rcnt_idle_win_time_3   << ','
            << +reg.rcnt_idle_win_time_4   << ','
            << +reg.pred_on_counter        << ','
            << +reg.pred_off_counter       << ','
            << +reg.rcnt_idle_win_idx_0    << ','
            << +reg.rcnt_idle_win_idx_1    << ','
            << +reg.rcnt_idle_win_idx_2    << ','
            << +reg.rcnt_idle_win_idx_3    << ','
            << +reg.rcnt_idle_win_idx_4    << ',';     
    for (size_t i = 0; i < ARRAY_SIZE(ppspi_reg::st_scr); ++i) {
        sstream << +reg.st_scr[i] << ',';
    }
    sstream << +reg.cyc_scr_2              << ','
            << +reg.cyc_scr_3              << ','
            << +reg.cyc_scr_4              << ','
            << +reg.cyc_scr_5              << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void PPSPIRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppspi =
                                &areg.regs.ppspi;

    IBDIAGNET_RETURN_VOID;

}

PPSPGCRegister::PPSPGCRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_PPSPGC,
               (const unpack_data_func_t)ppspgc_reg_unpack,
               ACC_REG_PPSPGC_INTERNAL_SECTION_NAME,
               ACC_REG_PPSPGC_NAME,
               ACC_REG_PPSPGC_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_PPSPGC_HEADER,
               SUPPORT_SW,
               true,
               false,
               VIA_GMP)

{

}

int PPSPGCRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void PPSPGCRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                             << '#'
            << handler_header                   << ','
            << "min_score_en"                   << ','
            << "min_score_dis"                  << ','
            << "max_score"                      << ','
            << "sample_rate"                    << ','
            << "too_long_pred_err_thr"          << ','
            << "l1_hw_inactive_cycles_pred_dis" << ','
            << "too_long_pred_err_iter_cnt"     ;

#endif
    IBDIAGNET_RETURN_VOID;
}

void PPSPGCRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const ppspgc_reg& reg = areg.regs.ppspgc;
    sstream << +reg.min_score_en                   << ','
            << +reg.min_score_dis                  << ','
            << +reg.max_score                      << ','
            << +reg.sample_rate                    << ','
            << +reg.too_long_pred_err_thr          << ','
            << +reg.l1_hw_inactive_cycles_pred_dis << ','
            << +reg.too_long_pred_err_iter_cnt     << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void PPSPGCRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppspgc =
                                &areg.regs.ppspgc;

    IBDIAGNET_RETURN_VOID;
}

MPPGCRRegister::MPPGCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_MPPGCR,
               (const unpack_data_func_t)mppgcr_reg_unpack,
               ACC_REG_MPPGCR_INTERNAL_SECTION_NAME,
               ACC_REG_MPPGCR_NAME,
               ACC_REG_MPPGCR_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_MPPGCR_HEADER,
               SUPPORT_SW,
               true,
               false,
               VIA_GMP)

{

}

int MPPGCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    if (p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int MPPGCRRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer)
{
    IBDIAGNET_ENTER;

    mppgcr_reg *reg = (mppgcr_reg *) data_to_unpack;
    unpack_data_func(data_to_unpack, unpacked_buffer);

    const AccRegKeyNode *key = dynamic_cast<const AccRegKeyNode*>(p_key);

    if (key) {
        PHYNodeData* data = this->m_phy_diag->GetPHYNodeData(key->node_guid);
        if (data) {
            data->curr_power_profile_id = reg->curr_profile_id;
        }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MPPGCRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                     << '#'
            << handler_header           << ','
            << "e"                      << ','
            << "next_profile_id"        << ','
            << "curr_profile_id"        << ','
            << "idle_power_a_factor"    << ','
            << "idle_power_b_factor"    << ','
            << "port_power_factor"      << ','
            << "active_power_factor"    << ','
            << "port_peak_c_dyn"        << ','
            << "shaper_adj_num_of_iter" << ','
            << "sampling_time"          << ','
            << "stc_efficiency"         << ','
            << "min_shaper"             ;

#endif
    IBDIAGNET_RETURN_VOID;
}

void MPPGCRRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const mppgcr_reg& reg = areg.regs.mppgcr;
    sstream << +reg.e                      << ','
            << +reg.next_profile_id        << ','
            << +reg.curr_profile_id        << ','
            << +reg.idle_power_a_factor    << ','
            << +reg.idle_power_b_factor    << ','
            << +reg.port_power_factor      << ','
            << +reg.active_power_factor    << ','
            << +reg.port_peak_c_dyn        << ','
            << +reg.shaper_adj_num_of_iter << ','
            << +reg.sampling_time          << ','
            << +reg.stc_efficiency         << ','
            << +reg.min_shaper             << std::endl;

    IBDIAGNET_RETURN_VOID;
}

void MPPGCRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_mppgcr =
                                &areg.regs.mppgcr;

    IBDIAGNET_RETURN_VOID;

}

MPPCRRegister::MPPCRRegister(PhyDiag * phy_diag)
    : Register(phy_diag,
               ACCESS_REGISTER_ID_MPPCR,
               (const unpack_data_func_t)mppcr_reg_unpack,
               ACC_REG_MPPCR_INTERNAL_SECTION_NAME,
               ACC_REG_MPPCR_NAME,
               ACC_REG_MPPCR_FIELDS_NUM,
               NSB::get(this),
               ACC_REG_MPPCR_HEADER,
               SUPPORT_SW,
               true,
               false,
               VIA_GMP)

{

}

int MPPCRRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
   IBDIAGNET_ENTER;

    if (p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct mppcr_reg reg;
    CLEAR_STRUCT(reg);

    PHYNodeData* phy_node_data = this->m_phy_diag->GetPHYNodeData(p_key->node_guid);
    if (phy_node_data) {
        reg.profile_id = phy_node_data->curr_power_profile_id;
    }

    mppcr_reg_pack(&reg, data);

   IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void MPPCRRegister::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                           << '#'
            << handler_header                                 << ','
            << "v"                                            << ','
            << "profile_id"                                   << ','
            << "param_b[0]"                                   << ','
            << "param_b[1]"                                   << ','
            << "param_b[2]"                                   << ','
            << "param_b[3]"                                   << ','
            << "profile_name"                                 << ','
            << "profile_config[0].v"                          << ','
            << "profile_config[0].power_allocation"           << ','
            << "profile_config[0].max_integral"               << ','
            << "profile_config[0].avg_pwr_num_of_sampling"    << ','
            << "profile_config[0].Kp_factor"                  << ','
            << "profile_config[0].Ki_factor"                  << ','
            << "profile_config[0].Kd_factor"                  << ','
            << "profile_config[0].param_a[0]"                 << ','
            << "profile_config[0].param_a[1]"                 << ','
            << "profile_config[0].param_a[2]"                 << ','
            << "profile_config[0].param_a[3]"                 << ','
            << "profile_config[0].param_b"                    << ','
            << "profile_config[0].param_c"                    << ','
            << "profile_config[0].param_d"                    << ','
            << "profile_config[1].v"                          << ','
            << "profile_config[1].power_allocation"           << ','
            << "profile_config[1].max_integral"               << ','
            << "profile_config[1].avg_pwr_num_of_sampling"    << ','
            << "profile_config[1].Kp_factor"                  << ','
            << "profile_config[1].Ki_factor"                  << ','
            << "profile_config[1].Kd_factor"                  << ','
            << "profile_config[1].param_a[0]"                 << ','
            << "profile_config[1].param_a[1]"                 << ','
            << "profile_config[1].param_a[2]"                 << ','
            << "profile_config[1].param_a[3]"                 << ','
            << "profile_config[1].param_b"                    << ','
            << "profile_config[1].param_c"                    << ','
            << "profile_config[1].param_d"                    ;
#endif
    IBDIAGNET_RETURN_VOID;
}

void MPPCRRegister::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    const mppcr_reg& reg = areg.regs.mppcr;
    sstream << +reg.v                                  << ','
            << +reg.profile_id                         << ','
            << +reg.param_b[0]                         << ','
            << +reg.param_b[1]                         << ','
            << +reg.param_b[2]                         << ','
            << +reg.param_b[3]                         << ','
            << QUOTED(DescToCsvDesc(reg.profile_name)) << ',';
    DumpProfileConfig(reg.profile_config[0], sstream);
    sstream << ',';
    DumpProfileConfig(reg.profile_config[1], sstream);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void MPPCRRegister::DumpProfileConfig(const power_profile_config& profile_conf, stringstream &sstream) const {
    sstream << +profile_conf.v                          << ','
            << +profile_conf.power_allocation           << ','
            << +profile_conf.max_integral               << ','
            << +profile_conf.avg_pwr_num_of_sampling    << ','
            << +profile_conf.Kp_factor                  << ','
            << +profile_conf.Ki_factor                  << ','
            << +profile_conf.Kd_factor                  << ','
            << +profile_conf.param_a[0]                 << ','
            << +profile_conf.param_a[1]                 << ','
            << +profile_conf.param_a[2]                 << ','
            << +profile_conf.param_a[3]                 << ','
            << +profile_conf.param_b                    << ','
            << +profile_conf.param_c                    << ','
            << +profile_conf.param_d                    ;
}

void MPPCRRegister::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_mppcr =
                                &areg.regs.mppcr;

    IBDIAGNET_RETURN_VOID;

}

int PPSLGRegister::PackData(AccRegKey *p_key, uint8_t *data, const IBNode *p_node)
{
    IBDIAGNET_ENTER;
    if (p_node->type == IB_SW_NODE && p_node->serdes_version < SerDes_5nm)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);
    
    if (p_node->type == IB_CA_NODE &&  p_node->ext_type != IB_GPU_NODE)
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DISABLED);

    struct ppslg_reg ppslg;
    CLEAR_STRUCT(ppslg);

    ppslg.pnat = m_pnat;
    ppslg.page_select = m_page_select;

    int rc = FillLocalPortAndPlaneInd(p_key, &ppslg.local_port, nullptr);

    if (rc)
        return rc;

    ppslg_reg_pack(&ppslg, data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int PPSLGRegister::UnpackData(const AccRegKey *p_key, void *data_to_unpack, u_int8_t *unpacked_buffer)
{
    IBDIAGNET_ENTER;
    ppslg_reg *reg = (ppslg_reg *) data_to_unpack;
    ppslg_reg_unpack(reg, unpacked_buffer);

    ppslg_RAW_page_data raw = reg->page_data.raw;
    unpack_data_func(&reg->page_data, raw.data);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

PPSLG_L1_Capabilities_And_Status::PPSLG_L1_Capabilities_And_Status(PhyDiag * phy_diag)
    : PPSLGRegister(phy_diag,
                    0x0,
                    (const unpack_data_func_t)ppslg_l1_capabilities_and_status_unpack,
                    ACC_REG_PPSLG_L1_CAPABILITIES_AND_STATUS_INTERNAL_SECTION_NAME,
                    ACC_REG_PPSLG_L1_CAPABILITIES_AND_STATUS_NAME,
                    ACC_REG_PPSLG_L1_CAPABILITIES_AND_STATUS_FIELDS_NUM,
                    ACC_REG_PPSLG_L1_CAPABILITIES_AND_STATUS_HEADER) {

}

void PPSLG_L1_Capabilities_And_Status::DumpFieldNames(stringstream &sstream) {
    sstream << "l1_hw_local_cap"                <<','
            << "l1_hw_local_req"                <<','
            << "l1_hw_remote_cap"               <<','
            << "l1_hw_remote_req"               <<','
            << "l1_hw_neg_res"                  <<','
            << "l1_force_local_cap"             <<','
            << "l1_force_local_req"             <<','
            << "l1_force_remote_cap"            <<','
            << "l1_force_remote_req"            <<','
            << "l1_remote_peq_duration_req"     <<','
            << "l1_remote_peq_recal_period_req" <<','
            << "l1_remote_min_linkup_time_req"  <<','
            << "l1_peq_duration_res"            <<','
            << "l1_peq_recal_period_res"        <<','
            << "l1_min_linkup_time_res"         ;

}

void PPSLG_L1_Capabilities_And_Status::DumpData(const ppslg_l1_capabilities_and_status& reg, stringstream &sstream) const {
    sstream << +reg.l1_hw_local_cap <<','
            << +reg.l1_hw_local_req <<','
            << +reg.l1_hw_remote_cap <<','
            << +reg.l1_hw_remote_req <<','
            << +reg.l1_hw_neg_res <<','
            << +reg.l1_force_local_cap <<','
            << +reg.l1_force_local_req <<','
            << +reg.l1_force_remote_cap <<','
            << +reg.l1_force_remote_req <<','
            << +reg.l1_remote_peq_duration_req <<','
            << +reg.l1_remote_peq_recal_period_req <<','
            << +reg.l1_remote_min_linkup_time_req <<','
            << +reg.l1_peq_duration_res <<','
            << +reg.l1_peq_recal_period_res <<','
            << +reg.l1_min_linkup_time_res ;

}

void PPSLG_L1_Capabilities_And_Status::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl           << '#'
            << handler_header << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPSLG_L1_Capabilities_And_Status::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    DumpData(areg.regs.ppslg.page_data.ppslg_l1_capabilities_and_status, sstream);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPSLG_L1_Capabilities_And_Status::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppslg_l1_capabilities_and_status =
                                &areg.regs.ppslg.page_data.ppslg_l1_capabilities_and_status;

    IBDIAGNET_RETURN_VOID;
}

PPSLG_L1_Configuration::PPSLG_L1_Configuration(PhyDiag * phy_diag)
    : PPSLGRegister(phy_diag,
                    0x1,
                    (const unpack_data_func_t)ppslg_l1_configuration_unpack,
                    ACC_REG_PPSLG_L1_CONFIGURATION_INTERNAL_SECTION_NAME,
                    ACC_REG_PPSLG_L1_CONFIGURATION_NAME,
                    ACC_REG_PPSLG_L1_CONFIGURATION_FIELDS_NUM,
                    ACC_REG_PPSLG_L1_CONFIGURATION_HEADER) {

}

void PPSLG_L1_Configuration::DumpFieldNames(stringstream &sstream) {
    sstream << "l1_local_peq_duration_req"     <<','
            << "l1_local_peq_recal_period_req" <<','
            << "l1_local_min_linkup_time_req"  ;

}

void PPSLG_L1_Configuration::DumpData(const ppslg_l1_configuration& reg, stringstream &sstream) const {
    sstream << +reg.l1_local_peq_duration_req <<','
            << +reg.l1_local_peq_recal_period_req <<','
            << +reg.l1_local_min_linkup_time_req ;

}

void PPSLG_L1_Configuration::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl           << '#'
            << handler_header << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPSLG_L1_Configuration::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    DumpData(areg.regs.ppslg.page_data.ppslg_l1_configuration, sstream);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPSLG_L1_Configuration::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppslg_l1_configuration =
                                &areg.regs.ppslg.page_data.ppslg_l1_configuration;

    IBDIAGNET_RETURN_VOID;

}

PPSLG_L1_Debug::PPSLG_L1_Debug(PhyDiag * phy_diag)
    : PPSLGRegister(phy_diag,
                    0x2,
                    (const unpack_data_func_t)ppslg_l1_debug_page_unpack,
                    ACC_REG_PPSLG_L1_DEBUG_PAGE_INTERNAL_SECTION_NAME,
                    ACC_REG_PPSLG_L1_DEBUG_PAGE_NAME,
                    ACC_REG_PPSLG_L1_DEBUG_PAGE_FIELDS_NUM,
                    ACC_REG_PPSLG_L1_DEBUG_PAGE_HEADER) {

}

void PPSLG_L1_Debug::DumpFieldNames(stringstream &sstream) {
    sstream << "const_quiet_time"                               <<','
            << "l1_desired_pu"                                  <<','
            << "l1_force_local_pu"                              <<','
            << "l1_force_rmt_pu"                                <<','
            << "l1_traffic_allowed_pu"                          <<','
            << "l1_pipe_empty_pu"                               <<','
            << "idle_status_pu"                                 <<','
            << "min_time_in_linkup"                             <<','
            << "min_sync_block_cnt_for_remote_l1_desire"        <<','
            << "min_time_in_linkup_tx_idle"                     <<','
            << "linkup_tx_empty_to_L1_sleep_num_of_sync_blocks" <<','
            << "min_time_in_linkup_tx_empty"                    <<','
            << "go_to_quiet_time"                               <<','
            << "retimer_align_time"                             <<','
            << "quiet_entry_time"                               <<','
            << "min_time_to_send_announce"                      <<','
            << "min_time_in_tx_hs"                              <<','
            << "reset_to_hw"                                    <<','
            << "set_mask"                                       ;

}

void PPSLG_L1_Debug::DumpData(const ppslg_l1_debug_page& reg, stringstream &sstream) const {
    sstream << +reg.const_quiet_time <<','
            << +reg.l1_desired_pu <<','
            << +reg.l1_force_local_pu <<','
            << +reg.l1_force_rmt_pu <<','
            << +reg.l1_traffic_allowed_pu <<','
            << +reg.l1_pipe_empty_pu <<','
            << +reg.idle_status_pu <<','
            << +reg.min_time_in_linkup <<','
            << +reg.min_sync_block_cnt_for_remote_l1_desire <<','
            << +reg.min_time_in_linkup_tx_idle <<','
            << +reg.linkup_tx_empty_to_L1_sleep_num_of_sync_blocks <<','
            << +reg.min_time_in_linkup_tx_empty <<','
            << +reg.go_to_quiet_time <<','
            << +reg.retimer_align_time <<','
            << +reg.quiet_entry_time <<','
            << +reg.min_time_to_send_announce <<','
            << +reg.min_time_in_tx_hs <<','
            << +reg.reset_to_hw <<','
            << +reg.set_mask ;

}

void PPSLG_L1_Debug::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPSLG_L1_Debug::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    DumpData(areg.regs.ppslg.page_data.ppslg_l1_debug_page, sstream);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPSLG_L1_Debug::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppslg_l1_debug_page =
                                &areg.regs.ppslg.page_data.ppslg_l1_debug_page;

    IBDIAGNET_RETURN_VOID;

}

PPSLG_L1_Statistical_Counters::PPSLG_L1_Statistical_Counters(PhyDiag * phy_diag)
    : PPSLGRegister(phy_diag,
                    0x3,
                    (const unpack_data_func_t)ppslg_l1_statistical_counters_unpack,
                    ACC_REG_PPSLG_L1_STATISTICAL_COUNTERS_INTERNAL_SECTION_NAME,
                    ACC_REG_PPSLG_L1_STATISTICAL_COUNTERS_NAME,
                    ACC_REG_PPSLG_L1_STATISTICAL_COUNTERS_FIELDS_NUM,
                    ACC_REG_PPSLG_L1_STATISTICAL_COUNTERS_HEADER) {

}

void PPSLG_L1_Statistical_Counters::DumpFieldNames(stringstream &sstream) {
    sstream << "sampling_time_interval"        <<','
            << "clr"                           <<','
            << "enable"                        <<','
            << "time_since_last_clear_high"    <<','
            << "time_since_last_clear_low"     <<','
            << "l1_entry_quiet_32b"            <<','
            << "l1_force_entry_32b"            <<','
            << "l1_exit_local_desired_32b"     <<','
            << "l1_exit_local_recal_32b"       <<','
            << "l1_exit_remote_32b"            <<','
            << "average_full_bw_exit"          <<','
            << "average_low_power_enter"       <<','
            << "average_local_low_power_exit"  <<','
            << "average_remote_low_power_exit" <<','
            << "average_low_power_exit"        <<','
            << "average_local_full_bw_enter"   <<','
            << "average_remote_full_bw_enter"  <<','
            << "average_full_bw_enter"         ;

}

void PPSLG_L1_Statistical_Counters::DumpData(const ppslg_l1_statistical_counters& reg, stringstream &sstream) const {
    sstream << +reg.sampling_time_interval <<','
            << +reg.clr <<','
            << +reg.enable <<','
            << +reg.time_since_last_clear_high <<','
            << +reg.time_since_last_clear_low <<','
            << +reg.l1_entry_quiet_32b <<','
            << +reg.l1_force_entry_32b <<','
            << +reg.l1_exit_local_desired_32b <<','
            << +reg.l1_exit_local_recal_32b <<','
            << +reg.l1_exit_remote_32b <<','
            << +reg.average_full_bw_exit <<','
            << +reg.average_low_power_enter <<','
            << +reg.average_local_low_power_exit <<','
            << +reg.average_remote_low_power_exit <<','
            << +reg.average_low_power_exit <<','
            << +reg.average_local_full_bw_enter <<','
            << +reg.average_remote_full_bw_enter <<','
            << +reg.average_full_bw_enter ;

}

void PPSLG_L1_Statistical_Counters::DumpRegisterHeader(stringstream &sstream, const string& handler_header) {
    IBDIAGNET_ENTER;
    Register::DumpRegisterHeader(sstream, handler_header);
#ifdef SHOW_PHY_HEADERS
    sstream << endl                                << '#'
            << handler_header                      << ',';
    DumpFieldNames(sstream);
#endif
    IBDIAGNET_RETURN_VOID;
}

void PPSLG_L1_Statistical_Counters::DumpRegisterData(const struct acc_reg_data & areg,
                                                         stringstream &sstream, const AccRegKey &key) const {
    IBDIAGNET_ENTER;
    ostream_flags_saver saver(sstream);
    DumpData(areg.regs.ppslg.page_data.ppslg_l1_statistical_counters, sstream);
    sstream << std::endl;
    IBDIAGNET_RETURN_VOID;
}

void PPSLG_L1_Statistical_Counters::ExportRegisterData(export_data_phy_port_t *p_export_data_phy_port,
                        export_data_phy_node_t *p_export_data_phy_node,
                        struct acc_reg_data &areg, AccRegKey *p_key) {
    IBDIAGNET_ENTER;

    if (p_export_data_phy_port)
        p_export_data_phy_port->p_ppslg_l1_statistical_counters =
                                &areg.regs.ppslg.page_data.ppslg_l1_statistical_counters;

    IBDIAGNET_RETURN_VOID;

}
//**************************************************
//***   Access Register Handler Implementation   ***
//**************************************************
AccRegHandler::AccRegHandler(Register *p_register, string hh)
    : phy_errors(&p_register->GetPhyDiag()->GetPhyErrors()),
          can_send_mads_by_lid(false),
          clbck_error_state(IBDIAG_SUCCESS_CODE),
          p_reg(p_register),
          handler_header(hh),
          data_map(keycomp)
{
}

AccRegHandler::~AccRegHandler()
{
    IBDIAGNET_ENTER;

    if (p_reg)
        delete p_reg;

    for (map_akey_areg::iterator it = this->data_map.begin(); it != this->data_map.end(); ++it) {
         delete(it->first);
    }
    this->data_map.clear();

    IBDIAGNET_RETURN_VOID;
}

int AccRegHandler::PCIAddressCollectGetClbck(const clbck_data_t &clbck_data,
                                  int rec_status,
                                  void *p_attribute_data)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    AccRegVia_t acc_reg_via = (AccRegVia_t)(uint64_t)clbck_data.m_data4;
    // invoke handler callback in order to handle return status and insert reply to the map
    if (acc_reg_via == VIA_SMP)
        rc = AccRegHandler::SMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);

    if (acc_reg_via == VIA_GMP)
        rc = AccRegHandler::GMPAccessRegisterHandlerGetClbck(clbck_data, rec_status, p_attribute_data);

    if (rc)
        IBDIAGNET_RETURN(rc)

    //in the case of success we are sure to get already unpacked and saved data
    IBNode *p_node = (IBNode *)clbck_data.m_data1;
    AccRegKey *p_key = (AccRegKey *)clbck_data.m_data2;

    std::map<AccRegKey *, struct acc_reg_data>::iterator it = this->data_map.find(p_key);
    if (it == this->data_map.end())
       IBDIAGNET_RETURN(1);

    const acc_reg_data& reg_data = it->second;
    p_node->addPCIAddress(PCI_Index(reg_data.regs.mpir.pcie_index,
                                    reg_data.regs.mpir.depth,
                                    reg_data.regs.mpir.node),
                          PCI_Address(reg_data.regs.mpir.bus,
                                      reg_data.regs.mpir.local_port,
                                      reg_data.regs.mpir.device),
                          reg_data.regs.mpir.sdm);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int AccRegHandler::SMPAccessRegisterHandlerGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAGNET_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (clbck_error_state)
    {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }
    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    if (rec_status &&
        IsSupportAccRegFailed(p_node,
                              NOT_SUPPORT_SMP_ACCESS_REGISTER,
                              this->p_reg->GetNotSupportedBit()))
    {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    if (rec_status) {
        IBPort* p_port = NULL;
        AccRegKeyPort* key_port = dynamic_cast<AccRegKeyPort*>((AccRegKey*)(clbck_data.m_data2));
        if (key_port) {
            p_port = this->GetPReg()->GetPhyDiag()->GetPort(key_port->node_guid, key_port->port_num);
        }
        switch(rec_status) {
            case IBIS_MAD_STATUS_UNSUP_METHOD_ATTR: {
                SET_NSB(p_node, this->p_reg->GetNotSupportedBit());

                stringstream ss;
                ss << "The firmware of this device does not support SMP register ID: 0x"
                   << std::hex << std::setw(4) << std::setfill('0') << this->p_reg->GetRegisterID()
                   << " [err=0x" << std::hex << std::setw(4) << std::setfill('0') << rec_status << "]";

                this->phy_errors->push_back(new FabricErrPhyNodeNotSupportCap(p_node, ss.str()));

                break;
            }

            case IBIS_MAD_STATUS_TIMEOUT: {
                SET_NSB(p_node, NOT_SUPPORT_SMP_ACCESS_REGISTER);
                if (p_port)
                    this->phy_errors->push_back(new FabricErrPhyPortNotRespond(p_port, "SMPAccessRegister [timeout]"));
                else
                    this->phy_errors->push_back(new FabricErrPhyNodeNotRespond(p_node, "SMPAccessRegister [timeout]"));

                break;
            }

            default: {
                stringstream ss;
                ss << "SMPAccessRegister [err=0x"
                   << std::hex << std::setw(4) << std::setfill('0') << rec_status << "]";
                if (p_port)
                    this->phy_errors->push_back(new FabricErrPhyPortNotRespond(p_port, ss.str()));
                else
                    this->phy_errors->push_back(new FabricErrPhyNodeNotRespond(p_node, ss.str()));

                break;
            }
        }

        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    struct SMP_AccessRegister *p_access_reg =
        (struct SMP_AccessRegister *)p_attribute_data;

    //already reported that this node doesn't support
    if (p_access_reg->status &&
        IsSupportAccRegFailed(p_node,
                              NOT_SUPPORT_SMP_ACCESS_REGISTER,
                              this->p_reg->GetNotSupportedBit()))
    {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    //check if the access register status, is it supported, etc.
    if (p_access_reg->status) {
        SET_NSB(p_node, this->p_reg->GetNotSupportedBit());

        this->phy_errors->push_back(new FabricNodeErrPhyRetrieveGeneral(p_node ,p_access_reg->status));

        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    AccRegKey *p_key = (AccRegKey *)clbck_data.m_data2;

    struct acc_reg_data register_unpack_data;
    CLEAR_STRUCT(register_unpack_data);

    if (this->p_reg->UnpackData(p_key, &register_unpack_data.data, p_access_reg->reg.data)) {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    if (!p_reg->IsValidPakcet(register_unpack_data)) {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    // insert the data into a map
    std::pair<std::map<AccRegKey *, struct acc_reg_data>::iterator,bool> ret;
    ret = this->data_map.insert(pair <AccRegKey *,
                                struct acc_reg_data>(p_key,
                                                     register_unpack_data));

    if ((!ret.second) || clbck_error_state) {
        this->p_reg->GetPhyDiag()->SetLastError("Failed to add %s data for node=%s," \
                " err=%s",
                (CHECK_NAME_RETRIEVE(this->
                                     GetPReg()->
                                     GetSectionName())).c_str(),
                p_node->getName().c_str(), this->p_reg->GetPhyDiag()->GetLastError());

        delete p_key;
        IBDIAGNET_RETURN(1);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


int AccRegHandler::GMPAccessRegisterHandlerGetClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    IBDIAGNET_ENTER;

    IBNode *p_node = ProgressBar::complete<IBNode>(clbck_data.m_p_progress_bar, clbck_data.m_data1);
    if (clbck_error_state) {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    if (rec_status && IsSupportAccRegFailed(p_node,
                                            NOT_SUPPORT_GMP_ACCESS_REGISTER,
                                            this->p_reg->GetNotSupportedBit()))
    {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    if (rec_status) {
        IBPort* p_port = NULL;
        AccRegKeyPort* key_port = dynamic_cast<AccRegKeyPort*>((AccRegKey*)(clbck_data.m_data2));
        if (key_port) {
            p_port = this->GetPReg()->GetPhyDiag()->GetPort(key_port->node_guid, key_port->port_num);
        }
        switch(rec_status) {
            case IBIS_MAD_STATUS_UNSUP_METHOD_ATTR:
            case MAD_STATUS_UNSUP_REGISTER_GMP: {
                SET_NSB(p_node, this->p_reg->GetNotSupportedBit());

                stringstream ss;
                ss << "The firmware of this device does not support GMP register ID: 0x"
                   << std::hex << std::setw(4) << std::setfill('0') << this->p_reg->GetRegisterID()
                   << " [err=0x" << std::hex << std::setw(4) << std::setfill('0') << rec_status << "]";

                this->phy_errors->push_back(new FabricErrPhyNodeNotSupportCap(p_node, ss.str()));

                break;
            }

            case IBIS_MAD_STATUS_TIMEOUT: {
                SET_NSB(p_node,  NOT_SUPPORT_GMP_ACCESS_REGISTER);
                if (p_port)
                    this->phy_errors->push_back(new FabricErrPhyPortNotRespond(p_port, "GMPAccessRegister [timeout]"));
                else
                    this->phy_errors->push_back(new FabricErrPhyNodeNotRespond(p_node, "GMPAccessRegister [timeout]"));

                break;
            }

            default: {
                stringstream ss;
                ss << "GMPAccessRegister [err=0x"
                   << std::hex << std::setw(4) << std::setfill('0') << rec_status << "]";
                if (p_port)
                    this->phy_errors->push_back(new FabricErrPhyPortNotRespond(p_port, ss.str()));
                else
                    this->phy_errors->push_back(new FabricErrPhyNodeNotRespond(p_node, ss.str()));

                break;
            }
        }

        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    struct GMP_AccessRegister *p_access_reg = (struct GMP_AccessRegister *)p_attribute_data;
    AccRegKey *p_key = (AccRegKey *)clbck_data.m_data2;

    struct acc_reg_data register_unpack_data;
    CLEAR_STRUCT(register_unpack_data);

    if (this->p_reg->UnpackData(p_key, &register_unpack_data.data, p_access_reg->reg.data)) {
        delete (AccRegKey*)clbck_data.m_data2;
        IBDIAGNET_RETURN(1);
    }

    // insert the data into the map
    pair<map<AccRegKey*, struct acc_reg_data>::iterator,bool> ret;
    ret = data_map.insert(pair <AccRegKey*, struct acc_reg_data>(p_key, register_unpack_data));

    if ((!ret.second) || clbck_error_state) {
        this->p_reg->GetPhyDiag()->SetLastError("Failed to add %s data for node=%s, err=%s",
                    (CHECK_NAME_RETRIEVE(GetPReg()->GetSectionName())).c_str(),
                    p_node->getName().c_str(),
                    this->p_reg->GetPhyDiag()->GetLastError());

        delete p_key;
        IBDIAGNET_RETURN(1);
    }

    // call virtual func of register to handle data
    // beside insert it to the map in the handler
    this->p_reg->HandleData(p_node, p_key, register_unpack_data);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void AccRegHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                               export_data_phy_node_t *p_export_data_phy_node)
{
}

void AccRegHandler::ExportDataImplNode(export_data_phy_node_t *p_export_data_phy_node)
{
    if (!p_export_data_phy_node)
        return;

    for (map_akey_areg::iterator nI = this->data_map.begin(); nI != this->data_map.end(); ++nI) {
        AccRegKey *p_key = nI->first;
        if (!p_key) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            IBDIAGNET_RETURN_VOID;
        }

        if (p_key->node_guid != p_export_data_phy_node->node_guid)
            continue;

        this->p_reg->ExportRegisterData(NULL, p_export_data_phy_node, nI->second, nI->first);
    }
}

template<typename KeyType>
void AccRegHandler::ExportDataImplPort(export_data_phy_port_t *p_export_data_phy_port)
{
    if (!p_export_data_phy_port)
        return;

    for (map_akey_areg::iterator nI = this->data_map.begin(); nI != this->data_map.end(); ++nI) {
        KeyType *p_key = (KeyType *)nI->first;
        if (!p_key) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            IBDIAGNET_RETURN_VOID;
        }

        if (p_key->port_guid != p_export_data_phy_port->port_guid ||
                p_key->port_num != p_export_data_phy_port->port_num)
            continue;

        this->p_reg->ExportRegisterData(p_export_data_phy_port, NULL, nI->second, nI->first);
    }
}

int AccRegGroupHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    map_pnode_to_max_group_num_t pnode_to_max_group_num;

    ProgressBarNodes progress_bar;

    for (map_str_pnode::iterator nI = this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.begin();
            nI != this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByName map for key = %s",
                                                    (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        //send group 0, to discover how many groups there are.
        AccRegKeyGroup *p_plkey = new AccRegKeyGroup(p_curr_node->guid_get(), 0);

        AccessRegister acc_reg;
        CLEAR_STRUCT(acc_reg);

        rc = this->SendAccReg(acc_reg_via,
                              p_curr_node,
                              0, // port 0
                              lid,
                              acc_reg,
                              p_plkey,
                              &progress_bar);

        if (rc == IBDIAG_ERR_CODE_DB_ERR || clbck_error_state)
            goto exit;
    }

    // need to get all the mads for group 0
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();

    // build the pnode_to_max_group_num map
    for (map_akey_areg::iterator nI = this->data_map.begin(); nI != this->data_map.end(); ++nI) {

        AccRegKey *p_aregkey = nI->first;
        if (!p_aregkey) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        map_guid_pnode::iterator ngI = this->p_reg->GetPhyDiag()->GetFabric()->
                NodeByGuid.find(((AccRegKeyGroup *)p_aregkey)->node_guid);

        if (ngI == this->p_reg->GetPhyDiag()->GetFabric()->NodeByGuid.end())
            goto exit;

        IBNode *p_curr_node = ngI->second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByGuid map for key = " U64H_FMT,
                                                    ((AccRegKeyGroup *)p_aregkey)->node_guid);
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        struct acc_reg_data areg = nI->second;
        struct ppll_reg *p_ppll = (struct ppll_reg *)&areg.data;
        pnode_to_max_group_num.insert(pair < IBNode *, u_int8_t >(p_curr_node, p_ppll->num_pll_groups));
    }

    // go over all the nodes and send for the rest of the group other then 0
    for (map_pnode_to_max_group_num_t::iterator nI = pnode_to_max_group_num.begin();
            nI != pnode_to_max_group_num.end(); ++nI) {

        IBNode *p_curr_node = nI->first;
        u_int8_t max_gr = nI->second;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        //starting from group 1, we already sent group 0
        for (u_int8_t gr = 1; gr < max_gr; gr++) {

            AccRegKeyGroup *p_plkey = new AccRegKeyGroup(p_curr_node->guid_get(), gr);

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            rc = this->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  0, // port 0
                                  lid,
                                  acc_reg,
                                  p_plkey,
                                  &progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR || clbck_error_state)
                goto exit;
        }
    }

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void AccRegGroupHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                    export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    this->ExportDataImplNode(p_export_data_phy_node);

    IBDIAGNET_RETURN_VOID;
}

int AccRegNodeSensorHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    for (map_akey_areg::iterator areg_iter = p_acc_reg_handler->data_map.begin();
            areg_iter != p_acc_reg_handler->data_map.end(); ++areg_iter) {

        AccRegKeyNode *p_nkey = (AccRegKeyNode*)areg_iter->first;
        IBNode *p_curr_node = p_reg->GetPhyDiag()->GetFabric()->NodeByGuid[p_nkey->node_guid];

        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in" \
                                                    " NodeByGuid map for key = " U64H_FMT,
                                                    p_nkey->node_guid);
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        list_uint8 sensors;
        rc = p_acc_reg_handler->GetPReg()->AvailableSensors(areg_iter->second, sensors);

        if (rc)
            goto exit;

        // go over all sensors of this node:
        for (list_uint8::iterator sensors_iter = sensors.begin();
                sensors_iter != sensors.end(); ++sensors_iter) {

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            AccRegKeyNodeSensor *p_ns_key =
                    new AccRegKeyNodeSensor(p_nkey->node_guid, *sensors_iter);

            rc = this->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  0, // port 0
                                  lid,
                                  acc_reg,
                                  p_ns_key,
                                  &progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR || clbck_error_state)
                goto exit;
        }
    }

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    IBDIAGNET_RETURN(rc);
}

void AccRegNodeSensorHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                         export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    this->ExportDataImplNode(p_export_data_phy_node);

    IBDIAGNET_RETURN_VOID;
}


int AccRegTypeGroupHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    for (map_str_pnode::iterator nI = p_reg->GetPhyDiag()->GetFabric()->NodeByName.begin();
         nI != p_reg->GetPhyDiag()->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByName map for key = %s",
                                                    nI->first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        for (int tp = 0; tp < max_type; tp++) {
            //send group 0, to discover how many groups there are.

            //Skip PCIe Group
            if(tp == 1)
                continue;

            AccRegKeyTypeGroup *p_gpkey = new AccRegKeyTypeGroup(
                    p_curr_node->guid_get(), tp, 0, 0, max_indices);

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            rc = this->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  0, // port 0
                                  lid,
                                  acc_reg,
                                  p_gpkey,
                                  &progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR || clbck_error_state)
                goto exit;
        }
    }

    // need to get all the mads for group 0
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();

    //go over all the nodes and send for the rest of the group other then 0
    for (map_akey_areg::iterator aregI = data_map.begin(); aregI != data_map.end(); ++aregI) {

        AccRegKeyTypeGroup* p_aregkey = (AccRegKeyTypeGroup*)aregI->first;
        if (!p_aregkey) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            goto exit;
        }

        // make sure that we are on group 0
        if (p_aregkey->group_num != 0)
            continue;

        map_guid_pnode::iterator ngI = this->p_reg->GetPhyDiag()->GetFabric()->
            NodeByGuid.find(p_aregkey->node_guid);
        if (ngI == this->p_reg->GetPhyDiag()->GetFabric()->NodeByGuid.end()) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - GUID " U64H_FMT
                                                    "not found in NodeByGuid map",
                                                    p_aregkey->node_guid);
            goto exit;
        }

        IBNode *p_curr_node = ngI->second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByGuid map for key = " U64H_FMT,
                                                    p_aregkey->node_guid);
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        u_int8_t opamp_gr_type = p_aregkey->group_type;

        struct acc_reg_data areg = aregI->second;
        u_int8_t max_opamp_gr;
        u_int16_t max_num_idx;

        this->p_reg->GetGroupData(areg, max_opamp_gr, max_num_idx);

        //start from group 1, we already sent group 0
        for (u_int8_t gr = 1; gr <= max_opamp_gr; gr++) {
            for (u_int16_t start_idx = 0; start_idx < max_num_idx; start_idx += max_indices) {

                u_int16_t idx_to_access =(((max_num_idx - start_idx) < max_indices)?
                        (max_num_idx % max_indices) : max_indices);

                AccRegKeyTypeGroup *p_gpkey = new AccRegKeyTypeGroup(
                        p_curr_node->guid_get(),
                        opamp_gr_type,
                        gr,
                        start_idx,
                        idx_to_access);

                AccessRegister acc_reg;
                CLEAR_STRUCT(acc_reg);

                rc = this->SendAccReg(acc_reg_via,
                                      p_curr_node,
                                      0, // port 0
                                      lid,
                                      acc_reg,
                                      p_gpkey,
                                      &progress_bar);

                if (rc == IBDIAG_ERR_CODE_DB_ERR || clbck_error_state)
                    goto exit;
            }
        }
    }

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}


int AccRegSpecificHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;
    rc = this->p_reg->BuildDB(this, phy_errors, &progress_bar);

    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void AccRegSpecificHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    for (map_akey_areg::iterator nI = this->data_map.begin(); nI != this->data_map.end(); ++nI) {
        AccRegKey *p_key = nI->first;
        if (!p_key) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            IBDIAGNET_RETURN_VOID;
        }

        if (this->p_reg->CheckRegisterKey(p_export_data_phy_port, p_export_data_phy_node, p_key))
            this->p_reg->ExportRegisterData(p_export_data_phy_port, p_export_data_phy_node, nI->second, p_key);
    }

    IBDIAGNET_RETURN_VOID;
}


int Register::PackDataSMP(AccRegKey *p_key, SMP_AccessRegister *acc_reg, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    acc_reg->register_id = this->GetRegisterID();
    acc_reg->len_reg = SMP_ACC_REG_TLV_LEN;

    IBDIAGNET_RETURN(PackData(p_key, acc_reg->reg.data, p_node));
}

int Register::PackDataGMP(AccRegKey *p_key, GMP_AccessRegister *acc_reg, const IBNode *p_node)
{
    IBDIAGNET_ENTER;

    IBDIAGNET_RETURN(PackData(p_key, acc_reg->reg.data, p_node));
}

bool Register::IsEnabledByPCAM(const IBNode* p_node) const {

    if (this->m_register_id >= 0x5000 || this->m_register_id < 0x507F)
        if (p_node) {
            PHYNodeData* data = dynamic_cast<PHYNodeData*>(p_node->p_phy_data);
            if (data && data->pcam)
                return data->pcam->port_access_reg_cap_mask[this->m_register_id - 0x5000];
        }

    return true;
}

int Register::HandleNodeNotSupportAccReg(PhyDiag *p_phy_diag,
                                         IBNode* p_node,
                                         uint64_t not_supported_bit)
{
    IBDIAGNET_ENTER;

    // check if warning was already set
    if (NSB_BITSET(p_node)[not_supported_bit])
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);

    SET_NSB(p_node, not_supported_bit);
    stringstream err_str;
    err_str << "This device does not support "
            << (not_supported_bit == NOT_SUPPORT_SMP_ACCESS_REGISTER ? "SMP" : "GMP")
            << " access register MAD capability";

    p_phy_diag->GetPhyErrors().push_back(new FabricErrPhyNodeNotSupportCap(p_node, err_str.str()));

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

AccRegVia_t Register::Validation(IBNode *p_node, int &rc)
{
    IBDIAGNET_ENTER;

    bool smp_sup = m_phy_diag->GetCapabilityModule()->IsSupportedSMPCapability(
            p_node, EnSMPCapIsAccessRegisterSupported) &&
                    !(NSB_BITSET(p_node)[NOT_SUPPORT_SMP_ACCESS_REGISTER]);

    bool gmp_sup = m_phy_diag->GetCapabilityModule()->IsSupportedGMPCapability(
            p_node, EnGMPCapIsAccessRegisterSupported) &&
                    !(NSB_BITSET(p_node)[NOT_SUPPORT_GMP_ACCESS_REGISTER]);

    bool smp_able = (GetRegisterType() == VIA_SMP) && (p_node->serdes_version < SerDes_7nm);

    if (GetRegisterPriority() == VIA_GMP) {
        if (gmp_sup)
            IBDIAGNET_RETURN(VIA_GMP);
        if (smp_sup && smp_able)
            IBDIAGNET_RETURN(VIA_SMP);
    } else { // VIA_SMP
        if (smp_sup && smp_able)
            IBDIAGNET_RETURN(VIA_SMP);
        if (gmp_sup)
            IBDIAGNET_RETURN(VIA_GMP);
    }

    rc = HandleNodeNotSupportAccReg(m_phy_diag, p_node, NOT_SUPPORT_GMP_ACCESS_REGISTER);

    if (!smp_sup)
        rc = HandleNodeNotSupportAccReg(m_phy_diag, p_node, NOT_SUPPORT_SMP_ACCESS_REGISTER);

    IBDIAGNET_RETURN(NOT_SUP_ACC_REG);
}

int AccRegHandler::SendSMPReg(IBNode *p_curr_node,
                              phys_port_t port_num,
                              SMP_AccessRegister *smp_reg,
                              AccRegKey *p_key,
                              ProgressBar *p_progress_bar,
                              clbck_data_t *call_back_data)
{
    IBDIAGNET_ENTER;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = this;
    clbck_data.m_data1 = p_curr_node;
    clbck_data.m_data2 = p_key;

    clbck_data_t *p_clbck_data;
    if (call_back_data) {
        call_back_data->m_p_progress_bar = p_progress_bar;
        p_clbck_data = call_back_data;
    } else {
        clbck_data.m_handle_data_func = &forwardClbck<AccRegHandler,
                &AccRegHandler::SMPAccessRegisterHandlerGetClbck>;
        clbck_data.m_p_progress_bar = p_progress_bar;
        p_clbck_data = &clbck_data;
    }

    // handle direct route
    direct_route_t *p_curr_direct_route = NULL;
    p_curr_direct_route = this->p_reg->GetPhyDiag()->GetIBDiag()->GetDR(p_curr_node);

    if (!p_curr_direct_route) {
        this->p_reg->GetPhyDiag()->SetLastError("DB error - can't find direct route "
                                                "to node=%s (node guid: " U64H_FMT ")",
                                                p_curr_node->getName().c_str(),
                                                p_curr_node->guid_get());
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    int rc = p_reg->PackDataSMP(p_key, smp_reg, p_curr_node);

    if (!rc)
        rc = this->p_reg->GetPhyDiag()->SMPAccRegGetByDirect(p_curr_direct_route,
                                                             port_num,
                                                             smp_reg,
                                                             p_clbck_data);

    IBDIAGNET_RETURN(rc);
}

int AccRegHandler::SendGMPReg(IBNode *p_curr_node,
                              uint16_t lid,
                              GMP_AccessRegister *gmp_reg,
                              AccRegKey *p_key,
                              ProgressBar *p_progress_bar,
                              clbck_data_t *call_back_data)
{
    IBDIAGNET_ENTER;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = this;
    clbck_data.m_data1 = p_curr_node;
    clbck_data.m_data2 = p_key;

    clbck_data_t *p_clbck_data;
    if (call_back_data) {
        call_back_data->m_p_progress_bar = p_progress_bar;
        p_clbck_data = call_back_data;
    } else {
        clbck_data.m_handle_data_func = &forwardClbck<AccRegHandler,
                &AccRegHandler::GMPAccessRegisterHandlerGetClbck>;
        clbck_data.m_p_progress_bar = p_progress_bar;
        p_clbck_data = &clbck_data;
    }

    int rc = p_reg->PackDataGMP(p_key, gmp_reg, p_curr_node);

    if (!rc)
        rc = this->p_reg->GetPhyDiag()->GMPAccRegGet(lid,
                                                     p_reg->GetRegisterID(),
                                                     gmp_reg,
                                                     p_clbck_data);

    IBDIAGNET_RETURN(rc);
}

int AccRegHandler::SendAccReg(AccRegVia_t acc_reg_via,
                              IBNode *p_curr_node,
                              phys_port_t port_num,
                              uint16_t lid,
                              AccessRegister &acc_reg,
                              AccRegKey *p_key,
                              ProgressBar *p_progress_bar,
                              clbck_data_t *call_back_data)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_ERR_CODE_DISABLED;

    if (this->p_reg->IsEnabledByPCAM(p_curr_node))
    {
        switch (acc_reg_via) {
            case VIA_SMP: {
                rc = this->SendSMPReg(p_curr_node,
                                      port_num,
                                      &acc_reg.smp_reg,
                                      p_key,
                                      p_progress_bar,
                                      call_back_data);

                break;
            }

            case VIA_GMP: {
                rc = this->SendGMPReg(p_curr_node,
                                      lid,
                                      &acc_reg.gmp_reg,
                                      p_key,
                                      p_progress_bar,
                                      call_back_data);
                break;
            }

            default:
                cerr << "Invalid Access Register type!" << endl;
        }
    }

    if (rc)
        delete p_key;

    IBDIAGNET_RETURN(rc == IBDIAG_ERR_CODE_DISABLED ? IBDIAG_SUCCESS_CODE : rc);
}

int AccRegPortLaneHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    for (map_str_pnode::iterator nI = p_reg->GetPhyDiag()->GetFabric()->NodeByName.begin();
         nI != p_reg->GetPhyDiag()->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByName map for key = %s",
                                                    nI->first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        // in case of SLRIP register - loop in lane is not relevant
        // if serdes version is 7nm
        if (p_reg->GetRegisterID() == ACCESS_REGISTER_ID_SLRIP) {
            if (p_curr_node->serdes_version >= SerDes_7nm)
                max_idx_in_lane = 1;
            else
                max_idx_in_lane = SLRIP_IB_SEL_NUM;
        }

        // go over all ports of this node:
        for (phys_port_t i = 1; i < p_curr_node->numPorts + 1; ++i) {

            IBPort *p_curr_port = p_curr_node->getPort(i);

            // port not exist
            if (!p_curr_port)
                continue;

            // skip ports that are not in sub fabric.
            if (!p_curr_port->getInSubFabric())
                continue;

            // validate disconnected
            if (p_curr_port->get_internal_state() <= IB_PORT_STATE_DOWN &&
                    (!p_reg->GetRetriveDisconnected() ||
                        p_curr_node->type == IB_CA_NODE))
                continue;

            // special port not supported
            if (p_curr_port->isSpecialPort())
                continue;

            for (u_int32_t ln = 0; ln < LANE_NUM; ln++) {
                for (u_int8_t lane_idx = 0; lane_idx < max_idx_in_lane; lane_idx++) {

                    AccessRegister acc_reg;
                    CLEAR_STRUCT(acc_reg);

                    AccRegKeyPortLane *p_plkey = new AccRegKeyPortLane(
                            p_curr_node->guid_get(),
                            p_curr_port->guid_get(),
                            i,
                            ln,
                            lane_idx);

                    rc = this->SendAccReg(acc_reg_via,
                                          p_curr_node,
                                          i,
                                          p_curr_port->base_lid,
                                          acc_reg,
                                          p_plkey,
                                          &progress_bar);

                    if (rc == IBDIAG_ERR_CODE_DB_ERR)
                        goto exit;

                } // idx_in_lane
            } // lane
        } // port
    } // node

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void AccRegPortLaneHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                       export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    this->ExportDataImplPort<AccRegKeyPortLane>(p_export_data_phy_port);

    IBDIAGNET_RETURN_VOID;
}

int AccRegNodeHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    for (map_str_pnode::iterator nI = this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.begin();
            nI != this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByName map for key = %s",
                                                    nI->first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        AccessRegister acc_reg;
        CLEAR_STRUCT(acc_reg);

        lid_t lid = p_curr_node->getFirstLid();
        if (lid == IB_LID_UNASSIGNED)
            continue;

        AccRegKeyNode *p_nkey = new AccRegKeyNode(p_curr_node->guid_get());

        rc = this->SendAccReg(acc_reg_via,
                              p_curr_node,
                              0, // port 0
                              lid,
                              acc_reg,
                              p_nkey,
                              &progress_bar);

        if (rc == IBDIAG_ERR_CODE_DB_ERR)
            goto exit;

    } // nodes loop

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void AccRegNodeHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                   export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    this->ExportDataImplNode(p_export_data_phy_node);

    IBDIAGNET_RETURN_VOID;
}

AccRegKey* AccRegPortHandler::CreateKey(uint64_t ng, uint64_t pg, uint8_t pn)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(new AccRegKeyPort(ng, pg, pn));
}

int AccRegPortHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->p_reg->GetPhyDiag()->GetIBDiag()->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarNodes progress_bar;

    for (map_str_pnode::iterator nI = this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.begin();
         nI != this->p_reg->GetPhyDiag()->GetFabric()->NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null node in"  \
                                                    " NodeByName map for key = %s",
                                                    (*nI).first.c_str());
            rc = IBDIAG_ERR_CODE_DB_ERR;
            goto exit;
        }

        if (!p_curr_node->getInSubFabric())
            continue;

        if (!p_reg->IsRegSupportNodeType(p_curr_node))
            continue;

        AccRegVia_t acc_reg_via = p_reg->Validation(p_curr_node, rc);

        if (acc_reg_via == NOT_SUP_ACC_REG)
            continue;

        // go over all ports of this node:
        for (u_int32_t i = 1; i < (unsigned)p_curr_node->numPorts + 1; ++i) {

            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);

            // port not exist
            if (!p_curr_port)
                continue;

            // skip ports that are not in sub fabric.
            if (!p_curr_port->getInSubFabric())
                continue;

            // validate disconnected
            if (p_curr_port->get_internal_state() <= IB_PORT_STATE_DOWN &&
                    (!p_reg->GetRetriveDisconnected() ||
                        p_curr_node->type == IB_CA_NODE))
                continue;

            // special port not supported
            if (p_curr_port->isSpecialPort())
                continue;

            AccessRegister acc_reg;
            CLEAR_STRUCT(acc_reg);

            AccRegKey *p_pkey = CreateKey(p_curr_node->guid_get(),
                                          p_curr_port->guid_get(),
                                          i);

            rc = this->SendAccReg(acc_reg_via,
                                  p_curr_node,
                                  i,
                                  p_curr_port->base_lid,
                                  acc_reg,
                                  p_pkey,
                                  &progress_bar);

            if (rc == IBDIAG_ERR_CODE_DB_ERR)
                goto exit;

        } // ports
    } // nodes

exit:
    this->p_reg->GetPhyDiag()->GetIbis()->MadRecAll();
    if (clbck_error_state)
        rc = clbck_error_state;
    else if (!phy_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAGNET_RETURN(rc);
}

void AccRegPortHandler::ExportData(export_data_phy_port_t *p_export_data_phy_port,
                                   export_data_phy_node_t *p_export_data_phy_node)
{
    IBDIAGNET_ENTER;

    this->ExportDataImplPort<AccRegKeyPort>(p_export_data_phy_port);

    IBDIAGNET_RETURN_VOID;
}


// ** AccRegPortIndexHandler **
AccRegKey* AccRegPortIndexHandler::CreateKey(uint64_t ng, uint64_t pg, uint8_t pn)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(new AccRegKeyPortIndex(ng, pg, pn, this->cur_index));
}

int AccRegPortIndexHandler::BuildDB(list_p_fabric_general_err &phy_errors)
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    for (uint64_t idx : indexes) {
        cur_index = idx;
        rc = AccRegPortHandler::BuildDB(phy_errors);
    }

    IBDIAGNET_RETURN(rc);
}


void AccRegHandler::DumpCSV(CSVOut &csv_out)
{
    IBDIAGNET_ENTER;
    stringstream sstream;

    // Skiped. At least PPLLRegister::DumpRegisterData and SLTPRegister::DumpRegisterData
    // write warnings to the ibdiagnet2.log file
    (void) csv_out.DumpStart(p_reg->GetSectionName().c_str());

    p_reg->DumpRegisterHeader(sstream, handler_header);
    sstream << endl;
    csv_out.WriteBuf(sstream.str());

    for (map_akey_areg::iterator nI = this->data_map.begin();
         nI != this->data_map.end(); ++nI) {
        AccRegKey * p_aregkey = (*nI).first;
        if (!p_aregkey)
        {
            this->p_reg->GetPhyDiag()->SetLastError("DB error - found null key in data_map");
            IBDIAGNET_RETURN_VOID;
        }

        sstream.str("");
        p_aregkey->DumpKeyData(sstream);
        struct acc_reg_data areg = (*nI).second;
        p_reg->DumpRegisterData(areg, sstream, *p_aregkey);
        csv_out.WriteBuf(sstream.str());
    }
    csv_out.DumpEnd(p_reg->GetSectionName().c_str());
    IBDIAGNET_RETURN_VOID;
}
