/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2020, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *  Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */


#include <errno.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <memory>

#include <../../ibis_tools/include/infiniband/misc/tool_trace/tool_trace.h>
#include "smparquery.h"


//============================================================================//
#define MAX_SUPPORTED_PLFT_NUMBER 8 //!!! TAMIR add to OSM


#define RN_GEN_STRING_TABLE_BLOCK_SIZE 32
#define RN_RCV_STRING_BLOCK_SIZE 16
#define RN_XMIT_PORT_MASK_BLOCK_SIZE 128

#define AR_LFT_TABLE_BLOCK_SIZE         32
#define AR_LFT_TABLE_BLOCK_SIZE_SX      16
#define AR_LFT_TABLE_NUM_BLOCKS         1536
#define AR_LFT_TABLE_NUM_BLOCKS_SX      3072

#define AR_GROUP_TABLE_BLOCK_SIZE       2
#define AR_GROUP_TABLE_NUM_BLOCKS       1024
//#define AR_GROUP_UNASSIGNED             0x0 //used only for is_group_per_leaf_sw
//#define AR_COPY_GROUP_BLOCK_SIZE        16

#define RN_XMIT_PORT_MASK_GENERATE_ARN 0X1
#define RN_XMIT_PORT_MASK_GENERATE_FRN 0X2
#define RN_XMIT_PORT_MASK_PASS_ON_RN   0X4

#define RN_SUB_GROUP_DIRECTION_BLOCK_SIZE 64
//#define RN_SUB_GROUP_DIRECTION_NUM_BLOCKS 64
//#define RN_SUB_GROUP_DIRECTION_UNASSIGNED 0xFF

#define ARMGR_MAX_GROUP_TABLES          8
//#define ARMGR_SW_NOT_UPDATED            0
//#define ARMGR_SW_UPDATED                1


enum SMP_AR_LID_STATE {
    AR_IB_LID_STATE_BOUNDED = 0x0,
    AR_IB_LID_STATE_FREE = 0x1,
    AR_IB_LID_STATE_STATIC = 0x2,
    AR_IB_LID_STATE_LAST
};


inline static const char * ConvertARIBLidStateToStr(IN u_int8_t lid_state)
{
    switch (lid_state) {
    case AR_IB_LID_STATE_BOUNDED:
        return "Bounded";
    case AR_IB_LID_STATE_FREE:
        return "Free";
    case AR_IB_LID_STATE_STATIC:
        return "Static";
    default:
        return "Unknown";
    }
}

inline static string ConvertBitMaskToStr(IN u_int64_t mask)
{
    char buffer[1024];
    string str = "";
    for (unsigned port_num = 1; port_num < 64; ++port_num) {
        if (mask & (1ull << port_num)) {
            snprintf(buffer, sizeof(buffer), "%u,", port_num);
            str += buffer;
        }
    }
    if (str != "")
        str.erase(str.end() - 1, str.end());
    return (str);
}

#ifndef __AR_REVERSED_LFT_MAD
    #define __AR_REVERSED_LFT_MAD
#endif

inline static u_int8_t FixLFTEntryNum(u_int8_t entry_num)
{
    u_int8_t res = entry_num;
#ifdef __AR_REVERSED_LFT_MAD
    res ^= 1;
#endif
    return res;
}

#define OSM_NO_PATH 0 // TODO::CHECK before release if it's correct value

/****************************************************/
/* globals */
static const match_rec_t match_tbl[] = {
        {"ARInfo", "ARI", &SMPARQuery::ARInfoProcess, NO_PARAMS},
        {"ARGroupTable", "ARGT", &SMPARQuery::ARGroupTableProcess, PARAM_PLFT | PARAM_GROUP_TBL | PARAM_BLOCK_NUM},
        {"ARLFTTable", "ARLT", &SMPARQuery::ARLFTTableProcess, PARAM_PLFT | PARAM_BLOCK_NUM},
        {"PLFTInfo", "PLFTI", &SMPARQuery::PLFTInfoProcess, NO_PARAMS},
        {"PLFTDef", "PLFTD", &SMPARQuery::PLFTDefProcess, PARAM_BLOCK_NUM},
        {"PLFTMap", "PLFTM", &SMPARQuery::PLFTMapProcess, PARAM_PLFT | PARAM_CONTROL_MAP},
        {"PortSLToPLFTMap", "PLFTP", &SMPARQuery::PortSLToPLFTMapProcess, PARAM_BLOCK_NUM},
        {"RNSubGroupDirectionTable", "DIRT", &SMPARQuery::RNSubGroupDirectionTableProcess, PARAM_BLOCK_NUM},
        {"RNGenStringTable", "GSTR", &SMPARQuery::RNGenStringTableProcess, PARAM_PLFT | PARAM_BLOCK_NUM},
        {"RNGenBySubGroupPriority", "GSGP", &SMPARQuery::RNGenBySubGroupPriorityProcess, NO_PARAMS},
        {"RNRcvString", "RSTR", &SMPARQuery::RNRcvStringProcess, PARAM_BLOCK_NUM},
        {"RNXmitPortMask", "RNXM", &SMPARQuery::RNXmitPortMaskProcess, PARAM_BLOCK_NUM},
        {"PortRNCounters", "RNPC", &SMPARQuery::PortRNCountersProcess, NO_PARAMS},
        {0, 0, 0, 0}
};


/****************************************************/
#include "smparquery_version.h"

const string get_tool_version()
{
    string version = "unknown";
#ifdef SMPARQUERY_VERSION
    version = SMPARQUERY_VERSION;
#endif
    return version;
}


/****************************************************/
/******************private methods*******************/
/****************************************************/
int SMPARQuery::ConvertStrToDR(direct_route_t *p_dr, string route_path)
{
    char *s, *str, *str2;
    str = str2 = new char[route_path.size() + 1];
    if (!str) {
        ERR_PRINT("Failed to allocate buff for dr\n");
        return 1;
    }
    strcpy(str, route_path.c_str());
    CLEAR_STRUCT(*p_dr);
    while (str && *str) {
        if ((s = strchr(str, ',')))
            *s = 0;
        p_dr->path.BYTE[p_dr->length] = (u_int8_t)atoi(str);
        ++p_dr->length;
        if (!s)
            break;
        str = s + 1;
    }
    delete [] str2;
    return 0;
}


void SMPARQuery::AddGeneralOptions()
{
    this->AddDescription(TOOL_DESC);
    this->AddOptions(OPTION_CA, OPTION_CA_SHORT, OPTION_CA_ARG_DESC, OPTION_CA_DESC);
    this->AddOptions(OPTION_PORT, OPTION_PORT_SHORT, OPTION_PORT_ARG_DESC, OPTION_PORT_DESC);
    this->AddOptions(OPTION_DR, OPTION_DR_SHORT, OPTION_DR_ARG_DESC, OPTION_DR_DESC);
    this->AddOptions(OPTION_LID, OPTION_LID_SHORT, OPTION_LID_ARG_DESC, OPTION_LID_DESC);
    this->AddOptions(OPTION_HELP, OPTION_HELP_SHORT, OPTION_HELP_ARG_DESC, OPTION_HELP_DESC);
    this->AddOptions(OPTION_VERSION, OPTION_VERSION_SHORT, OPTION_VERSION_ARG_DESC, OPTION_VERSION_DESC);
    this->AddOptions(OPTION_DEBUG, OPTION_DEBUG_SHORT, OPTION_DEBUG_ARG_DESC, OPTION_DEBUG_DESC);
}


void SMPARQuery::Usage()
{
    char buffer1[4096], buffer2[4096];
    string str = "\nUsage: ";
    const match_rec_t *r;
    unsigned n;

    snprintf(buffer2, sizeof(buffer2),
            "    %s argrouptable 3 1  \t\t# argrouptable by lid, with block modifier\n"  \
            "    %s -D arinfo 0,1,2,3 \t\t# arinfo by direct route\n" \
            "    %s PLFTMAP -L 24 1 0x1\t\t# PLFT Map with specified control map fields - 0x1\n" \
            "    %s PLFTMAP -L 24 1  \t\t# PLFT Map with no specified control map fields - set 0x3 as default value\n",
            TOOL_NAME, TOOL_NAME, TOOL_NAME, TOOL_NAME);

    n = snprintf(buffer1, sizeof(buffer1), "%s <op> <dest dr_path|lid> [op params]\n"
            "\nSupported ops (and aliases, case insensitive):\n",
            TOOL_NAME);
    for (r = match_tbl; r->name; ++r) {
        n += snprintf(buffer1 + n,
                sizeof(buffer1) - n,
                "    %s (%s) <addr>%s%s%s%s\n",
                r->name,
                r->alias ? r->alias : "",
                r->params & PARAM_PLFT ? " [<plft>]" : "",
                r->params & PARAM_GROUP_TBL ? " [<group_table>]" : "",
                r->params & PARAM_BLOCK_NUM ? " [<blocknum>]" : "",
                r->params & PARAM_CONTROL_MAP ? " [<control_map>]" : "");
        if (n >= sizeof(buffer1))
            exit(1);
    }
    str += buffer1;
    str += "\n\nOptions:\n";
    str += this->GetUsageOptions();
    str += "\nExamples:\n";
    str += buffer2;
    printf("%s", str.c_str());
}


//Returns: 0 - OK / 1 - parsing error / 2 - OK but need to exit
int SMPARQuery::HandleOption(string name, string value)
{
    //printf("%s %s\n", name.c_str(), value.c_str());
    if (name == OPTION_CA) {
        this->m_ca_name = value;
        return 0;
    } else if (name == OPTION_PORT) {
        this->m_port = (u_int8_t)strtoul(value.c_str(), NULL, 0);
        return 0;
    } else if (name == OPTION_DR) {
        if (this->m_to_use_lid) {
            ERR_PRINT("Can not use both LID and DR as dest addr\n");
            return 1;
        }
        this->m_to_use_dr = true;
        return 0;
    } else if (name == OPTION_LID) {
        if (this->m_to_use_dr) {
            ERR_PRINT("Can not use both LID and DR as dest addr\n");
            return 1;
        }
        this->m_to_use_lid = true;
        return 0;
    } else if (name == OPTION_VERSION) {
        INFO_PRINT("%s\n", get_tool_version().c_str());
        return 2;           //exit with no error
    } else if (name == OPTION_HELP) {
        this->Usage();
        return 2;           //exit with no error
    } else if (name == OPTION_DEBUG) {
        this->is_debug_log = true;
        return 0;
    }

    //if we reached here than unknown option
    return 1;
}


/****************************************************/
/******************public methods********************/
/****************************************************/
SMPARQuery::SMPARQuery() : CommandLineRequester("Main"), m_initialized(false),
    m_ca_name(""), m_port(0),
    m_to_use_dr(false), m_to_use_lid(false), is_debug_log(false),
    m_operation(""), m_dest(""), m_optional_args("")
{
    // Initialize main parser
    this->m_p_main_parser = new CommandLineParser(TOOL_NAME);
    if (!this->m_p_main_parser) {
        ERR_PRINT("Failed to allocate main parser\n");
        return;
    }

    // Add the known option to requester
    this->AddGeneralOptions();

    // Add the requester to the parser
    if (this->m_p_main_parser->AddRequester(this)) {
        ERR_PRINT("Failed add options to main parser, err=%s\n", this->m_p_main_parser->GetErrDesc());
        return;
    }

    this->m_initialized = true;
    return;
}


SMPARQuery::~SMPARQuery()
{
    tt_log_destroy();
    delete this->m_p_main_parser;
    return;
}


//Returns: 0 - OK / 1 - error / 2 - OK but need to exit
int SMPARQuery::GetOpt(int argc, const char **argv)
{
    int i, rc2, rc = 1;

    /*
     * finally parse all options
     */
    rc2 = this->m_p_main_parser->ParseOptions(argc, argv, true);
    if (rc2 == 1) {         //parse error and we need to dump usage
        goto usage_and_exit;
    } else if (rc2 == 2) {  //OK but need to exit
        rc = 2;
        goto exit;
    }

    /*
     * if no lid / direct route were given than mark by lid as default one
     */
    this->m_to_use_lid = true;

    /*
     * also parse operation + dest string + optional args
     */
    this->m_optional_args = this->m_p_main_parser->GetUnknownOptions();
    i = this->m_optional_args.find(' ');
    this->m_operation = this->m_optional_args.substr(0, i);
    this->m_optional_args.erase(0, i + 1);
    i = this->m_optional_args.find(' ');
    this->m_dest = this->m_optional_args.substr(0, i);
    this->m_optional_args.erase(0, i + 1);
    if (this->m_operation == "" || this->m_dest == "")
        goto usage_and_exit;

    printf("op = %s, dest = %s, rest = %s\n", this->m_operation.c_str(), this->m_dest.c_str(), this->m_optional_args.c_str());

    rc = 0;
    goto exit;

usage_and_exit:
    this->Usage();
exit:
    return rc;
}


void SMPARQuery::ActiveDebugLog()
{
    if (tt_log_construct(0xff,                   /* need_to_flush */
                         TT_LOG_MODULE_ALL,      /* module verbosity*/
                         (~TT_LOG_LEVEL_FUNCS),  /* LEVEL VERBOSITY */
                         NULL,                   /* log file - if NULL will be printed to stdout*/
                         DEFAULT_LOG_SIZE,       /* max size (in MB) */
                         0))                     /* append to prev log */
        printf("-E- Failed to print debug logs\n");
}


int SMPARQuery::Run()
{
    int rc;

    /*
     * init ibis
     */
    if (this->m_ibis_obj.Init()) {
        ERR_PRINT("Failed to initialize, err=%s\n", this->m_ibis_obj.GetLastError());
        return 1;
    }

    if (is_debug_log)
        this->ActiveDebugLog();

    /*
     * bind to port
     */
    if (this->m_ibis_obj.SetPort(this->m_ca_name.c_str(), this->m_port)) {
        ERR_PRINT("Failed to bind port, err=%s\n", this->m_ibis_obj.GetLastError());
        return 1;
    }

    /*
     * match the operation
     */
    const match_rec_t *r;
    op_fn_t fn = NULL;
    unsigned len = this->m_operation.size();
    for (r = match_tbl; r->name; ++r)
        if (!strncasecmp(r->name, this->m_operation.c_str(), len) ||
                (r->alias && !strncasecmp(r->alias, this->m_operation.c_str(), len)))
            fn = r->fn;

    if (!fn) {
        ERR_PRINT("Operation %s not supported\n", this->m_operation.c_str());
        return 1;
    }

    /*
     * decide if lid or direct route
     */
    if (this->m_to_use_dr) {
        direct_route_t dr;
        if (this->ConvertStrToDR(&dr, this->m_dest)) {
            ERR_PRINT("Can not resolve destination address: %s\n", this->m_dest.c_str());
            return 1;
        }
        rc = fn(&this->m_ibis_obj, 0, &dr, this->m_optional_args);
    } else {
        u_int16_t lid = (u_int16_t)strtoul(this->m_dest.c_str(), NULL, 0);

        if (!lid) {
            ERR_PRINT("Invalid lid!\n");
            return 1;
        }

        rc = fn(&this->m_ibis_obj, lid, NULL, this->m_optional_args);
    }
    return rc;
}


/****************************************************/
/****************************************************/
/****************************************************/
int SMPARQuery::GetNodeDevId(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        u_int16_t &dev_id) {
    struct SMP_NodeInfo node_info = {0};
    int rc = GetNodeInfo(p_ibis, lid, p_direct_route, node_info);
    if (rc)
        return rc;

    dev_id = node_info.DeviceID;
    INFO_PRINT("DevId:%d(0x%x):\n", node_info.DeviceID, node_info.DeviceID);
    return 0;
}

int SMPARQuery::GetNodeInfo(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        struct SMP_NodeInfo &node_info) {
    int rc = 0;

    if (lid) {  // send by lid
        INFO_PRINT("Getting NodeInfo from lid=%u\n",
                lid);
        rc = p_ibis->SMPNodeInfoMadGetByLid(lid,
                &node_info);
    } else {    // send by direct route
        INFO_PRINT("Getting NodeInfo from direct route=%s\n",
                Ibis::ConvertDirPathToStr(p_direct_route).c_str());
        rc = p_ibis->SMPNodeInfoMadGetByDirect(p_direct_route,
                &node_info);
    }
    rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get NodeInfo, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get NodeInfo, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }
    return 0;
}

int SMPARQuery::PLFTInfoProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string optional_args) {
    struct ib_private_lft_info plft_info = {0};
    int rc, i;

    if (optional_args != "") {
        ERR_PRINT("No need for more params for this operation\n");
        return 1;
    };
    if (lid) {  // send by lid
        INFO_PRINT("Getting PLFTInfo from lid=%u\n",
                   lid);
        rc = p_ibis->SMPPLFTInfoMadGetSetByLid(lid,
                IBIS_IB_MAD_METHOD_GET,
                &plft_info);
    } else {    // send by direct route
        INFO_PRINT("Getting PLFTInfo from direct route=%s\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str());
        rc = p_ibis->SMPPLFTInfoMadGetSetByDirect(p_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                &plft_info);
    }
    rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
            (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get PrivateLFTInfo, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get PrivateLFTInfo, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    INFO_PRINT("PrivateLFTInfo:\n");
    INFO_PRINT("PLFT Active_Mode.........................%u\n", plft_info.Active_Mode);
    INFO_PRINT("PLFT ModeCap.............................%u\n", plft_info.ModeCap);
    INFO_PRINT("PLFT Fallback............................%s\n", !plft_info.NoFallback ? "Enabled" : "Disabled");
    for (i = 0; i < 4; i++) {
        INFO_PRINT("PLFT TableSize%d..........................%u\n", i+1, plft_info.Description_Mode[i].LFTableSize);
        INFO_PRINT("PLFT Number Of PLFTs%d....................%u\n", i+1, plft_info.Description_Mode[i].NumOfLFTables);
    }
    printf("\n");
    return 0;
}

int SMPARQuery::PLFTDefProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string optional_args) {
    struct ib_private_lft_def plft_def = {0};
    int rc, i;

    i = optional_args.find(' ');
    string block_num_str = optional_args.substr(0, i);
    string group_table_str = optional_args.erase(0, i + 1);

    u_int32_t block_num = (u_int32_t)strtoul(block_num_str.c_str(), NULL, 0);

    if (block_num >= AR_LFT_TABLE_BLOCK_SIZE_SX) {
        ERR_PRINT("block number %u is out of range [0,%u]\n",
                  block_num, AR_LFT_TABLE_BLOCK_SIZE_SX - 1);
        return 1;
    }
    // u_int32_t attr_mod = 0x000000f & block_num; // removed because used ibis API

    if (lid) {  // send by lid
        INFO_PRINT("Getting PLFTDef from lid=%u\n",
                   lid);

        rc = p_ibis->SMPPLFTDefMadGetSetByLid(lid,
                IBIS_IB_MAD_METHOD_GET,
                block_num,
                &plft_def);
    } else {    // send by direct route
        INFO_PRINT("Getting PLFTDef from direct route=%s\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str());

        rc = p_ibis->SMPPLFTDefMadGetSetByDirect(p_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                block_num,
                &plft_def);
    }

    rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here
    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
            (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get PrivateLFTDef, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get PrivateLFTDef, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    INFO_PRINT("PrivateLFTDef:\n");
    for (i = 0; i < 16; i++) {
        INFO_PRINT("PLFT LFT%d Offset.........................%u\n", i+1, plft_def.LFT[i].FDB_LID_Offset);
        INFO_PRINT("PLFT LFT%d Space..........................%u\n", i+1, plft_def.LFT[i].FDB_LID_Space);
        INFO_PRINT("PLFT LFT%d TBLIndex.......................%u\n", i+1, plft_def.LFT[i].FDB_Table_IX);
    }
    printf("\n");
    return 0;
}

void SMPARQuery::printFDBPortGroupMask(struct ib_private_lft_map *plft_map)
{
    u_int16_t *buf;
    int size_of_buf = (int)(sizeof(plft_map->FDB_Port_Group_Mask) / sizeof(u_int32_t));

    printf("FDB Port Group Mask:");
    for (int i = 0; i < size_of_buf; i++) {
         if (!(i % 4))
             printf("\n\t\t");
         buf = (u_int16_t *)(plft_map->FDB_Port_Group_Mask + i);
         for (int j = 0; j < int(sizeof(u_int32_t)/sizeof(u_int16_t)); j += 2) {
              printf("%04hx %04hx ", buf[j + 1], buf[j]);
         }
    }
}

int SMPARQuery::PLFTMapProcess(Ibis *p_ibis,
            u_int16_t lid, direct_route_t *p_direct_route,
            string optional_args)
{
    int rc;
    struct ib_private_lft_map plft_map = {0};

    int i = optional_args.find(' ');
    string plft_id_str = optional_args.substr(0, i);
    string control_map_str = optional_args.erase(0, i + 1);

    u_int8_t plft_id = (u_int8_t)strtoul(plft_id_str.c_str(), NULL, 0);
    if (plft_id >= MAX_SUPPORTED_PLFT_NUMBER) {
        ERR_PRINT("plft id number %u is out of range [0,%u]\n",
                  plft_id, MAX_SUPPORTED_PLFT_NUMBER - 1);
        return 1;
    }

    u_int8_t control_map;
    if (control_map_str == "") {
        // Default value 0x3, bit 1 (LFT_TopEn) and bit 0 (Port_MaskEn) set to 1;
        control_map = PORT_MASK_ENABLE | LFT_TOP_ENABLE;
    } else {
        control_map = (u_int8_t)strtoul(control_map_str.c_str(), NULL, 0);
        if (control_map > MAX_VALID_CONTROL_MAP_VALUE) {
            ERR_PRINT("Control map value %u is out of range [0,0x%x]\n",
                      control_map, MAX_VALID_CONTROL_MAP_VALUE);
            return 1;
        }
    }

    CLEAR_STRUCT(plft_map);
    plft_map.Port_MaskEn = (control_map & PORT_MASK_ENABLE) ? 1 : 0;
    plft_map.LFT_TopEn = (control_map & LFT_TOP_ENABLE) ? 1 : 0;

    if (lid) {  // send by lid
        INFO_PRINT("Getting PLFTMap from lid=%u\n",
                   lid);
        rc = p_ibis->SMPPLFTMapMadGetSetByLid(lid,
                IBIS_IB_MAD_METHOD_GET,
                plft_id,
                &plft_map);
    } else {    // send by direct route
        INFO_PRINT("Getting PLFTMap from direct route=%s\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str());
        rc = p_ibis->SMPPLFTMapMadGetSetByDirect(p_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                plft_id,
                &plft_map);
    }

    rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here
    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
            (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get PrivateLFTMap, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get PrivateLFTMap, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    INFO_PRINT("PLFTMap:\n");
    printf("LFT Top...........................%u\n", plft_map.LFT_Top);
    printFDBPortGroupMask(&plft_map);
    printf("\n");

    return 0;
}

int SMPARQuery::PortSLToPLFTMapProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string optional_args) {
    struct ib_port_sl_to_private_lft_map plft_map = {0};
    struct ib_extended_node_info ext_node_info = {0};

    int rc, i;
    int maxSLVLMap = 16;

    i = optional_args.find(' ');
    string port_block_str = optional_args.substr(0, i);

    u_int32_t port_block = (u_int32_t)strtoul(port_block_str.c_str(), NULL, 0);

    if (lid) {  // send by lid
        rc = p_ibis->SMPExtendedNodeInfoMadGetSetByLid(lid, IBIS_IB_MAD_METHOD_GET, &ext_node_info, nullptr);

        rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

        if (rc) {
            if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
                ERR_PRINT("Failed to get ExtSWInfo, err=%s\n", p_ibis->GetLastError());
            else
                ERR_PRINT("Failed to get ExtSWInfo, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
            return 1;
        }

        INFO_PRINT("Extended Node Info:\n");
        INFO_PRINT("SL2VL_Cap..........................%u\n", ext_node_info.sl2vl_cap);
        INFO_PRINT("SL2VL_Act..........................%u\n", ext_node_info.sl2vl_act);
        INFO_PRINT("Num_PCIe...........................%u\n", ext_node_info.num_pcie);
        INFO_PRINT("Num_OOB............................%u\n", ext_node_info.num_oob);
        INFO_PRINT("AnycastLIDTop......................%u\n", ext_node_info.AnycastLIDTop);
        INFO_PRINT("AnycastLidCap......................%u\n", ext_node_info.AnycastLidCap);
        INFO_PRINT("Node_Type_Extended.................%u\n", ext_node_info.node_type_extended);

        INFO_PRINT("Getting PLFTMap from lid=%u\n", lid);
        //check if VL2VL supported
        if (ext_node_info.sl2vl_cap & 0x1)
            maxSLVLMap = 8;
        INFO_PRINT("Getting PLFTMap for Port Block=%d\n", port_block);

        rc = p_ibis->SMPPortSLToPrivateLFTMapGetSetByLid(lid,
                IBIS_IB_MAD_METHOD_GET,
                port_block,
                &plft_map);

        if (rc) {
            if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
                ERR_PRINT("Failed to get PortSLToPrivateLFTMap, err=%s\n", p_ibis->GetLastError());
            else
                ERR_PRINT("Failed to get PortSLToPrivateLFTMap, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
            return 1;
        }
        INFO_PRINT("PortSLToPrivateLFTMap:\n");
        for (i = 0; i < 4; i++) {
            INFO_PRINT("PLFT ID FOR Port[%d] SL[0]..........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL0);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[1]..........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL1);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[2]..........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL2);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[3]..........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL3);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[4]..........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL4);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[5]..........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL5);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[6]..........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL6);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[7]..........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL7);
            if(maxSLVLMap > 8) {
                INFO_PRINT("PLFT ID FOR Port[%d] SL[8]..........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL8);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[9]..........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL9);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[10].........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL10);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[11].........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL11);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[12].........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL12);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[13].........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL13);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[14].........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL14);
            }
            INFO_PRINT("PLFT ID FOR Port[%d] SL[15].........................%u\n", (port_block * 4) + i, plft_map.PortSLToPLFT[i].PLFTToPortSL15);
        }
    } else {    // send by direct route
        INFO_PRINT("Getting PLFTMap from direct route=%s\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str());

        rc = p_ibis->SMPExtendedNodeInfoMadGetSetByDirect(p_direct_route, IBIS_IB_MAD_METHOD_GET, &ext_node_info, nullptr);

        rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

        if (rc) {
            if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
                ERR_PRINT("Failed to get ARInfo, err=%s\n", p_ibis->GetLastError());
            else
                ERR_PRINT("Failed to get ARInfo, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
            return 1;
        }

        INFO_PRINT("Extended Node Info:\n");
        INFO_PRINT("SL2VL_Cap..........................%u\n", ext_node_info.sl2vl_cap);
        INFO_PRINT("SL2VL_Act..........................%u\n", ext_node_info.sl2vl_act);
        INFO_PRINT("Num_PCIe...........................%u\n", ext_node_info.num_pcie);
        INFO_PRINT("Num_OOB............................%u\n", ext_node_info.num_oob);
        INFO_PRINT("Node_Type_Extended.................%u\n", ext_node_info.node_type_extended);

        if (ext_node_info.sl2vl_cap & 0x1)
            maxSLVLMap = 8;
        INFO_PRINT("Getting PLFTMap for Port Block=%d\n", port_block);

        rc = p_ibis->SMPPortSLToPrivateLFTMapGetSetByDirect(p_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                port_block,
                &plft_map);

        if (rc) {
            if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
                ERR_PRINT("Failed to get ARInfo, err=%s\n", p_ibis->GetLastError());
            else
                ERR_PRINT("Failed to get ARInfo, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
            return 1;
        }
        INFO_PRINT("PortSLToPrivateLFTMap:\n");
        for (i = 0; i < 4; i++) {
            INFO_PRINT("PLFT ID FOR Port[%d] SL[0]..........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL0);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[1]..........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL1);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[2]..........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL2);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[3]..........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL3);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[4]..........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL4);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[5]..........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL5);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[6]..........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL6);
            INFO_PRINT("PLFT ID FOR Port[%d] SL[7]..........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL7);
            if(maxSLVLMap > 8) {
                INFO_PRINT("PLFT ID FOR Port[%d] SL[8]..........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL8);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[9]..........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL9);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[10].........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL10);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[11].........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL11);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[12].........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL12);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[13].........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL13);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[14].........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL14);
                INFO_PRINT("PLFT ID FOR Port[%d] SL[15].........................%u\n", i, plft_map.PortSLToPLFT[i].PLFTToPortSL15);
            }
        }
    }

    return 0;
}

int SMPARQuery::ARInfoProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string optional_args) {
    struct adaptive_routing_info ar_info = {0};
    int rc;

    if (optional_args != "") {
        ERR_PRINT("No need for more params for this operation\n");
        return 1;
    };
    if (lid) {  // send by lid
        INFO_PRINT("Getting ARInfo from lid=%u\n",
                lid);
        rc = p_ibis->SMPARInfoGetSetByLid(
            lid, IBIS_IB_MAD_METHOD_GET, false, &ar_info, NULL);
    } else {    // send by direct route
        INFO_PRINT("Getting ARInfo from direct route=%s\n",
                Ibis::ConvertDirPathToStr(p_direct_route).c_str());

        rc = p_ibis->SMPARInfoGetSetByDirect(
            p_direct_route, IBIS_IB_MAD_METHOD_GET, false, &ar_info, NULL);
    }
    rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get ARInfo, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get ARInfo, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    INFO_PRINT("ARInfo:\n");
    printf("AR Status.........................%s\n", ar_info.e ? "Enabled" : "Disabled");
    printf("Is ARN Supported..................%s\n", ar_info.is_arn_sup ? "Yes" : "No");
    printf("Is FRN Supported..................%s\n", ar_info.is_frn_sup ? "Yes" : "No");
    printf("Is FR Supported...................%s\n", ar_info.is_fr_sup ? "Yes" : "No");
    printf("FR Enabled........................%s\n", ar_info.fr_enabled ? "Yes" : "No");
    printf("RN Xmit Enabled...................%s\n", ar_info.rn_xmit_enabled ? "Yes" : "No");
    printf("Is AR Trials Supported............%s\n", ar_info.is_ar_trials_supported ? "Yes" : "No");
    printf("AR Sub Groups Active..............%d\n", ar_info.sub_grps_active);
    printf("AR Groups Copy Supported..........%d\n", ar_info.group_table_copy_sup);

    printf("Direction Num Supported...........%d\n", ar_info.direction_num_sup);
    printf("AR IS4 mode.......................%s\n", ar_info.is4_mode ? "Yes" : "No");
    printf("AR Glb Group......................%s\n", ar_info.glb_groups ? "Yes" : "No");
    printf("AR By SL Cap......................%s\n", ar_info.by_sl_cap ? "Yes" : "No");
    printf("AR By SL Enabled..................%s\n", ar_info.by_sl_en ? "Yes" : "No");
    printf("AR By Transport Cap...............%s\n", ar_info.by_transp_cap ? "Yes" : "No");
    printf("AR Dynamic Cap Calc...............%s\n", ar_info.dyn_cap_calc_sup ? "Yes" : "No");

    printf("AR Group Capability...............%u\n", ar_info.group_cap);
    printf("AR Group Top......................%u\n", ar_info.group_top);
    printf("AR Group Table Capability.........%u\n", ar_info.group_table_cap);
    printf("RN String Width Capability........%u\n", ar_info.string_width_cap);

    printf("AR Version........................%u\n", ar_info.ar_version_cap);
    printf("RN Version........................%u\n", ar_info.rn_version_cap);
    printf("AR Sub Groups Capability..........0x%01x\n", ar_info.sub_grps_supported);

    if (ar_info.by_sl_en && ar_info.e) {
        printf("AR By SL Mask Enable..............%u\n", ar_info.by_sl_en);
        printf("AR SL Mask........................0x%02x\n", ar_info.enable_by_sl_mask);
    } else {
         printf("AR By SL Mask Enable..............%s\n",
                ar_info.e ? "0 (All SLs enabled)" : "N/A");
         printf("AR SL Mask........................N/A\n");
    }

    printf("AR By TransportDisable............0x%01x\n", ar_info.by_transport_disable);
    printf("AR Ageing Time Value..............%u\n", ar_info.ageing_time_value);

    printf("PFRN Enabled......................%s\n", ar_info.pfrn_enabled ? "Yes" : "No");
    printf("By VL Enabled.....................%s\n", ar_info.by_vl_en ? "Yes" : "No");
    printf("WHBF Enabled......................%s\n", ar_info.whbf_en ? "Yes" : "No");
    printf("By SL HBF Enabled.................%s\n", ar_info.by_sl_hbf_en ? "Yes" : "No");

    printf("Is PFRN Supported.................%s\n", ar_info.is_pfrn_supported ? "Yes" : "No");
    printf("By VL Capability..................%s\n", ar_info.by_vl_cap ? "Yes" : "No");
    printf("Is BTH DQP Hash Supported.........%s\n", ar_info.is_bth_dqp_hash_supported ? "Yes" : "No");
    printf("Is DCETH Hash Supported...........%s\n", ar_info.is_dceth_hash_supported ? "Yes" : "No");
    printf("Is Symmetric Hash Supported.......%s\n", ar_info.is_symmetric_hash_supported ? "Yes" : "No");
    printf("Is WHBF Supported.................%s\n", ar_info.is_whbf_supported ? "Yes" : "No");
    printf("Is HBF Supported..................%s\n", ar_info.is_hbf_supported ? "Yes" : "No");

    if (ar_info.by_sl_hbf_en)
        printf("SL Mask HBF.......................0x%02x\n", ar_info.enable_by_sl_mask_hbf);
    else
        printf("SL Mask HBF.......................N/A\n");

    printf("WHBF Granularity..................%u\n", ar_info.whbf_granularity);

    if (ar_info.by_vl_en && ar_info.e) {
        printf("AR By VL Mask Enable..............%u\n", ar_info.by_vl_en);
        printf("AR VL Mask........................0x%02x\n", ar_info.enable_by_vl_mask);
    } else {
         printf("AR By VL Mask Enable..............%s\n",
                ar_info.e ? "0 (All VLs enabled)" : "N/A");
         printf("AR VL Mask........................N/A\n");
    }

    printf("SL VL Mode........................%u\n", ar_info.sl_vl_mode);

    // todo add 'if' for new bit 'by_vl_hbf_en' when available
    printf("VL Mask HBF.......................0x%02x\n", ar_info.enable_by_vl_mask_hbf);

    printf("\n");

    return 0;
}

#define AR_GROUP_TABLE_NUM_BLOCKS_MASK  0xfff

int SMPARQuery::ARGroupTableProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string optional_args) {
    struct ib_ar_grp_table ar_group_table = {0};
    struct adaptive_routing_info ar_info = {0};

    int rc;

    int i = optional_args.find(' ');
    string plft_id_str = optional_args.substr(0, i);
    optional_args.erase(0, i + 1);
    i = optional_args.find(' ');
    string group_table_str = optional_args.substr(0, i);
    string block_num_str = optional_args.erase(0, i + 1);

    u_int8_t  plft_id = (u_int32_t)strtoul(plft_id_str.c_str(), NULL, 0);
    u_int32_t block_num   = (u_int32_t)strtoul(block_num_str.c_str(), NULL, 0);
    u_int8_t  group_table = (u_int32_t)strtoul(group_table_str.c_str(), NULL, 0);

    if (plft_id >= MAX_SUPPORTED_PLFT_NUMBER) {
        ERR_PRINT("plft id number %u is out of range [0,%u]\n",
                  plft_id, MAX_SUPPORTED_PLFT_NUMBER - 1);
        return 1;
    }

    if (block_num >= AR_GROUP_TABLE_NUM_BLOCKS) {
        ERR_PRINT("block number %u is out of range [0,%u]\n",
                block_num, AR_GROUP_TABLE_NUM_BLOCKS - 1);
        return 1;
    }
    if (group_table >= ARMGR_MAX_GROUP_TABLES) {
        ERR_PRINT("group table number %u is out of range [0,%u]\n",
                block_num, ARMGR_MAX_GROUP_TABLES - 1);
        return 1;
    }

    if (lid) {  // send by lid
        rc = p_ibis->SMPARInfoGetSetByLid(
            lid, IBIS_IB_MAD_METHOD_GET, false, &ar_info, NULL);

        rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

        if (rc) {
            if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                    (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
                ERR_PRINT("Failed to get ARInfo, err=%s\n", p_ibis->GetLastError());
            else
                ERR_PRINT("Failed to get ARInfo, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
            return 1;
        }

        INFO_PRINT("Getting ARGroupTable from lid=%u, plft=%u, table=%u, block=%u \n",
                   lid, plft_id, group_table,
                   (AR_GROUP_TABLE_NUM_BLOCKS_MASK & block_num));

        rc = p_ibis->SMPARGroupTableGetSetByLid(lid,
                IBIS_IB_MAD_METHOD_GET,
                block_num,
                group_table,
                plft_id,
                &ar_group_table);

    } else {    // send by direct route
        // aquiring ARInfo for getting PSE bit
        rc = p_ibis->SMPARInfoGetSetByDirect(
            p_direct_route, IBIS_IB_MAD_METHOD_GET, false, &ar_info, NULL);

        rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

        if (rc) {
            if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                    (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
                ERR_PRINT("Failed to get ARInfo, err=%s\n", p_ibis->GetLastError());
            else
                ERR_PRINT("Failed to get ARInfo, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
            return 1;
        }

        INFO_PRINT("Getting ARGroupTable from direct route=%s, plft=%u, table=%u, block=%u\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str(), plft_id, group_table, block_num);

        rc = p_ibis->SMPARGroupTableGetSetByDirect(p_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                block_num,
                group_table,
                plft_id,
                &ar_group_table);

    }
    rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get ARGroupTable, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get ARGroupTable, err=%s\n\n"    \
                    "NOTE: make sure AR is Enabled\n",
                    Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    unsigned group_num;
    unsigned sub_group_num;
    INFO_PRINT("ARGroupTable (Sub groups in group: %d):\n", ar_info.sub_grps_active+1 );
    for (unsigned group_entry = 0; group_entry < AR_GROUP_TABLE_BLOCK_SIZE; ++group_entry) {
        u_int64_t bitmask = ar_group_table.Group[group_entry].SubGroup_0;

        group_num =  (group_entry + block_num * AR_GROUP_TABLE_BLOCK_SIZE) /
            (ar_info.sub_grps_active+1);
        sub_group_num = (group_entry + block_num * AR_GROUP_TABLE_BLOCK_SIZE) %
            (ar_info.sub_grps_active+1);

        if (bitmask) {
            printf("Group...%u\tSub Group...%u\tMask....."
                   U64H_FMT "\t\t[Ports...\"%s\"]\n",
                    group_num, sub_group_num, bitmask,
                    ConvertBitMaskToStr(bitmask).c_str());
        } else {
            printf("Group...%u\tSub Group...%u\tMask....."
                   U64H_FMT "\t\t[Ports...\"%s\"]\n",
                   group_num, sub_group_num, bitmask,
                    "Empty");
        }
    }
    return 0;
}

int SMPARQuery::ARLFTTableProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string optional_args) {

    int i = optional_args.find(' ');
    u_int32_t max_tbl_blocks;
    string plft_id_str = optional_args.substr(0, i);
    string block_num_str = optional_args.erase(0, i + 1);

    max_tbl_blocks = AR_LFT_TABLE_NUM_BLOCKS_SX;

    u_int32_t plft_id = (u_int32_t)strtoul(plft_id_str.c_str(), NULL, 0);
    u_int32_t block_num = (u_int32_t)strtoul(block_num_str.c_str(), NULL, 0);

    if (block_num >= max_tbl_blocks) {
        ERR_PRINT("block number %u is out of range [0,%u]\n",
                block_num, max_tbl_blocks - 1);
        return 1;
    }

    if (plft_id >= MAX_SUPPORTED_PLFT_NUMBER) {
        ERR_PRINT("plft id number %u is out of range [0,%u]\n",
                  plft_id, MAX_SUPPORTED_PLFT_NUMBER - 1);
        return 1;
    }

    return GetLFTTable_SX(p_ibis, lid, p_direct_route, plft_id, block_num);
}

int SMPARQuery::GetLFTTable(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        u_int32_t plft_id, u_int32_t block_num) {
    struct ib_ar_linear_forwarding_table ar_lft_table = {0};
    int rc;

    if (lid) {  // send by lid
        INFO_PRINT("Getting ARLFTTable from lid=%u, plft=%u, block=%u\n",
                   lid, plft_id, block_num);

        rc = p_ibis->SMPARLinearForwardingTableNoSXGetSetByLid(lid,
                IBIS_IB_MAD_METHOD_GET,
                block_num,
                plft_id,
                &ar_lft_table);
    } else {    // send by direct route
        INFO_PRINT("Getting ARLFTTable from direct route=%s, block=%u\n",
                Ibis::ConvertDirPathToStr(p_direct_route).c_str(), block_num);

        rc = p_ibis->SMPARLinearForwardingTableNoSXGetSetByDirect(p_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                block_num,
                plft_id,
                &ar_lft_table);
    }
    rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get ARLFTTable, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get ARLFTTable, err=%s\n\n"    \
                    "NOTE: make sure AR is Enabled\n",
                    Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    INFO_PRINT("ARLFTTable:\n");
    for (unsigned lft_entry = 0; lft_entry < AR_LFT_TABLE_BLOCK_SIZE; ++lft_entry) {
        unsigned fixed_lft_entry = FixLFTEntryNum(lft_entry);
        if (ar_lft_table.LidEntry[fixed_lft_entry].GroupNumber != OSM_NO_PATH) {
            printf("LID....." U16H_FMT "\t\tTable.....%-5u\t\tState.....%-10s\t\tGroup.....%-5u\n",
                    lft_entry + block_num * AR_LFT_TABLE_BLOCK_SIZE,
                    ar_lft_table.LidEntry[fixed_lft_entry].TableNumber,
                    ConvertARIBLidStateToStr(ar_lft_table.LidEntry[fixed_lft_entry].LidState),
                    ar_lft_table.LidEntry[fixed_lft_entry].GroupNumber);
        } else {
            printf("LID....." U16H_FMT "\t\tTable.....%-5u\t\tState.....%-10s\t\tGroup.....%s\n",
                    lft_entry + block_num * AR_LFT_TABLE_BLOCK_SIZE,
                    ar_lft_table.LidEntry[fixed_lft_entry].TableNumber,
                    ConvertARIBLidStateToStr(ar_lft_table.LidEntry[fixed_lft_entry].LidState),
                    "Unreachable");
        }
    }
    return 0;
}

int SMPARQuery::GetLFTTable_SX(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        u_int32_t plft_id, u_int32_t block_num) {
    struct ib_ar_linear_forwarding_table_sx ar_lft_table = {0};
    int rc;

    struct adaptive_routing_info ar_info = {0};
    rc = GetArInfo(p_ibis, lid, p_direct_route, ar_info);

    if (rc)
        return rc;

    if (!ar_info.e) {
        ERR_PRINT("AR/FR is not enabled.\n");
        return 1;
    }


    if (lid) {  // send by lid
        INFO_PRINT("Getting ARLFTTable from lid=%u, plft=%u, block=%u\n",
                   lid, plft_id, block_num);

        rc = p_ibis->SMPARLinearForwardingTableGetSetByLid(lid,
                IBIS_IB_MAD_METHOD_GET,
                block_num,
                plft_id,
                &ar_lft_table);

    } else {    // send by direct route
        INFO_PRINT("Getting ARLFTTable from direct route=%s, block=%u\n",
                Ibis::ConvertDirPathToStr(p_direct_route).c_str(), block_num);

        rc = p_ibis->SMPARLinearForwardingTableGetSetByDirect(p_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                block_num,
                plft_id,
                &ar_lft_table);
    }

    rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get ARLFTTable, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get ARLFTTable, err=%s\n\n"    \
                    "NOTE: make sure AR is Enabled\n",
                    Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    INFO_PRINT("ARLFTTable:\n");
    for (unsigned lft_entry = 0; lft_entry < AR_LFT_TABLE_BLOCK_SIZE_SX; ++lft_entry) {
        if (ar_lft_table.LidEntry[lft_entry].LidState == AR_IB_LID_STATE_BOUNDED ||
            ar_lft_table.LidEntry[lft_entry].LidState == AR_IB_LID_STATE_FREE ||
            ar_info.fr_enabled) { //if fr enabled group number
                                  //is is valid also for static
            printf("LID....." U16H_FMT "\t\tTable.....%-5u\t\tState.....%-10s\t\t"
                   "Default Port.....%-5u\t\tGroup.....%-5u\n",
                    lft_entry + block_num * AR_LFT_TABLE_BLOCK_SIZE_SX,
                    ar_lft_table.LidEntry[lft_entry].TableNumber,
                    ConvertARIBLidStateToStr(ar_lft_table.LidEntry[lft_entry].LidState),
                    ar_lft_table.LidEntry[lft_entry].DefaultPort,
                    ar_lft_table.LidEntry[lft_entry].GroupNumber);
        } else {
            printf("LID....." U16H_FMT "\t\tTable.....%-5u\t\tState.....%-10s\t\t"
                   "Default Port.....%-5u\t\tGroup.....%s\n",
                    lft_entry + block_num * AR_LFT_TABLE_BLOCK_SIZE_SX,
                    ar_lft_table.LidEntry[lft_entry].TableNumber,
                    ConvertARIBLidStateToStr(ar_lft_table.LidEntry[lft_entry].LidState),
                    ar_lft_table.LidEntry[lft_entry].DefaultPort,
                    "no group assigned");
        }
    }
    return 0;
}

int SMPARQuery::GetArInfo(Ibis *p_ibis, u_int16_t lid,
                          direct_route_t *p_direct_route,
                          struct adaptive_routing_info &ar_info) {

    memset(&ar_info, 0, sizeof(ar_info));

    int rc;

    if (lid) {  // send by lid
        // aquiring ARInfo for getting PSE bit
        rc = p_ibis->SMPARInfoGetSetByLid(
            lid, IBIS_IB_MAD_METHOD_GET, false, &ar_info, NULL);

        rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

        if (rc) {
            if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                    (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
                ERR_PRINT("Failed to get ARInfo, err=%s\n", p_ibis->GetLastError());
            else
                ERR_PRINT("Failed to get ARInfo, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
            return 1;
        }
    } else {    // send by direct route
        // aquiring ARInfo for getting PSE bit
        rc = p_ibis->SMPARInfoGetSetByDirect(
            p_direct_route, IBIS_IB_MAD_METHOD_GET, false, &ar_info, NULL);

        rc = rc & 0x00ff;   //bits [8-15] are class specific - don't need them here

        if (rc) {
            if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                    (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
                ERR_PRINT("Failed to get ARInfo, err=%s\n", p_ibis->GetLastError());
            else
                ERR_PRINT("Failed to get ARInfo, err=%s\n", Ibis::ConvertMadStatusToStr(rc).c_str());
            return 1;
        }
    }
    return 0;
}

int SMPARQuery::RNSubGroupDirectionTableProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string optional_args) {

    struct rn_sub_group_direction_tbl sub_group_direction_table = {0};
    struct adaptive_routing_info ar_info = {0};

    int rc = GetArInfo(p_ibis, lid, p_direct_route, ar_info);

    if (rc)
        return rc;

    if (!ar_info.is_arn_sup && !ar_info.is_frn_sup) {
        ERR_PRINT("RNSubGroupDirectionTable is not supported.\n");
        return 1;
    }

    int i = optional_args.find(' ');
    string block_num_str = optional_args.substr(0, i);
    u_int32_t block_num   = (u_int32_t)strtoul(block_num_str.c_str(), NULL, 0);

    u_int32_t top_block_num =
        ((ar_info.group_top + 1) * (ar_info.sub_grps_active + 1) - 1) /
        RN_SUB_GROUP_DIRECTION_BLOCK_SIZE;

    if (block_num > top_block_num) {
        ERR_PRINT("block number %u is greater than top block number %u\n",
                block_num, top_block_num);
        return 1;
    }

    if (lid) {  // send by lid
        INFO_PRINT("Getting RNSubGroupDirectionTable from lid=%u, block=%u \n",
                   lid, block_num);
        rc = p_ibis->SMPRNSubGroupDirectionTableGetSetByLid(
            lid, IBIS_IB_MAD_METHOD_GET, block_num,
            &sub_group_direction_table, NULL);

    } else {    // send by direct route
        INFO_PRINT("Getting RNSubGroupDirectionTable from direct route=%s, block=%u\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str(), block_num);

        rc = p_ibis->SMPRNSubGroupDirectionTableGetSetByDirect(
            p_direct_route, IBIS_IB_MAD_METHOD_GET, block_num,
            &sub_group_direction_table, NULL);

    }

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get RNSubGroupDirectionTable, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get RNSubGroupDirectionTable, err=%s\n",
                    Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    unsigned group_num;
    unsigned sub_group_num;
    INFO_PRINT("RNSubGroupDirectionTable (Sub groups in group: %d):\n", ar_info.sub_grps_active+1 );
    for (unsigned group_entry = 0;
         group_entry < RN_SUB_GROUP_DIRECTION_BLOCK_SIZE; ++group_entry) {

        group_num =  (group_entry + block_num * RN_SUB_GROUP_DIRECTION_BLOCK_SIZE) /
            (ar_info.sub_grps_active+1);
        sub_group_num = (group_entry + block_num * RN_SUB_GROUP_DIRECTION_BLOCK_SIZE) %
            (ar_info.sub_grps_active+1);

        if (group_num > ar_info.group_top)
            break;

        printf("Group...%u\tSub Group...%u\tDirection...%u\n",
               group_num, sub_group_num,
               sub_group_direction_table.direction_of_sub_group[group_entry]);

    }
    return 0;
}

int SMPARQuery::RNGenStringTableProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string optional_args) {

    struct rn_gen_string_tbl rn_gen_string_table = {0};
    struct adaptive_routing_info ar_info = {0};

    int rc = GetArInfo(p_ibis, lid, p_direct_route, ar_info);

    if (rc)
        return rc;

    if (!ar_info.is_arn_sup && !ar_info.is_frn_sup) {
        ERR_PRINT("RNGenStringTable is not supported.\n");
        return 1;
    }

    int i = optional_args.find(' ');
    string plft_id_str = optional_args.substr(0, i);
    string block_num_str = optional_args.erase(0, i + 1);
    u_int32_t block_num   = (u_int32_t)strtoul(block_num_str.c_str(), NULL, 0);
    u_int32_t plft_id = (u_int32_t)strtoul(plft_id_str.c_str(), NULL, 0);

    u_int32_t top_block_num = ar_info.direction_num_sup /
        RN_GEN_STRING_TABLE_BLOCK_SIZE;

    if (block_num > top_block_num) {
        ERR_PRINT("block number %u is greater than top block number %u\n",
                block_num, top_block_num);
        return 1;
    }

    if (lid) {  // send by lid
        INFO_PRINT("Getting RNGenStringTable from lid=%u, plft=%u,"
                   " block=%u \n",
                   lid, plft_id, block_num);
        rc = p_ibis->SMPRNGenStringTableGetSetByLid(lid,
                         IBIS_IB_MAD_METHOD_GET,
                         block_num,
                         plft_id,
                         &rn_gen_string_table,
                         NULL);

    } else {    // send by direct route
        INFO_PRINT("Getting RNGenStringTable from direct route=%s, plft=%u,"
                   " block=%u\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                   plft_id, block_num);
        rc = p_ibis->SMPRNGenStringTableGetSetByDirect(p_direct_route,
                         IBIS_IB_MAD_METHOD_GET,
                         block_num,
                         plft_id,
                         &rn_gen_string_table,
                         NULL);

    }

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get RNGenStringTable, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get RNGenStringTable, err=%s\n",
                    Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    unsigned direction;
    for (unsigned block_entry = 0;
         block_entry < RN_GEN_STRING_TABLE_BLOCK_SIZE; ++block_entry) {

        direction =  (block_entry + block_num * RN_GEN_STRING_TABLE_BLOCK_SIZE);

        if (direction > ar_info.direction_num_sup)
            break;

        printf("Direction...%u\tGenerated String...%u\n",direction,
               rn_gen_string_table.element[block_entry].rn_gen_string);
    }
    return 0;
}

int SMPARQuery::RNGenBySubGroupPriorityProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string /*optional_args*/) {

    struct rn_gen_by_sub_group_prio rn_gen_by_sg_priority = {0};
    struct adaptive_routing_info ar_info = {0};

    int rc = GetArInfo(p_ibis, lid, p_direct_route, ar_info);

    if (rc)
        return rc;

    if (!ar_info.is_arn_sup && !ar_info.is_frn_sup) {
        ERR_PRINT("RNGenBySubGroupPriorityProcess is not supported.\n");
        return 1;
    }

    if (lid) {  // send by lid
        INFO_PRINT("Getting RNGenBySubGroupPriority from lid=%u\n",lid);

        rc = p_ibis->SMPRNGenBySubGroupPriorityMadGetSetByLid(
            lid,
            IBIS_IB_MAD_METHOD_GET,
            &rn_gen_by_sg_priority,
            NULL);

    } else {    // send by direct route
        INFO_PRINT("Getting RNGenBySubGroupPriority from direct route=%s\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str());

        rc = p_ibis->SMPRNGenBySubGroupPriorityMadGetSetByDirect(
            p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            &rn_gen_by_sg_priority,
            NULL);

    }

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get RNGenBySubGroupPriority, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get RNGenBySubGroupPriority, err=%s\n",
                    Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    for (u_int8_t sg = 0;
         sg < ar_info.sub_grps_active + 1; ++sg) {

        printf("SubGroup Priority...%u\tgen_arn...%u\tgen_frn...%u\n",
               (int)sg,
               (int)rn_gen_by_sg_priority.element[sg].gen_arn,
               (int)rn_gen_by_sg_priority.element[sg].gen_frn);
    }
    return 0;
}

static const char *DecisionToStr(u_int8_t decision){
    switch (decision) {
        case 0:
            return "Discard";
        case 1:
            return "Consume ARN";
        case 2:
            return "Consume ARN/FRN";
        case 3:
            return "Pass-on";
    }

    return "Invalid";
}

int SMPARQuery::RNRcvStringProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string optional_args) {

    struct rn_rcv_string rcv_string = {0};
    struct adaptive_routing_info ar_info = {0};

    int rc = GetArInfo(p_ibis, lid, p_direct_route, ar_info);

    if (rc)
        return rc;

    if (!ar_info.is_arn_sup && !ar_info.is_frn_sup) {
        ERR_PRINT("RNRcvString is not supported.\n");
        return 1;
    }

    int i = optional_args.find(' ');
    string block_num_str = optional_args.substr(0, i);
    u_int32_t block_num   = (u_int32_t)strtoul(block_num_str.c_str(), NULL, 0);

    u_int32_t max_string = (0x1 << (ar_info.string_width_cap)) - 1;
    u_int32_t top_block_num = max_string /
        RN_RCV_STRING_BLOCK_SIZE;

    if (block_num > top_block_num) {
        ERR_PRINT("block number %u is greater than top block number %u\n",
                block_num, top_block_num);
        return 1;
    }

    if (lid) {  // send by lid
        INFO_PRINT("Getting RNRcvString from lid=%u, block=%u \n",
                   lid, block_num);
        rc = p_ibis->SMPRNRcvStringGetSetByLid(
            lid,
            IBIS_IB_MAD_METHOD_GET,
            block_num,
            &rcv_string,
            NULL);

    } else {    // send by direct route
        INFO_PRINT("Getting RNRcvString from direct route=%s, block=%u\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str(), block_num);
        rc = p_ibis->SMPRNRcvStringGetSetByDirect(
            p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            block_num,
            &rcv_string,
            NULL);

    }

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get RNRcvString, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get RNRcvString, err=%s\n",
                    Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    unsigned string_num;
    for (unsigned block_entry = 0;
         block_entry < RN_SUB_GROUP_DIRECTION_BLOCK_SIZE; ++block_entry) {

        string_num =  (block_entry + block_num * RN_RCV_STRING_BLOCK_SIZE);

        if (string_num > max_string)
            break;

        printf("String...%u\tDecision...%s\tpLFT...%u\tString to string...%u\n",
               string_num,
               DecisionToStr(rcv_string.element[block_entry].decision),
               rcv_string.element[block_entry].plft_id,
               rcv_string.element[block_entry].string2string);
    }
    return 0;
}

int SMPARQuery::RNXmitPortMaskProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string optional_args) {

    struct rn_xmit_port_mask xmit_port_mask = {0};
    struct adaptive_routing_info ar_info = {0};

    int rc = GetArInfo(p_ibis, lid, p_direct_route, ar_info);

    if (rc)
        return rc;

    if (!ar_info.is_arn_sup && !ar_info.is_frn_sup) {
        ERR_PRINT("RNXmitPortMask is not supported.\n");
        return 1;
    }

    struct SMP_NodeInfo node_info = {0};
    rc = GetNodeInfo(p_ibis, lid, p_direct_route, node_info);
    if (rc)
        return rc;

    int i = optional_args.find(' ');
    string block_num_str = optional_args.substr(0, i);
    u_int32_t block_num   = (u_int32_t)strtoul(block_num_str.c_str(), NULL, 0);

    u_int32_t top_block_num = node_info.NumPorts /
        RN_XMIT_PORT_MASK_BLOCK_SIZE;

    if (block_num > top_block_num) {
        ERR_PRINT("block number %u is greater than top block number %u\n",
                block_num, top_block_num);
        return 1;
    }

    if (lid) {  // send by lid
        INFO_PRINT("Getting RNXmitPortMask from lid=%u, block=%u \n",
                   lid, block_num);
        rc = p_ibis->SMPRNXmitPortMaskGetSetByLid(
            lid,
            IBIS_IB_MAD_METHOD_GET,
            block_num,
            &xmit_port_mask,
            NULL);

    } else {    // send by direct route
        INFO_PRINT("Getting RNXmitPortMask from direct route=%s, block=%u\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str(), block_num);

        rc = p_ibis->SMPRNXmitPortMaskGetSetByDirect(
            p_direct_route,
            IBIS_IB_MAD_METHOD_GET,
            block_num,
            &xmit_port_mask,
            NULL);

    }

    if (rc) {
        if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
            ERR_PRINT("Failed to get RNXmitPortMask, err=%s\n", p_ibis->GetLastError());
        else
            ERR_PRINT("Failed to get RNXmitPortMask, err=%s\n",
                    Ibis::ConvertMadStatusToStr(rc).c_str());
        return 1;
    }

    unsigned port_num;
    for (unsigned block_entry = 0;
         block_entry < RN_XMIT_PORT_MASK_BLOCK_SIZE; ++block_entry) {

        port_num =  (block_entry + block_num * RN_XMIT_PORT_MASK_BLOCK_SIZE);

        if (port_num > node_info.NumPorts)
            break;

        u_int8_t element =
            xmit_port_mask.element[block_entry];

        printf("Port...%u\tGenerate ARN...%u\tGenerate FRN...%u\tPass on RN...%u\n",
               port_num,
               (bool)(element & RN_XMIT_PORT_MASK_GENERATE_ARN),
               (bool)(element & RN_XMIT_PORT_MASK_GENERATE_FRN),
               (bool)(element & RN_XMIT_PORT_MASK_PASS_ON_RN));
    }
    return 0;
}

int SMPARQuery::PortRNCountersProcess(Ibis *p_ibis,
        u_int16_t lid, direct_route_t *p_direct_route,
        string /*optional_args*/) {

    struct port_rn_counters counters = {0};
    struct adaptive_routing_info ar_info = {0};

    if (!lid) {
        ERR_PRINT("PortRNCounters supports only lid route.\n");
        return 1;
    }

    int rc = GetArInfo(p_ibis, lid, p_direct_route, ar_info);

    if (rc)
        return rc;

    if (!ar_info.is_arn_sup && !ar_info.is_frn_sup) {
        ERR_PRINT("PortRNCounters is not supported.\n");
        return 1;
    }

    struct SMP_NodeInfo node_info = {0};
    rc = GetNodeInfo(p_ibis, lid, p_direct_route, node_info);
    if (rc)
        return rc;

    if (lid)
        INFO_PRINT("Getting PortRNCounters from lid=%u\n",lid);
    else
        INFO_PRINT("Getting PortRNCounters from direct route=%s\n",
                   Ibis::ConvertDirPathToStr(p_direct_route).c_str());

    for (u_int8_t port_num = 1;
         port_num <=  node_info.NumPorts;
         ++port_num) {

        rc = p_ibis->VSPortRNCountersGet(lid, port_num, &counters, NULL);

        if (rc) {
            if ((rc == IBIS_MAD_STATUS_GENERAL_ERR) || (rc == IBIS_MAD_STATUS_TIMEOUT) ||
                    (rc == IBIS_MAD_STATUS_RECV_FAILED) || (rc == IBIS_MAD_STATUS_SEND_FAILED))
                ERR_PRINT("Failed to get PortRNCounters, port:%u err=%s\n",
                          port_num, p_ibis->GetLastError());
            else
                ERR_PRINT("Failed to get PortRNCounters, port:%u err=%s\n",
                        port_num, Ibis::ConvertMadStatusToStr(rc).c_str());
            return 1;
        }

        cout << "Port..." << +port_num << endl
             << "\t Rcv rn pkt..." << counters.port_rcv_rn_pkt << endl
             << "\t Xmit rn pkt..." << counters.port_xmit_rn_pkt << endl
             << "\t Rcv rn error..." << counters.port_rcv_rn_error << endl
             << "\t Rcv switch relay rn error..." << counters.port_rcv_switch_relay_rn_error << endl;

        if (ar_info.is_ar_trials_supported)
            cout << "\t Port AR trials..." << counters.port_ar_trials << endl;

        if (ar_info.is_pfrn_supported)
            cout << "\t PFRN Received packet..."  << counters.pfrn_received_packet << endl
                 << "\t PFRN Received error..."   << counters.pfrn_received_error  << endl
                 << "\t PFRN Xmit packet..."      << counters.pfrn_xmit_packet     << endl
                 << "\t PFRN Start packet..."     << counters.pfrn_start_packet    << endl;
    }

    return 0;
}

/****************************************************/
int main(int argc, const char **argv)
{
    std::unique_ptr<SMPARQuery> smparquery_obj(new SMPARQuery());

    // check if init was done
    if (smparquery_obj->IsInitialized() == false)
        return 1;

    // get args
    int rc = smparquery_obj->GetOpt(argc, argv);
    if (rc) {
        if (rc == 2)    //2 - OK but need to exit
            return 0;
        return 1;
    }
    try {
        rc = smparquery_obj->Run();
    }
    catch(std::string & s) {
        std::cout << "Internal error: " << s << std::endl;
    }
    catch(...) {
        std::cout << "Internal error: unhandled exception..." << std::endl;
    }
    return rc;
}

